/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kill.c: KILL command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kill.c,v 1.7 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_kill(aClient *, aClient *, int, char **);
static char buf[BUFSIZE];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_kill.so",
	"/KILL command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("KILL", m_kill, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("KILL", m_kill);
}
#endif

/*
 * m_kill
 *	parv[0] = sender prefix
 *	parv[1] = target
 *	parv[2] = path
 */
int m_kill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr;
	char *killer, *p = NULL, *nick;
	char path[KILLLEN + 1];
	int chasing = 0, kcount = 0;

	if (!IsPrivileged(cptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KILL");
		return 0;
	}
	if (parc < 3 || *parv[1] == '\0') {
		parv[2] = "()";
	}

	if (MyClient(sptr)) {
		parv[1] = canonize(parv[1]);
	}
	if (strlen(parv[2]) > KILLLEN) {
		parv[2][KILLLEN] = '\0';
	}

	for (nick = strtoken(&p, parv[1], ","); nick; nick = strtoken(&p, NULL, ",")) {
		chasing = 0;
		acptr = find_client(nick, NULL);
		if (!acptr) {
			acptr = get_history(nick, (long)KILLCHASETIMELIMIT);
			if (!acptr) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
				continue;
			}
			sendto_one(sptr, ":%s NOTICE %s :KILL target changed from %s to %s", me.name,
				parv[0], nick, acptr->name);
			chasing = 1;
		}

		if ((!MyConnect(acptr) && MyClient(cptr) && !OPCanGKill(cptr)) ||
		  (MyConnect(acptr) && MyClient(cptr) && !OPCanLKill(cptr))) {
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
			continue;
		}
		if (IsServer(acptr) || IsMe(acptr)) {
			sendto_one(sptr, err_str(ERR_CANTKILLSERVER), me.name, parv[0]);
			ircdlog(LOG_KILL, "client %s attempted to kill server %s",
				get_client_name(cptr, FALSE), acptr->name);
			continue;
		}
		if (IsServices(acptr) && !IsServer(sptr) && !IsULine(sptr)) {
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
			ircdlog(LOG_KILL, "client %s attempted to kill service client %s",
				get_client_name(cptr, FALSE), get_client_name(acptr, FALSE));
			continue;
		}

		kcount++;
		if (!IsServer(sptr) && kcount > MAXKILLS) {
			sendto_one(sptr, ":%s NOTICE %s :Too many targets, kill list truncated (maximum is %d)",
				me.name, parv[0], MAXKILLS);
			break;
		}

		if (MyClient(sptr)) {
			char myname[HOSTLEN + 1], *s;
			int slen;

			strncpy(myname, me.name, 80);
			s = index(myname, '.');
			if (s) {
				*s = '\0';
			}

			slen = KILLLEN - (strlen(myname) +
			  strlen(IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host) +
			  strlen(sptr->name) + 8);
			if (slen < 0) {
				slen = 0;
			}

			if (strlen(parv[2]) > slen) {
				parv[2][slen] = '\0';
			}

			ircsprintf(path, "%s!%s!%s (%s)", myname,
				IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, sptr->name,
				parv[2]);
		}
		else {
			strncpy(path, parv[2], KILLLEN + 1);
		}

		sendto_ops_lev(IsOper(sptr) ? 0 : SKILL_LEV, "Received KILL message for %s!%s@%s. "
			"From: %s Path: %s", acptr->name, acptr->user ? acptr->user->username : "<unknown>",
			acptr->user ? IsMasked(acptr) ? acptr->user->maskedhost : acptr->user->host : "<unknown>",
			sptr->name, path);
		if (IsOper(sptr)) {
			ircdlog(LOG_KILL, "KILL for %s!%s@%s. From: %s Path: %s", acptr->name,
				acptr->user ? acptr->user->username : "<unknown>",
				acptr->user ? IsMasked(acptr) ? acptr->user->maskedhost : sptr->user->host : "<unknown>",
				sptr->name, path);
		}

		if (!MyConnect(acptr) || !MyConnect(sptr) || !IsOper(sptr)) {
			sendto_serv_butone(cptr, ":%s KILL %s :%s", parv[0], acptr->name, path);
			if (chasing && IsServer(cptr)) {
				sendto_one(cptr, ":%s KILL %s :%s", me.name, acptr->name, path);
			}
			acptr->flags |= FLAGS_KILLED;
		}

		if (MyConnect(acptr)) {
			sendto_prefix_one(acptr, sptr, ":%s KILL %s :%s", parv[0], acptr->name, path);
		}
		if (MyConnect(acptr) && MyConnect(sptr) && IsOper(sptr)) {
			ircsprintf(buf, "Local kill by %s (%s)", sptr->name, parv[2]);
		}
		else {
			killer = strchr(path, '(');
			if (!killer) {
				killer = "()";
			}
			ircsprintf(buf, "Killed (%s %s)", sptr->name, killer);
		}
		if (exit_client(cptr, acptr, sptr, buf) == FLUSH_BUFFER) {
			return FLUSH_BUFFER;
		}
	}
	return 0;
}
