/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_kline.c: KLINE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_kline.c,v 1.12 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_kline(aClient *, aClient *, int, char **);

extern void add_ConfigItem(ConfigItem *, ConfigItem **);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_kline.so",
	"/KLINE command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("KLINE", m_kline, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("KLINE", m_kline);
}
#endif

/*
 * m_kline
 *	parv[0] = sender prefix
 *	parv[1] = user@host or nickname
 *	parv[2] = reason
 */
int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *user, *host, *userhost, *p, *reason, tmp[HOSTLEN + 2];
	ConfigItem_ban *ban;
	aClient *acptr;
	int i = 0;

	if (!MyClient(sptr) || !IsOper(sptr) || !OPCanKline(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || BadPtr(parv[1])) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KLINE");
		return 0;
	}

	host = strchr(parv[1], '@');
	if (host || *parv[1] == '*') {
		if (host) {
			user = parv[1];
			*(host++) = '\0';
		}
		else {
			user = "*";
			host = parv[1];
		}

		if (strlen(user) > USERLEN) {
			user[USERLEN] = '\0';
		}
		if (strlen(host) > HOSTLEN) {
			host[HOSTLEN] = '\0';
		}

		if (!strcmp(user, "*")) {
			for (p = host; *p; p++) {
				if (*p != '*' && *p != '.') {
					i++;
				}
			}
			if (i < 1) {
				sendto_one(sptr, ":%s NOTICE %s :*** What a sweeping K:Line! If only your "
					"admin knew you tried that...", me.name, parv[0]);
				sendto_realops("%s (%s@%s) attempted to K:Line *@*", sptr->name,
					sptr->user->username,
					IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host);
				return 0;
			}
		}
	}
	else {
		acptr = find_chasing(sptr, parv[1], NULL);
		if (!acptr || !acptr->user || IsServer(acptr)) {
			return 0;
		}

		tmp[0] = '*';
		if (*acptr->user->username == '~') {
			strcpy(tmp + 1, (char *)acptr->user->username + 1);
		}
		else {
			strcpy(tmp + 1, acptr->user->username);
		}
		user = tmp;
		host = acptr->user->host;
	}

	if (parc > 2 && !BadPtr(parv[2])) {
		reason = parv[2];
	}
	else {
		reason = "No reason specified";
	}

	userhost = make_user_host(user, host);
	ban = find_ban(BAN_USERHOST, userhost);
	if (ban) {
		sendto_one(sptr, ":%s NOTICE %s :*** %s is already on the K:Line list as %s", me.name,
			parv[0], userhost, ban->mask);
		return 0;
	}
	ban = (ConfigItem_ban *)MyMalloc(sizeof(ConfigItem_ban));
	ban->type = BAN_USERHOST;
	DupString(ban->mask, userhost);
	DupString(ban->reason, reason);
	ban->flag.temporary = 1;
	add_ConfigItem((ConfigItem *)ban, (ConfigItem **)&config_ban);
	sendto_realops("%s (%s@%s) added temporary K:Line for %s (%s)", parv[0], sptr->user->username,
		IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, ban->mask, ban->reason);

	for (i = 0; i <= highest_fd; i++) {
		if (!(acptr = local[i]) || IsMe(acptr) || IsLog(acptr)) {
			continue;
		}
		if (IsPerson(acptr)) {
			if (acptr->user->username && ((acptr->sockhost && !match(host, acptr->sockhost) &&
			  !match(user, acptr->user->username)) || (acptr->hostip && !match(host, acptr->hostip)
			  && !match(user, acptr->user->username)))) {
				sendto_realops("K:Line active for %s (%s@%s) [%s]", acptr->name,
					acptr->user->username, acptr->user->host, ban->reason);
				sendto_one(acptr, err_str(ERR_YOUREBANNEDCREEP), me.name, acptr->name, "K:Lined");
				sendto_one(acptr, ":%s NOTICE %s :*** You are not welcome on this server: %s",
					me.name, acptr->name, ban->reason);
				sendto_one(acptr, ":%s NOTICE %s :*** For more information, mail %s and include "
					"everything shown here.", me.name, acptr->name, ConfigSet.localklineaddress);
				exit_client(acptr, acptr, &me, "K:Lined");
			}
		}
	}
	return 0;
}
