/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_message.c: PRIVMSG and NOTICE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_message.c,v 1.19 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_private(aClient *, aClient *, int, char **);
int m_notice(aClient *, aClient *, int, char **);
static int parse_message(aClient *, aClient *, int, char **, int);
static int msg_has_color(char *);
static char *strip_color(char *);
static int check_for_client_flood(int, aClient *, aClient *);
static int check_for_channel_flood(int, aClient *, aChannel *);
static int check_for_ctcp(char *, char **);

extern char *canonize(char *);
extern int check_dccsend(aClient *, aClient *, char *);
extern int check_message_accept(aClient *, aClient *);
extern int is_silenced(aClient *, aClient *);
#ifdef FLUD
extern int check_for_flud(aClient *, aClient *, aChannel *chptr, int);
#endif

#define CTCP_NONE	0
#define CTCP_YES	1
#define CTCP_DCC	2
#define CTCP_DCCSEND	3

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_message.so",
	"/PRIVMSG and /NOTICE commands",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("PRIVMSG", m_private, 0, 15, 1, 0, 1, 0L);
	add_Command("NOTICE", m_notice, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("PRIVMSG", m_private);
	del_Command("NOTICE", m_notice);
}
#endif

/*
 * m_private
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
int m_private(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 0);
}

/*
 * m_notice
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
int m_notice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	return parse_message(cptr, sptr, parc, parv, 1);
}

/*
 * parse_message
 *	parv[0] = sender prefix
 *	parv[1] = targets
 *	parv[2] = message
 */
static int parse_message(aClient *cptr, aClient *sptr, int parc, char *parv[], int notice)
{
	aClient *acptr = NULL;
	aChannel *chptr;
	char *command = notice ? "NOTICE" : "PRIVMSG";
	char *p = NULL, *s = NULL, *s2 = NULL;
	char *nick, *server, *dccmsg;
	int i, ret, ischan = 0, flag = 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NORECIPIENT), me.name, parv[0], command);
		return 0;
	}
	if (parc < 3 || *parv[2] == '\0') {
		sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0], command);
		return 0;
	}

	if (MyConnect(sptr)) {
#if defined(ANTI_SPAMBOT) && !defined(ANTI_SPAMBOT_WARN_ONLY)
		if (sptr->join_leave_count >= MAX_JOIN_LEAVE_COUNT) {
			return 0;
		}
#endif
		parv[1] = canonize(parv[1]);
	}

	for (nick = strtoken(&p, parv[1], ","), i = 0; nick && i < 20; nick = strtoken(&p, NULL, ",")) {
		if (i++ > 10 && !IsULine(sptr)) {
			if ((ConfigSet.nooperflood && !IsOper(sptr)) || !ConfigSet.nooperflood) {
				sptr->since += 4;
			}
		}

		ischan = IsChannelName(nick);
		chptr = find_channel(nick, NullChn);
		if (ischan && chptr) {
			if (!notice) {
				switch (check_for_ctcp(parv[2], NULL)) {
					case CTCP_NONE:
						break;
					case CTCP_DCCSEND:
					case CTCP_DCC:
						sendto_one(sptr, ":%s NOTICE %s :You may not send a DCC "
							"command to a channel (%s)", me.name, parv[0],
							nick);
						continue;
					default:
#ifdef FLUD
						if (check_for_flud(sptr, NULL, chptr, 1)) {
							return 0;
						}
#endif
						if (chptr->mode.mode & MODE_NOCTCP) {
							sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN),
								me.name, parv[0], nick);
							continue;
						}
						break;
				}
			}

			ret = IsULine(sptr) ? 0 : can_send(sptr, chptr);
			if (!ret) {
				char *text = parv[2];
				if (MyClient(sptr)) {
					if (check_for_channel_flood(notice, sptr, chptr)) {
						continue;
					}
					if (chptr->mode.mode & MODE_NOCOLOR && msg_has_color(text)) {
						if (!notice) {
							sendto_one(sptr, err_str(ERR_NOCOLORSONCHAN), me.name,
								parv[0], nick, parv[2]);
						}
						continue;
					}
					if (chptr->mode.mode & MODE_STRIPCOLOR) {
						text = strip_color(text);
					}
				}
				sendto_channel_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0],
					command, nick, text);
			}
			else if (!notice) {
				sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], nick);
			}
			continue;
		}

		acptr = find_person(nick, NULL);
		if (!ischan && acptr) {
#ifdef FLUD
			if (!notice && MyFludConnect(acptr)) {
#else
			if (!notice && MyConnect(acptr)) {
#endif

				switch (check_for_ctcp(parv[2], &dccmsg)) {
					case CTCP_NONE:
						break;
					case CTCP_DCCSEND:
#ifdef FLUD
						if (check_for_flud(sptr, acptr, NULL, 1)) {
							return 0;
						}
#endif
						if (check_dccsend(sptr, acptr, dccmsg)) {
							continue;
						}
						break;
					default:
#ifdef FLUD
						if (check_for_flud(sptr, acptr, NULL, 1)) {
							return 0;
						}
#endif
						break;
				}
			}
			if (!IsOper(sptr) && !IsULine(sptr) && !IsServer(sptr)) {
				if (check_message_accept(sptr, acptr)) {
					continue;
				}
			}
			if (MyClient(sptr) && check_for_client_flood(notice, sptr, acptr)) {
				continue;
			}
			if (!is_silenced(sptr, acptr)) {
				if (!notice && MyClient(acptr) && acptr->user && acptr->user->away) {
					sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], acptr->name,
						acptr->user->away);
				}
				sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0], command, nick,
					parv[2]);
			}
			continue;
		}

		s = strchr(nick, '#');
		if (s) {
			chptr = find_channel(s, NullChn);
			if (!chptr) {
				sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], s);
				continue;
			}
			if (can_send(sptr, chptr) && !IsULine(sptr)) {
				if (!notice) {
					sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, parv[0], s);
				}
				continue;
			}
			if (MyClient(sptr) && check_for_channel_flood(notice, sptr, chptr)) {
				continue;
			}

			if (s == nick) {
				sendto_channel_butone(cptr, sptr, chptr, ":%s %s %s :%s", parv[0], command,
					nick, parv[2]);
			}
			else {
				flag = 0;
				for (s2 = nick; s2 != s; s2++) {
					switch (*s2) {
						case '*':
							flag |= CHFL_CHANADMIN;
							break;
						case '@':
							flag |= CHFL_CHANOP;
							break;
						case '%':
							flag |= CHFL_HALFOP;
							break;
						case '+':
							flag |= CHFL_VOICE;
							break;
						default:
							break;
					}
				}
				sendto_channelflag_butone(flag, cptr, sptr, chptr, ":%s %s %s :%s", parv[0],
					command, nick, parv[2]);
			}
			continue;
		}

		if (IsOper(sptr)) {
			if (*nick == '$' || *nick == '#') {
				s = strrchr(nick, '.');
				if (!s) {
					sendto_one(sptr, err_str(ERR_NOTOPLEVEL), me.name, parv[0], nick);
					continue;
				}
				while (*++s) {
					if (*s == '.' || *s == '*' || *s == '?') {
						break;
					}
				}
				if (*s == '*' || *s == '?') {
					sendto_one(sptr, err_str(ERR_WILDTOPLEVEL), me.name, parv[0], nick);
					continue;
				}
				sendto_match_butone(IsServer(cptr) ? cptr : NULL, sptr, nick + 1,
					*nick == '#' ? MATCH_HOST : MATCH_SERVER, ":%s %s %s :%s", parv[0],
					command, nick, parv[2]);
				continue;
			}
		}

		server = strchr(nick, '@');
		if (!ischan && server) {
			if (*(server + 1) != '\0') {
				acptr = find_server(server + 1, NULL);
				if (acptr) {
					if (!IsMe(acptr)) {
						sendto_one(acptr, ":%s %s %s :%s", parv[0], command, nick,
							parv[2]);
						continue;
					}

					acptr = find_person(nick, NULL);
					if (acptr) {
						sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0],
							command, nick, parv[2]);
					}
				}
			}
			if (acptr) {
				continue;
			}
		}
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
	}
	if (i > 20 && sptr->user) {
		sendto_realops_lev(SPY_LEV, "%s (%s@%s) tried to message %d users", me.name, parv[0],
			sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, i);
	}
	return 0;
}

static int msg_has_color(char *message)
{
	char *s = message;
	while (*s) {
		if (*s == '\003' || *s == '\033') {
			break;
		}
		else {
			s++;
		}
	}
	if (*s) {
		return 1;
	}
	return 0;
}

static char *strip_color(char *message)
{
	static char in[512], out[512];
	int i = 0, j = 0;

	memset((char *)out, '\0', sizeof(out));
	memcpy(in, message, 512);

	while (in[i]) {
		if (in[i] == '\003' || in[i] == '\033') {
			i++;
			while (isdigit(in[i]) || in[i] == ',') {
				i++;
			}
			continue;
		}
		out[j] = in[i];
		i++;
		j++;
	}
	return (out);
}

static int check_for_client_flood(int notice, aClient *sptr, aClient *acptr)
{
	int delta = 0;

	if (!ConfigSet.floodcount || !sptr->user || !acptr->user || IsULine(acptr) || IsServer(acptr)) {
		return 0;
	}

	if (acptr->user->first_message_time + 1 < NOW) {
		delta = NOW - acptr->user->first_message_time;
		acptr->user->number_of_messages -= delta;
		acptr->user->first_message_time = NOW;

		if (acptr->user->number_of_messages <= 0) {
			acptr->user->number_of_messages = 0;
			acptr->user->flood_notice = 0;
		}
	}
	if (acptr->user->number_of_messages >= ConfigSet.floodcount || acptr->user->flood_notice) {
		if (!acptr->user->flood_notice) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s (%s@%s)",
				sptr->name, sptr->user->username,
				IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host,
				acptr->name, acptr->user->username,
				IsMasked(acptr) ? acptr->user->maskedhost : acptr->user->host);
			acptr->user->flood_notice = 1;
			acptr->user->number_of_messages += 2;
		}
		if (MyClient(sptr) && !notice) {
			sendto_one(sptr, ":%s NOTICE %s :Messages to %s throttled due to flooding",
				me.name, sptr->name, acptr->name);
		}
		return 1;
	}
	else {
		acptr->user->number_of_messages++;
	}
	return 0;
}

static int check_for_channel_flood(int notice, aClient *sptr, aChannel *chptr)
{
	int delta = 0;

	if (!ConfigSet.floodcount || !sptr->user || !chptr || IsULine(sptr) || IsServer(sptr)) {
		return 0;
	}

	if (chptr->first_message_time + 1 < NOW) {
		delta = NOW - chptr->first_message_time;
		chptr->number_of_messages -= delta;
		chptr->first_message_time = NOW;

		if (chptr->number_of_messages <= 0) {
			chptr->number_of_messages = 0;
			chptr->flood_notice = 0;
		}
	}
	if (chptr->number_of_messages >= ConfigSet.floodcount || chptr->flood_notice) {
		if (!chptr->flood_notice) {
			sendto_realops_lev(FLOOD_LEV, "Possible flooder: %s (%s@%s) target: %s",
				sptr->name, sptr->user->username,
				IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host,
				chptr->chname);
			chptr->flood_notice = 1;
			chptr->number_of_messages += 2;
		}
		if (MyClient(sptr) && !notice) {
			sendto_one(sptr, ":%s NOTICE %s :Messages to %s throttled due to flooding",
				me.name, sptr->name, chptr->chname);
		}
		return 1;
	}
	else {
		chptr->number_of_messages++;
	}
	return 0;
}

static int check_for_ctcp(char *str, char **dccptr)
{
	if (*str == 1) {
		if (!strncmp(&str[1], "DCC", 3)) {
			if (dccptr) {
				*dccptr = &str[1];
			}
			if (!strncmp(&str[4], " SEND", 5)) {
				return CTCP_DCCSEND;
			}
			return CTCP_DCC;
		}
		if (strncmp(&str[1], "ACTION", 6)) {
			return CTCP_YES;
		}
	}
	return CTCP_NONE;
}
