/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_mode.c: MODE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_mode.c,v 1.11 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_mode(aClient *, aClient *, int, char **);

extern int user_mode(aClient *, aClient *, int, char **);
extern void do_mode(aChannel *, aClient *, aClient *, int, char **, int, int);
extern void bounce_mode(aChannel *, aClient *, int, char **);
extern int check_channelmask(aClient *, aClient *, char *);
extern void clean_channelname(unsigned char *);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_mode.so",
	"/MODE command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("MODE", m_mode, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("MODE", m_mode);
}
char *_module_version = "100000000";
#endif

/*
 * m_mode
 *	parv[0] = sender prefix
 *	parv[1] = channel
 */
int m_mode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	long unsigned sendts;
	aChannel *chptr;
	aBan *ban;

	if (check_registered(sptr)) {
		return 0;
	}

	if (parc > 1) {
		chptr = find_channel(parv[1], NullChn);
		if (chptr == NullChn) {
			return user_mode(cptr, sptr, parc, parv);
		}
	}
	else {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
		return 0;
	}

	if (MyConnect(sptr)) {
		clean_channelname(parv[1]);
	}
	if (check_channelmask(sptr, cptr, parv[1])) {
		return 0;
	}

	if (parc < 3) {
		*mode_buf = *parabuf = '\0';
		mode_buf[1] = '\0';

		channel_modes(sptr, mode_buf, parabuf, chptr);
		sendto_one(sptr, rpl_str(RPL_CHANNELMODEIS), me.name, parv[0], chptr->chname,
			mode_buf, parabuf);
		sendto_one(sptr, rpl_str(RPL_CREATIONTIME), me.name, parv[0], chptr->chname, chptr->channelts);
		return 0;
	}

	sendts = 0;

	if (IsPerson(sptr)) {
		if (parc < 4 && ((*parv[1] == 'b' && parv[2][1] == '\0') || (parv[2][1] == 'b' &&
		  parv[2][2] == '\0' && (*parv[2] == '+' || *parv[2] == '-')))) {
			if (!IsMember(sptr, chptr)) {
				return 0;
			}
			for (ban = chptr->banlist; ban; ban = ban->next) {
				sendto_one(sptr, rpl_str(RPL_BANLIST), me.name, parv[0], chptr->chname,
					ban->banstr, ban->who, ban->when);
			}
			sendto_one(sptr, rpl_str(RPL_ENDOFBANLIST), me.name, parv[0], chptr->chname);
			return 0;
		}
		if (parc < 4 && ((*parv[1] == 'e' && parv[2][1] == '\0') || (parv[2][1] == 'e' &&
		  parv[2][2] == '\0' && (*parv[2] == '+' || *parv[2] == '-')))) {
			if (!IsMember(sptr, chptr)) {
				return 0;
			}
			for (ban = chptr->exbanlist; ban; ban = ban->next) {
				sendto_one(sptr, rpl_str(RPL_EXBANLIST), me.name, parv[0], chptr->chname,
					ban->banstr, ban->who, ban->when);
			}
			sendto_one(sptr, rpl_str(RPL_ENDOFEXBANLIST), me.name, parv[0], chptr->chname);
			return 0;
		}
		if (!IsULine(sptr) && !is_chan_op(sptr, chptr) && !is_half_op(sptr, chptr) &&
		  (cptr == sptr || !IsSAdmin(sptr))) {
			if (cptr == sptr) {
				sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0], chptr->chname);
				return 0;
			}
			if (IsMember(sptr, chptr)) {
				sendto_one(cptr, ":%s MODE %s -oh %s 0", me.name, chptr->chname, parv[0]);
			}
			else{
				sendto_one(cptr, "KICK %s %s :desynched user", chptr->chname, parv[0]);
			}
			bounce_mode(chptr, cptr, parc - 2, parv + 2);
			return 0;
		}
	}
	else if (IsServer(sptr)) {
		if ((sendts = atoi(parv[parc - 1])) && !IsULine(sptr) && chptr->channelts
		  && (sendts > chptr->channelts)) {
			if (!*parv[2] == '&') {
				bounce_mode(chptr, cptr, parc - 2, parv + 2);
				return 0;
			}
			sendto_ops_lev(DEBUG_LEV, "Double bounce for %s - %lu(ours) %lu(theirs)",
				chptr->chname, chptr->channelts, sendts);
		}
		if (!sendts && *parv[parc - 1] != '0') {
			sendts = -1;
		}
		if (sendts != -1) {
			parc--;	
		}
	}

	do_mode(chptr, cptr, sptr, parc - 2 , parv + 2, sendts, 0);
	return 0;
}
