/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_nick.c: NICK command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_nick.c,v 1.14 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_nick(aClient *, aClient *, int, char **);
static int do_nick_name(char *);

extern int do_user(char *, aClient *, aClient *, char *, char *, char *, unsigned long, unsigned int, char *);
extern int register_user(aClient *, aClient *, char *, char *);
extern int remove_accept_references(aClient *, int);
extern int user_modes[];

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_nick.so",
	"/NICK command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("NICK", m_nick, 0, 15, 1, 1, 0, 0L);
}
void _module_unload(void) {
	del_Command("NICK", m_nick);
}
#endif

/*
 * m_nick
 *	parv[0] = sender prefix
 *	parv[1] = nickname
 *	parv[2] = hopcount when new user; TS when nick change
 *	parv[3] = TS
 *      --- new user only below ---
 *	parv[4] = usermodes
 *	parv[5] = username
 *	parv[6] = hostname
 *	parv[7] = server
 *	parv[8] = services id
 *	if link is NICKIP:
 *		parv[9] = client ip
 *		parv[10] = realname
 *      else
 *		parv[9] = realname
 */
int m_nick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient *acptr, *uplink;
	Link *lp;
	char nick[NICKLEN + 2];
	ts_val newts = 0;
	int sameuser = 0, samenick = 0;

	if (parc < 2) {
		sendto_one(sptr, err_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
		return 0;
	}

	if (!IsServer(sptr) && IsServer(cptr) && parc > 2) {
		newts = atol(parv[2]);
	}
	else if (IsServer(sptr) && parc > 3) {
		newts = atol(parv[3]);
	}
	else {
		parc = 2;
	}

	if ((IsServer(sptr) || parc > 4) && ((IsNICKIP(sptr) && parc < 11) || (!IsNICKIP(sptr) && parc < 10))) {
		sendto_realops_lev(DEBUG_LEV, "IGNORING BAD NICK: %s[%s@%s] on %s (from %s)",
			parv[1], parc >= 6 ? parv[5] : "-", parc >= 7 ? parv[6] : "-",
			parc >= 8 ? parv[7] : "-", parv[0]);
		return 0;
	}

	strncpyzt(nick, parv[1], NICKLEN + 1);

	if (!do_nick_name(nick) || (IsServer(cptr) && strcmp(nick, parv[1]))) {
		sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name, parv[0], parv[1]);

		if (IsServer(cptr)) {
			ircstp->is_kill++;
			sendto_realops_lev(DEBUG_LEV, "Bad Nick: %s From: %s %s", parv[1], parv[0],
				get_client_name(cptr, FALSE));
			sendto_one(cptr, ":%s KILL %s :%s (%s <- %s[%s])", me.name, parv[0],
				me.name, parv[1], nick, cptr->name);

			if (sptr != cptr) {
				sendto_serv_butone(cptr, ":%s KILL %s :%s (%s <- %s!%s@%s)", me.name,
					parv[0], me.name, get_client_name(cptr, FALSE), parv[0],
					sptr->user ? sptr->username : "",
					sptr->user ? sptr->user->server : cptr->name);
				sptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, sptr, &me, "BadNick");
			}

		}
		return 0;
	}

	do {
		acptr = find_server(nick, NULL);
		if (acptr) {
			if (MyConnect(sptr)) {
				sendto_one(sptr, err_str(ERR_NICKNAMEINUSE), me.name,
					BadPtr(parv[1]) ? "*" : parv[0], nick);
				return 0;
			}

			sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)", sptr->name,
				acptr->from->name, get_client_name(cptr, FALSE));
			ircstp->is_kill++;
			sendto_one(cptr, ":%s KILL %s :%s (%s <- %s)", me.name, sptr->name, me.name,
				acptr->from->name, get_client_name(cptr, FALSE));
			sptr->flags |= FLAGS_KILLED;
			return exit_client(cptr, sptr, &me, "Nick/Server collision");
		}

		acptr = find_client(nick, NULL);
		if (!acptr) {
			break;
		}

		if (acptr == sptr) {
			if (strcmp(acptr->name, nick)) {
				break;
			}
			else {
				return 0;
			}
		}

		if (IsUnknown(acptr)) {
			if (MyConnect(acptr)) {
				exit_client(NULL, acptr, &me, "Overridden");
				break;
			}
			else {
				sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s(NOUSER) <- "
					"%s!%s@%s)(TS:%s)", acptr->name, acptr->from->name, parv[1],
					parv[5], parv[6], cptr->name);
				sendto_serv_butone(NULL, ":%s KILL %s :%s (%s(NOUSER) <- %s!%s@%s(TS:%s)",
					me.name, acptr->name, me.name, acptr->from->name, parv[1],
					parv[5], parv[6], cptr->name);
				acptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, acptr, &me, "Got TS NICK before NON-TS USER");
			}
		}

		if (!IsServer(cptr)) {
			sendto_one(sptr, err_str(ERR_NICKNAMEINUSE), me.name,
				BadPtr(parv[0]) ? "*" : parv[0], nick);
			return 0;
		}

		if (IsServer(sptr)) {
			if (!newts || !acptr->tsinfo || newts == acptr->tsinfo) {
				sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(both killed)",
					acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
				ircstp->is_kill++;
				sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
					acptr->name);
				sendto_serv_butone(NULL, ":%s KILL :%s (%s <- %s)", me.name,
					acptr->name, me.name, acptr->from->name,
					get_client_name(cptr, FALSE));
				acptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, acptr, &me, "Nick collision");
			}
			else {
				sameuser = (acptr->user) && mycmp(acptr->user->username, parv[5]) == 0
				  && mycmp(acptr->user->host, parv[6]) == 0;
				if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo)) {
					return 0;
				}
				else {
					sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(%s killed)",
						acptr->name, acptr->from->name,	get_client_name(cptr, FALSE),
						sameuser ? "older" : "newer");
					ircstp->is_kill++;
					sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
						acptr->name);
					sendto_serv_butone(sptr, ":%s KILL %s :%s (%s <- %s)", me.name,
						acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
					acptr->flags |= FLAGS_KILLED;
					exit_client(cptr, acptr, &me, "Nick collision");
					break;
				}
			}
		}

		if (!newts || !acptr->tsinfo || newts == acptr->tsinfo || !sptr->user) {
			sendto_ops_lev(SKILL_LEV, "Nick change collision from %s to %s(%s <- %s)"
				"(both killed)", sptr->name, acptr->name, acptr->from->name,
				get_client_name(cptr, FALSE));
			ircstp->is_kill++;
			sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name, acptr->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s(%s) <- %s)", me.name, sptr->name,
				me.name, acptr->from->name, acptr->name, get_client_name(cptr, FALSE));
			ircstp->is_kill++;
			sendto_one(sptr, err_str(ERR_NICKCOLLISION), me.name, sptr->name, sptr->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s <- %s(%s))", me.name, acptr->name,
				me.name, acptr->from->name, get_client_name(cptr, FALSE), sptr->name);
			acptr->flags |= FLAGS_KILLED;
			exit_client(NULL, acptr, &me, "Nick collision(new)");
			sptr->flags |= FLAGS_KILLED;
			return exit_client(cptr, sptr, &me, "Nick collision(old)");
		}
		else {
			sameuser = mycmp(acptr->user->username, sptr->user->username) == 0 &&
			  mycmp(acptr->user->host, sptr->user->host) == 0;
			if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo)) {
				sendto_ops_lev(SKILL_LEV, "Nick change collision from %s to %s(%s <- %s)"
					"(%s killed)", sptr->name, acptr->name, acptr->from->name,
					get_client_name(cptr, FALSE), sameuser ? "older" : "newer");
				ircstp->is_kill++;
				sendto_serv_butone(cptr, ":%s KILL %s :%s (%s(%s) <- %s)", me.name,
					sptr->name, me.name, acptr->from->name, acptr->name,
					get_client_name(cptr, FALSE));
				sptr->flags |= FLAGS_KILLED;
				return exit_client(cptr, sptr, &me,
					sameuser ? "Nick collision(old)" : "Nick collision (new)");
			}
			else {
				sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(%s killed)",
					acptr->name, acptr->from->name, get_client_name(cptr, FALSE),
					sameuser ? "older" : "newer");
				ircstp->is_kill++;
				sendto_one(acptr, err_str(ERR_NICKCOLLISION), me.name, acptr->name,
					acptr->name);
				sendto_serv_butone(sptr, ":%s KILL %s :%s (%s <- %s)", me.name,
					acptr->name, me.name, acptr->from->name,
					get_client_name(cptr, FALSE));
				acptr->flags |= FLAGS_KILLED;
				exit_client(cptr, acptr, &me, "Nick collision");
			}
		}
	} while (0);

	if (IsServer(sptr)) {
		uplink = find_server(parv[7], NULL);
		if (!uplink) {
			sendto_realops("Remote nick %s on UNKNOWN server %s", nick, parv[7]);
			return 0;
		}

		sptr = make_client(cptr, uplink);

		if (IsULine(uplink)) {
			sptr->flags |= FLAGS_ULINE;
		}

		add_client_to_list(sptr);
		if (parc > 2) {
			sptr->hopcount = atoi(parv[2]);
		}
		if (newts) {
			sptr->tsinfo = newts;
		}
		else {
			newts = sptr->tsinfo = (ts_val)timeofday;
			ts_warn("Remote nick %s introduced without a TS", nick);
		}

		strcpy(sptr->name, nick);
		add_to_client_hash_table(nick, sptr);

		if (parc >= 10) {
			int *s, flag;
			char *m;

			if (parc == 11 && !IsNICKIP(cptr)) {
				sendto_ops_lev(DEBUG_LEV, "Got extra parameters from non-NICKIP server %s",
					cptr->name);
			}

			m = &parv[4][1];
			while (*m) {
				for (s = user_modes; (flag = *s); s += 2) {
					if (*m == (char)(*(s + 1))) {
						if (flag == UMODE_o) {
							Count.oper++;
						}
						sptr->umode |= flag & SEND_UMODES;
						break;
					}
				}
				m++;
			}
			if (parc == 10) {
				return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7],
					strtoul(parv[8], NULL, 0), 0, parv[9]);
			}
			else if (parc == 11) {
				return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7],
					strtoul(parv[8], NULL, 0), strtoul(parv[9], NULL, 0),
					parv[10]);
			}
		}
	}
	else if (sptr->name[0])
	{
		ConfigItem_ban *ban = find_ban(BAN_NICKNAME, nick);
		if (ban) {
			if (!MyConnect(sptr)) {
				sendto_realops("Q:Lined nick %s from %s on %s", nick,
					*sptr->name && IsServer(sptr) ? sptr->name : "<unregistered>",
					!sptr->user ? IsServer(sptr) ? parv[6] : me.name : sptr->user->server);
			}
			if (MyConnect(sptr) && !IsServer(cptr) && !IsOper(cptr) && !IsULine(cptr)) {
				sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name,
					BadPtr(parv[0]) ? "*" : parv[0], nick, ban->reason, ban->mask);
				sendto_realops("Forbidding Q:Lined nickname %s from %s.", nick,
					get_client_name(cptr, FALSE));
				return 0;
			}
		}

		if (mycmp(parv[0], nick)) {
			sptr->tsinfo = newts ? newts : (ts_val)timeofday;
			sptr->umode &= ~UMODE_r;
		}

		if (MyConnect(sptr)) {
			if (IsRegisteredUser(sptr)) {
				for (lp = sptr->user->channel; lp; lp = lp->next) {
					if (can_send(sptr, lp->value.chptr)) {
						sendto_one(sptr, err_str(ERR_BANNICKCHANGE), me.name,
							sptr->name, lp->value.chptr->chname);
						return 0;
					}
					if (is_banned(sptr, lp->value.chptr)) {
						sendto_one(sptr, err_str(ERR_BANONCHAN), me.name,
							sptr->name, nick, lp->value.chptr->chname);
						return 0;
					}
					if (lp->value.chptr->mode.mode & MODE_NONICKCH && !IsULine(cptr) &&
					  !is_chan_admin(cptr, lp->value.chptr)) {
						sendto_one(sptr, ":%s NOTICE %s :*** Cannot change nickname: channel "
							"%s is set +N", me.name, sptr->name, lp->value.chptr->chname);
						return 0;
					}
				}

#ifdef ANTI_NICK_FLOOD
				if (sptr->last_nick_change + MAX_NICK_TIME < NOW) {
					sptr->number_of_nick_changes = 0;
				}
				sptr->last_nick_change = NOW;
				sptr->number_of_nick_changes++;

				if (sptr->number_of_nick_changes <= MAX_NICK_CHANGES || IsOper(sptr)) {
#endif
					sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
					if (sptr->user) {
						add_history(sptr, 1);
						sendto_serv_butone(cptr, ":%s NICK %s :%ld", parv[0],
							nick, sptr->tsinfo);
					}
#ifdef ANTI_NICK_FLOOD
				}
				else {
					sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Too many nickname "
						"changes. Wait %d seconds before trying again.", me.name,
						sptr->name, MAX_NICK_TIME);
					return 0;
				}
#endif
			}
		}
		else {
			sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
			if (sptr->user) {
				add_history(sptr, 1);
				sendto_serv_butone(cptr, ":%s NICK %s :%ld", parv[0], nick, sptr->tsinfo);
			}
		}
	}
	else {
		if (MyConnect(sptr)) {
			ConfigItem_ban *ban = find_ban(BAN_NICKNAME, nick);
			if (ban) {
				if (MyConnect(sptr) && !IsServer(sptr) && !IsOper(sptr) && !IsULine(cptr)) {
					sendto_one(sptr, err_str(ERR_ERRONEUSNICKNAME), me.name,
						BadPtr(parv[0]) ? "*" : parv[0], nick, ban->reason, ban->mask);
					sendto_realops("Forbidding Q:Lined nick %s from <unregistered>%s.",
						nick, get_client_name(cptr, FALSE));
					return 0;
				}
			}
		}

		strcpy(sptr->name, nick);
		sptr->tsinfo = timeofday;
		if (sptr->user) {
			if (register_user(cptr, sptr, nick, sptr->user->username) == FLUSH_BUFFER) {
				return FLUSH_BUFFER;
			}
		}
	}

	if (sptr->name[0]) {
		del_from_client_hash_table(sptr->name, sptr);
		samenick = mycmp(sptr->name, nick) ? 0 : 1;
		if (IsPerson(sptr) && !samenick) {
			hash_check_watch(sptr, RPL_LOGOFF);
			remove_accept_references(sptr, 1);
		}
	}

	strcpy(sptr->name, nick);
	add_to_client_hash_table(nick, sptr);

	if (IsPerson(sptr) && !samenick) {
		hash_check_watch(sptr, RPL_LOGON);
	}
	return 0;
}

static int do_nick_name(char *nick)
{
	char *ch;

	if (*nick == '-' || isdigit(*nick)) {
		return 0;
	}

	for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++) {
		if (!isvalid(*ch) || isspace(*ch)) {
			break;
		}
	}
	*ch = '\0';
	return (ch - nick);
}
