/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_oper.c: OPER command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_oper.c,v 1.20 2002/04/08 16:00:47 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_oper(aClient *, aClient *, int, char **);

extern void do_maskedhost(aClient *);

struct error_message {
	int numeric;
	char *message;
};
static struct error_message errors[0] = {
	{ ERR_NOOPERHOST, "UNKNOWN ERROR" },
	{ ERR_NOOPERHOST, "NO CONFIGURATION" },
	{ ERR_NOOPERHOST, "USERHOST MISMATCH" },
	{ ERR_PASSWDMISMATCH, "PASSWORD MISMATCH" }
};

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_oper.so",
	"/OPER command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("OPER", m_oper, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("OPER", m_oper);
}
#endif

/*
 * m_oper
 *	parv[0] = sender prefix
 *	parv[1] = login
 *	parv[2] = password
 */
int m_oper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	ConfigItem_oper *oper = NULL;
	char from[USERLEN + HOSTLEN + 2], *encr;
	extern char *crypt();
	int old = 0, send_error = 0;
	struct error_message *err = NULL;

	if (!IsServer(cptr) && (BadPtr(parv[1]) || BadPtr(parv[2]))) {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "OPER");
		return 0;
	}
	if ((IsServer(cptr) || IsMe(cptr)) && !IsOper(sptr)) {
		sptr->umode |= (UMODE_o|UMODE_h);
		sendto_serv_butone(cptr, ":%s MODE %s :+oh", parv[0], parv[0]);
		Count.oper++;
		if (IsMe(cptr)) {
			sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
		}
		return 0;
	}
	else if (IsOper(sptr)) {
		if (MyClient(sptr)) {
			sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
		}
		return 0;
	}

	ircsprintf(from, "%s@%s", sptr->user->username, sptr->user->host);
	for (oper = config_oper; oper; oper = (ConfigItem_oper *)oper->next) {
		if (strcasecmp(oper->name, parv[1])) {
			send_error = 1;
			continue;
		}
		if (match(oper->from, from) && match(from, oper->from)) {
			send_error = 2;
			continue;
		}
		if (ConfigSet.encryptedpasswords) {
			encr = (parv[2] && *oper->password) ? crypt(parv[2], oper->password) : "";
		}
		else {
			encr = parv[2];
		}
		if (strcmp(oper->password, encr)) {
			send_error = 3;
			continue;
		}
		break;
	}
	if (!oper) {
		err = &errors[send_error] ? &errors[send_error] : &errors[0];
		ircdlog(LOG_OPER, "failed oper attempt by %s [%s]", get_client_name(cptr, FALSE),
			err->message);
		sendto_one(sptr, err_str(err->numeric), me.name, parv[0]);
		if (ConfigSet.failopernotice) {
			sendto_ops_lev(SPY_LEV, "Failed OPER attempt by %s (%s@%s) [%s]",
				sptr->name, sptr->user->username,
				IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host,
				err->message);
		}
		return 0;
	}

	if (sptr->class && sptr->class->clients > 0) {
		sptr->class->clients--;
	}
	sptr->class = oper->class;
	sptr->class->clients++;
	sptr->pingval = sptr->class->pingfreq;
	old = (sptr->umode & ALL_UMODES);

	if (oper->flags & OFLAG_SADMIN) {
		sptr->umode |= UMODE_a;
	}
	if (oper->flags & OFLAG_NETADMIN) {
		sptr->umode |= UMODE_N;
	}
	if (oper->flags & OFLAG_ADMIN) {
		sptr->umode |= UMODE_A;
	}
	if (oper->flags & OFLAG_COADMIN) {
		sptr->umode |= UMODE_C;
	}
	sptr->umode |= (UMODE_o|UMODE_s|UMODE_g|UMODE_w|UMODE_n|UMODE_h);
	sptr->oflag = oper->flags;
	Count.oper++;

	send_umode_out(cptr, sptr, old);
	addto_fdlist(sptr->fd, &oper_fdlist);

	sendto_one(sptr, rpl_str(RPL_YOUREOPER), me.name, parv[0]);
	if (IsNetAdmin(sptr)) {
		sendto_ops("%s (%s@%s) is now a Network Administrator [%s]", sptr->name,
			sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
		sendto_serv_butone(&me, ":%s GLOBOPS :%s (%s@%s) is now a Network Administrator [%s]",
			me.name, sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
	}
	else if (IsAdmin(sptr)) {
		sendto_ops("%s (%s@%s) is now a Server Administrator [%s]", sptr->name,
			sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
		sendto_serv_butone(&me, ":%s GLOBOPS :%s (%s@%s) is now a Server Administrator [%s]",
			me.name, sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
	}
	else if (IsCoAdmin(sptr)) {
		sendto_ops("%s (%s@%s) is now a Server Co-Administrator [%s]", sptr->name,
			sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
		sendto_serv_butone(&me, ":%s GLOBOPS :%s (%s@%s) is now a Server Co-Administrator [%s]",
			me.name, sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
	}
	else {
		sendto_ops("%s (%s@%s) is now an IRC Operator [%s]", sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
		sendto_serv_butone(&me, ":%s GLOBOPS :%s (%s@%s) is now an IRC Operator [%s]", me.name,
			sptr->name, sptr->user->username,
			IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host, oper->name);
	}
	ircdlog(LOG_OPER, "%s successfully opered as %s", get_client_name(sptr, FALSE), oper->name);
	do_maskedhost(sptr);
	runHook1(HOOKTYPE_OPER, sptr);
	return 0;
}
