/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_sjoin.c: SJOIN protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_sjoin.c,v 1.23 2002/04/08 17:05:56 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_sjoin(aClient *, aClient *, int, char **);

typedef struct parv {
	int parc;
	char *parv[256];
} aParv;

extern void set_mode(aChannel *, aClient *, int, char **, u_int *, char [MAXMODEPARAMS][MODEBUFLEN + 3], int);
extern void add_user_to_channel(aChannel *, aClient *, int);
extern aChannel *get_channel(aClient *, char *, int);
extern aCtab cFlagTab[];

static aParv *mp2parv(char *, char *);

static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];
static aParv pparv;

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_sjoin.so",
	"SJOIN protocol",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("SJOIN", m_sjoin, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("SJOIN", m_sjoin);
}
#endif

/*
 * m_sjoin
 *	parv[0] = sender prefix
 *	parv[1] = timestamp
 *	parv[2] = channel name
 *	if (parc == 3) - cliSJOIN
 *	if (parc >= 6) - SJOIN
 *		parv[3] = modes
 *		parv[4 to parc - 2] = mode parameters
 *		parv[parc - 1] = user list with statuses and channel bans
 */
int m_sjoin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char nick[NICKLEN + 1], pvar[MAXMODEPARAMS][MODEBUFLEN + 3], userbuf[BUFSIZE];
	char parabuf2[1024], *s, *p, *ns;
	aChannel *chptr;
	chanMember *cm;
	aClient *acptr;
	aParv *ap;
	aCtab *tab;
	aBan *ban;
	Mode oldmode;
	int nomodes = 0, noparams = 0, wipeours = 0, wipetheirs = 0, copy = 0, isnew = 0;
	int merge = 0, pcount, i = 0, j = 0, fl = 0, keyerror = 0, x, y, z, clisjoin = 0;
	ts_val ts = 0, tstosend = 0;

	if (MyClient(sptr) || (parc < 6 && IsServer(sptr)) || (parc < 3 && IsPerson(sptr))) {
		return 0;
	}
	if (!IsChannelName(parv[2])) {
		return 0;
	}

	if (parc == 3) {
		clisjoin = 1;
	}
	else {
		if (parv[3][1] == '\0') {
			nomodes = 1;
		}
		if (!strncmp(parv[4], "<none>", 6)) {
			noparams = 1;
		}
	}

	isnew = ChannelExists(parv[2]) ? 0 : 1;
	chptr = get_channel(cptr, parv[2], CREATE);
	ts = atol(parv[1]);

	if (clisjoin) {
		if (isnew) {
			chptr->channelts = tstosend = ts;
		}
		else if (!ts || !chptr->channelts) {
			chptr->channelts = tstosend = 0;
		}
		else if (ts == chptr->channelts) {
			tstosend = chptr->channelts;
		}
		else if (ts < chptr->channelts) {
			chptr->channelts = tstosend = ts;
		}
		else {
			tstosend = chptr->channelts;
		}

		if (!IsMember(sptr, chptr)) {
			add_user_to_channel(chptr, sptr, 0);
			sendto_channel_butserv(chptr, sptr, ":%s JOIN :%s", parv[0], parv[2]);
		}
		sendto_ssjoin_servs_butone(chptr, sptr, ":%s SJOIN %ld :%s", parv[0], tstosend, parv[2]);
		return 0;
	}

	if (chptr->channelts > ts) {
		wipeours = copy = 1;
		sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :*** Notice -- TS for %s changed "
			"from %lu to %lu", me.name, chptr->chname, chptr->chname,
			chptr->channelts, ts);
		chptr->channelts = ts;
	}
	else if (chptr->channelts < ts && chptr->channelts != 0) {
		wipetheirs = 1;
	}
	else if (chptr->channelts == ts) {
		merge = 1;
	}
	if (!chptr->channelts) {
		chptr->channelts = ts;
	}

	modebuf[0] = '\0';
	parabuf[0] = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (wipeours) {
		modebuf[0] = '-';

		if (modebuf[1] != '\0') {
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
			sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
				chptr->chname, modebuf, parabuf);
		}

		for (cm = chptr->members; cm; cm = cm->next) {
			modebuf[0] = '-';
			modebuf[1] = '\0';
			parabuf[0] = '\0';
			j = 0;

			if (cm->flags & CHFL_CHANADMIN) {
				cm->flags &= ~CHFL_CHANADMIN;
				modebuf[++j] = 'a';
				strcat(parabuf, cm->cptr->name);
				strcat(parabuf, " ");
			}
			if (cm->flags & CHFL_CHANOP) {
				cm->flags &= ~CHFL_CHANOP;
				modebuf[++j] = 'o';
				strcat(parabuf, cm->cptr->name);
				strcat(parabuf, " ");
			}
			if (cm->flags & CHFL_HALFOP) {
				cm->flags &= ~CHFL_HALFOP;
				modebuf[++j] = 'h';
				strcat(parabuf, cm->cptr->name);
				strcat(parabuf, " ");
			}
			if (cm->flags & CHFL_VOICE) {
				cm->flags &= ~CHFL_VOICE;
				modebuf[++j] = 'v';
				strcat(parabuf, cm->cptr->name);
				strcat(parabuf, " ");
			}

			if (j > 0) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}

			if (chptr->topic[0] != '\0') {
				chptr->topic[0] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", me.name,
					chptr->chname, chptr->topic);
			}
		}

		for (ban = chptr->banlist; ban; ban = ban->next) {
			if (!j) {
				modebuf[0] = '-';
				modebuf[1] = '\0';
				parabuf[0] = '\0';
			}

			strcpy(userbuf, ban->banstr);
			if (del_ban_fromlist(chptr, ban->banstr, (aBan **)&chptr->banlist)) {
				modebuf[++j] = 'b';
				strcat(parabuf, userbuf);
				strcat(parabuf, " ");
			}

			if (j >= 6) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}
		}

		for (ban = chptr->exbanlist; ban; ban = ban->next) {
			if (!j) {
				modebuf[0] = '-';
				modebuf[1] = '\0';
				parabuf[0] = '\0';
			}

			strcat(userbuf, ban->banstr);
			if (del_ban_fromlist(chptr, ban->banstr, (aBan **)&chptr->exbanlist)) {
				modebuf[++j] = 'e';
				strcat(parabuf, userbuf);
				strcat(parabuf, " ");
			}

			if (j >= 6) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}
		}

		if (j > 0) {
			modebuf[++j] = '\0';
			sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
				chptr->chname, modebuf, parabuf);
			j = 0;
		}
	}

	modebuf[0] = '+';
	modebuf[1] = '\0';
	parabuf[0] = '\0';
	j = 0;

	strcpy(userbuf, parv[parc - 1]);
	for (s = strtoken(&p, userbuf, " "); s; s = strtoken(&p, NULL, " ")) {
		ns = s;
		fl = i = 0;

		while (*ns == '*' || *ns == '@' || *ns == '%' || *ns == '+' || *ns == '&' || *ns == '"') {
			switch (*(ns++)) {
				case '*':
					fl |= CHFL_CHANADMIN;
					break;
				case '@':
					fl |= CHFL_CHANOP;
					break;
				case '%':
					fl |= CHFL_HALFOP;
					break;
				case '+':
					fl |= CHFL_VOICE;
					break;
				case '&':
					fl |= MODE_BAN;
					goto getnick;
					break;
				case '"':
					fl |= MODE_EXBAN;
					goto getnick;
					break;
			}
		}

			getnick:
		while (*ns != '\0' && *ns != ' ') {
			nick[i++] = *(ns++);
		}
		nick[i] = '\0';
		if (*nick == '\0' || *nick == ' ') {
			continue;
		}

		if (!(fl & MODE_BAN) && !(fl & MODE_EXBAN)) {
			acptr = find_client(nick, NULL);
			if (!acptr) {
				sendto_realops("WARNING! Possible desynch: missing user %s in "
					"SJOIN for %s", nick, chptr->chname);
				continue;
			}
			if (acptr->from != sptr->from) {
				if (IsMember(acptr, chptr)) {
					continue;
				}
				sendto_realops("WARNING! Possible desynch: fake direction from user "
					"%s at %s in SJOIN for %s", acptr->name, acptr->user->server,
					chptr->chname);
				sendto_one(sptr, ":%s KICK %s %s :(SJOIN) fake direction", me.name,
					chptr->chname, acptr->name);
				continue;
			}
			if (wipetheirs) {
				fl = 0;
			}
			if (!IsMember(acptr, chptr)) {
				add_user_to_channel(chptr, acptr, fl);
				sendto_channel_butserv(chptr, acptr, ":%s JOIN :%s", acptr->name,
					chptr->chname);
			}

			cm = find_user_member(chptr->members, acptr);
			if (!cm) {
				sendto_realops("WARNING! Possible desynch: missing chamMember link for user "
					"%s in SJOIN for %s", acptr->name, chptr->chname);
				continue;
			}

			modebuf[0] = '+';
			modebuf[1] = '\0';
			parabuf[0] = '\0';
			j = 0;

			if (fl & CHFL_CHANADMIN) {
				cm->flags |= CHFL_CHANADMIN;
				modebuf[++j] = 'a';
				strcat(parabuf, acptr->name);
				strcat(parabuf, " ");
			}
			if (fl & CHFL_CHANOP) {
				cm->flags |= CHFL_CHANOP;
				modebuf[++j] = 'o';
				strcat(parabuf, acptr->name);
				strcat(parabuf, " ");
			}
			if (fl & CHFL_HALFOP) {
				cm->flags |= CHFL_HALFOP;
				modebuf[++j] = 'h';
				strcat(parabuf, acptr->name);
				strcat(parabuf, " ");
			}
			if (fl & CHFL_VOICE) {
				cm->flags |= CHFL_VOICE;
				modebuf[++j] = 'v';
				strcat(parabuf, acptr->name);
				strcat(parabuf, " ");
			}

			if (j > 0) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}
		}
		else {
			if (wipetheirs) {
				continue;
			}
			if (!j) {
				modebuf[0] = '+';
				modebuf[1] = '\0';
				parabuf[0] = '\0';
			}

			if (fl & MODE_BAN) {
				if (add_ban_tolist(sptr, chptr, nick, (aBan **)&chptr->banlist)) {
					modebuf[++j] = 'b';
					strcat(parabuf, nick);
					strcat(parabuf, " ");
				}
			}

			if (fl & MODE_EXBAN) {
				if (add_ban_tolist(sptr, chptr, nick, (aBan **)&chptr->exbanlist)) {
					modebuf[++j] = 'e';
					strcat(parabuf, nick);
					strcat(parabuf, " ");
				}
			}

			if (j >= 6) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}
		}
	}

	if (j > 0) {
		modebuf[++j] = '\0';
		sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name, chptr->chname,
			modebuf, parabuf);
		j = 0;
	}

	modebuf[0] = '\0';
	parabuf[0] = '\0';
	parabuf2[0] = '\0';
	if (!merge && !wipetheirs && !nomodes) {
		strcat(modebuf, parv[3]);

		if (!noparams) {
			for (j = 4; j <= (parc - 2); j++) {
				strcat(parabuf, parv[j]);
				strcat(parabuf, " ");
			}
		}
		strcpy(parabuf2, parabuf);

		ap = mp2parv(modebuf, parabuf);
		set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
		sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name, chptr->chname,
			modebuf, parabuf2);
	}

	modebuf[0] = '\0';
	parabuf[0] = '\0';
	parabuf2[0] = '\0';
	j = 0;

	if (merge && !nomodes) {
		strcpy(modebuf, parv[3]);
		if (noparams) {
			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
			sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
				chptr->chname, modebuf, parabuf);
		}
		else {
			memcpy(&chptr->mode, &oldmode, sizeof(Mode));
			for (j = 4; j <= (parc - 2); j++) {
				strcat(parabuf, parv[j]);
				strcat(parabuf, " ");
			}

			ap = mp2parv(modebuf, parabuf);
			set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
			modebuf[0] = '-';
			modebuf[1] = '\0';
			parabuf[0] = '\0';
			j = 0;

			if (oldmode.limit && !chptr->mode.limit) {
				modebuf[++j] = 'l';
				strcat(parabuf, (char *)my_itoa(oldmode.limit));
				strcat(parabuf, " ");
			}
			if (*oldmode.key && !*chptr->mode.key) {
				modebuf[++j] = 'k';
				strcat(parabuf, oldmode.key);
				strcat(parabuf, " ");
			}

			for (tab = cFlagTab; tab->mode; tab++) {
				if ((oldmode.mode & tab->mode) && !(chptr->mode.mode & tab->mode)
				  && !tab->param) {
					modebuf[++j] = tab->flag;
				}
			}

			if (j > 0) {
				modebuf[++j] = '+';
			}
			else {
				modebuf[0] = '+';
				modebuf[1] = '\0';
				j = 0;
			}

			if (!oldmode.limit && chptr->mode.limit) {
				modebuf[++j] = 'l';
				strcat(parabuf, (char *)my_itoa(chptr->mode.limit));
				strcat(parabuf, " ");
			}
			if (!*oldmode.key && *chptr->mode.key) {
				modebuf[++j] = 'k';
				strcat(parabuf, chptr->mode.key);
				strcat(parabuf, " ");
			}

			if (oldmode.limit && chptr->mode.limit && oldmode.limit != chptr->mode.limit) {
				chptr->mode.limit = oldmode.limit > chptr->mode.limit ? oldmode.limit : chptr->mode.limit;
				if (oldmode.limit != chptr->mode.limit) {
					modebuf[++j] = 'l';
					strcat(parabuf, (char *)my_itoa(chptr->mode.limit));
					strcat(parabuf, " ");
				}
			}
			if (*oldmode.key && *chptr->mode.key && strcmp(oldmode.key, chptr->mode.key)) {
				x = strlen(oldmode.key);
				y = strlen(chptr->mode.key);
				z = x > y ? x : y;

				if (x == y) {
					sendto_channel_butserv(chptr, sptr, ":%s NOTICE %s :WARNING! "
						"Unable to determine which key (+k) to use in SJOIN "
						"for %s, removing all keys.", me.name, chptr->chname,
						chptr->chname);
					keyerror = 1;
				}
				else {
					if (z == x) {
						strcpy(chptr->mode.key, oldmode.key);
					}
					else {
						modebuf[++j] = 'k';
						strcat(parabuf, chptr->mode.key);
						strcat(parabuf, " ");
					}
				}
			}

			if (j > 0) {
				modebuf[++j] = '\0';
				sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", me.name,
					chptr->chname, modebuf, parabuf);
				j = 0;
			}

			if (keyerror) {
				modebuf[0] = '-';
				modebuf[1] = '\0';
				parabuf[0] = '\0';
				j = 0;

				if (keyerror) {
					modebuf[++j] = 'k';
					strcat(parabuf, chptr->mode.key);
					strcat(parabuf, " ");
				}

				if (j > 0) {
					modebuf[++j] = '\0';
					ap = mp2parv(modebuf, parabuf);
					set_mode(chptr, cptr, ap->parc, ap->parv, &pcount, pvar, 0);
					sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
						me.name, chptr->chname, modebuf, parabuf);
					j = 0;
				}
			}
		}
	}

	if (j > 0) {
		modebuf[++j] = '\0';
		sendto_channel_butserv(chptr, cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf, parabuf);
		j = 0;
	}

	parabuf[0] = '\0';
	if (!noparams) {
		for (j = 4; j <= (parc - 2); j++) {
			strcat(parabuf, parv[j]);
			strcat(parabuf, " ");
		}
	}
	sendto_ssjoin_servs_butone(chptr, cptr, ":%s SJOIN %ld %s %s %s :%s", sptr->name, ts, chptr->chname,
		parv[3], parabuf[0] == '\0' ? "<none>" : parabuf, parv[parc - 1]);
	return 0;
}

static aParv *mp2parv(char *mbuf, char *pbuf)
{
	char *p, *s;
	int c = 1;

	pparv.parv[0] = mbuf;
	for (s = strtoken(&p, pbuf, " "); s; s = strtoken(&p, NULL, " ")) {
		pparv.parv[c++] = s;
	}
	pparv.parv[c] = '\0';
	return &pparv;
}
