/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_svinfo.c: SVINFO protocol module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_svinfo.c,v 1.12 2002/04/08 16:00:48 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include "patchlevel.h"
#include "zlink.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_svinfo(aClient *, aClient *, int, char **);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_svinfo.so",
	"SVINFO protocol",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("SVINFO", m_svinfo, 0, 15, 1, 1, 0, 0L);
}
void _module_unload(void) {
	del_Command("SVINFO", m_svinfo);
}
#endif

/*
 * m_svinfo
 *	parv[0] = sender prefix
 *
 *	Initializing compression
 *	if (parc == 2)
 *		parv[1] = ZIP
 *
 *	if (parc == 5) -- link is TSora (bahamut compatible)
 *		parv[1] = TS_CURRENT	(TSora: -2)
 *		parv[2] = TS_MIN	(TSora: -2)
 *		parv[3] = standalone or connected to non-TS
 *		parv[4] = UTC time 
 *
 *	if (parc > 5) -- link is rage2
 *		parv[5] = ircd codename
 *		parv[6] = ircnetwork (or "SERVICES" for services)
 *		parv[7] = masking keys (key1.key2.key3, or "0.0.0" for services)
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	char *inpath = get_client_name(cptr, HIDEME);
	time_t delta, tmptime, theirtime;
	char keys[50];
	int ts_cur = 0, ts_min = 0;

	if (!IsServer(sptr) || !MyConnect(sptr)) {
		return 0;
	}
	if (parc == 2 && !mycmp(parv[1], "ZIP")) {
		SetZipIn(sptr);
		sptr->serv->zip_in = zip_create_input_session();
		return ZIP_NEXT_BUFFER;
	}

	if (parc < 5) {
		sendto_gnotice("Link %s dropped, incorrect number of SVINFO parameters", inpath);
		sendto_one(sptr, "ERROR :Incorrect number of SVINFO parameters");
		return exit_client(sptr, sptr, sptr, "Incorrect number of SVINFO parameters");
	}

	if (parc > 5) {
		ts_cur = TS_CURRENT;
		ts_min = TS_MIN;
	}
	else {
		ts_cur = TSORA_CURRENT;
		ts_min = TSORA_MIN;
	}
	if (ts_cur < atoi(parv[2]) || atoi(parv[1]) < ts_min) {
		sendto_gnotice("Link %s dropped, wrong TS protocol version (%s, %s)",
			inpath, parv[1], parv[2]);
		sendto_one(sptr, "ERROR :Wrong TS protocol version");
		return exit_client(sptr, sptr, sptr, "Wrong TS protocl version");
	}

	tmptime = time(NULL);
	theirtime = atol(parv[4]);
	delta = abs(theirtime - tmptime);

	if (delta > TS_MAX_DELTA) {
		sendto_gnotice("Link %s dropped, excessive TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
		sendto_one(sptr, "ERROR :Excessive TS delta");
		return exit_client(sptr, sptr, sptr, "Excessive TS delta");
	}

	if (delta > TS_WARN_DELTA) {
		sendto_gnotice("Link %s has a notable TS delta (our TS=%d, their TS=%d, delta=%d)",
			inpath, tmptime, theirtime, delta);
	}

	if (parc > 5) {
		if (strcmp(CODENAME, parv[5])) {
			sendto_gnotice("Link %s dropped, incompatible server version (ours=%s, theirs=%s)",
				inpath, CODENAME, parv[5]);
			sendto_one(sptr, "ERROR :Incompatible server version (please upgrade)");
			return exit_client(sptr, sptr, sptr, "Incompatible server version (please upgrade)");
		}
		if (strcmp(ConfigSet.ircnetwork, parv[6])) {
			sendto_gnotice("Link %s dropped, irc network mismatch (ours=%s, theirs=%s)",
				inpath, ConfigSet.ircnetwork, parv[6]);
			sendto_one(sptr, "ERROR :IRC network mismatch");
			return exit_client(sptr, sptr, sptr, "IRC network mismatch");
		}
		ircsprintf(keys, "%ld.%ld.%ld", ConfigSet.mask_key1, ConfigSet.mask_key2, ConfigSet.mask_key3);
		if (strcmp(keys, parv[7])) {
			sendto_gnotice("Link %s dropped, masking key mismatch", inpath);
			sendto_one(sptr, "ERROR :Masking key mismatch");
			return exit_client(sptr, sptr, sptr, "Masking key mismatch");
		}
	}
	return 0;
}
