/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * modules/m_trace.c: TRACE command module
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: m_trace.c,v 1.5 2002/04/08 16:00:48 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "memory.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

int m_trace(aClient *, aClient *, int, char **);

#ifndef STATIC_MODULES
Module _module_header = {
	"modules/m_trace.so",
	"/TRACE command",
	30000,
	NULL,
	NULL
};
void _module_load(void) {
	add_Command("TRACE", m_trace, 0, 15, 1, 0, 0, 0L);
}
void _module_unload(void) {
	del_Command("TRACE", m_trace);
}
#endif

/*
 * m_trace
 *	parv[0] = sender prefix
 *	parv[1] = server name
 */
int m_trace(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static time_t last_used = 0L;
	aClient *acptr = NULL;
	ConfigItem_class *class = NULL;
	int i, doall, link_s[MAXCONNECTIONS], link_u[MAXCONNECTIONS];
	int cnt = 0, wilds = 0, dow = 0;
	char *tname = parc > 1 ? parv[1] : me.name;

	if (ConfigSet.hideulines) {
		acptr = next_client_double(client, tname);
		if (acptr) {
			if (!IsOper(sptr) && IsULine(acptr)) {
				sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
				return 0;
			}
			acptr = NULL;
		}
	}

	if (parc > 2) {
		if (hunt_server(cptr, sptr, ":%s TRACE %s :%s", 2, parc, parv)) {
			return 0;
		}
	}

	switch (hunt_server(cptr, sptr, ":%s TRACE :%s", 1, parc, parv)) {
		case HUNTED_PASS:
		{
			aClient *acptr2 = next_client_double(client, tname);
			sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0], version,
				debugmode, tname,
				acptr2->from->name ? acptr2->from->name : "acptr2_is_NULL!");
			return 0;
		}
		case HUNTED_ISME:
			break;
		default:
			return 0;
	}

	if (!IsOper(sptr)) {
		if (parv[1] && !strchr(parv[1], '.') && (strchr(parv[1], '*') || strchr(parv[1], '?'))) {
			sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], parv[1]);
			return 0;
		}
		if (last_used + ConfigSet.pacetime > NOW) {
			sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
			return 0;
		}
		else {
			last_used = NOW;
		}
		if (IsPerson(sptr)) {
			sendto_realops_lev(SPY_LEV, "TRACE requested by %s (%s@%s) [%s]", sptr->name,
				sptr->user->username, IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host,
				sptr->user->server);
		}
	}

	doall = (parv[1] && parc > 1) ? !match(tname, me.name) : TRUE;
	wilds = !parv[1] || strchr(tname, '*') || strchr(tname, '?');
	dow = wilds || doall;

	if (!IsOper(sptr) || !dow) {
		char *name, *class;
		acptr = hash_find_client(tname, (aClient *)NULL);
		if (!acptr || !IsPerson(acptr)) {
			sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
			return 0;
		}

		name = get_client_name(acptr, FALSE);
		class = acptr->class ? acptr->class->name : "default";

		sendto_one(sptr, rpl_str(IsOper(acptr) ? RPL_TRACEOPERATOR : RPL_TRACEUSER), me.name,
			parv[0], class, name, timeofday - acptr->lasttime);
		sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
		return 0;
	}

	if (dow && lifesux && !IsOper(sptr)) {
		sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
		return 0;
	}

	memset((char *)link_s, '\0', sizeof(link_s));
	memset((char *)link_u, '\0', sizeof(link_u));

	if (doall) {
		for (acptr = client; acptr; acptr = acptr->next) {
			if (IsPerson(acptr) && (!IsInvisible(acptr) || IsOper(sptr))) {
				link_u[acptr->from->fd]++;
			}
			else if (IsServer(acptr)) {
				if (ConfigSet.hideulines) {
					if (IsOper(sptr) || !IsULine(acptr)) {
						link_s[acptr->from->fd]++;
					}
				}
				else {
					link_s[acptr->from->fd]++;
				}
			}
		}
	}

	for (i = 0; i <= highest_fd; i++) {
		char *name, *class;
		if (!(acptr = local[i])) {
			continue;
		}
		if (ConfigSet.hideulines && !IsOper(sptr) && IsULine(acptr)) {
			continue;
		}
		if (IsInvisible(acptr) && dow && !(MyConnect(sptr) && IsOper(sptr)) &&
		  !IsOper(acptr) && acptr != sptr) {
			continue;
		}
		if (!dow && mycmp(tname, acptr->name)) {
			continue;
		}

		name = get_client_name(acptr, IsOper(sptr) ? FALSE : HIDEME);
		class = acptr->class ? acptr->class->name : "default";

		switch (acptr->status) {
			case STAT_CONNECTING:
				sendto_one(sptr, rpl_str(RPL_TRACECONNECTING), me.name, parv[0],
					class, name);
				cnt++;
				break;
			case STAT_HANDSHAKE:
				sendto_one(sptr, rpl_str(RPL_TRACEHANDSHAKE), me.name, parv[0],
					class, name);
				cnt++;
				break;
			case STAT_ME:
				break;
			case STAT_UNKNOWN:
				sendto_one(sptr, rpl_str(RPL_TRACEUNKNOWN), me.name, parv[0],
					class, name,
					acptr->firsttime ? timeofday - acptr->firsttime : -1);
				cnt++;
				break;
			case STAT_CLIENT:
				if (((IsOper(sptr) && MyClient(sptr)) || !(dow && IsInvisible(acptr)))
				  || !dow || IsOper(acptr)) {
					if (IsOper(acptr)) {
						sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR), me.name,
							parv[0], class, name, timeofday - acptr->lasttime);
					}
				}
				break;
			case STAT_SERVER:
				sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], class,
					link_s[i], link_u[i], name,
					*(acptr->serv->bynick) ? acptr->serv->bynick : "*",
					*(acptr->serv->byuser) ? acptr->serv->byuser : "*",
					*(acptr->serv->byhost) ? acptr->serv->byhost : me.name);
				cnt++;
				break;
			case STAT_LOG:
				sendto_one(sptr, rpl_str(RPL_TRACELOG), me.name, parv[0], LOGFILE,
					acptr->port);
				cnt++;
				break;
			default:
				sendto_one(sptr, rpl_str(RPL_TRACENEWTYPE), me.name, parv[0], name);
				cnt++;
				break;
		}
	}

	if (!SendWallops(sptr) || !cnt) {
		if (cnt) {
			sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
			return 0;
		}
		sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], 0, link_s[me.fd],
			link_u[me.fd], me.name, "*", "*", me.name, timeofday - acptr->lasttime);
		sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
		return 0;
	}

	if (!ConfigSet.hideulines || (ConfigSet.hideulines && IsOper(sptr))) {
		for (class = config_class; class; class = (ConfigItem_class *)class->next) {
			sendto_one(sptr, rpl_str(RPL_TRACECLASS), me.name, parv[0],
				class->name ? class->name : "<noname>", class->clients);
		}
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	return 0;
}
