/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/channel.c: channel management routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: channel.c,v 1.52 2002/04/08 16:00:50 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"
#include "memory.h"

aChannel *channel = NullChn;
BlockHeap *channel_heap;
BlockHeap *chanmember_heap;
BlockHeap *ban_heap;

int check_channelmask(aClient *, aClient *, char *);
void set_mode(aChannel *, aClient *, int, char **, u_int *, char [MAXMODEPARAMS][MODEBUFLEN + 3], int);
static void make_mode_str(aChannel *, u_int, u_int, int, char [MAXMODEPARAMS][MODEBUFLEN + 3], char *, char *, char);
static int do_mode_char(aChannel *, u_int, char, char *, u_int, aClient *, u_int *, char [MAXMODEPARAMS][MODEBUFLEN + 3], char);
void do_mode(aChannel *, aClient *, aClient *, int, char **, int, int);
void bounce_mode(aChannel *, aClient *, int, char **);
void clean_channelname(unsigned char *);
int can_join(aClient *, aChannel *, char *);
void channel_modes(aClient *, char *, char *, aChannel *);
static void sub1_from_channel(aChannel *);

int add_ban_tolist(aClient *, aChannel *, char *, aBan **);
int del_ban_fromlist(aChannel *, char *, aBan **);

#ifdef ORATIMING
struct timeval tsdnow, tsdthen;
unsigned long tsdms;
#endif

static char buf[BUFSIZE];
static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];
static int used_samode = 0;

#ifdef ANTI_SPAMBOT
extern int spam_num;
extern int spam_time;
#endif

aCtab cFlagTab[] = {
	{ MODE_BAN,		'b', 0, 1 },
	{ MODE_INVITEONLY,	'i', 1, 0 },
	{ MODE_KEY,		'k', 0, 1 },
	{ MODE_LIMIT,		'l', 0, 1 },
	{ MODE_MODERATED,	'm', 0, 0 },
	{ MODE_NOPRIVMSGS,	'n', 1, 0 },
	{ MODE_CHANOP,		'o', 0, 1 },
	{ MODE_PRIVATE,		'p', 1, 0 },
	{ MODE_REGISTERED,	'r', 0, 0 },
	{ MODE_REGONLY,		'R', 0, 0 },
	{ MODE_SECRET,		's', 1, 0 },
	{ MODE_TOPICLIMIT,	't', 1, 0 },
	{ MODE_VOICE,		'v', 1, 1 },
	{ MODE_NOCOLOR,		'c', 1, 0 },
	{ MODE_OPERONLY,	'O', 0, 0 },
	{ MODE_ADMINONLY,	'A', 0, 0 },
	{ MODE_HALFOP,		'h', 1, 1 },
	{ MODE_CHANADMIN,	'a', 0, 1 },
	{ MODE_STRIPCOLOR,	'S', 1, 0 },
	{ MODE_EXBAN,		'e', 0, 1 },
	{ MODE_NOCTCP,		'C', 1, 0 },
	{ MODE_NONICKCH,	'N', 1, 0 },
	{ 0x0,			0x0, 0x0, 0x0 }
};

void channel_heap_garbage_collect()
{
	BlockHeapGarbageCollect(channel_heap);
	BlockHeapGarbageCollect(chanmember_heap);
	BlockHeapGarbageCollect(ban_heap);
	return;
}

void init_channels()
{
	channel_heap = BlockHeapCreate((size_t)sizeof(aChannel), CHANNELS_PREALLOCATE);
	chanmember_heap = BlockHeapCreate((size_t)sizeof(chanMember), CHANMEMBERS_PREALLOCATE);
	ban_heap = BlockHeapCreate((size_t)sizeof(aBan), BANS_PREALLOCATE);
	addEvent("channel_heap_garbage_collect", channel_heap_garbage_collect, NULL, CHANNEL_HEAP_GARBAGE_COLLECT_FREQ, 1);
	return;
}

int check_halfop(u_int modetype)
{
	aCtab *tab = &cFlagTab[0];

	while (tab->mode != 0x0) {
		if (tab->mode == modetype) {
			return ((tab->halfop == 1) ? TRUE : FALSE);
		}
		tab++;
	}
	return (TRUE);
}

char chmode_string[30];
void build_chmode_string(void)
{
	char *p = &chmode_string[0];
	aCtab *tab = &cFlagTab[0];

	while (tab->mode != 0x0) {
		*p++ = tab->flag;
		tab++;
	}
	*p = '\0';
}


/* return the length (>= 0) of a chain of links */
static int list_length(Link *lp)
{
	int count = 0;

	for (; lp; lp = lp->next) {
		count++;
	}
	return (count);
}

/*
 * find_chasing 
 *   Find the client structure for a nick name (user) using history 
 *   mechanism if necessary. If the client is not found, an error message 
 *   (NO SUCH NICK) is generated. If the client was found through the 
 *   history, chasing will be 1 and otherwise 0.
 */
aClient *find_chasing(aClient *sptr, char *user, int *chasing)
{
	aClient *acptr = find_client(user, (aClient *)NULL);

	if (chasing) {
		*chasing = 0;
	}
	if (acptr) {
		return (acptr);
	}

	acptr = get_history(user, (long)KILLCHASETIMELIMIT);
	if (!acptr) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, user);
		return (aClient *)NULL;
	}

	if (chasing) {
		*chasing = 1;
	}
	return acptr;
}

int add_ban_tolist(aClient *cptr, aChannel *chptr, char *banid, aBan **list)
{
	aBan *bptr;
	int cnt = 0, len = 0;

	if (MyClient(cptr)) {
		collapse(banid);
	}
	for (bptr = *list; bptr; bptr = bptr->next) {
		len += strlen(bptr->banstr);
		if (MyClient(cptr)) {
			if (++cnt >= MAXBANS) {
				sendto_one(cptr, err_str(ERR_BANLISTFULL), me.name, cptr->name,
					chptr->chname, banid);
				return 0;
			}
			else {
				if (!match(bptr->banstr, banid) || !match(banid, bptr->banstr)) {
					return 0;
				}
			}
		}
		else if (!strcasecmp(bptr->banstr, banid)) {
			return 0;
		}
	}

	bptr = (aBan *)MyMallocX(sizeof(aBan));
	bptr->next = *list;
	bptr->banstr = (char *)MyMallocX(strlen(banid) + 1);
	strcpy(bptr->banstr, banid);
	bptr->who = (char *)MyMallocX(strlen(cptr->name) + 1);
	strcpy(bptr->who, cptr->name);
	bptr->when = timeofday;
	*list = bptr;
	return 1;
}

int del_ban_fromlist(aChannel *chptr, char *banid, aBan **list)
{
	aBan **bptr, *tmp;

	if (BadPtr(banid)) {
		return 0;
	}
	for (bptr = &(*list); *bptr; bptr = &((*bptr)->next)) {
		if (!strcasecmp(banid, (*bptr)->banstr)) {
			tmp = *bptr;
			*bptr = tmp->next;
			MyFree(tmp->banstr);
			MyFree(tmp->who);
			MyFree(tmp);
			return 1;
		}
	}
	return 0;
}

/*
 * is_banned - returns a pointer to the ban structure if banned else
 * NULL
 */
aBan *is_banned(aClient *cptr, aChannel *chptr)
{
	aBan *tmp;
	char *s;
	char host[NICKLEN+USERLEN+HOSTLEN+6], maskedhost[NICKLEN+USERLEN+HOSTLEN+6];
	int testem = 0;

	if (!IsPerson(cptr)) {
		return NULL;
	}
	if (strcmp(cptr->user->host, cptr->user->maskedhost)) {
		testem = 1;
	}

	s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->host);
	strcpy(host, s);
	s = make_nick_user_host(cptr->name, cptr->user->username, cptr->user->maskedhost);
	strcpy(maskedhost, s);

	for (tmp = chptr->banlist; tmp; tmp = tmp->next) {
		if ((match(tmp->banstr, host) == 0) || (testem && (match(tmp->banstr, maskedhost) == 0))) {
			break;
		}
	}
	return (tmp);
}

/*
 * adds a user to a channel by adding another link to the channels
 * member chain.
 */
void add_user_to_channel(aChannel *chptr, aClient *acptr, int flags)
{
	Link *ptr;
	chanMember *cm;

	if (acptr->user) {
		cm = make_chanmember();
		cm->flags = flags;
		cm->cptr = acptr;
		cm->next = chptr->members;
		chptr->members = cm;
		chptr->users++;

		ptr = make_link();
		ptr->value.chptr = chptr;
		ptr->next = acptr->user->channel;
		acptr->user->channel = ptr;
		acptr->user->joined++;
	}
}

void remove_user_from_channel(aClient *sptr, aChannel *chptr)
{
	chanMember **curr, *tmp;
	Link **lcurr, *ltmp;

	for (curr = &chptr->members; (tmp = *curr); curr = &tmp->next) {
		if (tmp->cptr == sptr) {
			*curr = tmp->next;
			free_chanmember(tmp);
			break;
		}
	}
	for (lcurr = &sptr->user->channel; (ltmp = *lcurr); lcurr = &ltmp->next) {
		if (ltmp->value.chptr == chptr) {
			*lcurr = ltmp->next;
			free_link(ltmp);
			break;
		}
	}

	sptr->user->joined--;
	sub1_from_channel(chptr);
}

int is_chan_admin(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;

	if (chptr) {
		cm = find_user_member(chptr->members, cptr);
		if (cm) {
			return (cm->flags & CHFL_CHANADMIN);
		}
	}
	return 0;
}

int is_chan_op(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;

	if (chptr) {
		cm = find_user_member(chptr->members, cptr);
		if (cm) {
			return (cm->flags & CHFL_CHANOP);
		}
	}
	return 0;
}

int is_half_op(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;

	if (chptr) {
		cm = find_user_member(chptr->members, cptr);
		if (cm) {
			return (cm->flags & CHFL_HALFOP);
		}
	}
	return 0;
}

int is_deopped(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;

	if (chptr) {
		cm = find_user_member(chptr->members, cptr);
		if (cm) {
			return (cm->flags & CHFL_DEOPPED);
		}
	}
	return 0;
}

int has_voice(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;

	if (chptr) {
		cm = find_user_member(chptr->members, cptr);
		if (cm) {
			return (cm->flags & CHFL_VOICE);
		}
	}
	return 0;
}

int can_send(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	int member;

	if (IsServer(cptr) || IsULine(cptr)) {
		return 0;
	}

	member = (cm = find_user_member(chptr->members, cptr)) ? 1 : 0;
	if (!member) {
		if (chptr->mode.mode & MODE_MODERATED) {
			return (MODE_MODERATED);
		}
		if (chptr->mode.mode & MODE_NOPRIVMSGS) {
			return (MODE_NOPRIVMSGS);
		}
		if (MyClient(cptr) && is_banned(cptr, chptr)) {
			return (MODE_BAN);
		}
	}
	else {
		if ((chptr->mode.mode & MODE_MODERATED) && !(cm->flags & (CHFL_CHANOP|CHFL_HALFOP|CHFL_VOICE))) {
			return (MODE_MODERATED);
		}
		if (is_banned(cptr, chptr) && !(cm->flags & (CHFL_CHANOP|CHFL_HALFOP|CHFL_VOICE))) {
			return (MODE_BAN);
		}
	}
	return 0;
}

/*
 * write the "simple" list of channel modes for channel chptr onto
 * buffer mbuf with the parameters in pbuf.
 */
void channel_modes(aClient *cptr, char *mbuf, char *pbuf, aChannel *chptr)
{
	u_int zode;
	aCtab *tab = &cFlagTab[0];
	char tmpbuf[1024];

	*mbuf++ = '+';
	while (tab->mode != 0x0) {
		if ((chptr->mode.mode & tab->mode)) {
			zode = chptr->mode.mode;

			if (!(zode & (MODE_LIMIT | MODE_KEY))) {
				if (!(zode & (MODE_CHANADMIN |MODE_CHANOP | MODE_HALFOP | MODE_VOICE))) {
					if (!(zode & (MODE_BAN | MODE_EXBAN))) {
						*mbuf++ = tab->flag;
					}
				}
			}
		}
		tab++;
	}

	if (chptr->mode.limit) {
		*mbuf++ = 'l';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr)) {
			ircsprintf(tmpbuf, "%d ", chptr->mode.limit);
			strcat(pbuf, tmpbuf);
		}
	}
	if (*chptr->mode.key) {
		*mbuf++ = 'k';
		if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr)) {
			ircsprintf(tmpbuf, "%s ", chptr->mode.key);
			strcat(pbuf, tmpbuf);
		}
	}

	*mbuf++ = '\0';
	return;
}

/* send "cptr" a full list of the modes for channel chptr using SSJv2. */
void send_channel_modes(aClient *cptr, aChannel *chptr)
{
	chanMember *cm;
	char *bufptr;
	aBan *ban;
	int n = 0;

	if (*chptr->chname != '#') {
		return;
	}
	if (!IsSSJOIN(cptr) && !IsSSJv2(cptr)) {
		sendto_ops_lev(DEBUG_LEV, "Sending SJOIN command to non-SSJOIN and non-SSJv2 server %s",
			cptr->name);
	}

	modebuf[0] = parabuf[0] = '\0';
	channel_modes(cptr, modebuf, parabuf, chptr);

	if (parabuf[0] == '\0') {
		strcat(parabuf, "<none>");
	}

	ircsprintf(buf, ":%s SJOIN %ld %s %s %s :", me.name, chptr->channelts, chptr->chname, modebuf,
		parabuf);
	bufptr = (buf + strlen(buf));

	for (cm = chptr->members; cm; cm = cm->next) {
		if (cm->flags & CHFL_CHANOP) {
			*bufptr++ = '@';
		}
		if (cm->flags & CHFL_VOICE) {
			*bufptr++ = '+';
		}

		/* If cptr supports SSJv2, send CHANADMIN and HALFOP statuses too */
		if (IsSSJv2(cptr)) {
			if (cm->flags & CHFL_CHANADMIN) {
				*bufptr++ = '*';
			}
			if (cm->flags & CHFL_HALFOP) {
				*bufptr++ = '%';
			}
		}
		strcpy(bufptr, cm->cptr->name);

		bufptr += strlen(bufptr);
		*bufptr++ = ' ';
		n++;

		if ((bufptr - buf) > (BUFSIZE - 80)) {
			*bufptr++ = '\0';
			if (bufptr[-1] == ' ') {
				bufptr[-1] = '\0';
			}
			sendto_one(cptr, "%s", buf);
			n = 0;
			bufptr = (buf + strlen(buf));
		}
	}

	/* If cptr supports SSJv2, send any bans and exbans we may have */
	if (IsSSJv2(cptr)) {
		for (ban = chptr->banlist; ban; ban = ban->next) {
			*bufptr++ = '&';
			strcpy(bufptr, ban->banstr);

			bufptr += strlen(bufptr);
			*bufptr++ = ' ';
			n++;

			if ((bufptr - buf) > (BUFSIZE - 80)) {
				*bufptr++ = '\0';
				if (bufptr[-1] == ' ') {
					bufptr[-1] = '\0';
				}
				sendto_one(cptr, "%s", buf);
				n = 0;
				bufptr = (buf + strlen(buf));
			}
		}
		for (ban = chptr->exbanlist; ban; ban = ban->next) {
			*bufptr++ = '"';
			strcpy(bufptr, ban->banstr);

			bufptr += strlen(bufptr);
			*bufptr++ = ' ';
			n++;

			if ((bufptr - buf) > (BUFSIZE - 80)) {
				*bufptr++ = '\0';
				if (bufptr[-1] == ' ') {
					bufptr[-1] = '\0';
				}
				sendto_one(cptr, "%s", buf);
				n = 0;
				bufptr = (buf + strlen(buf));
			}
		}
	}

	if (n) {
		*bufptr++ = '\0';
		if (bufptr[-1] == ' ') {
			bufptr[-1] = '\0';
		}
		sendto_one(cptr, "%s", buf);
	}

	return;
}


int check_channelmask(aClient *sptr, aClient *cptr, char *chname)
{
	char *s;

	if (*chname == '&' && IsServer(cptr)) {
		return -1;
	}

	s = rindex(chname, ':');
	if (!s) {
		return 0;
	}

	s++;
	if (match(s, me.name) || (IsServer(cptr) && match(s, cptr->name))) {
		if (MyClient(sptr)) {
			sendto_one(sptr, err_str(ERR_BADCHANMASK), me.name, sptr->name, chname);
		}
		return -1;
	}
	return 0;
}

/* bounce_mode -- written by binary
 *      User or server is NOT authorized to change the mode.  This takes care
 * of making the bounce string and bounce it.  Because of the 1 for the bounce
 * param (last param) of the calls to set_mode and make_mode_str, it will not
 * set the mode, but create the bounce string.
 */
void bounce_mode(aChannel *chptr, aClient *cptr, int parc, char *parv[])
{
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	int pcount;
	u_int oldm, oldl;


	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;

	set_mode(chptr, cptr, parc, parv, &pcount, pvar, 1);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf, parabuf, 1);

	sendto_one(cptr, ":%s MODE %s &%s %s %lu", me.name, chptr->chname, mode_buf, parabuf, chptr->channelts);
}

/* do_mode -- written by binary
 *      User or server is authorized to do the mode.  This takes care of
 * setting the mode and relaying it to other users and servers.
 */
void do_mode(aChannel *chptr, aClient *cptr, aClient *sptr, int parc, char *parv[], int sendts, int samode)
{
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3];
	char mode_buf[MODEBUFLEN], parabuf[MODEBUFLEN];
	int pcount;
	u_int oldm, oldl;
	char tschange = 0, isbounce = 0;

	if (**parv == '&') {
		isbounce = 1;
	}
	oldm = chptr->mode.mode;
	oldl = chptr->mode.limit;

	set_mode(chptr, sptr, parc, parv, &pcount, pvar, 0);
	make_mode_str(chptr, oldm, oldl, pcount, pvar, mode_buf, parabuf, 0);

	if (IsServer(sptr)) {
		if (sendts > 0) {
			if (sendts < chptr->channelts) {
				tschange = 1;
				chptr->channelts = sendts;
			}
			if (sendts > chptr->channelts && chptr->channelts) {
				sendts = chptr->channelts;
				sendto_one(cptr, ":%s MODE %s + %lu", me.name, chptr->chname, chptr->channelts);
			}
		}
		if (sendts == -1 && chptr->channelts) {
			sendts = chptr->channelts;
		}
	}

	if (*mode_buf == '\0' || (*(mode_buf + 1) == '\0' && (*mode_buf == '+' || *mode_buf == '-'))) {
		if (tschange || isbounce) {
			sendto_serv_butone(cptr, ":%s MODE %s %s+ %lu", me.name, chptr->chname,
				isbounce ? "&" : "", chptr->channelts);
		}
		return;
	}

	if (IsPerson(sptr) && samode && MyClient(sptr)) {
		sendto_serv_butone(NULL, ":%s GLOBOPS :%s used SAMODE %s (%s %s)", me.name, sptr->name,
			chptr->chname, mode_buf, parabuf);
		sendto_ops("%s used SAMODE %s (%s %s)", sptr->name, chptr->chname, mode_buf, parabuf);
		sendts = 0;
		used_samode = 1;
	}

	sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", sptr->name, chptr->chname, mode_buf, parabuf);

	if (IsServer(sptr) && sendts != -1) {
		sendto_serv_butone(cptr, ":%s MODE %s %s%s %s %lu", sptr->name, chptr->chname,
			isbounce ? "&" : "", mode_buf, parabuf);
	}
	else {
		sendto_serv_butone(cptr, ":%s MODE %s %s%s %s NoTS", sptr->name, chptr->chname,
			isbounce? "&" : "", mode_buf, parabuf);
	}
}

/* make_mode_str -- written by binary
 *      Reconstructs the mode string, to make it look clean.  mode_buf will
 *  contain the +x-y stuff, and the parabuf will contain the parameters.
 *  If bounce is set to 1, it will make the string it needs for a bounce.
 */
void make_mode_str(aChannel *chptr, u_int oldm, u_int oldl, int pcount, char pvar[MAXMODEPARAMS][MODEBUFLEN + 3],
	char *mode_buf, char *parabuf, char bounce)
{
	char *x = mode_buf, tmpbuf[MODEBUFLEN], *tmpstr;
	int what = 0, cnt, z;
	aCtab *tab = &cFlagTab[0];

	*tmpbuf = '\0';
	*mode_buf = '\0';
	*parabuf = '\0';

	/* + paramless mode */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0) {
		if ((chptr->mode.mode & tab->mode) && !(oldm & tab->mode)) {
			if (what != MODE_ADD) {
				*x++ = bounce ? '-' : '+';
				what = MODE_ADD;
			}
			*x++ = tab->flag;
		}
		tab++;
	}
	*x = '\0';

	/* - paramless mode */
	tab = &cFlagTab[0];
	while (tab->mode != 0x0) {
		if (!(chptr->mode.mode & tab->mode) && (oldm & tab->mode)) {
			if (what != MODE_DEL) {
				*x++ = bounce ? '+' : '-';
				what = MODE_DEL;
			}
			*x++ = tab->flag;
		}
		tab++;
	}
	*x = '\0';

	/* reconstruct bkohv chain */
	for (cnt = 0; cnt < pcount; cnt++) {
		if ((*(pvar[cnt]) == '+') && what != MODE_ADD) {
			*x++ = bounce ? '-' : '+';
			what = MODE_ADD;
		}
		if ((*(pvar[cnt]) == '-') && what != MODE_DEL) {
			*x++ = bounce ? '+' : '-';
			what = MODE_DEL;
		}
		*x++ = *(pvar[cnt] + 1);
		tmpstr = &pvar[cnt][2];
		strncat(parabuf, tmpstr, MODEBUFLEN - 1);
		parabuf[MODEBUFLEN - 1] = '\0';
		z = strlen(parabuf);
		parabuf[z] = ' ';
		parabuf[z + 1] = '\0';
	}

	/* user limit */
	if (chptr->mode.limit != oldl) {
		if ((!bounce && chptr->mode.limit == 0) || (bounce && chptr->mode.limit != 0)) {
			if (what != MODE_DEL) {
				*x++ = '-';
				what = MODE_DEL;
			}
			if (bounce) {
				chptr->mode.limit = 0;
			}
			*x++ = 'l';
			ircsprintf(parabuf, "%s%d", parabuf, oldl);
		}
		else {
			if (what != MODE_ADD) {
				*x++ = '+';
				what = MODE_ADD;
			}
			if (bounce) {
				chptr->mode.limit = oldl;
			}
			*x++ = 'l';
			ircsprintf(parabuf, "%s%d", parabuf, chptr->mode.limit);
		}
	}
	if (bounce) {
		chptr->mode.mode = oldm;
	}

	z = strlen(parabuf);
	if (parabuf[z - 1] == ' ') {
		parabuf[z - 1] = '\0';
	}
	*x = '\0';

	if (*mode_buf == '\0') {
		*mode_buf = '+';
		mode_buf++;
		*mode_buf = '\0';
	}
	return;
}

/* do_mode_char
 *  processes one mode character   
 *  returns 1 if it ate up a param, otherwise 0
 *      written by binary
 */
int do_mode_char(aChannel *chptr, u_int modetype, char modechar, char *param, u_int what, aClient *cptr,
	u_int *pcount, char pvar[MAXMODEPARAMS][MODEBUFLEN + 3], char bounce)
{
	unsigned int tmp = 0;
	int retval = 0, chasing = 0, x;
	chanMember *member;
	aClient *who;
	char tmpbuf[256], *tmpstr, *modename = "(huh?)", modeflag = ' ';
	aCtab *tab = &cFlagTab[0];

	if (!IsULine(cptr) && !IsServer(cptr) && !used_samode) {
		if (!is_chan_op(cptr, chptr) && is_half_op(cptr, chptr)) {
			if (check_halfop(modetype) == FALSE) {
				while (tab->mode != 0x0) {
					if (tab->mode == modetype) {
						sendto_one(cptr, err_str(ERR_ONLYHALFOP), me.name, cptr->name, tab->flag);
					}
					tab++;
				}
				return 0;
			}
		}
	}

	switch (modetype) {
		case MODE_REGISTERED:
			if (!IsServer(cptr)) {
				break;
			}
		case MODE_ADMINONLY:
			if (!IsNetAdmin(cptr) && !IsAdmin(cptr) && !IsCoAdmin(cptr)) {
				sendto_one(cptr, err_str(ERR_NOPRIVILEGES), me.name, cptr->name);
				break;
			}
			if (chptr->mode.mode & MODE_OPERONLY) {
				chptr->mode.mode &= ~MODE_OPERONLY;
			}
		case MODE_OPERONLY:
			if (!IsOper(cptr)) {
				sendto_one(cptr, err_str(ERR_NOPRIVILEGES), me.name, cptr->name);
				break;
			}
			if (chptr->mode.mode & MODE_ADMINONLY) {
				sendto_one(cptr, ":%s NOTICE %s :*** You cannot set %s Oper Only as it is currently "
					"Admin Only.", me.name, cptr->name, chptr->chname);
				break;
			}
		case MODE_PRIVATE:
		case MODE_SECRET:
			if (modetype == MODE_SECRET && (chptr->mode.mode & MODE_PRIVATE)) {
				chptr->mode.mode &= ~MODE_PRIVATE;
			}
			if (modetype == MODE_PRIVATE && (chptr->mode.mode & MODE_SECRET)) {
				chptr->mode.mode &= ~MODE_SECRET;
			}
		case MODE_NOCOLOR:
		case MODE_STRIPCOLOR:
			if (modetype == MODE_STRIPCOLOR && (chptr->mode.mode & MODE_NOCOLOR)) {
				chptr->mode.mode &= ~MODE_NOCOLOR;
			}
			if (modetype == MODE_NOCOLOR && (chptr->mode.mode & MODE_STRIPCOLOR)) {
				chptr->mode.mode &= ~MODE_STRIPCOLOR;
			}
		case MODE_MODERATED:
		case MODE_TOPICLIMIT:
		case MODE_NOPRIVMSGS:
		case MODE_INVITEONLY:
		case MODE_NONICKCH:
		case MODE_NOCTCP:
			retval = 0;
			if (what == MODE_ADD) {
				chptr->mode.mode |= modetype;
			}
			else {
				chptr->mode.mode &= ~modetype;
			}
			break;
		case MODE_CHANADMIN:
		case MODE_CHANOP:
		case MODE_HALFOP:
		case MODE_VOICE:
			if (!param || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			if (!(who = find_chasing(cptr, param, &chasing))) {
				break;
			}
			if (!(member = find_user_member(chptr->members, who))) {
				sendto_one(cptr, err_str(ERR_USERNOTINCHANNEL), me.name, cptr->name,
					who->name, chptr->chname);
				break;
			}

			switch (modetype) {
				case MODE_CHANADMIN:
					modename = "deadmin";
					modeflag = 'a';
					break;
				case MODE_CHANOP:
					modename = "deop";
					modeflag = 'o';
					break;
				case MODE_HALFOP:
					modename = "dehalfop";
					modeflag = 'h';
					break;
				case MODE_VOICE:
					modename = "devoice";
					modeflag = 'v';
					break;
			}

			if (modetype == MODE_CHANADMIN) {
				if (!IsServer(cptr) && !IsULine(cptr) && who != cptr && !IsNetAdmin(cptr)) {
					break;
				}
			}

			if (is_chan_admin(who, chptr) && what == MODE_DEL) {
				if (!IsULine(cptr) && !IsServer(cptr) && who != cptr && !IsNetAdmin(cptr)) {
					sendto_one(cptr, ":%s NOTICE %s :*** Cannot %s %s - user is channel admin.",
						me.name, cptr->name, modename, who->name);
					break;
				}
			}

			tmp = member->flags;
			if (what == MODE_ADD) {
				member->flags |= modetype;
			}
			else {
				member->flags &= ~modetype;
			}
			if (tmp == member->flags && (bounce || !IsULine(cptr))) {
				break;
			}
			if (bounce) {
				member->flags = tmp;
			}

			ircsprintf(pvar[*pcount], "%c%c%s", what == MODE_ADD ? '+' : '-', modeflag, who->name);
			(*pcount)++;
			break;
		case MODE_LIMIT:
			if (what == MODE_ADD) {
				if (!param) {
					retval = 0;
					break;
				}
				retval = 1;
				tmp = atoi(param);

				if (chptr->mode.limit == tmp) {
					break;
				}
				chptr->mode.limit = tmp;
			}
			else {
				retval = 0;
				if (!chptr->mode.limit) {
					break;
				}
				chptr->mode.limit = 0;
			}
			break;
		case MODE_KEY:
			if (!param || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}
			retval = 1;

			for (x = 0; x < *pcount; x++) {
				if (pvar[x][1] == 'k') {
					retval = 0;
					break;
				}
			}
			if (retval == 0) {
				break;
			}

			if (what == MODE_ADD) {
				if (!bounce) {
					strncpyzt(chptr->mode.key, param, sizeof(chptr->mode.key));
				}
				tmpstr = param;
			}
			else {
				if (!*chptr->mode.key) {
					break;
				}

				strncpyzt(tmpbuf, chptr->mode.key, sizeof(tmpbuf));
				tmpstr = tmpbuf;
				if (!bounce) {
					strcpy(chptr->mode.key, "");
				}
			}
			retval = 1;
			ircsprintf(pvar[*pcount], "%ck%s", what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;
			break;
		case MODE_BAN:
			if (!param || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			tmpstr = pretty_mask(param);

			if (!bounce) {
				if (what == MODE_ADD && !add_ban_tolist(cptr, chptr, tmpstr, (aBan **)&chptr->banlist)) {
					break;
				}
				else if (what == MODE_DEL && !del_ban_fromlist(chptr, tmpstr, (aBan **)&chptr->banlist)) {
					break;
				}
			}

			ircsprintf(pvar[*pcount], "%cb%s", what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;
			break;
		case MODE_EXBAN:
			if (!param || *pcount >= MAXMODEPARAMS) {
				retval = 0;
				break;
			}

			retval = 1;
			tmpstr = pretty_mask(param);

			if (!bounce) {
				if (what == MODE_ADD && !add_ban_tolist(cptr, chptr, tmpstr, (aBan **)&chptr->exbanlist)) {
					break;
				}
				else if (what == MODE_DEL && !del_ban_fromlist(chptr, tmpstr, (aBan **)&chptr->exbanlist)) {
					break;
				}
			}

			ircsprintf(pvar[*pcount], "%ce%s", what == MODE_ADD ? '+' : '-', tmpstr);
			(*pcount)++;
			break;
	}
	return (retval);
}

/* set_mode
 *      written by binary
 */
void set_mode(aChannel *chptr, aClient *cptr, int parc, char *parv[], u_int *pcount,
	char pvar[MAXMODEPARAMS][MODEBUFLEN + 3], int bounce)
{
	char *curchr, *param = NULL;
	u_int what = MODE_ADD;
	int modetype = 0, found = 0, paracount = 1;
	aCtab *tab = &cFlagTab[0], foundat;

	*pcount = 0;

	for (curchr = parv[0]; *curchr; curchr++) {
		switch (*curchr) {
			case '+':
				what = MODE_ADD;
				break;
			case '-':
				what = MODE_DEL;
				break;
			default:
				tab = &cFlagTab[0];
				found = 0;

				while (tab->mode != 0x0 && found == 0) {
					if (tab->flag == *curchr) {
						found = 1;
						foundat = *tab;
					}
					tab++;
				}

				if (found == 1) {
					modetype = foundat.mode;
				}
				if (found == 0) {
					modetype = 0;
				}

				if (modetype == 0) {
					if (!MyClient(cptr)) {
						break;
					}
					sendto_one(cptr, err_str(ERR_UNKNOWNMODE), me.name, cptr->name, *curchr);
					break;
				}

				param = (parc - paracount > 1) ? parv[paracount] : NULL;
				if (param && strlen(param) >= MODEBUFLEN) {
					param[MODEBUFLEN - 1] = '\0';
				}
				paracount += do_mode_char(chptr, modetype, *curchr, param, what, cptr,
					pcount, pvar, bounce);
				break;
		}
	}
}

int can_join(aClient *sptr, aChannel *chptr, char *key)
{
	Link *lp;
	int invited = 0;

	for (lp = sptr->user->invited; lp; lp = lp->next) {
		if (lp->value.chptr == chptr) {
			invited = 1;
			break;
		}
	}

	if (invited || IsULine(sptr)) {
		return 0;
	}

	if (is_banned(sptr, chptr)) {
		if (IsOper(sptr)) {
			sendto_ops("[OperWalk] - [Type: ban] - [IRCop: %s] - [Channel: %s]",
				sptr->name, chptr->chname);
			sendto_serv_butone(&me, ":%s GLOBOPS :[OperWalk] - [Type: ban] - [IRCop: %s] "
				"- [Channel: %s]", me.name, sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_BANNEDFROMCHAN);
	}

	if (chptr->mode.mode & MODE_INVITEONLY) {
		if (IsOper(sptr)) {
			sendto_ops("[OperWalk] - [Type: invite] - [IRCop: %s] - [Channel: %s]",
				sptr->name, chptr->chname);
			sendto_serv_butone(&me, ":%s GLOBOPS :[OperWalk] - [Type: invite] - [IRCop: %s] "
				"- [Channel: %s]", me.name, sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_INVITEONLYCHAN);
	}

	if (chptr->mode.mode & MODE_REGONLY && !IsRegNick(sptr)) {
		if (IsOper(sptr)) {
			sendto_ops("[OperWalk] - [Type: regonly] - [IRCop: %s] - [Channel: %s]",
				sptr->name, chptr->chname);
			sendto_serv_butone(&me, ":%s GLOBOPS :[OperWalk] - [Type: regonly] - [IRCop: %s] "
				"- [Channel: %s]", me.name, sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_NEEDREGGEDNICK);
	}

	if (chptr->mode.mode & MODE_OPERONLY && !IsOper(sptr)) {
		return (ERR_NOPRIVILEGES);
	}
	if (chptr->mode.mode & MODE_ADMINONLY && !IsNetAdmin(sptr) && !IsAdmin(sptr) && !IsCoAdmin(sptr)) {
		return (ERR_NOPRIVILEGES);
	}

	if (*chptr->mode.key && (BadPtr(key) || mycmp(chptr->mode.key, key))) {
		if (IsOper(sptr)) {
			sendto_ops("[OperWalk] - [Type: key] - [IRCop: %s] - [Channel: %s]",
				sptr->name, chptr->chname);
			sendto_serv_butone(&me, ":%s GLOBOPS :[OperWalk] - [Type: key] - [IRCop: %s] "
				"- [Channel: %s]", me.name, sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_BADCHANNELKEY);
	}

	if (chptr->mode.limit && chptr->users >= chptr->mode.limit) {
		if (IsOper(sptr)) {
			sendto_ops("[OperWalk] - [Type: limit] - [IRCop: %s] - [Channel: %s]",
				sptr->name, chptr->chname);
			sendto_serv_butone(&me, ":%s GLOBOPS :[OperWalk] - [Type: limit] - [IRCop: %s] "
				"- [Channel: %s]", me.name, sptr->name, chptr->chname);
			return 0;
		}
		return (ERR_BADCHANNELKEY);
	}
	return 0;
}

/*
 * Remove bells and commas from channel name
 */
void clean_channelname(unsigned char *cn)
{
	for (; *cn; cn++) {
		if (*cn < 33 || *cn == ',' || *cn == 160) {
			*cn = '\0';
			return;
		}
	}
	return;
}

/* we also tell the client if the channel is invalid. */
int check_channelname(aClient *cptr, unsigned char *cn)
{
	if (!MyClient(cptr)) {
		return 1;
	}

	for (; *cn; cn++) {
		if (*cn < 33 || *cn == ',' || *cn == 160) {
			sendto_one(cptr, err_str(ERR_BADCHANNAME), me.name, cptr->name, cn);
			return 0;
		}
	}
	return 1;
}

/*
 * Get Channel block for chname (and allocate a new channel *
 * block, if it didn't exist before).
 */
aChannel *get_channel(aClient *cptr, char *chname, int flag)
{
	aChannel *chptr;
	int len;

	if (BadPtr(chname)) {
		return (NULL);
	}

	len = strlen(chname);
	if (MyClient(cptr) && len > CHANNELLEN) {
		len = CHANNELLEN;
		*(chname + CHANNELLEN) = '\0';
	}

	chptr = find_channel(chname, NullChn);
	if (chptr) {
		return (chptr);
	}

	if (flag == CREATE) {
		chptr = make_channel();
		strncpyzt(chptr->chname, chname, len + 1);

		if (channel) {
			channel->prevch = chptr;
		}

		chptr->prevch = NULL;
		chptr->nextch = channel;
		channel = chptr;
		chptr->channelts = timeofday;
		add_to_channel_hash_table(chname, chptr);
		Count.chan++;
	}
	return chptr;
}

/* Create invite block and add to channel invite list and client invite list */
void add_invite(aClient *cptr, aChannel *chptr)
{
	Link *inv, **tmp;

	del_invite(cptr, chptr);

	if (list_length(cptr->user->invited) >= ConfigSet.maxchansperuser) {
		del_invite(cptr, cptr->user->invited->value.chptr);
	}

	inv = make_link();
	inv->value.cptr = cptr;
	inv->next = chptr->invites;
	chptr->invites = inv;

	for (tmp = &(cptr->user->invited); *tmp; tmp = &((*tmp)->next));
	inv = make_link();
	inv->value.chptr = chptr;
	inv->next = NULL;
	(*tmp) = inv;
}

/* Delete invite block from channel invite list and client invite list */
void del_invite(aClient *cptr, aChannel *chptr)
{
	Link **inv, *tmp;

	for (inv = &(chptr->invites); (tmp = *inv); inv = &tmp->next) {
		if (tmp->value.cptr == cptr) {
			*inv = tmp->next;
			free_link(tmp);
			break;
		}
	}
	for (inv = &(cptr->user->invited); (tmp = *inv); inv = &tmp->next) {
		if (tmp->value.chptr == chptr) {
			*inv = tmp->next;
			free_link(tmp);
			break;
		}
	}
	return;
}

/*
 * Subtract one user from channel i (and free channel *  block, if
 * channel became empty).
 */
static void sub1_from_channel(aChannel *chptr)
{
	Link *tmp;
	aBan *bp, *next;

	if (--chptr->users <= 0) {
		while ((tmp = chptr->invites)) {
			del_invite(tmp->value.cptr, chptr);
		}

		for (bp = chptr->banlist; bp; bp = next) {
			next = bp->next;
			MyFree(bp->banstr);
			MyFree(bp->who);
			MyFree(bp);
		}

		if (chptr->prevch) {
			chptr->prevch->nextch = chptr->nextch;
		}
		else {
			channel = chptr->nextch;
		}
		if (chptr->nextch) {
			chptr->nextch->prevch = chptr->prevch;
		}
		del_from_channel_hash_table(chptr->chname, chptr);
#ifdef FLUD
		free_fluders(NULL, chptr);
#endif
		free_channel(chptr);
		Count.chan--;
	}
	return;
}

int count_channels(aClient *sptr)
{
	aChannel *chptr;
	int count = 0;

	for (chptr = channel; chptr; chptr = chptr->nextch) {
		count++;
	}
	return (count);
}

void send_topic_burst(aClient *cptr)
{
	aChannel *chptr;
	aClient *acptr;

	for (chptr = channel; chptr; chptr = chptr->nextch) {
		if (*chptr->topic != '\0') {
			sendto_one(cptr, ":%s TOPIC %s %s %ld :%s", me.name, chptr->chname,
				chptr->topic_nick, chptr->topic_time, chptr->topic);
		}
	}
	for (acptr = client; acptr; acptr = acptr->next) {
		if (!IsPerson(acptr) || acptr->from == cptr) {
			continue;
		}
		if (acptr->user->away) {
			sendto_one(cptr, ":%s AWAY :%s", acptr->name, acptr->user->away);
		}
	}
	return;
}

void send_names(aClient *sptr, aChannel *chptr)
{
	chanMember *cm;
	aClient *acptr;
	char *s;
	int idx = 1, spos, flag = 1, mlen = strlen(me.name) + NICKLEN + 7;

	if (PubChannel(chptr)) {
		buf[0] = '=';
	}
	else if (SecretChannel(chptr)) {
		buf[0] = '@';
	}
	else {
		buf[0] = '*';
	}

	buf[idx++] = ' ';
	for (s = chptr->chname; *s; s++) {
		buf[idx++] = *s;
	}
	buf[idx++] = ' ';
	buf[idx++] = ':';
	buf[idx] = '\0';
	spos = idx;

	for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;

		if (IsInvisible(acptr) && !IsMember(sptr, chptr)) {
			continue;
		}

		if (cm->flags & CHFL_CHANOP) {
			buf[idx++] = '@';
		}
		else if (cm->flags & CHFL_HALFOP) {
			buf[idx++] = '%';
		}
		else if (cm->flags & CHFL_VOICE) {
			buf[idx++] = '+';
		}
		for (s = acptr->name; *s; s++) {
			buf[idx++] = *s;
		}
		buf[idx++] = ' ';
		buf[idx] = '\0';
		flag = 1;

		if (mlen + idx + NICKLEN > BUFSIZE - 3) {
			sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, sptr->name, buf);
			idx = spos;
			flag = 0;
		}
	}
	if (flag) {
		sendto_one(sptr, rpl_str(RPL_NAMREPLY), me.name, sptr->name, buf);
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFNAMES), me.name, sptr->name, chptr->chname);
	return;
}


/*
 * The function which sends the actual channel list back to the user.
 * Operates by stepping through the hashtable, sending the entries back if
 * they match the criteria.
 * cptr = Local client to send the output back to.
 * numsend = Number (roughly) of lines to send back. Once this number has
 * been exceeded, send_list will finish with the current hash bucket,
 * and record that number as the number to start next time send_list
 * is called for this user. So, this function will almost always send
 * back more lines than specified by numsend (though not by much,
 * assuming CH_MAX is was well picked). So be conservative in your choice
 * of numsend. -Rak
 */

void send_list(aClient *sptr, int numsend)
{
	aChannel *chptr;
	LOpts *lopt = sptr->user->lopt;
	int hashnum;
	char modestuff[TOPICLEN + 3 + KEYLEN + 8 + 3 + 1], tmpbuf[3 + KEYLEN + 8 + 3 + 1];

	for (hashnum = lopt->starthash; hashnum < CH_MAX; hashnum++) {
		if (numsend > 0) {
			for (chptr = (aChannel *)hash_get_chan_bucket(hashnum); chptr;
			  chptr = chptr->hnextch) {
				if (SecretChannel(chptr) && !IsMember(sptr, chptr) && !IsSAdmin(sptr))
					continue;
				if ((!lopt->showall) && ((chptr->users < lopt->usermin) ||
				  ((lopt->usermax >= 0) && (chptr->users > lopt->usermax)) ||
				  ((chptr->channelts||1) < lopt->chantimemin) ||
				  (chptr->topic_time < lopt->topictimemin) ||
				  (chptr->channelts > lopt->chantimemax) ||
				  (chptr->topic_time > lopt->topictimemax) ||
				  (lopt->nolist && find_str_link(lopt->nolist, chptr->chname)) ||
				  (lopt->yeslist && !find_str_link(lopt->yeslist, chptr->chname))))
					continue;

				if (IsSAdmin(sptr)) {
					*modebuf = *parabuf = '\0';
					channel_modes(sptr, modebuf, parabuf, chptr);

					ircsprintf(tmpbuf, " [%s%s%s]", modebuf,
						*parabuf ? " " : "",
						*parabuf ? parabuf : "");
					ircsprintf(modestuff, "%-2s %s", tmpbuf, chptr->topic);

					sendto_one(sptr, rpl_str(RPL_LIST), me.name, sptr->name,
						chptr->chname, chptr->users, modestuff);
				}
				else {
					if (ShowChannel(sptr, chptr)) {
						sendto_one(sptr, rpl_str(RPL_LIST), me.name, sptr->name,
							chptr->chname, chptr->users, chptr->topic);
					}
				}
				numsend--;
			}
		}
		else {
			break;
		}
	}

	if (hashnum == CH_MAX) {
		Link *lp, *next;

		sendto_one(sptr, rpl_str(RPL_LISTEND), me.name, sptr->name);
		for (lp = lopt->yeslist; lp; lp = next) {
			next = lp->next;
			MyFree(lp->value.cp);
			free_link(lp);
		}
		for (lp = lopt->nolist; lp; lp = next) {
			next = lp->next;
			MyFree(lp->value.cp);
			free_link(lp);
		}

		MyFree(sptr->user->lopt);
		sptr->user->lopt = NULL;
		return;
	}

	lopt->starthash = hashnum;
	return ;
}

void send_user_joins(aClient *cptr, aClient *user)
{
	Link *lp;
	aChannel *chptr;
	int cnt = 0, len = 0, clen;
	char *mask;

	*buf = ':';
	strcpy(buf + 1, user->name);
	strcat(buf, " JOIN ");
	len = strlen(user->name) + 7;

	for (lp = user->user->channel; lp; lp = lp->next) {
		chptr = lp->value.chptr;
		if (*chptr->chname == '&') {
			continue;
		}

		mask = strchr(chptr->chname, ':');
		if (mask) {
			if (match(++mask, cptr->name)) {
				continue;
			}
		}

		clen = strlen(chptr->chname);
		if (clen > BUFSIZE - 7 - len) {
			if (cnt) {
				sendto_one(cptr, "%s", buf);
			}

			*buf = ':';
			strcpy(buf + 1, user->name);
			strcat(buf, " JOIN ");
			len = strlen(user->name) + 7;
			cnt = 0;
		}

		strcpy(buf + len, chptr->chname);
		cnt++;
		len += clen;

		if (lp->next) {
			len++;
			strcat(buf, ",");
		}
	}

	if (*buf && cnt) {
		sendto_one(cptr, "%s", buf);
	}

	return;
}

char *pretty_mask(char *mask)
{
	char *cp, *user, *host;

	if ((user = index((cp = mask), '!'))) {
		*user++ = '\0';
	}
	if ((host = rindex(user ? user : cp, '@'))) {
		*host++ = '\0';
		if (!user) {
			return (make_nick_user_host(NULL, cp, host));
		}
	}
	else if (!user && index(cp, '.')) {
		return (make_nick_user_host(NULL, NULL, cp));
	}
	return (make_nick_user_host(cp, user, host));
}
