/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/dbuf.c: data buffer management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: dbuf.c,v 1.4 2002/04/08 16:00:50 rageircd Exp $
 */

#include "dbuf.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include <stdlib.h>
#include "memory.h"

int DBufCount = 0;
int DBufUsedCount = 0;

static DBufBuffer *dbufFreeList = NULL;
static DBufBlock *dbuf_blocks = NULL;

void count_dbuf_memory(size_t *allocated, size_t *used)
{
	*allocated = DBufCount * sizeof(DBufBuffer);
	*used = DBufUsedCount * sizeof(DBufBuffer);
	return;
}

void dbuf_allocblock(int num)
{
	DBufBuffer *dfree, *dbp;
	DBufBlock *dblock;
	int i;

	dblock = (DBufBlock *)MyMalloc(sizeof(DBufBlock));
	dblock->num = num;
	dblock->next = dbuf_blocks;
	dbuf_blocks = dblock;

	dbp = (DBufBuffer *)MyMalloc(sizeof(DBufBuffer) * num);
	dblock->buf = dbp;
	dfree = dbp;

	for (i = 0; i < (num - 1); i++) {
		dfree->next = dfree + 1;
		dfree++;
	}
	dfree->next = dbufFreeList;
	dbufFreeList = dbp;
	DBufCount += num;
	return;
}

void dbuf_init()
{
	dbuf_allocblock(INITIAL_DBUFS);
	return;
}

static DBufBuffer *dbuf_alloc()
{
	DBufBuffer *db = dbufFreeList;

	if (!db) {
		dbuf_allocblock(INITIAL_DBUFS);
		db = dbufFreeList;
		if (!db) {
			return NULL;
		}
	}

	dbufFreeList = dbufFreeList->next;
	DBufUsedCount++;

	db->next = 0;
	db->start = db->end = db->data;
	return db;
}

static void dbuf_free(DBufBuffer *ptr)
{
	DBufUsedCount--;
	ptr->next = dbufFreeList;
	dbufFreeList = ptr;
	return;
}

static int dbuf_malloc_error(DBuf *dyn)
{
	DBufBuffer *db = dyn->head;

	dyn->length = 0;
	while (db) {
		dyn->head = db->next;
		dbuf_free(db);
		db = dyn->head;
	}
	dyn->tail = 0;
	return 0;
}

int dbuf_put(DBuf *dyn, const char *buf, size_t length)
{
	DBufBuffer **h, *d;
	int chunk;

	if (!dyn->length) {
		h = &(dyn->head);
	}
	else {
		h = &(dyn->tail);
	}
	dyn->length += length;

	for (; length > 0; h = &(d->next)) {
		d = *h;
		if (!d) {
			d = dbuf_alloc();
			if (!d) {
				return dbuf_malloc_error(dyn);
			}
			dyn->tail = d;
			*h = d;
		}
		chunk = (d->data + DBUF_SIZE) - d->end;
		if (chunk) {
			if (chunk > length) {
				chunk = length;
			}
			memcpy(d->end, buf, chunk);
			length -= chunk;
			buf += chunk;
			d->end += chunk;
		}
	}
	return 1;
}

char *dbuf_map(DBuf *dyn, size_t *length)
{
	if (!dyn->length) {
		*length = 0;
		return 0;
	}
	*length = dyn->head->end - dyn->head->start;
	return dyn->head->start;
}

void dbuf_delete(DBuf *dyn, size_t length)
{
	DBufBuffer *db;
	size_t chunk;

	if (length > dyn->length) {
		length = dyn->length;
	}
	while (length > 0) {
		db = dyn->head;
		if (!db) {
			break;
		}

		chunk = db->end - db->start;
		if (chunk > length) {
			chunk = length;
		}

		length -= chunk;
		dyn->length -= chunk;
		db->start += chunk;

		if (db->start == db->end) {
			dyn->head = db->next;
			dbuf_free(db);
		}
	}
	if (!dyn->head) {
		dyn->tail = 0;
		dyn->length = 0;
	}
	return;
}

size_t dbuf_get(DBuf *dyn, char *buf, size_t length)
{
	size_t moved = 0, chunk;
	const char *b;

	while (length > 0 && (b = dbuf_map(dyn, &chunk)) != 0) {
		if (chunk > length) {
			chunk = length;
		}
		memcpy(buf, b, chunk);
		dbuf_delete(dyn, chunk);
		buf += chunk;
		length -= chunk;
		moved += chunk;
	}
	return moved;
}

static size_t dbuf_flush(DBuf *dyn)
{
	DBufBuffer *db = dyn->head;

	if (!db) {
		return 0;
	}

	while (IsEol(*db->start)) {
		if (++db->start == db->end) {
			dyn->head = db->next;
			dbuf_free(db);

			db = dyn->head;
			if (!db) {
				dyn->tail = 0;
				dyn->length = 0;
				break;
			}
		}
		--dyn->length;
	}
	return dyn->length;
}

int dbuf_getmsg(DBuf *dyn, char *buf, size_t length)
{
	DBufBuffer *db;
	char *start, *end;
	size_t count, copied = 0;

	if (!dbuf_flush(dyn)) {
		return 0;
	}

	db = dyn->head;
	start = db->start;

	if (length > dyn->length) {
		length = dyn->length;
	}

	while (length > 0) {
		end = IRCD_MIN(db->end, start + length);
		while (start < end && !IsEol(*start)) {
			*buf++ = *start++;
		}

		count = start - db->start;
		if (start < end) {
			*buf = '\0';
			copied += count;
			dbuf_delete(dyn, copied);
			dbuf_flush(dyn);
			return copied;
		}

		db = db->next;
		if (!db) {
			break;
		}
		copied += count;
		length -= count;
		start = db->start;
	}
	return 0;
}
