/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/event.c: internal events management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: event.c,v 1.6 2002/04/08 16:00:50 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "h.h"
#include "memory.h"

Event EventList[MAXEVENTS];
static int event_count = 0;
static const char *last_event_run = NULL;
static time_t event_time_min = -1;

static int findEvent(void (*)(), void *);

void addEvent(const char *name, void (*func)(), void *arg, time_t when, int loop)
{
	int i;

	for (i = 0; i < event_count; i++) {
		if (!EventList[i].active) {
			break;
		}
	}
	if (i >= event_count) {
		event_count = i + 1;
	}

	EventList[i].name = name;
	EventList[i].func = func;
	EventList[i].arg = arg;
	EventList[i].when = timeofday + when;
	EventList[i].frequency = when;
	EventList[i].loop = loop;
	EventList[i].active = 1;

	if (EventList[i].when < event_time_min || event_time_min == -1) {
		event_time_min = EventList[i].when;
	}
	return;
}

void delEvent(void (*func)(), void *arg)
{
	int i = findEvent(func, arg);
	if (i == -1) {
		return;
	}

	EventList[i].name = NULL;
	EventList[i].func = NULL;
	EventList[i].arg = NULL;
	EventList[i].loop = 0;
	EventList[i].active = 0;
	return;
}

void checkEvent(void)
{
	int i;

	if (event_count == 0) {
		return;
	}
	for (i = 0; i < event_count; i++) {
		if (EventList[i].active && EventList[i].when <= timeofday) {
			last_event_run = EventList[i].name;
			EventList[i].func(EventList[i].arg);
			if (EventList[i].loop) {
				EventList[i].when = timeofday + EventList[i].frequency;
			}
			else {
				delEvent(EventList[i].func, EventList[i].arg);
			}
			event_time_min = -1;
		}
	}
	return;
}

time_t next_checkEvent(void)
{
	int i;

	if (event_count == 0) {
		return timeofday + 1;
	}
	else if (event_time_min == -1) {
		for (i = 0; i < event_count; i++) {
			if (EventList[i].active && (EventList[i].when < event_time_min || event_time_min == -1)) {
				event_time_min = EventList[i].when;
			}
		}
	}
	return event_time_min;
}

void show_events(aClient *cptr)
{
	int i, shown = 0;

	if (last_event_run != NULL) {
		sendto_one(cptr, ":%s NOTICE %s :*** The last event to be executed was: %s", me.name,
			cptr->name, last_event_run);
	}
	for (i = 0; i < event_count; i++) {
		if (EventList[i].active) {
			sendto_one(cptr, ":%s NOTICE %s :*** %s (next execution in %d seconds, %s)", me.name,
				cptr->name, EventList[i].name, (int)(EventList[i].when - timeofday),
				EventList[i].loop ? "continuous" : "once only");
			shown++;
		}
	}
	if (shown == 0) {
		sendto_one(cptr, ":%s NOTICE %s :*** There are currently no pending events.", me.name,
			cptr->name);
	}
	else {
		sendto_one(cptr, ":%s NOTICE %s :*** Displayed a total of %d pending event%s.", me.name,
			cptr->name, shown, shown == 1 ? "" : "s");
	}
	return;
}

static int findEvent(void (*func)(), void *arg)
{
	int i;
	for (i = 0; i < event_count; i++) {
		if (EventList[i].active && EventList[i].func == func && EventList[i].arg == arg) {
			return i;
		}
	}
	return -1;
}

