/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/flud.c: fludbot detection and provention system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: flud.c,v 1.3 2002/04/08 16:00:50 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "h.h"
#include "blalloc.h"
#include "memory.h"
#if defined(HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef FLUD

BlockHeap *fludbot_heap;

int flud_num = FLUD_NUM;
int flud_time = FLUD_TIME;
int flud_block = FLUD_BLOCK;

struct fludbot *remove_fluder_reference(struct fludbot **, aClient*);
Link *remove_fludee_reference(Link **, void *);
void free_fluders(aClient *, aChannel *);
void free_fludees(aClient *);

void fludbot_heap_garbage_collect()
{
	BlockHeapGarbageCollect(fludbot_heap);
	return;
}

void init_fludbot()
{
	fludbot_heap = BlockHeapCreate((size_t)sizeof(struct fludbot), MAXCONNECTIONS);
	addEvent("fludbot_heap_garbage_collect", fludbot_heap_garbage_collect, NULL, FLUDBOT_HEAP_GARBAGE_COLLECT_FREQ, 1);
	return;
}

struct fludbot *remove_fluder_reference(struct fludbot **fluders, aClient *fluder)
{
	struct fludbot *current, *prev = NULL, *next = NULL;

	for (current = *fluders; current; current = next) {
		next = current->next;
		if (current->fluder == fluder) {
			if (prev) {
				prev->next = current->next;
			}
			else {
				*fluders = current->next;
			}
			if (BlockHeapFree(fludbot_heap, current)) {
				ircdlog(LOG_ERROR, "BlockHeapFree failed in remove_fluder_reference [current=%lX]",
					current);
			}
		}
		else {
			prev = current;
		}
	}
        return *fluders;
}

Link *remove_fludee_reference(Link **fludees, void *fludee)
{
	Link *current, *prev = NULL, *next = NULL;

	for (current = *fludees; current; current = next) {
		next = current->next;
		if (current->value.cptr == (aClient *)fludee) {
			if (prev) {
				prev->next = current->next;
			}
			else {
				*fludees = current->next;
			}
			if (BlockHeapFree(link_heap, current)) {
				ircdlog(LOG_ERROR, "BlockHeapFree failed in remove_fludee_reference [current=%lX]",
					current);
			}
		}
		else {
			prev = current;
		}
	}
	return *fludees;
}

int check_for_flud(aClient *fluder, aClient *cptr, aChannel *chptr)
{
	time_t now;
	struct fludbot *current, *prev = NULL, *next = NULL;
	int blocking, count = 0, found = 0;
	Link *newfludee;

	if (!flud_block || (!cptr && !chptr)) {
		return 0;
	}
	if (cptr && !MyFludConnect(cptr)) {
		ircdlog(LOG_ERROR, "encountered non-local client %s in check_for_flud",
			get_client_name(cptr, FALSE));
		return 0;
	}

	time(&now);
	blocking = cptr ? cptr->fludblock : chptr->fludblock > now - flud_block;
	if (!blocking) {
		for (current = cptr ? cptr->fluders : chptr->fluders; current; current = next) {
			next = current->next;
			if (current->last_msg < now - flud_time) {
				remove_fludee_reference(&current->fluder->fludees, cptr ? (void *)cptr : (void *)chptr);
				if (prev) {
					prev->next = current->next;
				}
				else {
					cptr ? cptr->fluders : chptr->fluders = current->next;
				}
				if (BlockHeapFree(fludbot_heap, current)) {
					ircdlog(LOG_ERROR, "BlockHeapFree failed in check_for_flud [current=%lX]",
						current);
				}
			}
			else {
				prev = current;
			}
		}
	}

	count = found = 0;
	for (current = cptr ? cptr->fluders : chptr->fluders; current; current = current->next) {
		if (current->fluder == fluder) {
			current->last_msg = now;
			current->count++;
			found = 1;
		}
		if (current->first_msg < now - flud_time) {
			count++;
		}
		else {
			count += current->count;
		}
	}
	if (!found) {
		current = (struct fludbot *)BlockHeapAlloc(fludbot_heap);
		if (!current) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in check_for_flud (current "
				"allocation failed)");
			outofmemory();
		}
		current->fluder = fluder;
		current->count = 1;
		current->first_msg = current->last_msg = now;
		current->next = cptr ? cptr->fluders : chptr->fluders;
		cptr ? cptr->fluders : chptr->fluders = current;
		count++;

		newfludee = (Link *)BlockHeapAlloc(link_heap);
		if (!newfludee) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in check_for_flud (newfludee "
				"allocation failed)");
			outofmemory();
		}
		newfludee->flags = cptr ? 0 : 1;
		if (cptr) {
			newfludee->value.cptr = cptr;
		}
		else {
			newfludee->value.chptr = chptr;
		}
		newfludee->next = fluder->fludees;
		fluder->fludees = newfludee;

		if (blocking) {
			sendto_ops_lev(FLOOD_LEV, "Flooder %s (%s@%s) on %s target: %s", fluder->name,
				fluder->user->username,
				IsMasked(fluder) ? fluder->user->maskedhost : fluder->user->host,
				fluder->user->server, cptr ? cptr->name : chptr->chname);
		}
	}
	if (!blocking && found > flud_num) {
		blocking = 1;
		ircstp->is_flud++;
		if (cptr) {
			sendto_one(cptr, ":%s NOTICE %s :*** Notice -- Server flood protection for %s activated.",
				me.name, cptr->name, cptr->name);
		}
		else {
			sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :*** Notice -- Server flood protection "
				"for %s activated.", me.name, chptr->chname, chptr->chname);
		}

		for (current = cptr ? cptr->fluders : chptr->fluders; current; current = current->next) {
			sendto_ops_lev(FLOOD_LEV, "Flooder %s (%s@%s) on %s target: %s", current->fluder->name,
				current->fluder->user->username,
				IsMasked(current->fluder) ? current->fluder->user->maskedhost : current->fluder->user->host,
				current->fluder->user->server, cptr ? cptr->name : chptr->chname);
		}
	}
	if (blocking) {
		cptr ? cptr->fludblock : chptr->fludblock = now;
	}
	return blocking;
}

void free_fluders(aClient *cptr, aChannel *chptr)
{
	struct fludbot *fluders, *next = NULL;

	if (!cptr && !chptr) {
		return;
	}
	if (cptr && !MyFludConnect(cptr)) {
		ircdlog(LOG_ERROR, "encountered non-local client %s in free_fluders",
			get_client_name(cptr, FALSE));
		return;
	}
	for (fluders = cptr ? cptr->fluders : chptr->fluders; fluders; fluders = next) {
		next = fluders->next;
		remove_fludee_reference(&fluders->fluder->fludees, cptr ? (void *)cptr : (void *)chptr);
		if (BlockHeapFree(fludbot_heap, fluders)) {
			ircdlog(LOG_ERROR, "BlockHeapFree failed in free_fluders [fluders=%lX]", fluders);
		}
	}
	return;
}

void free_fludees(aClient *cptr)
{
	Link *fludees, *next = NULL;

	if (!cptr) {
		return;
	}
	for (fludees = cptr->fludees; fludees; fludees = next) {
		next = fludees->next;
		if (fludees->flags) {
			remove_fluder_reference(&fludees->value.chptr->fluders, cptr);
		}
		else {
			if (!MyFludConnect(fludees->value.cptr)) {
				ircdlog(LOG_ERROR, "encountered non-local client %s in free_fludees",
					get_client_name(fludees->value.cptr, FALSE));
			}
			else {
				remove_fluder_reference(&fludees->value.cptr->fluders, cptr);
			}
		}
		if (BlockHeapFree(link_heap, fludees)) {
			ircdlog(LOG_ERROR, "BlockHeapFree failed in free_fludees [fludees=%lX]", fludees);
		}
	}
	return;
}

#endif /* FLUD */
