/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/iphash.c: ip hash system
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: iphash.c,v 1.6 2002/04/08 16:00:50 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "inet.h"
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/wait.h>
#ifdef __hpux
# include "inet.h"
#endif
#if defined(AIX) || defined(SVR3)
# include <time.h>
#endif
#include <signal.h>
#include "h.h"
#include "memory.h"

#define IP_HASH_SIZE 0x1000

IP_ENTRY *ip_hash_table[IP_HASH_SIZE];
static IP_ENTRY *free_ip_entries;

void init_ip_hash(void);
IP_ENTRY *find_or_add_ip(unsigned long);
int hash_ip(unsigned long);

/*
 * init_ip_hash()
 *
 * INPUT:        none
 * OUTPUT:       none
 * SIDE EFFECTS: clear the ip hash table
 */
void clear_ip_hash_table()
{
	void *block_IP_ENTRIES;
	IP_ENTRY *new_IP_ENTRY, *last_IP_ENTRY;
	int size, n_left_to_allocate = MAXCONNECTIONS;

	/*
	 * ok. if the sizeof the structure isn't aligned with that of the
	 * smallest guaranteed valid pointer (void *), then align it ya. you
	 * could just turn 'size' into a #define. do it. :-)
	 *
	 * -Dianora
	 */

	size = sizeof(IP_ENTRY) + (sizeof(IP_ENTRY) & (sizeof(void *) - 1));
	block_IP_ENTRIES = (void *)MyMalloc((size * n_left_to_allocate));
	free_ip_entries = (IP_ENTRY *)block_IP_ENTRIES;
	last_IP_ENTRY = free_ip_entries;

	/*
	 * *shudder pointer arithmatic
	 */
	while (--n_left_to_allocate) {
		block_IP_ENTRIES = (void *)((unsigned long)block_IP_ENTRIES + (unsigned long)size);
		new_IP_ENTRY = (IP_ENTRY *)block_IP_ENTRIES;
		last_IP_ENTRY->next = new_IP_ENTRY;
		new_IP_ENTRY->next = (IP_ENTRY *)NULL;
		last_IP_ENTRY = new_IP_ENTRY;
	}
	memset((char *)ip_hash_table, '\0', sizeof(ip_hash_table));
}

/*
 * find_or_add_ip()
 *
 * INPUTS:       unsigned long IP address value
 * OUTPUTS:      pointer to an IP_ENTRY element
 * SIDE EFFECTS: If the IP # was not found, a new IP_ENTRY is created, and
 *               the IP count is set to 0.
 *
 * Minor modifications ... -TD (September 2001)
 */
IP_ENTRY *find_or_add_ip(unsigned long ip_in)
{
	int hash_index;
	IP_ENTRY *ptr = ip_hash_table[hash_index = hash_ip(ip_in)];
	IP_ENTRY *newptr = (IP_ENTRY *)NULL;

	while (ptr) {
		if (ptr->ip == ip_in) {
			return (ptr);
		}
		else {
			ptr = ptr->next;
		}
	}

	ptr = ip_hash_table[hash_index];
	if (ptr != (IP_ENTRY *)NULL) {
		if (free_ip_entries == (IP_ENTRY *)NULL) {
			sendto_ops("iphash.c free_ip_entries was found NULL in find_or_add_ip()");
			sendto_ops("rehash_ip was done, this is an error.");
			sendto_ops("Please report this to the RageIRCd Development Team! "
				"rageircd-bugs@lists.sourceforge.net");
			if (free_ip_entries == (IP_ENTRY *)NULL) {
				outofmemory();
			}
		}

		newptr = ip_hash_table[hash_index] = free_ip_entries;
		free_ip_entries = newptr->next;
		newptr->ip = ip_in;
		newptr->count = 0;
		newptr->next = ptr;
		return (newptr);
	}
	else {
		if (free_ip_entries == (IP_ENTRY *)NULL) {
			sendto_ops("iphash.c free_ip_entries was found NULL in find_or_add_ip()");
			sendto_ops("rehash_ip was done, this is an error.");
			sendto_ops("Please report this to the RageIRCd Development Team! "
				"rageircd-bugs@lists.sourceforge.net");
			if (free_ip_entries == (IP_ENTRY *)NULL) {
				outofmemory();
			}
		}

		ptr = ip_hash_table[hash_index] = free_ip_entries;
		free_ip_entries = ptr->next;
		ptr->ip = ip_in;
		ptr->count = 0;
		ptr->next = (IP_ENTRY *)NULL;
		return (ptr);
	}
}

/*
 * remove_one_ip()
 *
 * INPUTS:       unsigned long IP address value
 * OUTPUTS:      none
 * SIDE EFFECTS: ip address listed, is looked up in ip hash table and number of
 *               ip addresses for that ip decremented. if ip # count reaches 0,
 *               the IP_ENTRY is returned to the free_ip_entries link list.
 */
void remove_one_ip(unsigned long ip_in)
{
	int hash_index;
	IP_ENTRY *last_ptr, *ptr, *old_free_ip_entries;

	last_ptr = ptr = ip_hash_table[hash_index = hash_ip(ip_in)];

	while (ptr) {
		if (ptr->ip == ip_in) {
			if (ptr->count != 0) {
				ptr->count--;
			}
			if (ptr->count == 0) {
				if (ip_hash_table[hash_index] == ptr) {
					ip_hash_table[hash_index] = ptr->next;
				}
				else {
					last_ptr->next = ptr->next;
				}

				if (free_ip_entries != (IP_ENTRY *)NULL) {
					old_free_ip_entries = free_ip_entries;
					free_ip_entries = ptr;
					ptr->next = old_free_ip_entries;
				}
				else {
					free_ip_entries = ptr;
					ptr->next = (IP_ENTRY *)NULL;
				}
			}
			return ;
		}
		else {
			last_ptr = ptr;
			ptr = ptr->next;
		}
	}
	sendto_ops("iphash.c couldn't find ip# in hash table in remove_one_ip()");
	sendto_ops("Please report this to the RageIRCd Development Team! "
		"rageircd-bugs@lists.sourceforge.net");
	return;
}

/*
 * hash_ip()
 *
 * INPUTS:       unsigned long ip address
 * OUTPUTS:      integer value used as index into hash table
 * SIDE EFFECTS: none
 */
int hash_ip(unsigned long ip_in)
{
	int hash;

	ip_in = ntohl(ip_in);
	hash = ((ip_in >>= 12) + ip_in) & (IP_HASH_SIZE - 1);
	return (hash);
}

/*
 * count_ip_hash()
 *
 * INPUTS:       pointer to counter of number of ips hashed, pointer to memory
 *               used for ip hash
 * OUTPUTS:      returned via pointers input
 * SIDE EFFECTS: none
 *
 * number of hashed ip #'s is counted up, plus the amount of memory used in
 * the hash. Added so s_debug.c coudl check memory usage in here. -Dianora
 */
void count_ip_hash(int *number_ips_stored, u_long *mem_ips_stored)
{
	IP_ENTRY *ip_hash_ptr;
	int i;

	*number_ips_stored = 0;
	*mem_ips_stored = 0;

	for (i = 0; i < IP_HASH_SIZE; i++) {
		ip_hash_ptr = ip_hash_table[i];
		while (ip_hash_ptr) {
			*number_ips_stored = *number_ips_stored + 1;
			*mem_ips_stored = *mem_ips_stored + sizeof(IP_ENTRY);

			ip_hash_ptr = ip_hash_ptr->next;
		}
	}
}

/*
 * rehash_ip_hash()
 *
 * This function clears the ip hash table, then re-enters all the ip's
 * found in local clients.
 *
 * NB. This should never have to be called, and hopefully, we can
 *     remove this function in future versions of hybrid. i.e. if
 *     everything is working right, there should never ever be a
 *     case where an IP is still in the ip_hash table and the
 *     corresponding client isn't.
 */
void rehash_ip_hash()
{
	IP_ENTRY *ip_hash_ptr, *tmp_ip_hash_ptr, *old_free_ip_entries;
	int i;

	for (i = 0; i < IP_HASH_SIZE; i++) {
		ip_hash_ptr = ip_hash_table[i];
		while (ip_hash_ptr) {
			tmp_ip_hash_ptr = ip_hash_ptr->next;
			if (free_ip_entries) {
				old_free_ip_entries = free_ip_entries;
				free_ip_entries = ip_hash_ptr;
				ip_hash_ptr->next = old_free_ip_entries;
			}
			else {
				free_ip_entries = ip_hash_ptr;
				ip_hash_ptr->next = (IP_ENTRY *)NULL;
			}
			ip_hash_ptr = tmp_ip_hash_ptr;
		}
		ip_hash_table[i] = (IP_ENTRY *)NULL;
	}

	for (i = highest_fd; i >= 0; i--) {
		if (local[i] && MyClient(local[i])) {
			if (local[i]->fd >= 0 && (local[i]->flags & FLAGS_IPHASH)) {
				ip_hash_ptr = find_or_add_ip(local[i]->ip.s_addr);
				ip_hash_ptr->count++;
			}
		}
	}
}
