/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/ircd.c: core ircd functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: ircd.c,v 1.61 2002/04/08 16:00:50 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "memory.h"
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HPUX
#define _KERNEL
#endif
#include <sys/resource.h>
#ifdef HPUX
#undef _KERNEL
#endif
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <fcntl.h>
#include "inet.h"
#include "h.h"
#include "patchlevel.h"
#include "dh.h"
#include "fdlist.h"
#include "setup.h"

fdlist      serv_fdlist;
fdlist      oper_fdlist;
fdlist      listen_fdlist;

#ifndef NO_PRIORITY
fdlist      busycli_fdlist;	/* high-priority clients */
#endif
fdlist      default_fdlist;	/* just the number of the entry */

struct Counter Count;
int         R_do_dns, R_fin_dns, R_fin_dnsc, R_fail_dns, R_do_id, R_fin_id, R_fail_id;

time_t      	 NOW;
time_t		 last_stat_save;
aClient     	 me;		/* That's me */
aClient    	*client = &me;	/* Pointer to beginning of Client list */

void        	server_reboot();
void        	restart(char *);
void		save_stats();
static void 	open_debugfile(), setup_signals();
static void 	io_loop();

/* externally needed functions */

extern int	unload_all_modules();
extern void	build_chmode_string();
extern void	build_usermode_string();
extern void init_clients();
extern void init_channels();
extern void init_link();
#ifdef FLUD
extern void init_fludbot();
#endif
extern void	init_Modules();
extern void	init_config2();
extern void	run_listeners();
extern void	init_whowas();
extern void 	init_fdlist(fdlist *);	    /* defined in fdlist.c */
extern void	dbuf_init();	            /* defined in dbuf.c */
#ifdef HAVE_THROTTLE_ON
extern void throttle_init(void);
extern void throttle_timer(time_t);
#endif

char      **myargv;
int         debuglevel = -1;	   /* Server debug level */
int         bootopt = 0;	   /* Server boot option flags */
char       *debugmode = "";	   /* -"-    -"-   -"-  */
char       *sbrk0;		   /* initial sbrk(0) */
static int  dorehash = 0;
static char *ircd_path = IRCD_PREFIX;
int         noisy_htm = NOISY_HTM; /* Is high traffic mode noisy or not? */ 
time_t      nextconnect = 1;	   /* time for next try_connections call */
time_t      nextping = 1;	   /* same as above for check_pings() */
time_t      nextdnscheck = 0;	   /* next time to poll dns to force timeout */
time_t      nextexpire = 1;	   /* next expire run on the dns cache */

#ifdef	PROFIL
extern      etext();

VOIDSIG s_monitor() 
{
	static int  mon = 0;
# ifdef	POSIX_SIGNALS
	struct sigaction act;
# endif
   	(void) moncontrol(mon);
  	mon = 1 - mon;
# ifdef	POSIX_SIGNALS
   	act.sa_handler = s_rehash;
   	act.sa_flags = 0;
   	(void) sigemptyset(&act.sa_mask);
   	(void) sigaddset(&act.sa_mask, SIGUSR1);
   	(void) sigaction(SIGUSR1, &act, NULL);
# else
   	(void) signal(SIGUSR1, s_monitor);
# endif
}
#endif /* PROFIL */

VOIDSIG s_die()
{
	unload_all_modules();
	dump_connections(me.fd);
	ircdlog(LOG_DEFAULT, "server killed by SIGTERM");
	exit(-1);
}

static  VOIDSIG s_rehash() 
{
#ifdef	POSIX_SIGNALS
	struct sigaction act;
#endif
   	dorehash = 1;
#ifdef	POSIX_SIGNALS
   	act.sa_handler = s_rehash;
   	act.sa_flags = 0;
   	(void) sigemptyset(&act.sa_mask);
   	(void) sigaddset(&act.sa_mask, SIGHUP);
   	(void) sigaction(SIGHUP, &act, NULL);
#else
   	(void) signal(SIGHUP, s_rehash);	/* sysV -argv */
#endif
}

void restart(char *mesg) 
{
   static int  was_here = NO;	/* redundant due to restarting flag below */
   if (was_here)
      abort();
   was_here = YES;
   ircdlog(LOG_DEFAULT, "Server restarting: %s, sbrk(0)-etext: %d", mesg, (u_int)sbrk((size_t)0) - (u_int)sbrk0);
   server_reboot();
}

VOIDSIG s_restart() 
{
	static int  restarting = 0;
	
   ircdlog(LOG_DEFAULT, "Server restarting: SIGINT");
   if (restarting == 0) 
   {
      /* Send (or attempt to) a dying scream to oper if present */
      restarting = 1;
      server_reboot();
   }
}

void server_reboot() 
{
   int     i;
   sendto_ops("Aieeeee!!!  Restarting server... sbrk(0)-etext: %d",
				  (u_int) sbrk((size_t) 0) - (u_int) sbrk0);
	
   Debug((DEBUG_NOTICE, "Restarting server..."));
   dump_connections(me.fd);
   /*
    * fd 0 must be 'preserved' if either the -d or -i options have
    * been passed to us before restarting.
    */
   for (i = 3; i < MAXCONNECTIONS; i++)
      (void) close(i);

   if (!(bootopt & (BOOT_TTY | BOOT_DEBUG)))
      (void) close(2);

   (void) close(1);

   if ((bootopt & BOOT_CONSOLE) || isatty(0))
      (void) close(0);

   if (!(bootopt & BOOT_OPER))
      (void) execv(IRCD_EXE, myargv);

   ircdlog(LOG_ERROR, "execv(%s, %s) failed", IRCD_EXE, myargv[0]);
   Debug((DEBUG_FATAL, "Couldn't restart server: %s", strerror(errno)));
   unload_all_modules();
   exit(-1);
}

/*
 * try_connections 
 * 
 *      Scan through configuration and try new connections. 
 *   Returns  the calendar time when the next call to this 
 *      function should be made latest. (No harm done if this 
 *      is called earlier or later...)
 */
static time_t try_connections(time_t currenttime)
{
	ConfigItem_link *link;
	ConfigItem_ban *ban;
	aClient *cptr;
	int connfreq;
	time_t next = 0;

	Debug((DEBUG_NOTICE, "Connection check at : %s", myctime(currenttime)));
	for (link = config_link; link; link = (ConfigItem_link *)link->next) {
		if (!(link->options & LINK_AUTOCONNECT)) {
			continue;
		}
		if (link->hold > currenttime) {
			if (next > link->hold || next == 0) {
				next = link->hold;
			}
			continue;
		}

		connfreq = link->class->connfreq;
		link->hold = currenttime + connfreq;
		cptr = find_name(link->servername, (aClient *)NULL);

		if (!cptr && link->class->clients < link->class->maxclients) {
			ban = find_ban(BAN_SERVER, link->servername);
			if (ban) {
				sendto_gnotice("Passing autoconnect for banned server %s (%s).",
					link->servername, ban->reason);
				continue;
			}
			if (connect_server(link, (aClient *)NULL, (struct hostent *)NULL) == 0) {
				sendto_gnotice("Connection to %s activated.", link->servername);
			}
		}
		if (next > link->hold || next == 0) {
			next = link->hold;
		}
	}
	Debug((DEBUG_NOTICE, "Next connection check at : %s", myctime(next)));
	return (next);
}

/* dianora's code in the new checkpings is slightly wasteful.
 * however, upon inspection (thanks seddy), when we close a connection,
 * the ordering of local[i] is NOT reordered; simply local[highest_fd] becomes
 * local[i], so we can just i--;  - lucas
 */
time_t check_pings(time_t currenttime)
{
	aClient *cptr;
	time_t oldest = 0;
	int ping = 0, i;

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsMe(cptr) || IsLog(cptr)) {
			continue;
		}
		if (cptr->flags & FLAGS_DEADSOCKET) {
			exit_client(cptr, cptr, &me, (cptr->flags & FLAGS_SENDQEX) ? "SendQ exceeded" : "Dead socket");
			i--;
			continue;
		}

		ping = IsRegistered(cptr) ? cptr->pingval : CONNECTTIMEOUT;
		if (ping < currenttime - cptr->lasttime) {
			if ((cptr->flags & FLAGS_PINGSENT && currenttime - cptr->lasttime >= ping * 2) ||
			  (!IsRegistered(cptr) && currenttime - cptr->since >= ping)) {
				if (!IsRegistered(cptr) && (DoingDNS(cptr) || DoingAuth(cptr))) {
					if (cptr->authfd >= 0) {
						close(cptr->authfd);
						cptr->authfd = -1;
						cptr->count = 0;
						*cptr->buffer = '\0';
					}

					if (ConfigSet.showheaders) {
						if (DoingDNS(cptr)) {
							send(cptr->fd, REPORT_FAIL_DNS, R_fail_dns, 0);
						}
						if (DoingAuth(cptr)) {
							send(cptr->fd, REPORT_FAIL_ID, R_fail_id, 0);
						}
					}

					Debug((DEBUG_NOTICE, "DNS/AUTH timeout %s", get_client_name(cptr, TRUE)));
					del_queries((char *)cptr);
					ClearAuth(cptr);
					ClearDNS(cptr);
					SetAccess(cptr);
					cptr->since = currenttime;
					continue;
				}

				if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr)) {
					char *path = get_client_name(cptr, HIDEME);
					sendto_gnotice("No responce from %s, closing link.", path);
					sendto_serv_butone(cptr, ":%s GNOTICE :No responce from %s, closing link.",
						me.name, path);
				}

				exit_client(cptr, cptr, &me, "Ping timeout");
				i--;
				continue;
			}
			else if (!(cptr->flags & (FLAGS_PINGSENT|FLAGS_BURST))) {
				cptr->flags |= FLAGS_PINGSENT;
				cptr->lasttime = currenttime - ping;
				sendto_one(cptr, "PING :%s", me.name);
			}
		}

		if (IsUnknown(cptr)) {
			if (cptr->firsttime ? (timeofday - cptr->firsttime) > 100 : 0) {
				exit_client(cptr, cptr, &me, "Connection timed out");
			}
		}
	}
	oldest = currenttime + 9;
	Debug((DEBUG_NOTICE, "Next check_ping() call at: %s, %d %d %d", myctime(oldest), ping,
		oldest, currenttime));
	return oldest;
}

/*
 * * bad_command 
 *    This is called when the commandline is not acceptable. 
 *    Give error message and exit without starting anything.
 */
static int
bad_command()
{
   (void) printf(
		   "Usage: ircd %s[-h servername] [-x loglevel] [-s] [-t]\n",
#ifdef CMDLINE_CONFIG
		   "[-f config] "
#else
		   ""
#endif
      );
   (void) printf("Server not started\n\n");
   return (-1);
}
#ifndef TRUE
#define TRUE 1
#endif

/*
 * code added by mika nystrom (mnystrom@mit.edu) 
 * this flag is used to signal globally that the server is heavily
 * loaded, something which can be taken into account when processing
 * e.g. user commands and scheduling ping checks
 * Changed by Taner Halicioglu (taner@CERF.NET) 
 */

int lifesux = 1;
int LRV = LOADRECV;
int LCF = LOADCFREQ;
int currlife = 0;
int HTMLOCK = NO;

char REPORT_DO_DNS[256], REPORT_FIN_DNS[256], REPORT_FIN_DNSC[256], REPORT_FAIL_DNS[256];
char REPORT_DO_ID[256], REPORT_FIN_ID[256], REPORT_FAIL_ID[256];

FILE *dumpfp=NULL;

int
main(int argc, char *argv[])
{
   uid_t         uid, euid;
   struct rlimit corelim;

   if ((timeofday = time(NULL)) == -1) 
   {
      (void) fprintf(stderr, "ERROR: Clock Failure (%d)\n", errno);
      exit(errno);
   }

   corelim.rlim_cur = corelim.rlim_max = RLIM_INFINITY;
   if (setrlimit(RLIMIT_CORE, &corelim)) {
	printf("unlimit core size failed; errno = %d\n", errno);
   }
	
   build_version();

   Count.server = 1;		/* us */
   Count.oper = 0;
   Count.chan = 0;
   Count.local = 0;
   Count.total = 0;
   Count.invisi = 0;
   Count.unknown = 0;
   Count.max_loc = 0;
   Count.max_tot = 0;
   Count.today = 0;
   Count.weekly = 0;
   Count.monthly = 0;
   Count.yearly = 0;
   Count.start = NOW;
   Count.day = NOW;
   Count.week = NOW;
   Count.month = NOW;
   Count.year = NOW;

   /*
    * this code by mika@cs.caltech.edu 
    * it is intended to keep the ircd from being swapped out. BSD
    * swapping criteria do not match the requirements of ircd
    */
	
#ifdef INITIAL_DBUFS
   dbuf_init();			/* set up some dbuf stuff to control paging */
#endif

   sbrk0 = (char *) sbrk((size_t) 0);
   uid = getuid();
   euid = geteuid();
#ifdef	PROFIL
   (void) monstartup(0, etext);
   (void) moncontrol(1);
   (void) signal(SIGUSR1, s_monitor);
#endif
	
   myargv = argv;
   (void) umask(077);		/* better safe than sorry --SRB  */
   memset((char *) &me, '\0', sizeof(me));
	
   setup_signals();

   /*
    * * All command line parameters have the syntax "-fstring"  or "-f
    * string" (e.g. the space is optional). String may  be empty. Flag
    * characters cannot be concatenated (like "-fxyz"), it would
    * conflict with the form "-fstring".
    */
   while (--argc > 0 && (*++argv)[0] == '-') 
   {
	char       *p = argv[0] + 1;
	int         flag = *p++;
		
        if (flag == '\0' || *p == '\0') 
	{
	   if (argc > 1 && argv[1][0] != '-') 
	   {
		p = *++argv;
		argc -= 1;
	   }
	   else
		p = "";
	   }
		
      switch (flag) 
      {
		 case 'a':
			bootopt |= BOOT_AUTODIE;
			break;
		 case 'c':
			bootopt |= BOOT_CONSOLE;
			break;
		 case 'q':
			bootopt |= BOOT_QUICK;
			break;
		 case 'd':
			(void) setuid((uid_t) uid);
			ircd_path = p;
			break;
		 case 'o':		/* Per user local daemon... */
			(void) setuid((uid_t) uid);
			bootopt |= BOOT_OPER;
			break;
		 case 'h':
			strncpyzt(me.name, p, sizeof(me.name));
			break;
		 case 's':
			bootopt |= BOOT_STDERR;
			break;
		 case 't':
			(void) setuid((uid_t) uid);
			bootopt |= BOOT_TTY;
			break;
		 case 'v':
			(void) printf("ircd %s\n", version);
			exit(0);
		 case 'x':
#ifdef	DEBUGMODE
			(void) setuid((uid_t) uid);
			debuglevel = atoi(p);
			debugmode = *p ? p : "0";
			bootopt |= BOOT_DEBUG;
			break;
#else
			(void) fprintf(stderr,
				"%s: DEBUGMODE must be defined for -x y\n",
								myargv[0]);
			exit(0);
#endif
		 default:
			bad_command();
			break;
      }
   }
	
   if (chdir(ircd_path)) 
   {
      perror("chdir");
      fprintf(stderr, "Please ensure that your configuration directory exists and is accessable.\n");
      exit(-1);
   }
   if ((uid != euid) && !euid) 
   {
      (void) fprintf(stderr,
	"ERROR: do not run ircd setuid root. Make it setuid a normal user.\n");
      exit(-1);
   }
	
   if (argc > 0)
	  return bad_command();	/* This should exit out  */

#ifdef HAVE_OPENSSL_ON
   if (dh_init() == -1) {
	fprintf(stderr, "Error initializing OpenSSL encryption layer - RageIRCd not started.\n");
	return 0;
   }
#endif

   chmod(CONFIG_FILE, 0600);

   init_CommandHash();
   init_HookTable();
   init_Modules();
   init_config2();
   build_chmode_string();
   build_usermode_string();

   clear_client_hash_table();
   clear_channel_hash_table();
   clear_scache_hash_table();	/* server cache name table */
   clear_ip_hash_table();	/* client host ip hash table */

   init_clients();
   init_channels();
   init_link();
#ifdef HAVE_THROTTLE_ON
   throttle_init();
#endif
#ifdef FLUD
   init_fludbot();
#endif
#ifndef STATIC_MODULES
   init_whowas();
#endif
   initstats();
   init_send();
   NOW = time(NULL);
   open_debugfile();
   NOW = time(NULL);
   init_fdlist(&serv_fdlist);
   init_fdlist(&oper_fdlist);
   init_fdlist(&listen_fdlist);
	
#ifndef NO_PRIORITY
   init_fdlist(&busycli_fdlist);
#endif
	
   init_fdlist(&default_fdlist); {
	  int i;
	  for (i = MAXCONNECTIONS + 1; i > 0; i--) {
		  default_fdlist.entry[i] = i - 1;
	  }
   }

   if ((timeofday = time(NULL)) == -1) 
   {
      ircdlog(LOG_ERROR, "Clock Failure (%d), TS can be corrupted", errno);
      sendto_ops("Clock Failure (%d), TS can be corrupted", errno);
   }

   (void)init_sys();

   strncpyzt(me.name, config_servinfo->name, sizeof(me.name));
   strncpyzt(me.info, config_servinfo->info, sizeof(me.info));
   strncpyzt(me.sockhost, config_listen->ip, sizeof(me.sockhost) - 1);

   me.flags = FLAGS_LISTEN;
   me.port = config_listen->port;
   me.class = (ConfigItem_class *)config_listen;
   me.fd = -1;

   if (ConfigSet.hub) {
	lifesux = 0;
   }

   if (inetport(&me, config_listen->ip, config_listen->port)) {
	if (bootopt & BOOT_STDERR) {
		fprintf(stderr, "Couldn't bind to primary port %d\n", config_listen->port);
	}
	Debug((DEBUG_FATAL, "Couldn't bind to primary port %d", config_listen->port));
	exit(1);
   }

   config_listen->options |= (LISTEN_BOUND|LISTEN_MAINPORT);
   config_listen->clients++;
   run_listeners();

   me.hopcount = 0;
   me.authfd = -1;
   me.confs = NULL;
   me.next = NULL;
   me.user = NULL;
   me.from = &me;
   SetMe(&me);
   make_server(&me);
   me.serv->up = me.name;
   me.lasttime = me.since = me.firsttime = NOW;
   (void) add_to_client_hash_table(me.name, &me);
	
   /* We don't want to calculate these every time they are used :) */
	
   sprintf(REPORT_DO_DNS, REPORT_DO_DNS_, me.name);
   sprintf(REPORT_FIN_DNS, REPORT_FIN_DNS_, me.name);
   sprintf(REPORT_FIN_DNSC, REPORT_FIN_DNSC_, me.name);
   sprintf(REPORT_FAIL_DNS, REPORT_FAIL_DNS_, me.name);
   sprintf(REPORT_DO_ID, REPORT_DO_ID_, me.name);
   sprintf(REPORT_FIN_ID, REPORT_FIN_ID_, me.name);
   sprintf(REPORT_FAIL_ID, REPORT_FAIL_ID_, me.name);

   R_do_dns = strlen(REPORT_DO_DNS);
   R_fin_dns = strlen(REPORT_FIN_DNS);
   R_fin_dnsc = strlen(REPORT_FIN_DNSC);
   R_fail_dns = strlen(REPORT_FAIL_DNS);
   R_do_id = strlen(REPORT_DO_ID);
   R_fin_id = strlen(REPORT_FIN_ID);
   R_fail_id = strlen(REPORT_FAIL_ID);

   if (bootopt & BOOT_OPER) 
   {
      aClient    *tmp = add_connection(&me, 0);
		
      if (!tmp)
		  exit(1);
      SetMaster(tmp);
   }
   else
	  write_pidfile();

   Debug((DEBUG_NOTICE, "Server ready..."));
   ircdlog(LOG_DEFAULT, "Server ready...");
   NOW = time(NULL);
	
#ifndef NO_PRIORITY
   check_fdlists();
#endif
	
   if ((timeofday = time(NULL)) == -1) 
   {
      ircdlog(LOG_ERROR, "Clock Failure (%d), TS can be corrupted", errno);
      sendto_ops("Clock Failure (%d), TS can be corrupted", errno);
   }

#ifdef DUMP_DEBUG
   dumpfp=fopen("dump.log", "w");
#endif

   io_loop();
   return 0;
}

void io_loop()
{
char to_send[200];
time_t lasttime = 0;
long lastrecvK = 0;
int  lrv = 0;
time_t lasttimeofday = 0;
int delay = 0;
time_t delayEvent;

 while(1)
 {
   /*
    * Update irc statistics every second
    */

	delayEvent = next_checkEvent();
	if (delayEvent <= timeofday) {
		checkEvent();
	}

   if (lasttimeofday < timeofday) {
	save_stats();
   }

   lasttimeofday = timeofday;

   if ((timeofday = time(NULL)) == -1) 
   {
      ircdlog(LOG_ERROR, "Clock Failure (%d), TS can be corrupted", errno);
      sendto_ops("Clock Failure (%d), TS can be corrupted", errno);
   }

   if (timeofday < lasttimeofday) 
   {
      ircsprintf(to_send, "System clock is running backwards - (%d < %d)",
		 timeofday, lasttimeofday);
      report_error(to_send, &me);
   }

   NOW = timeofday;

   /*
    * This chunk of code determines whether or not "life sucks", that
    * is to say if the traffic level is so high that standard server
    * commands should be restricted
    * 
    * Changed by Taner so that it tells you what's going on as well as
    * allows forced on (long LCF), etc...
    */

   if (!ConfigSet.hub) {
   if ((timeofday - lasttime) >= LCF) 
   {
      lrv = LRV * LCF;
      lasttime = timeofday;
      currlife = (me.receiveK - lastrecvK) / LCF;
      if ((me.receiveK - lrv) > lastrecvK || HTMLOCK == YES) 
      {
	 if (!lifesux) 
 	 {
	    /*
	     * In the original +th code Taner had
	     * 
	     * LCF << 1;  / * add hysteresis * /
	     * 
	     * which does nothing... so, the hybrid team changed it to
	     * 
	     * LCF <<= 1;  / * add hysteresis * /
	     * 
	     * suddenly, there were reports of clients mysteriously just
	     * dropping off... Neither rodder or I can see why it makes
	     * a difference, but lets try it this way...
	     * 
	     * The original dog3 code, does not have an LCF variable
	     * 
	     * -Dianora
	     * 
	     */
	    lifesux = 1;

	    if (noisy_htm) 
	       sendto_ops("Entering high-traffic mode - (%dk/s > %dk/s)", currlife, LRV);
	 }
	 else 
	 {
	    lifesux++;		/* Ok, life really sucks! */
	    LCF += 2;		/* Wait even longer */
	    if (noisy_htm) 
	       sendto_ops("Still high-traffic mode %d%s (%d delay): %dk/s",
			  lifesux, (lifesux > 9) ? " (TURBO)" : "", (int) LCF, currlife);

	   /* Reset htm here, because its been on a little too long.
	    * Bad Things(tm) tend to happen with HTM on too long -epi */

	    if (lifesux>15) 
	    {
	       if (noisy_htm) 
		  sendto_ops("Resetting HTM and raising limit to: %dk/s\n", LRV + 5);
	       LCF=LOADCFREQ;
	       lifesux=0;
	       LRV+=5;
	    }
	 }
      }
      else 
      {
	 LCF = LOADCFREQ;
	 if (lifesux) 
	 {
	    lifesux = 0;
	    if (noisy_htm)
	       sendto_ops("Resuming standard operation...");
	 }
      }
      lastrecvK = me.receiveK;
   }
   } /* ConfigSet.hub */
   /*
    * * We only want to connect if a connection is due, not every
    * time through.  Note, if there are no active C lines, this call
    * to Tryconnections is made once only; it will return 0. - avalon
    */

   if (nextconnect && timeofday >= nextconnect)
      nextconnect = try_connections(timeofday);

   /* DNS checks. One to timeout queries, one for cache expiries.*/

   if (timeofday >= nextdnscheck)
      nextdnscheck = timeout_query_list(timeofday);
   if (timeofday >= nextexpire)
      nextexpire = expire_cache(timeofday);

   /*
    * * take the smaller of the two 'timed' event times as the time
    * of next event (stops us being late :) - avalon WARNING -
    * nextconnect can return 0!
    */

   if (nextconnect)
      delay = MIN(nextping, nextconnect);
   else
      delay = nextping;
   delay = MIN(nextdnscheck, delay);
   delay = MIN(nextexpire, delay);
   delay -= timeofday;

   /*
    * * Adjust delay to something reasonable [ad hoc values] (one
    * might think something more clever here... --msa) 
    * We don't really need to check that often and as long 
    * as we don't delay too long, everything should be ok. 
    * waiting too long can cause things to timeout... 
    * i.e. PINGS -> a disconnection :( 
    * - avalon
    */
   if (delay < 1)
      delay = 1;
   else
      delay = MIN(delay, TIMESEC);
   /*
    * We want to read servers on every io_loop, as well as "busy"
    * clients (which again, includes servers. If "lifesux", then we
    * read servers AGAIN, and then flush any data to servers. -Taner
    */

#ifndef NO_PRIORITY
   read_message(0, &serv_fdlist);
   read_message(1, &busycli_fdlist);
   if (lifesux) 
   {
      (void) read_message(1, &serv_fdlist);
      if (lifesux > 9) 		/* life really sucks */
      {
	 (void) read_message(1, &busycli_fdlist);
	 (void) read_message(1, &serv_fdlist);
      }
      flush_fdlist_connections(&serv_fdlist);
   }

   if ((timeofday = time(NULL)) == -1) 
   {
	ircdlog(LOG_ERROR, "Clock Failure (%d), TS can become corrupted", errno);
      sendto_ops("Clock Failure (%d), TS can be corrupted", errno);
   }
   /*
    * CLIENT_SERVER = TRUE: If we're in normal mode, or if "lifesux"
    * and a few seconds have passed, then read everything.
    * CLIENT_SERVER = FALSE: If it's been more than lifesux*2 seconds
    * (that is, at most 1 second, or at least 2s when lifesux is != 0)
    * check everything. -Taner
    */
   { 
      static time_t lasttime = 0;

# ifdef CLIENT_SERVER
      if (!lifesux || (lasttime + lifesux) < timeofday) {
# else
      if ((lasttime + (lifesux + 1)) < timeofday) {
# endif
	 (void) read_message(delay ? delay : 1, NULL);	/* check everything! */
	 lasttime = timeofday;
      }
   }
#else
   (void) read_message(delay, NULL);	/* check everything! */
#endif
   /*
    * * ...perhaps should not do these loops every time, but only if
    * there is some chance of something happening (but, note that
    * conf->hold times may be changed elsewhere--so precomputed next
    * event time might be too far away... (similarly with ping
    * times) --msa
    */

   if ((timeofday >= nextping))
      nextping = check_pings(timeofday);

   if (dorehash && !lifesux) 
   {
      (void) rehash(&me, &me, 1);
      dorehash = 0;
   }
   /*
    * 
    * Flush output buffers on all connections now if they 
    * have data in them (or at least try to flush)  -avalon
    *
    * flush_connections(me.fd);
    *
    * avalon, what kind of crack have you been smoking? why
    * on earth would we flush_connections blindly when
    * we already check to see if we can write (and do)
    * in read_message? There is no point, as this causes
    * lots and lots of unnecessary sendto's which 
    * 99% of the time will fail because if we couldn't
    * empty them in read_message we can't empty them here.
    * one effect: during htm, output to normal lusers
    * will lag.
    */

    /* Now we've made this call a bit smarter. */
    /* Only flush non-blocked sockets. */

    flush_connections(me.fd);

#ifndef NO_PRIORITY
   check_fdlists();
#endif
#ifdef HAVE_THROTTLE_ON
   throttle_timer(NOW);
#endif

 }
}
/*
 * open_debugfile
 * 
 * If the -t option is not given on the command line when the server is
 * started, all debugging output is sent to the file set by LPATH in
 * config.h Here we just open that file and make sure it is opened to
 * fd 2 so that any fprintf's to stderr also goto the logfile.  If the
 * debuglevel is not set from the command line by -x, use /dev/null as
 * the dummy logfile as long as DEBUGMODE has been defined, else dont
 * waste the fd.
 */
static void
open_debugfile()
{
#ifdef	DEBUGMODE
int         fd;
aClient    *cptr;

   if (debuglevel >= 0) 
   {
      cptr = make_client(NULL, NULL);
      cptr->fd = 2;
      SetLog(cptr);
      cptr->port = debuglevel;
      cptr->flags = 0;
      cptr->listener = cptr;
      local[2] = cptr;
      (void) strcpy(cptr->sockhost, me.sockhost);

      (void) printf("isatty = %d ttyname = %#x\n",
		    isatty(2), (u_int) ttyname(2));
      if (!(bootopt & BOOT_TTY)) 	/* leave debugging output on fd */ 
      {
	 (void) truncate(LOGFILE, 0);
	 if ((fd = open(LOGFILE, O_WRONLY | O_CREAT, 0600)) < 0)
	    if ((fd = open("/dev/null", O_WRONLY)) < 0)
	       exit(-1);
	 if (fd != 2) 
    	 {
	    (void) dup2(fd, 2);
	    (void) close(fd);
	 }
	 strncpyzt(cptr->name, LOGFILE, sizeof(cptr->name));
      }
      else if (isatty(2) && ttyname(2))
	 strncpyzt(cptr->name, ttyname(2), sizeof(cptr->name));
      else
	 (void) strcpy(cptr->name, "FD2-Pipe");
      Debug((DEBUG_FATAL, "Debug: File <%s> Level: %d at %s",
	     cptr->name, cptr->port, myctime(time(NULL))));
   }
   else
      local[2] = NULL;
#endif
   return;
}

static void
setup_signals()
{
#ifdef	POSIX_SIGNALS
struct sigaction act;

   act.sa_handler = SIG_IGN;
   act.sa_flags = 0;
   (void) sigemptyset(&act.sa_mask);
   (void) sigaddset(&act.sa_mask, SIGPIPE);
   (void) sigaddset(&act.sa_mask, SIGALRM);
# ifdef	SIGWINCH
   (void) sigaddset(&act.sa_mask, SIGWINCH);
   (void) sigaction(SIGWINCH, &act, NULL);
# endif
   (void) sigaction(SIGPIPE, &act, NULL);
   act.sa_handler = dummy;
   (void) sigaction(SIGALRM, &act, NULL);
   act.sa_handler = s_rehash;
   (void) sigemptyset(&act.sa_mask);
   (void) sigaddset(&act.sa_mask, SIGHUP);
   (void) sigaction(SIGHUP, &act, NULL);
   act.sa_handler = s_restart;
   (void) sigaddset(&act.sa_mask, SIGINT);
   (void) sigaction(SIGINT, &act, NULL);
   act.sa_handler = s_die;
   (void) sigaddset(&act.sa_mask, SIGTERM);
   (void) sigaction(SIGTERM, &act, NULL);

#else
# ifndef	HAVE_RELIABLE_SIGNALS
   (void) signal(SIGPIPE, dummy);
#  ifdef	SIGWINCH
   (void) signal(SIGWINCH, dummy);
#  endif
# else
#  ifdef	SIGWINCH
   (void) signal(SIGWINCH, SIG_IGN);
#  endif
   (void) signal(SIGPIPE, SIG_IGN);
# endif
   (void) signal(SIGALRM, dummy);
   (void) signal(SIGHUP, s_rehash);
   (void) signal(SIGTERM, s_die);
   (void) signal(SIGINT, s_restart);
#endif 

#ifdef RESTARTING_SYSTEMCALLS
   /*
    * * At least on Apollo sr10.1 it seems continuing system calls 
    * after signal is the default. The following 'siginterrupt' 
    * should change that default to interrupting calls.
    */
   (void) siginterrupt(SIGALRM, 1);
#endif
}

#ifndef NO_PRIORITY
/*
 * This is a pretty expensive routine -- it loops through all the fd's,
 * and finds the active clients (and servers and opers) and places them
 * on the "busy client" list
 */
void check_fdlists()
{
#ifdef CLIENT_SERVER
#define BUSY_CLIENT(x)	(((x)->priority < 55) || (!lifesux && ((x)->priority < 75)))
#else
#define BUSY_CLIENT(x)	(((x)->priority < 40) || (!lifesux && ((x)->priority < 60)))
#endif
#define FDLISTCHKFREQ  2

aClient *cptr;
int i, j;

   j = 0;
   for (i = highest_fd; i >= 0; i--) 
   {
      if (!(cptr = local[i]))
	 continue;
      if (IsServer(cptr) || IsListening(cptr) || IsOper(cptr)) 
      {
	 busycli_fdlist.entry[++j] = i;
	 continue;
      }
      if (cptr->receiveM == cptr->lastrecvM) 
      {
	 cptr->priority += 2;	/* lower a bit */
	 if (cptr->priority > 90)
	    cptr->priority = 90;
	 else if (BUSY_CLIENT(cptr))
	    busycli_fdlist.entry[++j] = i;
	 continue;
      }
      else 
      {
	 cptr->lastrecvM = cptr->receiveM;
	 cptr->priority -= 30;	/* active client */
	 if (cptr->priority < 0) 
	 {
	    cptr->priority = 0;
	    busycli_fdlist.entry[++j] = i;
	 }
	 else if (BUSY_CLIENT(cptr))
	    busycli_fdlist.entry[++j] = i;
      }
   }
   busycli_fdlist.last_entry = j;	/* rest of the fdlist is garbage */
/*   return (now + FDLISTCHKFREQ + (lifesux + 1)); */
}
#endif

void build_version(void)
{
	char *s = PATCHES;
	ircsprintf(version, "%s(%s)-%d.%d.%d%s", BASENAME, CODENAME, MAJOR, MINOR, REVISION,
		*s != 0 ? PATCHES : "");
}

void save_stats(void)
{
	FILE *fd = fopen(COUNT_FILE, "w+");
	if (!fd) {
		sendto_realops_lev(DEBUG_LEV, "Failed to save current irc statistics");
		return;
	}

	fprintf(fd, "%d\n%d\n%li\n%li\n%li\n%ld\n%ld\n%ld\n%ld", Count.max_loc, Count.max_tot, Count.weekly,
		Count.monthly, Count.yearly, Count.start, Count.week, Count.month, Count.year);
	fclose(fd);
}

int load_stats()
{
	FILE *fd = fopen(COUNT_FILE, "r");
	char buffer[1024];

	if (!fd) {
		return -1;
	}

	fgets(buffer, 1024, fd);
	Count.max_loc = atoi(buffer);
	fgets(buffer, 1024, fd);
	Count.max_tot = atoi(buffer);
	fgets(buffer, 1024, fd);
	Count.weekly = atoi(buffer);
	fgets(buffer, 1024, fd);
	Count.monthly = atoi(buffer);
	fgets(buffer, 1024, fd);
	Count.yearly = atoi(buffer);
	fgets(buffer, 1024, fd);
	Count.start = atol(buffer);
	fgets(buffer, 1024, fd);
	Count.week = atol(buffer);
	fgets(buffer, 1024, fd);
	Count.month = atol(buffer);
	fgets(buffer, 1024, fd);
	Count.year = atol(buffer);

	fclose(fd);
	return 0;
}

void ircdlog(char *filename, char *pattern, ...)
{
	FILE *fd;
	char buffer[1024];
	va_list vl;

	if (ConfigSet.logging) {
		fd = fopen(filename, "a");
		if (!fd) {
			Debug((DEBUG_ERROR, "Failed to open/create %s for log", filename));
			return;
		}
		va_start(vl, pattern);
		ircvsprintf(buffer, pattern, vl);
		va_end(vl);
		fprintf(fd, "[%s] %s\n", myctime(timeofday), buffer);
		fclose(fd);
	}
	return;
}
