/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/list.c: linked list management
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: list.c,v 1.19 2002/04/09 17:03:54 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "blalloc.h"
#include "dh.h"
#include "zlink.h"
#include "setup.h"
#include "memory.h"

int numclients = 0;
BlockHeap *link_heap;

void link_heap_garbage_collect()
{
	BlockHeapGarbageCollect(link_heap);
	return;
}

void init_link()
{
	link_heap = BlockHeapCreate((size_t)sizeof(Link), LINKS_PREALLOCATE);
	addEvent("link_heap_garbage_collect", link_heap_garbage_collect, NULL, LINK_HEAP_GARBAGE_COLLECT_FREQ, 1);
	return;
}

aClient *make_client(aClient *from, aClient *uplink)
{
	aClient *cptr = NULL;

	if (from == NULL) {
		cptr = (aClient *)BlockHeapAlloc(local_client_heap);
		if (cptr == NULL) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_client where from is NULL");
			outofmemory();
		}
		memset((char *)cptr, '\0', CLIENT_LOCAL_SIZE);

		cptr->from = cptr;
		cptr->status = STAT_UNKNOWN;
		cptr->fd = -1;
		cptr->authfd = -1;
		cptr->sockerr = -1;
		cptr->uplink = uplink;
		cptr->class = (ConfigItem_class *)NULL;
		strcpy(cptr->username, "unknown");
		cptr->since = cptr->lasttime = cptr->firsttime = timeofday;
		cptr->sendqlen = MAXSENDQLENGTH;
	}
	else {
		cptr = (aClient *)BlockHeapAlloc(remote_client_heap);
		if (cptr == NULL) {
			ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_client where from is non-NULL");
			outofmemory();
		}
		memset((char *)cptr, '\0', CLIENT_REMOTE_SIZE);

		cptr->from = from;
		cptr->status = STAT_UNKNOWN;
		cptr->fd = -1;
		cptr->uplink = uplink;
	}
	return cptr;
}

void free_client(aClient *cptr)
{
	if (BlockHeapFree(cptr->fd == -2 ? local_client_heap : remote_client_heap, cptr)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_client (cptr=%lX)", cptr);
	}
	return;
}

aChannel *make_channel()
{
	aChannel *chan = (aChannel *)BlockHeapAlloc(channel_heap);
	if (chan == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_channel");
		outofmemory();
	}
	memset((char *)chan, '\0', sizeof(aChannel));
	return chan;
}

void free_channel(aChannel *chptr)
{
	if (BlockHeapFree(channel_heap, chptr)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_channel [chptr=%lX]", chptr);
	}
	return;
}

anUser *make_user(aClient *cptr)
{
	anUser *user = cptr->user;
	if (user == NULL) {
		if (cptr->fd == -2 || cptr->fd >= 0) {
			user = (anUser *)BlockHeapAlloc(local_user_heap);
			if (user == NULL) {
				ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_user for local "
					"client [cptr=%lX]", cptr);
				outofmemory();
			}
			memset((char *)user, '\0', sizeof(anUser));
		}
		else {
			user = (anUser *)BlockHeapAlloc(remote_user_heap);
			if (user == NULL) {
				ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_user for remote "
					"client [cptr=%lX]", cptr);
				outofmemory();
			}
			memset((char *)user, '\0', sizeof(anUser));
		}
		cptr->user = user;

	}
	return user;
}

aServer *make_server(aClient *cptr)
{
	aServer *serv = cptr->serv;
	if (serv == NULL) {
		serv = (aServer *)MyMallocX(sizeof(aServer));
		*serv->bynick = '\0';
		*serv->byuser = '\0';
		*serv->byhost = '\0';
		serv->up = NULL;
		cptr->serv = serv;
	}
	return serv;
}

void free_user(anUser *user, aClient *cptr)
{
	if (user->away) {
		MyFree(user->away);
	}
	if (user->maskedhost) {
		MyFree(user->maskedhost);
	}
	if (user->joined || user->invited || user->channel) {
		Debug((DEBUG_ERROR, "%#x user (%s!%s@%s) %#x %#x %#x %d *", cptr,
			cptr ? cptr->name : "<noname>", user->username, user->host,
			user, user->invited, user->channel, user->joined));
	}

	if (cptr->fd == -2 || cptr->fd >= 0) {
		if (BlockHeapFree(local_user_heap, user)) {
			ircdlog(LOG_ERROR, "BlockHeapFree failed in free_user for local client [user=%lX]", user);
		}
	}
	else {
		if (BlockHeapFree(remote_user_heap, user)) {
			ircdlog(LOG_ERROR, "BlockHeapFree failed in free_user for remote client [user=%lX]", user);
		}
	}
	return;
}

void remove_client_from_list(aClient *cptr)
{
	if (IsServer(cptr)) {
		Count.server--;
	}
	else if (IsClient(cptr)) {
		Count.total--;
		if (IsOper(cptr)) {
			Count.oper--;
		}
		if (IsInvisible(cptr)) {
			Count.invisi--;
		}
	}

	checklist();

	if (cptr->prev) {
		cptr->prev->next = cptr->next;
	}
	else {
		client = cptr->next;
		client->prev = NULL;
	}
	if (cptr->next) {
		cptr->next->prev = cptr->prev;
	}
	if (IsPerson(cptr) && cptr->user) {
		add_history(cptr, 0);
		off_history(cptr);
	}
#ifdef FLUD
	if (MyFludConnect(cptr)) {
		free_fluders(cptr, NULL);
	}
	free_fludees(cptr);
#endif
	if (cptr->user) {
		free_user(cptr->user, cptr);
	}
	if (cptr->serv) {
#ifdef HAVE_OPENSSL_ON
		if (cptr->serv->sessioninfo_in) {
			dh_end_session(cptr->serv->sessioninfo_in);
		}
		if (cptr->serv->sessioninfo_out) {
			dh_end_session(cptr->serv->sessioninfo_out);
		}
		if (cptr->serv->rc4_in) {
			rc4_destroystate(cptr->serv->rc4_in);
		}
		if (cptr->serv->rc4_out) {
			rc4_destroystate(cptr->serv->rc4_out);
		}
#endif
		if (cptr->serv->zip_in) {
			zip_destroy_input_session(cptr->serv->zip_in);
		}
		if (cptr->serv->zip_out) {
			zip_destroy_output_session(cptr->serv->zip_out);
		}
		MyFree(cptr->serv);
	}
	free_client(cptr);
	return;
}

void add_client_to_list(aClient *cptr)
{
	cptr->next = client;
	client = cptr;
	if (cptr->next) {
		cptr->next->prev = cptr;
	}
	return;
}

chanMember *find_user_member(chanMember *cm, aClient *cptr)
{
	if (cptr) {
		for (; cm; cm = cm->next) {
			if (cm->cptr == cptr) {
				return (cm);
			}
		}
	}
	return (NULL);
}

Link *find_channel_link(Link *lp, aChannel *chptr)
{
	if (chptr) {
		for (; lp; lp = lp->next) {
			if (lp->value.chptr == chptr) {
				return (lp);
			}
		}
	}
	return (NULL);
}

Link *find_str_link(Link *lp, char *charptr)
{
	if (charptr) {
		for (; lp; lp = lp->next) {
			if (!match(lp->value.cp, charptr)) {
				return (lp);
			}
		}
	}
	return (NULL);
}

Link *make_link()
{
	Link *lp = (Link *)BlockHeapAlloc(link_heap);
	if (lp == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_link");
		outofmemory();
	}
	lp->next = NULL;
	return lp;
}

void free_link(Link *lp)
{
	if (BlockHeapFree(link_heap, lp)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_link [lp=%lX]", lp);
	}
	return;
}

chanMember *make_chanmember()
{
	chanMember *cm = (chanMember *)BlockHeapAlloc(chanmember_heap);
	if (cm == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_chanmember");
		outofmemory();
	}
	return cm;
}

void free_chanmember(chanMember *cm)
{
	if (BlockHeapFree(chanmember_heap, cm)) {
		ircdlog(LOG_ERROR, "BlockHEapFree failed in free_chanmember [cm=%lX]", cm);
	}
	return;
}

#ifdef HAVE_THROTTLE_ON
throttle *make_throttle()
{
	throttle *tp = (throttle *)BlockHeapAlloc(throttle_heap);
	if (tp == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_throttle");
		outofmemory();
	}
	return tp;
}

void free_throttle(throttle *tp)
{
	if (BlockHeapFree(throttle_heap, tp)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_throttle [tp=%lX]", tp);
	}
	return;
}

hashent *make_hashent()
{
	hashent *h = (hashent *)BlockHeapAlloc(hashent_heap);
	if (h == NULL) {
		ircdlog(LOG_ERROR, "FATAL: BlockHeapAlloc failed in make_hashent");
		outofmemory();
	}
	return h;
}

void free_hashent(hashent *h)
{
	if (BlockHeapFree(hashent_heap, h)) {
		ircdlog(LOG_ERROR, "BlockHeapFree failed in free_hashent [h=%lX]", h);
	}
	return;
}
#endif
