/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/modules.c: dynamic module routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: modules.c,v 1.55 2002/04/08 16:00:50 rageircd Exp $
 */

#include "struct.h"
#include "setup.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "msg.h"
#include "channel.h"
#include "send.h"
#include "memory.h"
#include <sys/types.h>
#include <sys/stat.h>   
#include <time.h>
#include <stdlib.h>
#include <string.h>
#ifndef STATIC_MODULES
#include <dlfcn.h>
#include <dirent.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <limits.h>	/* for PATH_MAX */
#include "patchlevel.h"

#ifndef STATIC_MODULES
static ModuleT *Modules = NULL;
static int loaded_modules = 0;
static void addModule(ModuleT *);
static ModuleT *delModule(ModuleT *);
#endif

extern void config_report(short, short, char *, ...);

#ifndef RTLD_NOW
#define RTLD_NOW RTLD_LAZY	/* OpenBSD deficiency */
#endif

void init_Modules(void)
{
	add_Command("MODULE", m_module, 0, 15, 1, 0, 0, 0L);
	return;
}

#ifndef STATIC_MODULES
static void addModule(ModuleT *mp)
{
	mp->prev = NULL;
	mp->next = Modules;
	if (mp->next) {
		mp->next->prev = mp;
	}
	Modules = mp;
	return;
}

static ModuleT *delModule(ModuleT *mp)
{
	ModuleT *mp2, *mp3;
	for (mp2 = Modules; mp2; mp2 = mp2->next) {
		if (mp2 == mp) {
			mp3 = mp2->next;
			if (mp2->prev) {
				mp2->prev->next = mp2->next;
			}
			else {
				Modules = mp2->next;
			}
			if (mp2->next) {
				mp2->next->prev = mp2->prev;
			}
			MyFree(mp2);
			return mp3;
		}
	}
	return NULL;
}
#endif

int load_one_module(int config_loadmod, char *name)
{
#ifndef STATIC_MODULES
	void *ptr, (*loadfunc)(), (*unloadfunc)();
	Module *m_header;
	ModuleT *mp;

	for (mp = Modules; mp; mp = mp->next) {
		if (!strcmp(mp->header->name, name)) {
			if (!config_loadmod) {
				sendto_realops("Module %s is already loaded", name);
			}
			return 0;
		}
	}

	ptr = dlopen(name, RTLD_NOW);
	if (!ptr) {
		char *error = (char *)dlerror();
		if (config_loadmod) {
			config_report(0, 1, "loadmodule->%s: failed (%s); ignoring command",
				name, error ? error : "cause unknown");
		}
		else {
			sendto_realops("Error loading module %s: %s", name, error ? error : "cause unknown");
		}
		return 0;
	}

	m_header = dlsym(ptr, "_module_header");
	if (!m_header) {
		m_header = dlsym(ptr, "__module_header");
		if (!m_header) {
			if (config_loadmod) {
				config_report(0, 1, "loadmodule->%s: failed (missing _module_header); "
					"ignoring command", name);
			}
			else {
				sendto_realops("Error loading module %s: missing _module_header", name);
			}
			dlclose(ptr);
			return 0;
		}
	}

	if (!m_header->name || !m_header->info) {
		if (config_loadmod) {
			config_report(0, 1, "loadmodule->%s: failed (invalid _module_header); "
				"ignoring command", name);
		}
		else {
			sendto_realops("Error loading module %s: invalid _module_header", name);
		}
		dlclose(ptr);
		return 0;
	}

	if (m_header->version < MIN_MODULE_VERSION) {
		if (config_loadmod) {
			config_report(0, 1, "loadmodule->%s: failed (version too old); ignoring command",
				name);
		}
		else {
			sendto_realops("Error loading module %s: version too old", name);
		}
		dlclose(ptr);
		return 0;
	}

	loadfunc = dlsym(ptr, "_module_load");
	if (!loadfunc) {
		loadfunc = dlsym(ptr, "__module_load");
		if (!loadfunc) {
			if (config_loadmod) {
				config_report(0, 1, "loadmodule->%s: failed (missing _module_load function); "
					"ignoring command", name);
			}
			else {
				sendto_realops("Error loading module %s: missing _module_load function");
			}
			dlclose(ptr);
			return 0;
		}
	}

	unloadfunc = dlsym(ptr, "_module_unload");
	if (!unloadfunc) {
		unloadfunc = dlsym(ptr, "__module_unload");
		if (!unloadfunc) {
			if (config_loadmod) {
				config_report(0, 1, "loadmodule->%s: failed (missing _module_unload function); "
					"ignoring command", name);
			}
			else {
				sendto_realops("Error loading module %s: missing _module_unload function", name);
			}
			dlclose(ptr);
			return 0;
		}
	}

	mp = (ModuleT *)MyMallocX(sizeof(ModuleT));
	mp->header = m_header;
	mp->header->ptr = ptr;
	mp->header->unloadfunc = unloadfunc;
	addModule(mp);
	loadfunc();
	loaded_modules++;

	if (!config_loadmod) {
		sendto_realops("Loaded module %s (%s) [version: %d]", mp->header->name, mp->header->info,
			mp->header->version);
	}
#endif
	return 1;
}

int unload_one_module(char *name)
{
#ifndef STATIC_MODULES
	ModuleT *mp;

	for (mp = Modules; mp; mp = mp->next) {
		if (!strcmp(mp->header->name, name)) {
			break;
		}
	}
	if (!mp) {
		sendto_realops("Module %s is not loaded", name);
		return 0;
	}

	(*mp->header->unloadfunc)();
	dlclose(mp->header->ptr);
	delModule(mp);
	MyFree(mp);

	loaded_modules--;
	sendto_realops("Unloaded module %s", name);
#endif
	return 1;
}

void unload_all_modules()
{
#ifndef STATIC_MODULES
	ModuleT *mp;

	for (mp = Modules; mp; mp = mp->next) {
		(*mp->header->unloadfunc)();
		dlclose(mp->header->ptr);
		delModule(mp);
		MyFree(mp);
		loaded_modules--;
	}
#endif
	return;
}

/*
 * m_module
 *	parv[0] = sender prefix
 *	parv[1] = function
 *	parv[2] = module
 */
int m_module(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#ifdef STATIC_MODULES
	sendto_one(sptr, ":%s NOTICE %s :Dynamic module support is not enabled on this server",
		me.name, parv[0]);
#else
	ModuleT *mp;

	if (!IsAdmin(sptr) && !IsCoAdmin(sptr) && !IsNetAdmin(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}
	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODULE");
		return 0;
	}

	if (!strcasecmp(parv[1], "LOAD")) {
		if (parc < 3 || *parv[2] == '\0') {
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODULE LOAD");
			return 0;
		}
		load_one_module(0, parv[2]);
		return 0;
	}
	if (!strcasecmp(parv[1], "UNLOAD")) {
		if (parc < 3 || *parv[2] == '\0') {
			sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODULE UNLOAD");
			return 0;
		}
		unload_one_module(parv[2]);
		return 0;
	}
	if (!strcasecmp(parv[1], "LIST")) {
		sendto_one(sptr, ":%s NOTICE %s :There %s currently %d module%s loaded", me.name, parv[0],
			loaded_modules == 1 ? "is" : "are", loaded_modules, loaded_modules == 1 ? "" : "s");
		sendto_one(sptr, ":%s NOTICE %s :-----------------------------------------", me.name,
			parv[0]);
		for (mp = Modules; mp; mp = mp->next) {
			sendto_one(sptr, ":%s NOTICE %s :%s (%s) [version: %d]", me.name, parv[0],
				mp->header->name, mp->header->info, mp->header->version);
		}
		sendto_one(sptr, ":%s NOTICE %s :-----------------------------------------", me.name,
			parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :End of module list", me.name, parv[0]);
		return 0;
	}
	sendto_one(sptr, ":%s NOTICE %s :Syntax: /MODULE <LOAD/UNLOAD/LIST> [path.to.module.so]",
		me.name, parv[0]);
#endif
	return 0;
}
