/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/motd.c: motd file handling functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: motd.c,v 1.13 2002/04/08 16:00:50 rageircd Exp $
 */

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <time.h>
#include "h.h"
#include "memory.h"

aMotd *motd = NULL;
aMotd *shortmotd = NULL;
aMotd *opermotd = NULL;
aMotd *rules = NULL;

char motd_changed_date[34];
char shortmotd_changed_date[34];
char opermotd_changed_date[34];
char rules_changed_date[34];

struct tm *motd_tm = NULL;
struct tm *shortmotd_tm = NULL;
struct tm *opermotd_tm = NULL;
struct tm *rules_tm = NULL;

void send_motd(aClient *);
void send_opermotd(aClient *);
void send_rules(aClient *);
void rehash_motd();
void rehash_opermotd();
void rehash_rules();

aMotd *load_file(aMotd **ptr, struct tm *tm, char changed_date[34],
  char *filename, int sm)
{
	aMotd *tmp, *last, *new;
	struct stat sb;
	char buffer[MOTDLINELEN];
	int lines = 0, fd = -1;

	while (*ptr) {
		tmp = (*ptr)->next;
		MyFree(*ptr);
		*ptr = tmp;
	}

	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		return NULL;
	}
	fstat(fd, &sb);
	dgets(-1, NULL, 0);

	tm = localtime(&sb.st_mtime);
	tmp = new = last = NULL;

	while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) {
		if (sm && lines >= 5) {
			break;
		}
		lines++;

		strip(buffer);

		tmp = (aMotd *)MyMalloc(sizeof(aMotd));
		if (!tmp) {
			outofmemory();
		}
		strncpyzt(tmp->line, buffer, MOTDLINELEN);
		tmp->next = NULL;

		if (!new) {
			new = tmp;
		}
		else {
			last->next = tmp;
		}
		last = tmp;
	}
	close(fd);

	if (tm) {
		ircsprintf(changed_date, "%d:%02d, %d/%d/%d", tm->tm_hour, tm->tm_min,
			tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year);
	}

	return (new);
}

void send_file(aClient *sptr, aMotd **ptr, struct tm *tm, char changed_date[34],
  char *shorttype, char *longtype, int sm)
{
	aMotd *tmp;

	if (!sm && *ptr == NULL) {
		sendto_one(sptr, err_str(ERR_NOMOTD), me.name, sptr->name, shorttype);
		return;
	}

	if (strlen(changed_date) > 1) {
		sendto_one(sptr, ":%s NOTICE %s :*** The %s was last changed at %s", me.name,
			sptr->name, shorttype, changed_date);
	}
	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, sptr->name, me.name, longtype);

	if (sm && *ptr == NULL) {
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, sptr->name, "*** This is the Short "
			"Message of the Day ***");
	}
	if (tm) {
		sendto_one(sptr, ":%s %d %s :- %d:%02d, %d/%d/%d", me.name, RPL_MOTD, sptr->name,
			tm->tm_hour, tm->tm_min, tm->tm_mday, tm->tm_mon + 1,
			1900 + tm->tm_year);
	}

	tmp = *ptr;
	while (tmp) {
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, sptr->name, tmp->line);
		tmp = tmp->next;
	}

	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, sptr->name, shorttype);
	return;
}

void free_file(aMotd **ptr, char changed_date[34])
{
	aMotd *tmp;
	while (*ptr) {
		tmp = (*ptr)->next;
		MyFree(*ptr);
		*ptr = tmp;
	}
	changed_date[0] = '\0';
	return;
}

void init_motd(int load)
{
	if (load == 1) {
		motd = load_file(&motd, motd_tm, motd_changed_date, MOTD_FILE, 0);
		shortmotd = load_file(&shortmotd, shortmotd_tm, shortmotd_changed_date, SHORTMOTD_FILE, 1);
		addHook(HOOKTYPE_LOCAL_CONNECT, NULL, send_motd);
		addHook(HOOKTYPE_REHASH, NULL, rehash_motd);
	}
	else {
		delHook(HOOKTYPE_LOCAL_CONNECT, NULL, send_motd);
		delHook(HOOKTYPE_REHASH, NULL, rehash_motd);
		free_file(&motd, motd_changed_date);
		free_file(&shortmotd, shortmotd_changed_date);
	}
	return;
}

void send_motd(aClient *sptr)
{
	if (ConfigSet.shortmotd) {
		sendto_one(sptr, ":%s NOTICE %s :*** Please read the full MOTD before continuing with "
			"this service.", me.name, sptr->name);
		send_file(sptr, &shortmotd, shortmotd_tm, shortmotd_changed_date, "SMOTD", "Short Message of the Day", 1);
	}
	else {
		send_file(sptr, &motd, motd_tm, motd_changed_date, "MOTD", "Message of the Day", 0);
	}
	return;
}

void rehash_motd()
{
	free_file(&motd, motd_changed_date);
	motd = load_file(&motd, motd_tm, motd_changed_date, MOTD_FILE, 0);
	return;
}

void init_opermotd(int load)
{
	if (load == 1) {
		opermotd = load_file(&opermotd, opermotd_tm, opermotd_changed_date, OPERMOTD_FILE, 0);
		addHook(HOOKTYPE_OPER, NULL, send_opermotd);
		addHook(HOOKTYPE_REHASH, NULL, rehash_opermotd);
	}
	else {
		delHook(HOOKTYPE_OPER, NULL, send_opermotd);
		delHook(HOOKTYPE_REHASH, NULL, rehash_opermotd);
		free_file(&opermotd, opermotd_changed_date);
	}
	return;
}

void send_opermotd(aClient *sptr)
{
	if (ConfigSet.showopermotd) {
		send_file(sptr, &opermotd, opermotd_tm, opermotd_changed_date, "OPERMOTD",
			"IRC Operator Message of the Day", 0);
	}
	return;
}

void rehash_opermotd()
{
	free_file(&opermotd, opermotd_changed_date);
	opermotd = load_file(&opermotd, opermotd_tm, opermotd_changed_date, OPERMOTD_FILE, 0);
	return;
}

void init_rules(int load)
{
	if (load == 1) {
		rules = load_file(&rules, rules_tm, rules_changed_date, RULES_FILE, 0);
		addHook(HOOKTYPE_REHASH, NULL, rehash_rules);
	}
	else {
		delHook(HOOKTYPE_REHASH, NULL, rehash_rules);
		free_file(&rules, rules_changed_date);
	}
	return;
}

void rehash_rules()
{
	free_file(&rules, rules_changed_date);
	rules = load_file(&rules, rules_tm, rules_changed_date, RULES_FILE, 0);
	return;
}
