/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/openssl.c: client-to-client OpenSSL functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: openssl.c,v 1.3 2002/04/08 16:00:50 rageircd Exp $
 */

#include "setup.h"

#ifdef HAVE_OPENSSL_ON

#include "config.h"
#include "struct.h"
#include "openssl.h"
#include "h.h"
#include "memory.h"
#include <string.h>
#include <unistd.h>

SSL_CTX *ctx;

extern void config_report(short, short, char *, ...);

static int openssl_pem_passwd_cb(char *, int, int, void *);
static int init_ctx();
static char *openssl_get_init_error_message(int);

static int openssl_pem_passwd_cb(char *buf, int size, int rwflag, void *password)
{
	char *pass;
	static int before = 0;
	static char beforebuf[1024];

	if (before) {
		strncpy(buf, (char *)beforebuf, size);
		buf[size - 1] = '\0';
		return strlen(buf);
	}

	pass = getpass("* Password for OpenSSL Private Key: ");
	if (pass) {
		int bsz = sizeof(beforebuf);
		strncpy(buf, (char *)pass, size);
		strncpy(beforebuf, (char *)pass, bsz);
		beforebuf[bsz - 1] = '\0';
		buf[size - 1] = '\0';
		before = 1;
		return strlen(buf);
	}
	return 0;
}

static int init_ctx()
{
	ctx = SSL_CTX_new(SSLv23_server_method());
	if (!ctx) {
		return -4;
	}

	SSL_CTX_set_default_passwd_cb(ctx, openssl_pem_passwd_cb);
	if (SSL_CTX_use_certificate_file(ctx, OPENSSL_CERT_FILE, SSL_FILETYPE_PEM) <= 0) {
		return -3;
	}
	if (SSL_CTX_use_PrivateKey_file(ctx, OPENSSL_KEY_FILE, SSL_FILETYPE_PEM) <= 0) {
		return -2;
	}
	if (!SSL_CTX_check_private_key(ctx)) {
		return -1;
	}
	return 0;
}

static char *openssl_get_init_error_message(int error)
{
	switch (error) {
		case -4:
			return "create new OpenSSL client ctx";
		case -3:
			return "open OpenSSL certificate file";
		case -2:
			return "open OpenSSL key file";
		case -1:
			return "check OpenSSL key file";
	}
	return NULL;
}

void init_openssl()
{
	int i;

	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();

	i = init_ctx();
	if (i < 0) {
		char *error = openssl_get_init_error_message(i);
		config_report(1, 0, "Failed to %s", error);
		ircdlog(LOG_ERROR, "FATAL: failed to %s", error);
	}
	else {
		config_report(0, 0, "Loaded client OpenSSL permission files.");
	}
	return;
}

int openssl_handshake(aClient *cptr)
{
	cptr->openssl = (struct SSL *)SSL_new(ctx);
	if (cptr->openssl == NULL) {
		return 0;
	}

	SSL_set_fd((SSL *)cptr->openssl, cptr->fd);
	set_non_blocking(cptr->fd, cptr);
	if (SSL_accept((SSL *)cptr->openssl) > 0) {
		return 0;
	}
	return 1;
}

void openssl_remap_fd(SSL *s, int fd)
{
	BIO_set_fd(SSL_get_rbio(s), fd, BIO_NOCLOSE);
	BIO_set_fd(SSL_get_wbio(s), fd, BIO_NOCLOSE);
	return;
}

char *openssl_get_cipher(SSL *s)
{
	static char buf[400], *ver = NULL;
	int bits;
	SSL_CIPHER *c;

	switch (s->session->ssl_version) {
		case SSL2_VERSION:
			ver = "SSLv2";
			break;
		case SSL3_VERSION:
			ver = "SSLv3";
			break;
		case TLS1_VERSION:
			ver = "TLSv1";
			break;
	}
	c = SSL_get_current_cipher(s);
	SSL_CIPHER_get_bits(c, &bits);

	buf[0] = '\0';
	ircsprintf(buf, "%s-%s-%dbits", ver ? ver : "UNKNOWN", SSL_get_cipher(s), bits);
        return (buf);
}

#endif /* HAVE_OPENSSL_ON */
