/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/parse.c: internal parsing functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: parse.c,v 1.72 2002/04/09 17:03:54 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "msg.h"
#include "sys.h"
#include "numeric.h"
#include "h.h"
#include "memory.h"

#ifdef HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

extern void init_motd(int);
extern void init_opermotd(int);
extern void init_rules(int);

/*
 * NOTE: parse() should not be called recursively by other functions!
 */
aCommand *CommandHash[256];
static char *para[MAXPARA + 1];
static char sender[HOSTLEN + 1];
static int  cancel_clients(aClient *, aClient *, char *);
static void remove_unknown(aClient *, char *, char *);

inline aCommand *find_Command(char *);
inline aCommand *find_CommandF(char *, int (*)());

/* Modular stuff -- based on the modular command system from UnrealIRCd v3.2 */

void init_CommandHash(void)
{
	memset(CommandHash, '\0', sizeof(CommandHash));
#ifdef STATIC_MODULES
	add_Command("CHANSERV", m_chanserv, 0, 1, 1, 0, 0, 0L);
	add_Command("CS", m_chanserv, 0, 1, 1, 0, 0, 0L);
	add_Command("NICKSERV", m_nickserv, 0, 1, 1, 0, 0, 0L);
	add_Command("NS", m_nickserv, 0, 1, 1, 0, 0, 0L);
	add_Command("MEMOSERV", m_memoserv, 0, 1, 1, 0, 0, 0L);
	add_Command("MS", m_memoserv, 0, 1, 1, 0, 0, 0L);
	add_Command("OPERSERV", m_operserv, 0, 1, 1, 0, 0, 0L);
	add_Command("OS", m_operserv, 0, 1, 1, 0, 0, 0L);
	add_Command("STATSERV", m_statserv, 0, 1, 1, 0, 0, 0L);
	add_Command("SS", m_statserv, 0, 1, 1, 0, 0, 0L);
	add_Command("SERVICES", m_services, 0, 1, 1, 0, 0, 0L);
	add_Command("IDENTIFY", m_identify, 0, 1, 1, 0, 0, 0L);
	add_Command("AWAY", m_away, 0, 15, 1, 0, 1, 0L);
	add_Command("MAP", m_map, 0, 15, 1, 0, 0, 0L);
	add_Command("MKPASSWD", m_mkpasswd, 0, 15, 1, 0, 0, 0L);
	add_Command("OPER", m_oper, 0, 15, 1, 0, 0, 0L);
	add_Command("VERSION", m_version, 0, 15, 1, 1, 0, 0L);
	add_Command("SVSMODE", m_svsmode, 0, 15, 1, 0, 0, 0L);
	add_Command("SVSNICK", m_svsnick, 0, 15, 1, 0, 0, 0L);
	add_Command("JOIN", m_join, 0, 15, 1, 0, 1, 0L);
	add_Command("PART", m_part, 0, 15, 1, 0, 1, 0L);
	add_Command("KLINE", m_kline, 0, 15, 1, 0, 0, 0L);
	add_Command("UNKLINE", m_unkline, 0, 15, 1, 0, 0, 0L);
	add_Command("DIE", m_die, 0, 15, 1, 0, 0, 0L);
	add_Command("RESTART", m_restart, 0, 15, 1, 0, 0, 0L);
	add_Command("ZLINE", m_zline, 0, 15, 1, 0, 0, 0L);
	add_Command("SQLINE", m_sqline, 0, 15, 1, 0, 0, 0L);
	add_Command("UNSQLINE", m_unsqline, 0, 15, 1, 0, 0, 0L);
	add_Command("WHOIS", m_whois, 0, 15, 1, 0, 0, 0L);
	add_Command("WHOWAS", m_whowas, 0, 15, 1, 0, 0, 0L);
	add_Command("ADMIN", m_admin, 0, 15, 1, 1, 0, 0L);
	add_Command("INFO", m_info, 0, 15, 1, 0, 0, 0L);
	add_Command("DALINFO", m_dalinfo, 0, 15, 1, 0, 0, 0L);
	add_Command("LINKS", m_links, 0, 15, 1, 0, 0, 0L);
	add_Command("KICK", m_kick, 0, 15, 1, 0, 0, 0L);
	add_Command("SENDVHOST", m_sendvhost, 0, 15, 1, 0, 0, 0L);
	add_Command("VHOST", m_vhost, 0, 15, 1, 0, 0, 0L);
	add_Command("MODE", m_mode, 0, 15, 1, 0, 0, 0L);
	add_Command("TOPIC", m_topic, 0, 15, 1, 0, 0, 0L);
	add_Command("INVITE", m_invite, 0, 15, 1, 0, 0, 0L);
	add_Command("NAMES", m_names, 0, 15, 1, 0, 0, 0L);
	add_Command("TIME", m_time, 0, 15, 1, 0, 0, 0L);
	add_Command("PING", m_ping, 0, 15, 1, 0, 0, 0L);
	add_Command("PONG", m_pong, 0, 15, 1, 0, 0, 0L);
	add_Command("QUIT", m_quit, 0, 15, 1, 0, 0, 0L);
	add_Command("PASS", m_pass, 0, 15, 1, 1, 0, 0L);
	add_Command("USERHOST", m_userhost, 0, 15, 1, 1, 0, 0L);
	add_Command("ISON", m_ison, 0, 15, 1, 0, 0, 0L);
	add_Command("MOTD", m_motd, 0, 15, 1, 0, 0, 0L);
	add_Command("RULES", m_rules, 0, 15, 1, 0, 0, 0L);
	add_Command("OPERMOTD", m_opermotd, 0, 15, 1, 0, 0, 0L);
	add_Command("SVSNOOP", m_svsnoop, 0, 15, 1, 0, 0, 0L);
	add_Command("LIST", m_list, 0, 15, 1, 0, 0, 0L);
	add_Command("SAMODE", m_samode, 0, 15, 1, 0, 0, 0L);
	add_Command("CAPAB", m_capab, 0, 15, 1, 1, 0, 0L);
	add_Command("SJOIN", m_sjoin, 0, 15, 1, 0, 0, 0L);
	add_Command("REHASH", m_rehash, 0, 15, 1, 0, 0, 0L);
	add_Command("HTM", m_htm, 0, 15, 1, 0, 0, 0L);
	add_Command("BURST", m_burst, 0, 15, 1, 0, 0, 0L);
	add_Command("SVINFO", m_svinfo, 0, 15, 1, 1, 0, 0L);
	add_Command("USERS", m_users, 0, 15, 1, 0, 0, 0L);
	add_Command("ERROR", m_error, 0, 15, 1, 1, 0, 0L);
	add_Command("LUSERS", m_lusers, 0, 15, 1, 0, 0, 0L);
	add_Command("CONNECT", m_connect, 0, 15, 1, 0, 0, 0L);
	add_Command("WALLOPS", m_wallops, 0, 15, 1, 0, 0, 0L);
	add_Command("LOCOPS", m_locops, 0, 15, 1, 0, 0, 0L);
	add_Command("GLOBOPS", m_globops, 0, 15, 1, 0, 0, 0L);
	add_Command("GOPER", m_goper, 0, 15, 1, 0, 0, 0L);
	add_Command("GNOTICE", m_gnotice, 0, 15, 1, 0, 0, 0L);
	add_Command("SQUIT", m_squit, 0, 15, 1, 0, 0, 0L);
	add_Command("CHATOPS", m_chatops, 0, 15, 1, 0, 0, 0L);
	add_Command("CLOSE", m_close, 0, 15, 1, 0, 0, 0L);
	add_Command("NICK", m_nick, 0, 15, 1, 1, 0, 0L);
	add_Command("USER", m_user, 0, 15, 1, 1, 0, 0L);
	add_Command("PRIVMSG", m_private, 0, 15, 1, 0, 1, 0L);
	add_Command("NOTICE", m_notice, 0, 15, 1, 0, 0, 0L);
	add_Command("KILL", m_kill, 0, 15, 1, 0, 0, 0L);
	add_Command("SVSKILL", m_svskill, 0, 15, 1, 0, 0, 0L);
	add_Command("AKILL", m_akill, 0, 15, 1, 0, 0, 0L);
	add_Command("RAKILL", m_rakill, 0, 15, 1, 0, 0, 0L);
	add_Command("DNS", m_dns, 0, 15, 1, 0, 0, 0L);
	add_Command("DKEY", m_dkey, 0, 15, 1, 1, 0, 0L);
	add_Command("TRACE", m_trace, 0, 15, 1, 0, 0, 0L);
	add_Command("WATCH", m_watch, 0, 15, 1, 0, 0, 0L);
	add_Command("DCCALLOW", m_dccallow, 0, 15, 1, 1, 0, 0L);
	add_Command("SILENCE", m_silence, 0, 15, 1, 0, 0, 0L);
	add_Command("STATS", m_stats, 0, 15, 1, 0, 0, 0L);
	add_Command("WHO", m_who, 0, 15, 1, 0, 0, 0L);
	add_Command("SERVER", m_server, 0, 15, 1, 1, 0, 0L);
	add_Command("KNOCK", m_knock, 0, 15, 1, 0, 0, 0L);
	add_Command("SET", m_set, 0, 15, 1, 0, 0, 0L);
	add_Command("SGLINE", m_sgline, 0, 15, 1, 0, 0, 0L);
	add_Command("UNSGLINE", m_unsgline, 0, 15, 1, 0, 0, 0L);
	add_Command("ACCEPT", m_accept, 0, 15, 1, 1, 0, 0L);

	init_motd(1);
	init_opermotd(1);
	init_rules(1);
#endif /* STATIC_MODULES */
}

void add_Command(char *cmd, int (*func)(), unsigned int count, int parameters, int flags,
	int unreg_use, int reset_idle, unsigned long bytes)
{
	aCommand *newcmd = find_Command(cmd);
	if (!newcmd) {
		newcmd = (aCommand *)MyMalloc(sizeof(aCommand));
		DupString(newcmd->cmd, cmd);
		newcmd->func = func;
		newcmd->count = count;
		newcmd->parameters = parameters;
		newcmd->flags = flags;
		newcmd->unreg_use = unreg_use;
		newcmd->reset_idle = reset_idle;
		newcmd->bytes = bytes;
		newcmd->prev = NULL;
		newcmd->next = CommandHash[toupper(*cmd)];
		if (newcmd->next) {
			newcmd->next->prev = newcmd;
		}
		CommandHash[toupper(*cmd)] = newcmd;
	}
	return;
}

void del_Command(char *cmd, int (*func)())
{
	aCommand *p = find_CommandF(cmd, func);
	if (p) {
		if (p->prev) {
			p->prev->next = p->next;
		}
		else {
			CommandHash[toupper(*cmd)] = p->next;
		}
		if (p->next) {
			p->next->prev = p->prev;
		}
		MyFree(p->cmd);
		MyFree(p);
	}
	return;
}

inline aCommand *find_Command(char *cmd)
{
	aCommand *p;
	for (p = CommandHash[toupper(*cmd)]; p; p = p->next) {
		if (!mycmp(p->cmd, cmd)) {
			return (p);
		}
	}
	return (NULL);
}

inline aCommand *find_CommandF(char *cmd, int (*func)())
{
	aCommand *p;
	for (p = CommandHash[toupper(*cmd)]; p; p = p->next) {
		if (!mycmp(p->cmd, cmd) && p->func == func) {
			return (p);
		}
	}
	return (NULL);
}

/*
 * parse a buffer.
 * 
 * NOTE: parse() should not be called recusively by any other functions!
 */
int parse(aClient *cptr, char *buffer, char *bufend)
{
   aClient *from = cptr;
   char *ch, *s;
   int i, numeric = 0, paramcount;
   aCommand *cmptr = NULL;

#ifdef DUMP_DEBUG
   if(dumpfp!=NULL) 
   {
      fprintf(dumpfp, "<- %s: %s\n", (cptr->name ? cptr->name : "*"), buffer);
      fflush(dumpfp);
   }
#endif
   Debug((DEBUG_DEBUG, "Parsing %s: %s", get_client_name(cptr, TRUE), buffer));

   if (IsDead(cptr))
      return -1;

   s = sender;
   *s = '\0';

   for (ch = buffer; *ch == ' '; ch++);	/* skip spaces */

   para[0] = from->name;
   if (*ch == ':') 
   {
      /*
       * Copy the prefix to 'sender' assuming it terminates with
       * SPACE (or NULL, which is an error, though).
       */

      for (++ch; *ch && *ch != ' '; ++ch)
	 if (s < (sender + HOSTLEN))
	    *s++ = *ch;		
      *s = '\0';

      /*
       * Actually, only messages coming from servers can have the
       * prefix--prefix silently ignored, if coming from a user
       * client...
       * 
       * ...sigh, the current release "v2.2PL1" generates also null
       * prefixes, at least to NOTIFY messages (e.g. it puts
       * "sptr->nickname" as prefix from server structures where it's
       * null--the following will handle this case as "no prefix" at
       * all --msa  (": NOTICE nick ...")
       */

      if (*sender && IsServer(cptr)) 
      {
         from = find_client(sender, (aClient *) NULL);

         /*
          * okay, this doesn't seem to do much here.
          * from->name _MUST_ be equal to sender. That's what find_client does.
          * find_client will find servers too, and since we don't use server
          * masking, the find server call is useless (and very wasteful).
          * now, there HAS to be a from and from->name and sender have to be the same
          * for us to get to the next if. but the next if starts out with if(!from)
          * so this is UNREACHABLE CODE! AGH! - lucas
          *
          *  if (!from || mycmp(from->name, sender))
          *     from = find_server(sender, (aClient *) NULL);
          *  else if (!from && strchr(sender, '@'))
	  *     from = find_nickserv(sender, (aClient *) NULL);
          */

	 para[0] = sender;
	 /*
	  * Hmm! If the client corresponding to the prefix is not
	  * found--what is the correct action??? Now, I will ignore the
	  * message (old IRC just let it through as if the prefix just
	  * wasn't there...) --msa
	  */
	 if (!from) 
         {
            Debug((DEBUG_ERROR, "Unknown prefix (%s)(%s) from (%s)",
                   sender, buffer, cptr->name));

            ircstp->is_unpf++;
            remove_unknown(cptr, sender, buffer);
		 
            return -1;
	 }

	 if (from->from != cptr) 
         {
	    ircstp->is_wrdi++;
	    Debug((DEBUG_ERROR, "Message (%s) coming from (%s)",
		   buffer, cptr->name));

	    return cancel_clients(cptr, from, buffer);
	 }
      }
      while (*ch == ' ')
	 ch++;
   }

   if (*ch == '\0') 
   {
      ircstp->is_empt++;
      Debug((DEBUG_NOTICE, "Empty message from host %s:%s",
	     cptr->name, from->name));
      return (-1);
   }
   /*
    * Extract the command code from the packet.  Point s to the end
    * of the command code and calculate the length using pointer
    * arithmetic.  Note: only need length for numerics and *all*
    * numerics must have parameters and thus a space after the command
    * code. -avalon
    * 
    * ummm???? - Dianora
    */

   /* check for numeric */
   if (*(ch + 3) == ' ' && isdigit(*ch) && isdigit(*(ch + 1)) && isdigit(*(ch + 2))) 
   {
      cmptr = (aCommand *)NULL;
      numeric = (*ch - '0') * 100 + (*(ch + 1) - '0') * 10 + (*(ch + 2) - '0');
      paramcount = MAXPARA;
      ircstp->is_num++;
      s = ch + 3;
      *s++ = '\0';
   }
   else 
   {
      s = strchr(ch, ' ');

      if (s)
	 *s++ = '\0';

      cmptr = find_Command(ch);

      if (!cmptr || !cmptr->cmd)
      {
	 /*
          * only send error messages to things that actually sent buffers to us
          * and only people, too.
	  */
	 if (buffer[0] != '\0') 
         {
	    if (IsPerson(from))
	       sendto_one(from, ":%s %d %s %s :Unknown command",
			  me.name, ERR_UNKNOWNCOMMAND, from->name, ch);
	    Debug((DEBUG_ERROR, "Unknown (%s) from %s",
		   ch, get_client_name(cptr, TRUE)));
	 }
	 ircstp->is_unco++;
	 return -1;
      }

      paramcount = cmptr->parameters;
      i = bufend - ((s) ? s : ch);
      cmptr->bytes += i;
      /*
       * Allow only 1 msg per 2 seconds (on average) to prevent
       * dumping. to keep the response rate up, bursts of up to 5 msgs
       * are allowed -SRB Opers can send 1 msg per second, burst of ~20
       * -Taner
       */
      if ((cmptr->flags == 1) && !(IsServer(cptr))) 
      {
	if (ConfigSet.nooperflood && IsOper(cptr)) {
		cptr->since += (cptr->receiveM % 10) ? 1 : 0;
	}
	else {
		cptr->since += (2 + i / 120);
	}
      }
   }
   /*
    * Must the following loop really be so devious? On surface it
    * splits the message to parameters from blank spaces. But, if
    * paramcount has been reached, the rest of the message goes into
    * this last parameter (about same effect as ":" has...) --msa
    */

   /*
    * Note initially true: s==NULL || *(s-1) == '\0' !! 
    */

   i = 1;
   if (s) 
   {
      if (paramcount > MAXPARA)
	 paramcount = MAXPARA;
      for (;;) 
      {
	 while (*s == ' ')
	    *s++ = '\0';

	 if (*s == '\0')
	    break;
	 if (*s == ':') 
         {
	    /* The rest is a single parameter */
	    para[i++] = s + 1;
	    break;
	 }
	 para[i++] = s;
	 if (i >= paramcount)
	    break;

         while(*s && *s != ' ')
            s++;
      }
   }

   para[i] = NULL;
   if (cmptr == (aCommand *) NULL)
      return (do_numeric(numeric, cptr, from, i, para));

   cmptr->count++;

   /*
    * patch to avoid server flooding from unregistered connects 
    */
   /*
    * check allow_unregistered_use flag I've set up instead of function
    * comparing *yech* - Dianora
    */

   if (!IsRegistered(cptr) && cmptr->unreg_use == 0) {
      sendto_one(from, ":%s %d %s %s :Register first.",
		 me.name, ERR_NOTREGISTERED, from->name, ch);
      return -1;
   }

   if (IsRegisteredUser(cptr) && cmptr->reset_idle == 1)
      from->user->last = timeofday;

   return (*cmptr->func) (cptr, from, i, para);
}

static int
cancel_clients(aClient *cptr,
	       aClient *sptr,
	       char *cmd)
{
   /*
    * kill all possible points that are causing confusion here, I'm not
    * sure I've got this all right... - avalon
    * 
    * knowing avalon, probably not.
    */
   /*
    * * with TS, fake prefixes are a common thing, during the * connect
    * burst when there's a nick collision, and they * must be ignored
    * rather than killed because one of the * two is surviving.. so we
    * don't bother sending them to * all ops everytime, as this could
    * send 'private' stuff * from lagged clients. we do send the ones
    * that cause * servers to be dropped though, as well as the ones
    * from * non-TS servers -orabidoo
    */
   /*
    * Incorrect prefix for a server from some connection.  If it is a
    * client trying to be annoying, just QUIT them, if it is a server
    * then the same deal.
    */
   if (IsServer(sptr) || IsMe(sptr)) {
		/* Sorry, but at the moment this is just too much for even opers 
		 to see. -Rak */
		/* or we could just take out the message. <EG>  -wd */
		 sendto_realops_lev(DEBUG_LEV, "Message for %s[%s] from %s",
								  sptr->name, sptr->from->name,
								  get_client_name(cptr, (IsServer(cptr) ? HIDEME : TRUE)));
		if (IsServer(cptr)) {
			sendto_realops_lev(DEBUG_LEV,
									 "Not dropping server %s (%s) for Fake Direction",
									 cptr->name, sptr->name);
			return -1;
		}
		
		if (IsClient(cptr))
		  sendto_realops_lev(DEBUG_LEV,
									"Would have dropped client %s (%s@%s) [%s from %s]",
									cptr->name, cptr->user->username, cptr->user->host,
									cptr->user->server, cptr->from->name);
		return -1;
		/*
		 * return exit_client(cptr, cptr, &me, "Fake Direction");
		 */
	}
	/*
	 * Ok, someone is trying to impose as a client and things are
	 * confused.  If we got the wrong prefix from a server, send out a
	 * kill, else just exit the lame client.
	 */
	if (IsServer(cptr)) {
	  /*
		* * If the fake prefix is coming from a TS server, discard it *
		* silently -orabidoo
		* also drop it if we're gonna kill services by not doing so }:/
		*/
		if (DoesTS(cptr)) {
			if (sptr->user)
			  sendto_realops_lev(DEBUG_LEV,
										"Message for %s[%s@%s!%s] from %s (TS, ignored)",
										sptr->name, sptr->user->username, sptr->user->host,
										sptr->from->name, get_client_name(cptr, TRUE));
			return 0;
		}
		else {
			if (sptr->user)
			  sendto_realops_lev(DEBUG_LEV,
										"Message for %s[%s@%s!%s] from %s",
										sptr->name, sptr->user->username, sptr->user->host,
										sptr->from->name,
										get_client_name(cptr, (IsServer(cptr) ? HIDEME : TRUE)));
			if(IsULine(sptr)) {
				sendto_realops_lev(DEBUG_LEV,
										 "Would have killed U:lined client %s for fake direction",
										 sptr->name);
				return 0;
			}
			sendto_serv_butone(NULL,
									 ":%s KILL %s :%s (%s[%s] != %s, Fake Prefix)",
									 me.name, sptr->name, me.name,
									 sptr->name, sptr->from->name,
									 get_client_name(cptr, (IsServer(cptr) ? HIDEME : TRUE)));
			sptr->flags |= FLAGS_KILLED;
			return exit_client(cptr, sptr, &me, "Fake Prefix");
		}
	}
	return exit_client(cptr, cptr, &me, "Fake prefix");
}

static void
remove_unknown(aClient *cptr,
	       char *sender,
	       char *buffer)
{
   if (!IsRegistered(cptr))
      return;

   if (IsClient(cptr)) {
      sendto_realops_lev(DEBUG_LEV,
	 "Weirdness: Unknown client prefix (%s) from %s, Ignoring %s",
			 buffer,
			 get_client_name(cptr, FALSE), sender);
      return;
   }
   /*
    * Not from a server so don't need to worry about it.
    */
   if (!IsServer(cptr))
      return;
   /*
    * Do kill if it came from a server because it means there is a
    * ghost user on the other server which needs to be removed. -avalon
    * Tell opers about this. -Taner
    */
   if (!strchr(sender, '.'))
      sendto_one(cptr, ":%s KILL %s :%s (%s(?) <- %s)",
		 me.name, sender, me.name, sender,
		 get_client_name(cptr, FALSE));
   else {
      sendto_realops_lev(DEBUG_LEV, "Unknown prefix (%s) from %s, Squitting %s", buffer, get_client_name(cptr, HIDEME), sender);
      sendto_one(cptr, ":%s SQUIT %s :(Unknown prefix (%s) from %s)",
	       me.name, sender, buffer, get_client_name(cptr, HIDEME));
   }
}
