/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_bsd.c: network functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_bsd.c,v 1.53 2002/04/08 16:00:51 rageircd Exp $
 */

#include "memory.h"
#include "setup.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "res.h"
#include "numeric.h"
#include "patchlevel.h"
#include "zlink.h"
#include "openssl.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(SOL20)
# include <sys/filio.h>
# include <sys/select.h>
# include <unistd.h>
#endif
#include "inet.h"
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <utmp.h>
#include <sys/resource.h>

#ifdef HAVE_POLL_ON
#include <poll.h>
#if defined(SOL20) || defined(AIX)
#define CONNECTFAST
#endif /* SOL20 || AIX */
#endif /* HAVE_POLL_ON */

#ifdef	AIX
# include <time.h>
# include <arpa/nameser.h>
#else
# include "nameser.h"
#endif
#include "resolv.h"

#include "h.h"
#include "fdlist.h"
extern fdlist serv_fdlist;

extern int AllowClient(aClient *, struct hostent *, char *);
extern void config_cleanup();

#ifdef HAVE_THROTTLE_ON
extern int throttle_check(char *, int, time_t);
#endif
#ifdef HAVE_OPENSSL_ON
extern int openssl_handshake(aClient *);
#endif

#ifndef NO_PRIORITY
extern fdlist busycli_fdlist;
#endif

extern fdlist default_fdlist;

#ifndef IN_LOOPBACKNET
# define IN_LOOPBACKNET	0x7f
#endif

#if defined(MAXBUFFERS)
int rcvbufmax = 0, sndbufmax = 0;
#endif

#ifdef MAXBUFFERS
void reset_sock_opts(int, int);
#endif

aClient *local[MAXCONNECTIONS];
int highest_fd = 0, resfd = -1;
time_t timeofday;
static struct sockaddr_in mysk;

static struct sockaddr *connect_inet(ConfigItem_link *, aClient *, int *);
static int completed_connection(aClient *);
static int check_init(aClient *, char *);
static void do_dns_async(void), set_sock_opts(int, aClient *);
struct sockaddr_in vserv;
char specific_virtual_host;

#if defined(MAXBUFFERS)
static char *readbuf;
#else
static char readbuf[8192];
#endif

/*
 * Try and find the correct name to use with getrlimit() for setting
 * the max. number of files allowed to be open by this process.
 */

#ifdef RLIMIT_FDMAX
# define RLIMIT_FD_MAX   RLIMIT_FDMAX
#else
# ifdef RLIMIT_NOFILE
#  define RLIMIT_FD_MAX RLIMIT_NOFILE
# else
#  ifdef RLIMIT_OPEN_MAX
#   define RLIMIT_FD_MAX RLIMIT_OPEN_MAX
#  else
#   undef RLIMIT_FD_MAX
#  endif
# endif
#endif


/*
 * add_local_domain() 
 * Add the domain to hostname, if it is missing
 * (as suggested by eps@TOASTER.SFSU.EDU)
 */

void add_local_domain(char *hname, int size)
{
#ifdef RES_INIT
   /* try to fix up unqualified name */
   if (!strchr(hname, '.')) {
      if (!(_res.options & RES_INIT)) {
	 Debug((DEBUG_DNS, "res_init()"));
	 res_init();
      }
      if (_res.defdname[0]) {
	 (void) strncat(hname, ".", size - 1);
	 (void) strncat(hname, _res.defdname, size - 2);
      }
   }
#endif
   return;
}

/*
 * Cannot use perror() within daemon. stderr is closed in 
 * ircd and cannot be used. And, worse yet, it might have 
 * been reassigned to a normal connection...
 */

/*
 * report_error 
 * This a replacement for perror(). Record error to log and 
 * also send a copy to all *LOCAL* opers online. 
 * text    is a *format* string for outputting error. It must
 * contain only two '%s', the first will be replaced by the
 * sockhost from the cptr, and the latter will be taken from 
 * sys_errlist[errno].
 * 
 * cptr, if not NULL, is the *LOCAL* client associated with
 * the error.
 */

void report_error(char *text, aClient * cptr)
{
   int errtmp = errno;		/* debug may change 'errno' */
   char *host;
   int err, len = sizeof(err);
   extern char *strerror();

   host =
       (cptr) ? get_client_name(cptr,
				(IsServer(cptr) ? HIDEME : FALSE)) : "";

   Debug((DEBUG_ERROR, text, host, strerror(errtmp)));
   /* 
    * Get the *real* error from the socket (well try to anyway..). This
    * may only work when SO_DEBUG is enabled but its worth the gamble
    * anyway.
    */

#ifdef	SO_ERROR
   if (!IsMe(cptr) && cptr->fd >= 0)
      if (!getsockopt(cptr->fd, SOL_SOCKET, SO_ERROR, (char *) &err, &len))
	 if (err)
	    errtmp = err;
#endif
   sendto_realops_lev(DEBUG_LEV, text, host, strerror(errtmp));
   ircdlog(LOG_ERROR, text, host, strerror(errtmp));
   return;
}

/* inetport
 *
 * Create a socket in the AF_INET domain, bind it to the port given in
 * 'port' and listen to it.  Connections are accepted to this socket
 * depending on the IP# mask given by 'name'.  Returns the fd of the
 * socket created or -1 on error.
 */
int inetport(aClient * cptr, char *name, int port)
{
   static struct sockaddr_in server;
   int ad[4], len = sizeof(server);
   char ipname[20];

   if (BadPtr(name)) {
        name = "*";
   }

   ad[0] = ad[1] = ad[2] = ad[3] = 0;
   /*
    * do it this way because building ip# from separate values for each
    * byte requires endian knowledge or some nasty messing. Also means
    * easy conversion of "*" 0.0.0.0 or 134.* to 134.0.0.0 :-)
    */
   (void) sscanf(name, "%d.%d.%d.%d", &ad[0], &ad[1], &ad[2], &ad[3]);
   (void) ircsprintf(ipname, "%d.%d.%d.%d", ad[0], ad[1], ad[2], ad[3]);
 
   if (cptr != &me) {
      (void) ircsprintf(cptr->sockhost, "%-.42s.%.u",
                        name, (unsigned int) port);
      (void) strcpy(cptr->name, me.name);
   }

   /* At first, open a new socket */
   
   if (cptr->fd == -1) {
      cptr->fd = socket(AF_INET, SOCK_STREAM, 0);
      if (cptr->fd < 0 && errno == EAGAIN) {
         sendto_realops("opening stream socket %s: No more sockets",
                        get_client_name(cptr, HIDEME));
         return -1;
      }
   }
   if (cptr->fd < 0) {
      report_error("opening stream socket %s:%s", cptr);
      return -1;
   } else if (cptr->fd >= (HARD_FDLIMIT - 10)) {
      sendto_realops("No more connections allowed (%s)", cptr->name);
      (void) close(cptr->fd);
      return -1;
   }
   set_sock_opts(cptr->fd, cptr);
      
   /*
    * Bind a port to listen for new connections if port is non-null,
    * else assume it is already open and try get something from it.
    */
   if (port) {
      server.sin_family = AF_INET;
      server.sin_addr.s_addr = inet_addr(ipname);
      server.sin_port = htons(port);
      /*
       * Try 10 times to bind the socket with an interval of 20
       * seconds. Do this so we dont have to keepp trying manually to
       * bind. Why ? Because a port that has closed often lingers
       * around for a short time. This used to be the case.  Now it no
       * longer is. Could cause the server to hang for too long -
       * avalon
       */
      if (bind(cptr->fd, (struct sockaddr *) &server,
               sizeof(server)) == -1) {
         report_error("binding stream socket %s:%s", cptr);
         (void) close(cptr->fd);
         return -1;
      }
   }
   if (getsockname(cptr->fd, (struct sockaddr *) &server, &len)) {
      report_error("getsockname failed for %s:%s", cptr);
      (void) close(cptr->fd);
      return -1;
   }  
   
   if (cptr == &me) {
      /* KLUDGE to get it work... */
      char buf[1024];
        
      (void) ircsprintf(buf, rpl_str(RPL_MYPORTIS), me.name, "*",
                        ntohs(server.sin_port));
      (void) write(0, buf, strlen(buf));
   }
   if (cptr->fd > highest_fd)
      highest_fd = cptr->fd;
   cptr->ip.s_addr = name ? inet_addr(ipname) : me.ip.s_addr;
   cptr->port = (int) ntohs(server.sin_port);
   /*
    * If the operating system has a define for SOMAXCONN, use it,
    * otherwise use HYBRID_SOMAXCONN -Dianora
    */
       
#ifdef SOMAXCONN
   (void) listen(cptr->fd, SOMAXCONN);
#else
   (void) listen(cptr->fd, HYBRID_SOMAXCONN);
#endif
   local[cptr->fd] = cptr;
   
   return 0;
}

/*
 * add_listener
 *
 * Create a new client which is essentially the stub like 'me' to be used
 * for a socket that is passive (listen'ing for connections to be
 * accepted).
 */
int add_listener(ConfigItem_listen *listen)
{
	aClient *cptr = make_client(NULL, &me);
	cptr->flags = FLAGS_LISTEN;
	cptr->listener = cptr;
	cptr->from = cptr;
	cptr->class = (ConfigItem_class *)listen;
	cptr->umode = listen->options;

	SetMe(cptr);
	strncpyzt(cptr->name, listen->ip, sizeof(cptr->name));

	if (inetport(cptr, listen->ip, listen->port)) {
		cptr->fd = -2;
	}
	if (cptr->fd >= 0) {
		set_non_blocking(cptr->fd, cptr);
		return 1;
	}
	else {
		free_client(cptr);
		return -1;
	}
	return 0;
}

/*
 * close_listeners
 *
 * Close and free all clients which are marked as having their socket open
 * and in a state where they can accept connections.  Unix sockets have
 * the path to the socket unlinked for cleanliness.
 */
void close_listeners()
{
	ConfigItem_listen *listen = NULL;
	aClient *cptr;
	int i, loop = 1;

	while (loop) {
		loop = 0;
		for (i = highest_fd; i >= 0; i--) {
			if (!(cptr = local[i]) || !IsListening(cptr)) {
				continue;
			}
			listen = (ConfigItem_listen *)cptr->class;
			if (listen->flag.temporary && listen->clients <= 0) {
				close_connection(cptr);
				loop = 1;
			}
		}
	}
}

#ifdef HAVE_FD_ALLOC
fd_set *write_set, *read_set;
#endif

/* init_sys */
void init_sys()
{
   int fd;

#ifdef RLIMIT_FD_MAX
   struct rlimit limit;

   if (!getrlimit(RLIMIT_FD_MAX, &limit)) {
	 if (limit.rlim_max < MAXCONNECTIONS)
	 {
	    (void) fprintf(stderr, "ircd fd table too big\n");
	    (void) fprintf(stderr, "Hard Limit: %ld IRC max: %d\n",
			   (long) limit.rlim_max, MAXCONNECTIONS);
	    (void) fprintf(stderr, "Fix MAXCONNECTIONS\n");
	    exit(-1);
	 }
      limit.rlim_cur = limit.rlim_max;	/* make soft limit the max */
      if (setrlimit(RLIMIT_FD_MAX, &limit) == -1) {
	 (void) fprintf(stderr, "error setting max fd's to %ld\n",
			(long) limit.rlim_cur);
	 exit(-1);
      }

# ifndef HAVE_POLL_ON
      if (MAXCONNECTIONS > FD_SETSIZE) {
	 (void) fprintf(stderr,
			"FD_SETSIZE = %d MAXCONNECTIONS = %d\n",
			FD_SETSIZE, MAXCONNECTIONS);
	 (void) fprintf(stderr,
			"Make sure your kernel supports a larger FD_SETSIZE then recompile with -DFD_SETSIZE=%d\n",
			MAXCONNECTIONS);
	 exit(-1);
      }
# endif

# ifdef HAVE_FD_ALLOC
      read_set = FD_ALLOC(MAXCONNECTIONS);
      write_set = FD_ALLOC(MAXCONNECTIONS);
# endif
   }
#endif

# if !defined(SOL20)
   (void) setlinebuf(stderr);
# endif

   for (fd = 3; fd < MAXCONNECTIONS; fd++) {
      (void) close(fd);
      local[fd] = NULL;
   }
   local[1] = NULL;

   if (bootopt & BOOT_TTY) {
      /* debugging is going to a tty */
      resfd = init_resolver(0x1f);
      return;
   }
   (void) close(1);
   if (!(bootopt & BOOT_DEBUG) && !(bootopt & BOOT_STDERR))
      (void) close(2);

   if (((bootopt & BOOT_CONSOLE) || isatty(0)) &&
       !(bootopt & (BOOT_INETD | BOOT_OPER)) && !(bootopt & BOOT_STDERR)) {
      int pid;

      if ((pid = fork()) < 0) {
	 if ((fd = open("/dev/tty", O_RDWR)) >= 0)
	    write(fd, "Couldn't fork!\n", 15);	/* crude, but effective */
	 exit(0);
      } else if (pid > 0)
	 exit(0);
#ifdef TIOCNOTTY
      if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
	 (void) ioctl(fd, TIOCNOTTY, (char *) NULL);
	 (void) close(fd);
      }
#endif
#if defined(SOL20) || defined(DYNIXPTX) || \
    defined(_POSIX_SOURCE) || defined(SVR4)
      (void) setsid();
#else
      (void) setpgrp(0, (int) getpid());
#endif
      (void) close(0);		/* fd 0 opened by inetd */
      local[0] = NULL;
   }

   resfd = init_resolver(0x1f);
   return;
}

void write_pidfile()
{
#ifdef PID_FILE
   int fd;
   char buff[20];

   if ((fd = open(PID_FILE, O_CREAT | O_WRONLY, 0600)) >= 0) {
      (void) ircsprintf(buff, "%5d\n", (int) getpid());
      if (write(fd, buff, strlen(buff)) == -1)
	 Debug((DEBUG_NOTICE, "Error writing to pid file %s",
		PID_FILE));
      (void) close(fd);
      return;
   }
# ifdef	DEBUGMODE
   else
      Debug((DEBUG_NOTICE, "Error opening pid file %s", PID_FILE));
# endif
#endif
}

/*
 * Initialize the various name strings used to store hostnames. This is
 * set from either the server's sockhost (if client fd is a tty or
 * localhost) or from the ip# converted into a string. 0 = success, -1
 * = fail.
 */
static int check_init(aClient * cptr, char *sockn)
{
   struct sockaddr_in sk;
   int len = sizeof(struct sockaddr_in);

   /* If descriptor is a tty, special checking... * IT can't EVER be a tty */

   if (getpeername(cptr->fd, (struct sockaddr *) &sk, &len) == -1) {

/*
 * This fills syslog, if on, and is just annoying. Nobody needs it. -lucas
 *    report_error("connect failure: %s %s", cptr);
 */
      return -1;
   }
   (void) strcpy(sockn, (char *) inetntoa((char *) &sk.sin_addr));
   if (inet_netof(sk.sin_addr) == IN_LOOPBACKNET) {
      cptr->hostp = NULL;
      strncpyzt(sockn, me.sockhost, HOSTLEN);
   }
   memcpy((char *) &cptr->ip, (char *) &sk.sin_addr,

	  sizeof(struct in_addr));

   cptr->port = (int) (ntohs(sk.sin_port));

   return 0;
}

/*
 * Ordinary client access check. Look for conf lines which have the
 * same status as the flags passed. 0 = Success -1 = Access denied -2 =
 * Bad socket.
 */
int check_client(aClient *cptr)
{
   static char sockname[HOSTLEN + 1];
   struct hostent *hp = NULL;
   int i;

   ClearAccess(cptr);
   Debug((DEBUG_DNS, "ch_cl: check access for %s[%s]",
	  cptr->name, inetntoa((char *) &cptr->ip)));

   if (check_init(cptr, sockname))
      return -2;

   hp = cptr->hostp;
   /* 
    * Verify that the host to ip mapping is correct both ways and that
    * the ip#(s) for the socket is listed for the host.
    */
   if (hp) {
      for (i = 0; hp->h_addr_list[i]; i++)
	 if (!memcmp(hp->h_addr_list[i], (char *) &cptr->ip,
		     sizeof(struct in_addr))) break;

      if (!hp->h_addr_list[i]) {
	 sendto_one(cptr,
		    "NOTICE AUTH :*** Your forward and reverse DNS do not match, ignoring hostname.");
	 hp = NULL;
      }
   }

   if ((i = AllowClient(cptr, hp, sockname))) {
      Debug((DEBUG_DNS, "ch_cl: access denied: %s[%s]",
	     cptr->name, sockname));
      return i;
   }

   Debug((DEBUG_DNS, "ch_cl: access ok: %s[%s]", cptr->name, sockname));

   if (inet_netof(cptr->ip) == IN_LOOPBACKNET ||
       inet_netof(cptr->ip) == inet_netof(mysk.sin_addr)) {
      ircstp->is_loc++;
      cptr->flags |= FLAGS_LOCAL;
   }
   return 0;
}

/*
 * completed_connection 
 * Complete non-blocking
 * connect()-sequence. Check access and *       terminate connection,
 * if trouble detected. *
 *
 *      Return  TRUE, if successfully completed *               FALSE,
 * if failed and ClientExit
 */
static int completed_connection(aClient *cptr)
{
	ConfigItem_link *link = cptr->serv ? cptr->serv->conf : NULL;

	SetHandshake(cptr);

	if (!link) {
		sendto_gnotice("Lost link configuration for %s", get_client_name(cptr, HIDEME));
		return -1;
	}

	if (!BadPtr(link->password)) {
		sendto_one(cptr, "PASS %s :TS", link->password);
	}

	sendto_one(cptr, "CAPAB NOQUIT BURST UNCONNECT ZIP SSJOIN SSJv2 NICKIP SENDVHOST");
#ifdef HAVE_OPENSSL_ON
	if (link->options & LINK_DKEY) {
		sendto_one(cptr, "CAPAB DKEY");
	}
#endif
	sendto_one(cptr, "SERVER %s 1 :%s", me.name, me.info);

	if (ConfigSet.doidentd && !IsDead(cptr)) {
		start_auth(cptr);
	}
	return (IsDead(cptr)) ? -1 : 0;
}

/*
 * close_connection
 *
 * Close the physical connection. This function must make
 * MyConnect(cptr) == FALSE, and set cptr->from == NULL.
 */
void close_connection(aClient *cptr)
{
	int i, j;
	int empty = cptr->fd;

	if (IsServer(cptr)) {
		ircstp->is_sv++;
		ircstp->is_sbs += cptr->sendB;
		ircstp->is_sbr += cptr->receiveB;
		ircstp->is_sks += cptr->sendK;
		ircstp->is_skr += cptr->receiveK;
		ircstp->is_sti += timeofday - cptr->firsttime;
		if (ircstp->is_sbs > 2047) {
			ircstp->is_sks += (ircstp->is_sbs >> 10);
			ircstp->is_sbs &= 0x3ff;
		}
		if (ircstp->is_sbr > 2047) {
			ircstp->is_skr += (ircstp->is_sbr >> 10);
			ircstp->is_sbr &= 0x3ff;
		}
	} else if (IsClient(cptr)) {
		ircstp->is_cl++;
		ircstp->is_cbs += cptr->sendB;
		ircstp->is_cbr += cptr->receiveB;
		ircstp->is_cks += cptr->sendK;
		ircstp->is_ckr += cptr->receiveK;
		ircstp->is_cti += timeofday - cptr->firsttime;
		if (ircstp->is_cbs > 2047) {
			ircstp->is_cks += (ircstp->is_cbs >> 10);
			ircstp->is_cbs &= 0x3ff;
		}
		if (ircstp->is_cbr > 2047) {
			ircstp->is_ckr += (ircstp->is_cbr >> 10);
			ircstp->is_cbr &= 0x3ff;
		}
	}
	else {
		ircstp->is_ni++;
	}

	del_queries((char *) cptr);

	if (IsServer(cptr) && !cptr->serv->conf->flag.temporary && (cptr->serv->conf->options & LINK_AUTOCONNECT)) {
		ConfigItem_link *link = cptr->serv->conf;
		link->hold = time(NULL);
		link->hold += (link->hold - cptr->since > HANGONGOODLINK) ? HANGONRETRYDELAY : link->class->connfreq;
		if (nextconnect > link->hold) {
			nextconnect = link->hold;
		}
	}

	if (cptr->authfd >= 0) {
		close(cptr->authfd);
	}
	if (cptr->fd >= 0) {
		dump_connections(cptr->fd);
		local[cptr->fd] = NULL;
		close(cptr->fd);
		cptr->fd = -2;
		DBufClear(&cptr->sendQ);
		DBufClear(&cptr->recvQ);
		memset(cptr->passwd, '\0', sizeof(cptr->passwd));
	}

	for (; highest_fd > 0; highest_fd--) {
		if (local[highest_fd]) {
			break;
		}
	}

	cptr->from = NULL;
#ifdef HAVE_OPENSSL_ON
	if (cptr->flags & FLAGS_OPENSSL && cptr->openssl != NULL) {
		SSL_shutdown((SSL *)cptr->openssl);
		SSL_free((SSL *)cptr->openssl);
	}
#endif

	if (empty > 0) {
		if ((j = highest_fd) > (i = empty) && !IsLog(local[j]) && !IsMe(local[j])) {
			if (dup2(j, i) == -1) {
				return;
			}
			local[i] = local[j];
			local[i]->fd = i;
#ifdef HAVE_OPENSSL_ON
			if (local[i]->flags & FLAGS_OPENSSL && local[i]->openssl != NULL) {
				openssl_remap_fd((SSL *)local[i]->openssl, local[i]->fd);
			}
#endif
			local[j] = NULL;

			if (IsServer(local[i])) {
#ifndef NO_PRIORITY
				delfrom_fdlist(j, &busycli_fdlist);
#endif
				delfrom_fdlist(j, &serv_fdlist);
#ifndef NO_PRIORITY
				addto_fdlist(i, &busycli_fdlist);
#endif
				addto_fdlist(i, &serv_fdlist);
			}

			if (IsOper(local[i])) {
#ifndef NO_PRIORITY
				delfrom_fdlist(j, &busycli_fdlist);
#endif
				delfrom_fdlist(j, &oper_fdlist);
#ifndef NO_PRIORITY
				addto_fdlist(i, &busycli_fdlist);
#endif
				addto_fdlist(i, &oper_fdlist);
			}

			close(j);
			while (!local[highest_fd]) {
				highest_fd--;
			}
		}
	}
	return;
}

#ifdef MAXBUFFERS

/*
 * reset_sock_opts type =  0 = client, 1 = server
 */
void reset_sock_opts(int fd, int type)
{
# define CLIENT_BUFFER_SIZE	4096
# define SEND_BUF_SIZE		2048
   int opt;

   opt = type ? rcvbufmax : CLIENT_BUFFER_SIZE;
   if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &opt, sizeof(opt)) <
       0) sendto_realops("REsetsockopt(SO_RCVBUF) for fd %d (%s) failed",
			 fd, type ? "server" : "client");
   opt = type ? (SEND_BUF_SIZE * 4) : SEND_BUF_SIZE;
   if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &opt, sizeof(opt)) <
       0) sendto_realops("REsetsockopt(SO_SNDBUF) for fd %d (%s) failed",
			 fd, type ? "server" : "client");
}

#endif				/* 
				 * MAXBUFFERS
				 */

/*
 * * set_sock_opts
 */
static void set_sock_opts(int fd, aClient * cptr)
{
   int opt;

#ifdef SO_REUSEADDR
   opt = 1;
   if (setsockopt
       (fd, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof(opt)) < 0)
      report_error("setsockopt(SO_REUSEADDR) %s:%s", cptr);
#endif
#if  defined(SO_DEBUG) && defined(DEBUGMODE) && 0
   /* 
    * Solaris with SO_DEBUG writes to syslog by default
    */
# if !defined(SOL20)
   opt = 1;
   if (setsockopt(fd, SOL_SOCKET, SO_DEBUG, (char *) &opt, sizeof(opt)) <
       0) report_error("setsockopt(SO_DEBUG) %s:%s", cptr);
# endif				/* 
				 * SOL20
				 */
#endif
#ifdef	SO_USELOOPBACK
   opt = 1;
   if (setsockopt
       (fd, SOL_SOCKET, SO_USELOOPBACK, (char *) &opt, sizeof(opt)) < 0)
      report_error("setsockopt(SO_USELOOPBACK) %s:%s", cptr);
#endif
#ifdef	SO_RCVBUF
# if defined(MAXBUFFERS)
   if (rcvbufmax == 0) {
      int optlen;

      optlen = sizeof(rcvbufmax);
      getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &rcvbufmax, &optlen);
      while ((rcvbufmax < 16385)
	     &&
	     (setsockopt
	      (fd, SOL_SOCKET, SO_RCVBUF, (char *) (char *) &rcvbufmax,
	       optlen) >= 0)) rcvbufmax += 1024;
      getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &rcvbufmax, &optlen);
      readbuf = (char *) malloc(rcvbufmax * sizeof(char));
   }
   if (IsServer(cptr))
      opt = rcvbufmax;
   else
      opt = 4096;
# else
   opt = 8192;
# endif
   if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *) &opt, sizeof(opt)) <
       0) report_error("setsockopt(SO_RCVBUF) %s:%s", cptr);
#endif
#ifdef	SO_SNDBUF
# if defined(MAXBUFFERS)
   if (sndbufmax == 0) {
      int optlen;

      optlen = sizeof(sndbufmax);
      getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &sndbufmax, &optlen);
      while ((sndbufmax < 16385)
	     &&
	     (setsockopt
	      (fd, SOL_SOCKET, SO_SNDBUF, (char *) &sndbufmax,
	       optlen) >= 0)) sndbufmax += 1024;
      getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &sndbufmax, &optlen);
   }
   if (IsServer(cptr))
      opt = sndbufmax;
   else
     opt = 4096;
# else
   opt = 8192;
# endif
   if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *) &opt, sizeof(opt)) <
       0) report_error("setsockopt(SO_SNDBUF) %s:%s", cptr);
#endif
#if defined(IP_OPTIONS) && defined(IPPROTO_IP)
   {
# if defined(MAXBUFFERS)
      char *s = readbuf, *t = readbuf + (rcvbufmax * sizeof(char)) / 2;
      opt = (rcvbufmax * sizeof(char)) / 8;

# else
      char *s = readbuf, *t = readbuf + sizeof(readbuf) / 2;

      opt = sizeof(readbuf) / 8;
# endif
      if (getsockopt(fd, IPPROTO_IP, IP_OPTIONS, t, &opt) < 0)
	 report_error("getsockopt(IP_OPTIONS) %s:%s", cptr);
      else if (opt > 0) {
	 for (*readbuf = '\0'; opt > 0; opt--, s += 3)
	    (void) ircsprintf(s, "%02.2x:", *t++);
	 *s = '\0';
	 sendto_realops("Connection %s using IP opts: (%s)",
			get_client_name(cptr, HIDEME), readbuf);
      }
      if (setsockopt(fd, IPPROTO_IP, IP_OPTIONS, (char *) NULL, 0) < 0)
	 report_error("setsockopt(IP_OPTIONS) %s:%s", cptr);
   }
#endif
}

int get_sockerr(aClient * cptr)
{
   int errtmp = errno, err = 0, len = sizeof(err);

#ifdef	SO_ERROR
   if (cptr->fd >= 0)
      if (!getsockopt(cptr->fd, SOL_SOCKET, SO_ERROR, (char *) &err, &len))
	 if (err)
	    errtmp = err;
#endif
   return errtmp;
}

char *irc_get_sockerr(aClient *cptr)
{
   if(cptr->sockerr == 0)
      return "No error";

   if(cptr->sockerr > 0)
      return strerror(cptr->sockerr);

   switch(cptr->sockerr)
   { 
      case -1: /* this is the default */
         return "Unset error message [this is a bug!]";
      case IRCERR_BUFALLOC:
         return "dbuf allocation error";
      case IRCERR_ZIP:
         return "compression general failure";
      default:
         return "Unknown error!";
   }
   
   /* unreachable code, but the compiler is complaining.. */
   return NULL;
}

/* set_blocking - set the client connection into non-blocking mode.
 * If your system doesn't support this, you're screwed, ircd will run
 * like crap.
 *
 *returns true (1)if successful, false (0) is otherwise
 */
int set_blocking(int fd)
{
	int flags;

	if ((flags = fcntl(fd, F_GETFL, 0)) < 0 || fcntl(fd, F_SETFL, flags & ~O_NONBLOCK) < 0) {
		return 0;
	}
	else {
		return 1;
	}
}

/*
 * * set_non_blocking *       Set the client connection into non-blocking
 * mode. If your *      system doesn't support this, you can make this
 * a dummy *    function (and get all the old problems that plagued the *
 * blocking version of IRC--not a problem if you are a *        lightly
 * loaded node...)
 */
void set_non_blocking(int fd, aClient * cptr)
{
   int res, nonb = 0;

   /* 
    * * NOTE: consult ALL your relevant manual pages *BEFORE* changing *
    * hese ioctl's.  There are quite a few variations on them, *
    * s can be seen by the PCS one.  They are *NOT* all the same. *
    * eed this well. - Avalon.
    */
#ifdef	NBLOCK_POSIX
   nonb |= O_NONBLOCK;
#endif
#ifdef	NBLOCK_BSD
   nonb |= O_NDELAY;
#endif
#ifdef	NBLOCK_SYSV
   /* 
    * This portion of code might also apply to NeXT.  -LynX
    */
   res = 1;

   if (ioctl(fd, FIONBIO, &res) < 0)
      report_error("ioctl(fd,FIONBIO) failed for %s:%s", cptr);
#else
   if ((res = fcntl(fd, F_GETFL, 0)) == -1)
      report_error("fcntl(fd, F_GETFL) failed for %s:%s", cptr);
   else if (fcntl(fd, F_SETFL, res | nonb) == -1)
      report_error("fcntl(fd, F_SETL, nonb) failed for %s:%s", cptr);
#endif
   return;
}

/*
 * Creates a client which has just connected to us on the given fd. The
 * sockhost field is initialized with the ip# of the host. The client
 * is added to the linked list of clients but isnt added to any hash
 * tables yuet since it doesnt have a name.
 */
aClient *add_connection(aClient* cptr, int fd)
{
	Link lin;
	aClient *acptr = make_client(NULL, &me);

	if (isatty(fd)) {
		get_sockhost(acptr, cptr->sockhost);
	}
	else {
		struct sockaddr_in addr;
		int len = sizeof(struct sockaddr_in);

		if (getpeername(fd, (struct sockaddr *)&addr, &len) == -1) {
			ircstp->is_ref++;
			acptr->fd = -2;
			free_client(acptr);
			close(fd);
			return NULL;
		}

		get_sockhost(acptr, (char *)inetntoa((char *)&addr.sin_addr));
		memcpy((char *)&acptr->ip, (char *)&addr.sin_addr, sizeof(struct in_addr));
		acptr->port = htons(addr.sin_port);

#ifdef HAVE_OPENSSL_ON
		if (ConfigSet.showheaders && !(cptr->umode & LISTEN_OPENSSL)) {
#else
		if (ConfigSet.showheaders) {
#endif
			write(fd, REPORT_DO_DNS, R_do_dns);
		}

		lin.flags = ASYNC_CLIENT;
		lin.value.cptr = acptr;
		Debug((DEBUG_DNS, "lookup %s", inetntoa((char *)&addr.sin_addr)));
		acptr->hostp = gethost_byaddr((char *)&acptr->ip, &lin);
		if (!acptr->hostp) {
			SetDNS(acptr);
		}
#ifdef HAVE_OPENSSL_ON
		else if (ConfigSet.showheaders && !(cptr->umode & LISTEN_OPENSSL)) {
#else
		else if (ConfigSet.showheaders) {
#endif
			write(fd, REPORT_FIN_DNSC, R_fin_dnsc);
		}
		nextdnscheck = 1;
	}

	acptr->fd = fd;
	if (fd > highest_fd) {
		highest_fd = fd;
	}
	local[fd] = acptr;
	acptr->listener = cptr;
	if (acptr->listener->class) {
		((ConfigItem_listen *)acptr->listener->class)->clients++;
	}
#ifdef HAVE_OPENSSL_ON
	if (acptr->listener->umode & LISTEN_OPENSSL) {
		openssl_handshake(acptr);
		acptr->flags |= FLAGS_OPENSSL;
	}
#endif
	add_client_to_list(acptr);
	set_non_blocking(acptr->fd, acptr);
	set_sock_opts(acptr->fd, acptr);
	if (ConfigSet.doidentd) {
		start_auth(acptr);
	}
	return acptr;
}

/* handle taking care of the client's recvq here */
static int do_client_queue(aClient *cptr)
{
   int dolen = 0, done;

   while (DBufLength(&cptr->recvQ) && !NoNewLine(cptr) &&
	  ((cptr->status < STAT_UNKNOWN) || (cptr->since - timeofday < 10) || IsNegoServer(cptr))) 
   {
      /* If it's become registered as a server, just parse the whole block */
      if (IsServer(cptr) || IsNegoServer(cptr)) 
      {
#if defined(MAXBUFFERS)
         dolen = dbuf_get(&cptr->recvQ, readbuf, rcvbufmax * sizeof(char));
#else
         dolen = dbuf_get(&cptr->recvQ, readbuf, sizeof(readbuf));
#endif
         if (dolen <= 0)
            break;
         if ((done = dopacket(cptr, readbuf, dolen)))
            return done;
         break;
      }

#if defined(MAXBUFFERS)
      dolen = dbuf_getmsg(&cptr->recvQ, readbuf, rcvbufmax * sizeof(char));
#else
      dolen = dbuf_getmsg(&cptr->recvQ, readbuf, sizeof(readbuf));
#endif

      if (dolen <= 0) 
      {
         if (dolen < 0)
            return exit_client(cptr, cptr, cptr, "dbuf_getmsg fail");

         if (DBufLength(&cptr->recvQ) < 510) 
         {
            cptr->flags |= FLAGS_NONL;
            break;
         }
         /* The buffer is full (more than 512 bytes) and it has no \n
          * Some user is trying to trick us. Kill their recvq. */
         DBufClear(&cptr->recvQ);
         break;
      }
      else if(client_dopacket(cptr, readbuf, dolen) == FLUSH_BUFFER)
         return FLUSH_BUFFER;
   }
   return 1;
}

/*
 * read_packet
 *
 * Read a 'packet' of data from a connection and process it.  Read in 8k 
 * chunks to give a better performance rating (for server connections). 
 * Do some tricky stuff for client connections to make sure they don't
 * do any flooding >:-) -avalon
 */

#define MAX_CLIENT_RECVQ 8192	/* 4 dbufs */

static int read_packet(aClient *cptr)
{
   int length = 0, done;

   /* If data is ready, and the user is either not a person or
    * is a person and has a recvq of less than MAX_CLIENT_RECVQ,
    * read from this client
    */ 
   if (!(IsPerson(cptr) && DBufLength(&cptr->recvQ) > MAX_CLIENT_RECVQ)) 
   {
      errno = 0;
#ifdef HAVE_OPENSSL_ON
	if (cptr->flags & FLAGS_OPENSSL) {
		length = SSL_read((SSL *)cptr->openssl, readbuf, sizeof(readbuf));
	}
	else
#endif
	{
		length = recv(cptr->fd, readbuf, sizeof(readbuf), 0);
	}

      cptr->lasttime = timeofday;
      if (cptr->lasttime > cptr->since)
	 cptr->since = cptr->lasttime;
      cptr->flags &= ~(FLAGS_PINGSENT | FLAGS_NONL);
      /* 
       * If not ready, fake it so it isnt closed
       */
      if (length == -1 && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
	 return 1;
      if (length <= 0)
      {
         cptr->sockerr = length ? errno : 0;
	 return length;
      }
   }

   /* 
    * For server connections, we process as many as we can without
    * worrying about the time of day or anything :)
    */
   if (IsServer(cptr) || IsConnecting(cptr) || IsHandshake(cptr) || IsNegoServer(cptr)) 
   {
      if (length > 0)
	 if ((done = dopacket(cptr, readbuf, length)))
	    return done;
   } 
   else 
   {
      /* 
       * Before we even think of parsing what we just read, stick 
       * it on the end of the receive queue and do it when its turn
       * comes around. */
      if (dbuf_put(&cptr->recvQ, readbuf, length) < 0)
	 return exit_client(cptr, cptr, cptr, "dbuf_put fail");

      if (IsPerson(cptr) && ((ConfigSet.nooperflood && !IsOper(cptr)) || !ConfigSet.nooperflood) &&
	DBufLength(&cptr->recvQ) > CLIENT_FLOOD) {
		sendto_realops_lev(FLOOD_LEV, "Flood -- %s!%s@%s (%d) Exceeds %d RecvQ",
			cptr->name, cptr->user->username,
			IsMasked(cptr) ? cptr->user->maskedhost : cptr->user->host,
			DBufLength(&cptr->recvQ), CLIENT_FLOOD);
	return exit_client(cptr, cptr, cptr, "Excess Flood");
      }
      return do_client_queue(cptr);
   }
   return 1;
}

void accept_connection(aClient *cptr)
{
	static struct sockaddr_in addr;
	char host[HOSTLEN + 2];
	int newfd, addrlen = sizeof(struct sockaddr_in);

	cptr->lasttime = timeofday;
	if ((newfd = accept(cptr->fd, (struct sockaddr *)&addr, &addrlen)) < 0) {
#ifdef EPROTO
		if (errno != EPROTO)
#endif
		report_error("Cannot accept connections %s:%s", cptr);
		return;
	}
	strncpyzt(host, (char *)inetntoa((char *)&addr.sin_addr), sizeof(host));

#ifdef HAVE_THROTTLE_ON
	if (!throttle_check(host, newfd, NOW)) {
		ircstp->is_ref++;
		close(newfd);
		return;
	}
#endif

	if (newfd >= HARD_FDLIMIT - 10) {
		ircstp->is_ref++;
		sendto_ops("All connections in use. (%s)", get_client_name(cptr, HIDEME));
		send(newfd, "ERROR :All connections in use\r\n", 32, 0);
		close(newfd);
		return;
	}

	ircstp->is_ac++;
	add_connection(cptr, newfd);
#ifdef PINGNAZI
	nextping = timeofday;
#endif
	if (!cptr->listener) {
		cptr->listener = &me;
	}
}

static void read_error_exit(aClient *cptr, int length, int err)
{
	char errmsg[512], *clipath = get_client_name(cptr, HIDEME);

	if (IsServer(cptr) || IsHandshake(cptr) || IsConnecting(cptr)) {
		if (length == 0) {
			sendto_gnotice("Server %s closed the connection", clipath);
			sendto_serv_butone(cptr, ":%s GNOTICE :Server %s closed the connection",
				me.name, clipath);
		}
		else {
			sendto_gnotice("Read error from %s, closing link (%s)", clipath, strerror(err));
			sendto_serv_butone(cptr, ":%s GNOTICE :Read error from %s, closing link (%s)",
				me.name, clipath, strerror(err));
		}
	}
	if (err) {
		ircsprintf(errmsg, "Read error: %s", strerror(err));
	}
	else {
		ircsprintf(errmsg, "Client closed connection");
	}
	exit_client(cptr, cptr, &me, errmsg);
	return;
}

/*
 * USE_FAST_FD_ISSET
 *
 * The idea with this, is to save the compute over and over again of
 * nearly the same thing.
 *
 * In SUNOS and BSD the following is done for a FD_ISSET
 *
 * (p)->fd_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
 *
 *
 * Thats one divide, one left shift, and one AND for every case of
 * FD_ISSET used. What I have done here, is saved the last value of
 * offset, and mask and increment them for use on the next fd.
 * hopefully adding up to some CPU savings.
 *
 * The caveats are the following:
 *
 * 1) sizeof(NFDBITS) != sizeof(int)
 *
 * 2) structure element fd_bits might not be, i.e. its linux or some
 * variant.
 *
 * i.e. basically, we now have carnal knowledge of the internals of what
 * happens in an FD_ISSET()
 *
 * 3) the fd list has to be scanned in a linear order, not as it was using
 * the fdlist.
 *
 * -Dianora
 */

#ifdef FAST_FD_ISSET
# define MYFD_ISSET_READ (read_set->fds_bits[fd_read_offset] & fd_read_mask)
# define MYFD_ISSET_WRITE (write_set->fds_bits[fd_write_offset] & fd_write_mask)
# define MYFD_CLR_READ read_set->fds_bits[fd_read_offset] &= ~fd_read_mask;
# define FAST_FD_INC fd_read_mask <<= 1; if (!fd_read_mask) { fd_read_offset++; fd_read_mask = 1; } fd_write_mask <<= 1; if (!fd_write_mask) { fd_write_offset++; fd_write_mask = 1; }
#else
# define MYFD_ISSET_READ (FD_ISSET(i, read_set))
# define MYFD_ISSET_WRITE (FD_ISSET(i, write_set))
# define MYFD_CLR_READ FD_CLR(i, read_set);
# define FAST_FD_INC
#endif

/*
 * Check all connections for new connections and input data that is to
 * be processed. Also check for connections with data queued and
 * whether we can write it out.
 */
#ifndef HAVE_POLL_ON

int read_message(time_t delay, fdlist *listp)
{
	aClient *cptr;
	int nfds, res, length, auth = 0, i, j;
	struct timeval wait;
#ifndef HAVE_FD_ALLOC
	fd_set readset, writeset;
	fd_set *read_set = &readset, *write_set = &writeset;
#endif
	time_t delay2 = delay, now;
	char errmsg[512];
#ifdef USE_FAST_FD_ISSET
	int fd_read_mask, fd_read_offset, fd_write_mask, fd_write_offset;
#endif

	if (!listp) {
		listp = &default_fdlist;
		listp->last_entry = highest_fd + 1;
	}

	now = timeofday;
	for (res = 0;;) {
		FD_ZERO(read_set);
		FD_ZERO(write_set);
		for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j]) {
			if (!(cptr = local[i]) || IsLog(cptr)) {
				continue;
			}
			if (DoingAuth(cptr)) {
				auth++;
				Debug((DEBUG_NOTICE, "auth on %x %d", cptr, i));
				FD_SET(cptr->authfd, read_set);
				if (cptr->flags & FLAGS_WRAUTH) {
					FD_SET(cptr->authfd, write_set);
				}
			}
			if (DoingDNS(cptr) || DoingAuth(cptr)) {
				continue;
			}
			if ((IsMe(cptr) && IsListening(cptr)) || IsConnecting(cptr)) {
				FD_SET(i, read_set);
			}
			else if (!IsMe(cptr)) {
				if (DBufLength(&cptr->recvQ) && delay2 > 2) {
					delay2 = 1;
				}
				if (DBufLength(&cptr->recvQ) < 4088) {
					FD_SET(i, read_set);
				}
			}

			length = DBufLength(&cptr->sendQ);

			if (DoList(cptr) && IsSendable(cptr)) {
				send_list(cptr, 64);
				length = DBufLength(&cptr->sendQ);
			}

			if (length || IsConnecting(cptr) || (ZipOut(cptr) && zip_is_data_out(cptr->serv->zip_out))) {
				FD_SET(i, write_set);
			}
		}

		if (resfd >= 0) {
			FD_SET(resfd, read_set);
		}

		wait.tv_sec = MIN(delay2, delay);
		wait.tv_usec = 0;
		nfds = select(MAXCONNECTIONS, read_set, write_set, 0, &wait);

		if ((timeofday = time(NULL)) == -1) {
			ircdlog(LOG_ERROR, "Clock Failure (%d), TS can be corrupted", errno);
			sendto_ops("Clock Failiure (%d), TS can be corrupted", errno);
		}

		if (nfds == -1 && errno == EINTR) {
			return -1;
		}
		else if (nfds >= 0) {
			break;
		}

		report_error("select %s:%s", &me);
		res++;
		if (res > 5) {
			restart("too many select errors");
		}
		sleep(10);
	}

	if (resfd >= 0 && FD_ISSET(resfd, read_set)) {
		do_dns_async();
		nfds--;
		FD_CLR(resfd, read_set);
	}

#ifdef USE_FAST_FD_ISSET
	fd_read_mask = fd_write_mask = 1;
	fd_read_offset = fd_write_offset = 0;

	for (i = 0; i <= highest_fd; i++)
#else
	for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j])
#endif
	{
		if (!(cptr = local[i])) {
			FAST_FD_INC
			continue;
		}

		if (auth > 0 && cptr->authfd >= 0) {
			auth--;
			if (nfds > 0 && FD_ISSET(cptr->authfd, write_set)) {
				nfds--;
				send_authports(cptr);
			}
			else if (nfds > 0 && FD_ISSET(cptr->authfd, read_set)) {
				nfds--;
				read_authports(cptr);
			}
		}

		if (IsListening(cptr) && MYFD_ISSET_READ) {
			MYFD_CLR_READ
			nfds--;
			accept_connection(cptr);
			FAST_FD_INC
			continue;
		}
		if (IsMe(cptr)) {
			FAST_FD_INC
			continue;
		}
		if (MYFD_ISSET_WRITE) {
			int write_err = 0;
			nfds--;

			if (IsConnecting(cptr)) {
				write_err = completed_connection(cptr);
			}
			if (!write_err) {
				send_queued(cptr);
			}
			if (IsDead(cptr) || write_err) {
				if (MYFD_ISSET_READ) {
					MYFD_CLR_READ
				}
				ircsprintf(errmsg, "Write Error: %s",
					cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
				exit_client(cptr, cptr, &me, errmsg);
				FAST_FD_INC
				continue;
			}
		}

		length = 1;
		if (MYFD_ISSET_READ) {
			length = read_packet(cptr);
		}
		else if (IsPerson(cptr) && !NoNewLine(cptr)) {
			length = do_client_queue(cptr);
		}

#ifdef DEBUGMODE
		readcalls++;
#endif

		if (length != FLUSH_BUFFER && IsDead(cptr)) {
			if (MYFD_ISSET_READ) {
				MYFD_CLR_READ
			}
			ircsprintf(errmsg, "Read/Dead Error: %s",
				cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
			exit_client(cptr, cptr, &me, errmsg);
			FAST_FD_INC
			continue;
		}
		if (length > 0) {
			if (MYFD_ISSET_READ) {
				nfds--;
			}
			FAST_FD_INC
			continue;
		}
		if (length != FLUSH_BUFFER) {
			Debug((DEBUG_ERROR, "READ ERROR: fd = %d %d %d", i, errno, length));
			read_error_exit(cptr, length, cptr->sockerr);
		}
		FAST_FD_INC
	}
	return 0;
}

#else /* HAVE_POLL_ON */

# ifdef AIX
#  define POLLREADFLAGS (POLLIN|POLLMSG)
# endif
# if defined(POLLMSG) && defined(POLLIN) && defined(POLLRDNORM)
#  define POLLREADFLAGS (POLLMSG|POLLIN|POLLRDNORM)
# endif
# if defined(POLLIN) && defined(POLLRDNORM) && !defined(POLLMSG)
#  define POLLREADFLAGS (POLLIN|POLLRDNORM)
# endif
# if defined(POLLIN) && !defined(POLLRDNORM) && !defined(POLLMSG)
#  define POLLREADFLAGS POLLIN
# endif
# if defined(POLLRDNORM) && !defined(POLLIN) && !defined(POLLMSG)
#  define POLLREADFLAGS POLLRDNORM
# endif

# if defined(POLLOUT) && defined(POLLWRNORM)
#  define POLLWRITEFLAGS (POLLOUT|POLLWRNORM)
# else
#  if defined(POLLOUT)
#   define POLLWRITEFLAGS POLLOUT
#  else
#   if defined(POLLWRNORM)
#    define POLLWRITEFLAGS POLLWRNORM
#   endif
#  endif
# endif

# if defined(POLLERR) && defined(POLLHUP)
#  define POLLERRORS (POLLERR|POLLHUP)
# else
#  define POLLERRORS POLLERR
# endif

# define PFD_SETR(thisfd) { CHECK_PFD(thisfd); pfd->events |= POLLREADFLAGS; }
# define PFD_SETW(thisfd) { CHECK_PFD(thisfd); pfd->events |= POLLWRITEFLAGS; }
# define CHECK_PFD(thisfd) if (pfd->fd != thisfd) { pfd = &poll_fdarray[nbr_pfds++]; pfd->fd = thisfd; pfd->events = 0; }

int read_message(time_t delay, fdlist *listp)
{
	aClient *cptr;
	int nfds, nbr_pfds = 0, res, length, fd, auth, rr, rw, i, j;
	static struct pollfd poll_fdarray[MAXCONNECTIONS];
	struct pollfd *pfd = poll_fdarray, *res_pfd = NULL;
	u_long waittime;
	time_t delay2 = delay;
	static char errmsg[512];
	static aClient *authclnts[MAXCONNECTIONS];

	if (!listp) {
		listp = &default_fdlist;
		listp->last_entry = highest_fd + 1;
	}

	for (res = 0;;) {
		nbr_pfds = 0;
		pfd = poll_fdarray;
		pfd->fd = -1;
		res_pfd = NULL;
		auth = 0;

		for (i = listp->entry[j = 1]; j <= listp->last_entry; i = listp->entry[++j]) {
			if (!(cptr = local[i]) || IsLog(cptr)) {
				continue;
			}
			if (DoingAuth(cptr)) {
				if (auth == 0) {
					memset((char *)&authclnts, '\0', sizeof(authclnts));
				}
				auth++;
				Debug((DEBUG_NOTICE, "auth on %x %d", cptr, i));
				PFD_SETR(cptr->authfd);
				if (cptr->flags & FLAGS_WRAUTH) {
					PFD_SETW(cptr->authfd);
				}
				authclnts[cptr->authfd] = cptr;
				continue;
			}
			if (DoingDNS(cptr) || DoingAuth(cptr)) {
				continue;
			}
			if (IsMe(cptr) && IsListening(cptr)) {
#ifdef CONNECTFAST
				PFD_SETR(i);
#else
				if (timeofday > cptr->lasttime + 2) {
					PFD_SETR(i);
				}
				else if (delay2 > 2) {
					delay2 = 2;
				}
#endif
			}
			else if (!IsMe(cptr)) {
				if (DBufLength(&cptr->recvQ) && delay2 > 2) {
					delay2 = 1;
				}
				if (DBufLength(&cptr->recvQ) < 4088) {
					PFD_SETR(i);
				}
			}

			length = DBufLength(&cptr->sendQ);
			if (DoList(cptr) && IsSendable(cptr)) {
				send_list(cptr, 64);
				length = DBufLength(&cptr->sendQ);
			}
			if (length || IsConnecting(cptr) || (ZipOut(cptr) && zip_is_data_out(cptr->serv->zip_out))) {
				PFD_SETW(i);
			}
		}

		if (resfd >= 0) {
			PFD_SETR(resfd);
			res_pfd = pfd;
		}

		waittime = MIN(delay2, delay) * 1000;
		nfds = poll(poll_fdarray, nbr_pfds, waittime);
		if (nfds == -1 && (errno == EINTR || errno == EAGAIN)) {
			return -1;
		}
		else if (nfds >= 0) {
			break;
		}
		report_error("poll %s:%s", &me);
		res++;
		if (res > 5) {
			restart("too many poll errors");
		}
		sleep(10);
	}

	if (res_pfd && (res_pfd->revents & (POLLREADFLAGS | POLLERRORS))) {
		do_dns_async();
		nfds--;
	}

	for (pfd = poll_fdarray, i = 0; i < nbr_pfds; i++, pfd++) {
		fd = pfd->fd;
		if (pfd == res_pfd) {
			continue;
		}

		if (nfds && pfd->revents) {
			nfds--;
			rr = pfd->revents & (POLLREADFLAGS | POLLERRORS);
			rw = pfd->revents & POLLWRITEFLAGS;

			if (auth > 0 && (cptr = authclnts[fd]) && cptr->authfd == fd) {
				auth--;
				if (rr) {
					read_authports(cptr);
				}
				if (rw) {
					send_authports(cptr);
				}
				continue;
			}

			if (!(cptr = local[fd])) {
				continue;
			}
			if (rr && IsListening(cptr)) {
				accept_connection(cptr);
				continue;
			}
			if (IsMe(cptr)) {
				continue;
			}
			if (rw) {
				int write_err = 0;

				if (IsConnecting(cptr)) {
					write_err = completed_connection(cptr);
				}
				if (!write_err) {
					send_queued(cptr);
				}
				if (IsDead(cptr) || write_err) {
					ircsprintf(errmsg, "Write Error: %s",
						cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
					exit_client(cptr, cptr, &me, errmsg);
					continue;
				}
			}

			length = 1;
			if (rr) {
				length = read_packet(cptr);
			}
			else if (IsPerson(cptr) && !NoNewLine(cptr)) {
				length = do_client_queue(cptr);
			}
		}
		else {
			if (!(cptr = local[i])) {
				continue;
			}
			if (IsPerson(cptr) && !NoNewLine(cptr)) {
				length = do_client_queue(cptr);
			}
			else {
				continue;
			}
		}

#ifdef DEBUGMODE
		readcalls++;
#endif

		if (length == FLUSH_BUFFER) {
			continue;
		}

		if (IsDead(cptr)) {
			ircsprintf(errmsg, "Read/Dead Error: %s",
				cptr->flags & FLAGS_SENDQEX ? "SendQ Exceeded" : irc_get_sockerr(cptr));
			exit_client(cptr, cptr, &me, errmsg);
			continue;
		}

		if (length > 0) {
			continue;
		}

		read_error_exit(cptr, length, cptr->sockerr);
	}
	return 0;
}

#endif /* HAVE_POLL_ON */

/*
 * connect_server
 */
int connect_server(ConfigItem_link *link, aClient *by, struct hostent *hp)
{
	struct sockaddr *svp;
	aClient *cptr;
	int errtmp, len;

	Debug((DEBUG_NOTICE, "Connect to %s[%s] @%s", link->servername, link->fromip,
		inetntoa((char *)&link->ipnum)));

	if (!link->ipnum.s_addr) {
		Link lin;

		lin.flags = ASYNC_CONNECT;
		lin.value.conf = (ConfigItem *)link;
		nextdnscheck = 1;

		if ((link->ipnum.s_addr = inet_addr(link->fromip)) == -1) {
			link->ipnum.s_addr = 0;
			hp = gethost_byname(link->fromip, &lin);
			Debug((DEBUG_NOTICE, "co_sv: hp %x li %x se %s ho %s", hp, link,
				link->servername, link->fromip));
			if (!hp) {
				return -2;
			}
			memcpy((char *)&link->ipnum, hp->h_addr, sizeof(struct in_addr));
		}
	}

	cptr = make_client(NULL, &me);
	cptr->hostp = hp;

	strncpyzt(cptr->name, link->servername, sizeof(cptr->name));
	strncpyzt(cptr->sockhost, link->fromip, HOSTLEN + 1);

	svp = connect_inet(link, cptr, &len);
	if (!svp) {
		if (cptr->fd != -1) {
			close(cptr->fd);
		}
		cptr->fd = -2;
		free_client(cptr);
		return -1;
	}

	set_non_blocking(cptr->fd, cptr);
	set_sock_opts(cptr->fd, cptr);
	signal(SIGALRM, dummy);

	if (connect(cptr->fd, svp, len) < 0 && errno != EINPROGRESS) {
		errtmp = errno;
		report_error("Connect to host %s failed: %s", cptr);
		if (by && IsPerson(by) && !MyClient(by)) {
			sendto_one(by, ":%s NOTICE %s :Connect to server %s failed.", me.name,
				by->name, cptr->name);
		}
		close(cptr->fd);
		cptr->fd = -2;
		free_client(cptr);
		errno = errtmp;
		if (errno == EINTR) {
			errno = ETIMEDOUT;
		}
		return -1;
	}

	make_server(cptr);
	cptr->serv->conf = link;
	cptr->class = cptr->serv->conf->class;
	if (by && IsPerson(by)) {
		strcpy(cptr->serv->bynick, by->name);
		strcpy(cptr->serv->byuser, by->user->username);
		strcpy(cptr->serv->byhost, by->user->host);
	}
	else {
		strcpy(cptr->serv->bynick, "AutoConn.");
		*cptr->serv->byuser = '\0';
		*cptr->serv->byhost = '\0';
	}
	cptr->serv->up = me.name;
	if (cptr->fd > highest_fd) {
		highest_fd = cptr->fd;
	}
	local[cptr->fd] = cptr;
	cptr->listener = &me;
	cptr->sendqlen = cptr->serv->conf->class->sendq;
	SetConnecting(cptr);

	get_sockhost(cptr, link->fromip);
	add_client_to_list(cptr);
#ifdef PINGNAZI
	nextping = timeofday;
#endif
	return 0;
}

static struct sockaddr *connect_inet(ConfigItem_link *link, aClient *cptr, int *lenp)
{
	static struct sockaddr_in server;
	struct hostent *hp;

	cptr->fd = socket(AF_INET, SOCK_STREAM, 0);
	if (cptr->fd < 0) {
		report_error("opening stream socket to server %s:%s", cptr);
		return NULL;
	}
	if (cptr->fd >= (HARD_FDLIMIT - 10)) {
		sendto_realops("No more connections allowed (%s)", cptr->name);
		return NULL;
	}

	memset((char *)&server, '\0', sizeof(server));
	server.sin_family = AF_INET;
	get_sockhost(cptr, link->fromip);

	server.sin_port = 0;
	server.sin_addr = me.ip;
	server.sin_family = AF_INET;

	if (link->bindip && strcmp("*", link->bindip)) {
		server.sin_addr.s_addr = inet_addr(link->bindip);
	}
	if (bind(cptr->fd, (struct sockaddr *)&server, sizeof(server)) == -1) {
		report_error("error binding to local port for %s:%s", cptr);
		return NULL;
	}

	memset((char *)&server, '\0', sizeof(struct sockaddr_in));
	server.sin_family = AF_INET;

	if (isdigit(*link->fromip) && (link->ipnum.s_addr == -1)) {
		link->ipnum.s_addr = inet_addr(link->fromip);
	}
	if (link->ipnum.s_addr == -1) {
		hp = cptr->hostp;
		if (!hp) {
			Debug((DEBUG_FATAL, "%s: unknown host", link->fromip));
			return NULL;
		}
		memcpy((char *)&link->ipnum, hp->h_addr, sizeof(struct in_addr));
	}

	memcpy((char *)&server.sin_addr, (char *)&link->ipnum, sizeof(struct in_addr));
	memcpy((char *)&cptr->ip, (char *)&link->ipnum, sizeof(struct in_addr));

	server.sin_port = htons(link->port);
	*lenp = sizeof(server);
	return (struct sockaddr *)&server;
}

/*
 * * find the real hostname for the host running the server (or one
 * which * matches the server's name) and its primary IP#.  Hostname is
 * stored * in the client structure passed as a pointer.
 */
void get_my_name(aClient * cptr, char *name, int len)
{
   static char tmp[HOSTLEN + 1];
   struct hostent *hp;

   /* 
    * The following conflicts with both AIX and linux prototypes oh
    * well, we can put up with the errors from other systems -Dianora
    */
   /* 
    * extern int        gethostname(char *, int);
    */

   char *cname = cptr->name;

   /* 
    * * Setup local socket structure to use for binding to.
    */
   memset((char *) &mysk, '\0', sizeof(mysk));
   mysk.sin_family = AF_INET;

   if (gethostname(name, len) == -1)
      return;
   name[len] = '\0';

   /* 
    * assume that a name containing '.' is a FQDN
    */
   if (!strchr(name, '.'))
      add_local_domain(name, len - strlen(name));
   /* 
    * * If hostname gives another name than cname, then check if there
    * is * a CNAME record for cname pointing to hostname. If so accept *
    * cname as our name.   meLazy
    */
   if (BadPtr(cname))
      return;
   if ((hp = gethostbyname(cname)) || (hp = gethostbyname(name))) {
      char *hname;
      int i = 0;

      for (hname = hp->h_name; hname; hname = hp->h_aliases[i++]) {
	 strncpyzt(tmp, hname, sizeof(tmp));
	 add_local_domain(tmp, sizeof(tmp) - strlen(tmp));
	 /* 
	  * * Copy the matching name over and store the * 'primary' IP#
	  * as 'myip' which is used * later for making the right one is
	  * used * for connecting to other hosts.
	  */
	 if (!mycmp(me.name, tmp))
	    break;
      }
      if (mycmp(me.name, tmp))
	 strncpyzt(name, hp->h_name, len);
      else
	 strncpyzt(name, tmp, len);
      memcpy((char *) &mysk.sin_addr, hp->h_addr, sizeof(struct in_addr));

      Debug((DEBUG_DEBUG, "local name is %s", get_client_name(&me, TRUE)));
   }
   return;
}

/*
 * do_dns_async
 *
 * Called when the fd returned from init_resolver() has been selected for
 * reading.
 */
static void do_dns_async()
{
	static Link ln;
	aClient *cptr;
	ConfigItem_link *link;
	struct hostent *hp;
	int bytes, packets = 0;

	do {
		ln.flags = -1;
		hp = get_res((char *)&ln);
		Debug((DEBUG_DNS, "%#x = get_res(%d,%#x)", hp, ln.flags, ln.value.cptr));

		switch (ln.flags) {
			case ASYNC_NONE:
				break;
			case ASYNC_CLIENT:
				if ((cptr = ln.value.cptr)) {
					del_queries((char *)cptr);
#ifdef HAVE_OPENSSL_ON
					if (ConfigSet.showheaders && !(cptr->listener->umode & LISTEN_OPENSSL)) {
#else
					if (ConfigSet.showheaders) {
#endif
						sendto_one(cptr, REPORT_FIN_DNS);
					}
					ClearDNS(cptr);
					cptr->hostp = hp;
					if (!DoingAuth(cptr)) {
						SetAccess(cptr);
					}
				}
				break;
			case ASYNC_CONNECT:
				link = (ConfigItem_link *)ln.value.conf;
				if (hp && link) {
					memcpy((char *)&link->ipnum, hp->h_addr, sizeof(struct in_addr));
					connect_server(link, NULL, hp);
				}
				else {
					sendto_gnotice("Connect to %s failed: host lookup",
						link ? link->servername : "<unknown>");
				}
				break;
			case ASYNC_CONF:
				link = (ConfigItem_link *)ln.value.conf;
				if (hp && link) {
					memcpy((char *)&link->ipnum, hp->h_addr, sizeof(struct in_addr));
				}
				break;
			default:
				break;
		}
		if (ioctl(resfd, FIONREAD, &bytes) == -1) {
			bytes = 0;
		}
		packets++;
	} while (bytes > 0 && packets < 10);
}
