/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_conf2.c: server configuration
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_conf2.c,v 1.84 2002/04/09 17:03:54 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "h.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "patchlevel.h"
#include "dh.h"
#include "openssl.h"
#include "conf2.h"
#include "config.h"
#include "memory.h"
#include <time.h>
#include <signal.h>
#include <sys/stat.h> 
#include <stdlib.h>
#include <string.h>
#include <utmp.h>
#include <fcntl.h>

static void parse_file(char *);
static void parse_line(char *);
static short config_set(char *);
static void check_config();

static char buf[BUFSIZE];
static int commentstart = 0;
static int opencomment = 0;

#ifndef STATIC_MODULES
extern void load_one_module(int, char *);
#endif
extern IP_ENTRY *find_or_add_ip(unsigned long);
extern time_t check_pings(time_t);
extern long crc32(char *);
extern char *strerror();
extern int errno;
extern int user_modes[];

void config_report(short, short, char *, ...);
void add_ConfigItem(ConfigItem *, ConfigItem **);
ConfigItem *del_ConfigItem(ConfigItem *, ConfigItem **);
void run_listeners(void);
void config_cleanup(void);

short rehashing = 0;
short config_errors = 0;
int current_file = 0;

static ConfigFile *ConfigFileTable[INCLUDE_DEPTH];

void parse_servinfo(char **, int);
void parse_admin(char **, int);
void parse_class(char **, int);
void parse_allow(char **, int);
void parse_password(char **, int);
void parse_oper(char **, int);
void parse_uline(char **, int);
void parse_link(char **, int);
void parse_ban(char **, int);
void parse_listen(char **, int);
void parse_except(char **, int);
void parse_vhost(char **, int);
void parse_set(char **, int);
void parse_include(char **, int);
void parse_loadmodule(char **, int);

ConfigItem_servinfo *config_servinfo = NULL;
ConfigItem_admin *config_admin = NULL;
ConfigItem_admin *config_admin_tail = NULL;
ConfigItem_class *config_class = NULL;
ConfigItem_allow *config_allow = NULL;
ConfigItem_password *config_password = NULL;
ConfigItem_oper *config_oper = NULL;
ConfigItem_uline *config_uline = NULL;
ConfigItem_link *config_link = NULL;
ConfigItem_ban *config_ban = NULL;
ConfigItem_listen *config_listen = NULL;
ConfigItem_except *config_except = NULL;
ConfigItem_vhost *config_vhost = NULL;
ConfigItem_set ConfigSet;

static ConfigBanType ConfigBanTypes[] = {
	{ "userhost", BAN_USERHOST },
	{ "ip", BAN_IP },
	{ "nickname", BAN_NICKNAME },
	{ "gcos", BAN_GCOS },
	{ "server", BAN_SERVER },
	{ "file", BAN_FILE },
	{ "channel", BAN_CHANNEL },
	{ NULL, 0 }
};

static int OperFlags[] = {
	~(OFLAG_NETADMIN|OFLAG_ADMIN|OFLAG_COADMIN|OFLAG_SADMIN|OFLAG_RSTAFF|OFLAG_ZLINE), '*',
	OFLAG_REHASH, 'r',
	OFLAG_RSTAFF, 'S',
	OFLAG_DIE, 'D',
	OFLAG_RESTART, 'R',
	OFLAG_GLOBOP, 'g',
	OFLAG_WALLOP, 'w',
	OFLAG_LOCOP, 'l',
	OFLAG_LROUTE, 'x',
	OFLAG_GROUTE, 'X',
	OFLAG_LKILL, 'k',
	OFLAG_GKILL, 'K',
	OFLAG_KLINE, 'b',
	OFLAG_UNKLINE, 'B',
	OFLAG_LNOTICE, 'q',
	OFLAG_GNOTICE, 'Q',
	OFLAG_ADMIN, 'A',
	OFLAG_SADMIN, 'a',
	OFLAG_UMODEC, 'c',
	OFLAG_UMODEF, 'f',
	OFLAG_ZLINE, 'z',
	OFLAG_NETADMIN, 'N',
	OFLAG_COADMIN, 'C',
	0, 0
};

static int LinkFlags[] = {
	LINK_AUTOCONNECT, 'A',
	LINK_ZIP, 'Z',
	LINK_DKEY, 'E',
	0, 0
};

static int ListenFlags[] = {
	LISTEN_CLIENTONLY, 'C',
	LISTEN_SERVERONLY, 'S',
#ifdef HAVE_OPENSSL_ON
	LISTEN_OPENSSL, 'E',
#endif
	0, 0
};

static ConfigCommand ConfigCommands[] = {
	{ "servinfo", parse_servinfo, 2 },
	{ "admin", parse_admin, 1 },
	{ "class", parse_class, 5 },
	{ "allow", parse_allow, 4 },
	{ "password", parse_password, 2 },
	{ "oper", parse_oper, 4 },
	{ "uline", parse_uline, 1 },
	{ "link", parse_link, 6 },
	{ "ban", parse_ban, 2 },
	{ "listen", parse_listen, 2 },
	/*{ "except", parse_except, 2 }, -- temporarily disabled */
	{ "vhost", parse_vhost, 4 },
	{ "set", parse_set, 2 },
	{ "include", parse_include, 1 },
	{ "loadmodule", parse_loadmodule, 1 },
	{ NULL, NULL, 0 }
};

void config_report(short error, short filetag, char *pattern, ...)
{
	va_list vl;
	char buffer[1024];

	va_start(vl, pattern);
	ircvsprintf(buffer, pattern, vl);
	va_end(vl);

	if (rehashing) {
		if (filetag) {
			sendto_realops("[%s:%d] %s", ConfigFileTable[current_file - 1]->filename,
				ConfigFileTable[current_file - 1]->linenum, buffer);
		}
		else {
			sendto_realops("%s", buffer);
		}
	}
	else {
		if (filetag) {
			fprintf(stderr, "* [%s:%d] %s\n", ConfigFileTable[current_file - 1]->filename,
				ConfigFileTable[current_file - 1]->linenum, buffer);
		}
		else {
			fprintf(stderr, "* %s\n", buffer);
		}
	}
	if (error) {
		config_errors++;
	}
	return;
}

void add_ConfigItem(ConfigItem *item, ConfigItem **list)
{
	item->prev = NULL;
	item->next = *list;
	if (item->next) {
		item->next->prev = item;
	}
	*list = item;
	return;
}

ConfigItem *del_ConfigItem(ConfigItem *item, ConfigItem **list)
{
	ConfigItem *ci, *c;

	for (ci = *list; ci; ci = ci->next) {
		if (ci == item) {
			c = ci->next;
			if (ci->prev) {
				ci->prev->next = ci->next;
			}
			else {
				*list = ci->next;
			}
			if (ci->next) {
				ci->next->prev = ci->prev;
			}
			return c;
		}
	}
	return NULL;
}

static int check_ipaddr(char *string)
{
	char *s;
	for (s = string; *s; s++) {
		if (!isdigit(*s) && *s != '.' && *s != '*') {
			return 0;
		}
	}
	return 1;
}

static short config_set(char *string)
{
	if (!mycmp(string, "YES") || !mycmp(string, "TRUE") || !mycmp(string, "ON")
	  || !mycmp(string, "ENABLE") || !mycmp(string, "1")) {
		return 1;
	}
	return 0;
}

static void parse_line(char *buffer)
{
	ConfigCommand *c = NULL;
	char *paras[MAX_CONFIG_PARAMS], *p;
	int paracount = 0, pos = 0;
	short openquote = 0;

	buf[0] = '\0';

	while (*buffer == ' ') {
		buffer++;
	}
	if (*buffer == '#' || *buffer == '\r' || *buffer == '\n' || *buffer == '\0') {
		return;
	}
	for (p = buffer; *p; p++) {
		if (c && paracount >= MAX_CONFIG_PARAMS) {
			config_report(0, 1, "exceeded maximum parameter count for directive '%s'", c->cmd);
			break;
		}
		switch (*p) {
			case '/':
				if (!openquote) {
					if (*(p + 1) == '/') {
						while (*p) {
							p++;
						}
						break;
					}
					else if (*(p + 1) == '*') {
						p++;
						opencomment++;
						commentstart = ConfigFileTable[current_file - 1]->linenum;
						break;
					}
				}
				buf[pos++] = *p;
				break;
			case '\\':
				if (openquote && *(p + 1) == '"') {
					p++;
				}
				buf[pos++] = *p;
				break;
			case '*': 
				if (!openquote) {
					if (*(p + 1) == '/') {
						p++;
						if (opencomment > 0) {
							opencomment--;
							while (*(p + 1) == ' ') {
								(*(p + 1))++;
							}
						}
						else {
							config_report(0, 1, "ignoring extra close-comment");
						}
						break;
					}
				}
				buf[pos++] = *p;
				break;
			case '"':
				if (!opencomment) {
					if (!openquote) {
						openquote = 1;
					}
					else {
						openquote = 0;
					}
				}
				break;
			default:
				if (!opencomment) {
					if (openquote && (*p == '\n' || *p == '\r' || *p == '\0')) {
						config_report(0, 1, "ignoring line containing unterminated "
							"double-quoted string");
						for (pos = 0; pos < paracount; pos++) {
							MyFree(paras[pos]);
						}
						return;
					}
					else if (!openquote && (*p == ' ' || *p == '\n' || *p == '\r' || *p == '\0')) {
						buf[pos] = '\0';
						if (*buf == '\0') {
							return;
						}

						if (!c || BadPtr(c->cmd)) {
							for (c = ConfigCommands; c->cmd; c++) {
								if (!mycmp(c->cmd, buf)) {
									break;
								}
							}
							if (!c || BadPtr(c->cmd)) {
								config_report(0, 1, "ignoring unknown directive '%s'",
									buf);
								return;
							}
						}
						else {
							DupString(paras[paracount++], buf);
						}

						*buf = '\0';
						pos = 0;
						break;
					}
					buf[pos++] = *p;
				}
				break;
		}
	}

	if (!c || BadPtr(c->cmd)) {
		return;
	}

	paras[paracount] = NULL;

	if (paracount < c->paras - 1) {
		int missingparas = c->paras - paracount;
		config_report(0, 1, "%s: missing %d parameter%s; ignoring directive", c->cmd,
			missingparas, missingparas == 1 ? "" : "s");
		return;
	}

	c->func(paras, paracount);

	for (pos = 0; pos < paracount; pos++) {
		MyFree(paras[pos]);
	}
	return;
}

static void check_config()
{
	if (!config_errors) {
		if (!config_servinfo) {
			config_report(1, 0, "servinfo: not configured");
		}
		if (!config_listen) {
			config_report(1, 0, "listen: not configured");
		}
		if (!config_allow) {
			ConfigItem_allow *allow = (ConfigItem_allow *)MyMallocX(sizeof(ConfigItem_allow));
			DupString(allow->hostname, "*");
			DupString(allow->ip, "*");
			allow->maxperip = 3;
			allow->class = find_class("default");
			add_ConfigItem((ConfigItem *)allow, (ConfigItem **)&config_allow);
			config_report(0, 0, "allow: not configured; created default entry");
		}

		ConfigSet.connectmodes &= ~(UMODE_o|UMODE_c|UMODE_r|UMODE_R|UMODE_n|UMODE_f|UMODE_y|UMODE_d);
		ConfigSet.connectmodes &= ~(UMODE_a|UMODE_A|UMODE_b|UMODE_h|UMODE_m|UMODE_e|UMODE_N|UMODE_C);

		if (BadPtr(ConfigSet.ircnetwork)) {
			config_report(0, 0, "set->networkname: not configured; using 'IRC'");
			DupString(ConfigSet.ircnetwork, "IRC");
		}
		if (BadPtr(ConfigSet.klineaddress)) {
			config_report(0, 0, "set->klineaddress: not configured; using 'admin@poorly-configured-network.com'");
			DupString(ConfigSet.klineaddress, "admin@poorly-configured-network.com");
		}
		if (BadPtr(ConfigSet.helpchannel)) {
			config_report(0, 0, "set->helpchannel: not configured; using '#Chat'");
			DupString(ConfigSet.helpchannel, "#Chat");
		}
		if (BadPtr(ConfigSet.defaultserver)) {
			config_report(0, 0, "set->defaultserver: not configured; using 'irc.poorly-configured-network.com'");
			DupString(ConfigSet.defaultserver, "irc.poorly-configured-network.com");
		}
		if (BadPtr(ConfigSet.services)) {
			config_report(0, 0, "set->services: not configured; using 'services.poorly-configured-network.com'");
			DupString(ConfigSet.services, "services.poorly-configured-network.com");
		}
		if (BadPtr(ConfigSet.stats)) {
			config_report(0, 0, "set->stats: not configured; using 'stats.poorly-configured-network.com'");
			DupString(ConfigSet.stats, "stats.poorly-configured-network.com");
		}
		if (BadPtr(ConfigSet.chanserv)) {
			config_report(0, 0, "set->chanserv: not configured; using 'ChanServ'");
			DupString(ConfigSet.chanserv, "ChanServ");
		}
		if (BadPtr(ConfigSet.nickserv)) {
			config_report(0, 0, "set->nickserv: not configured; using 'NickServ'");
			DupString(ConfigSet.nickserv, "NickServ");
		}
		if (BadPtr(ConfigSet.memoserv)) {
			config_report(0, 0, "set->memoserv: not configured; using 'MemoServ'");
			DupString(ConfigSet.memoserv, "MemoServ");
		}
		if (BadPtr(ConfigSet.operserv)) {
			config_report(0, 0, "set->operserv: not configured; using 'OperServ'");
			DupString(ConfigSet.operserv, "OperServ");
		}
		if (BadPtr(ConfigSet.statserv)) {
			config_report(0, 0, "set->statserv: not configured; using 'StatServ'");
			DupString(ConfigSet.statserv, "StatServ");
		}
		if (BadPtr(ConfigSet.netadminmask)) {
			config_report(0, 0, "set->netadminmask: not configured; using 'network-administrator'");
			DupString(ConfigSet.netadminmask, "network-administrator");
		}
		if (BadPtr(ConfigSet.adminmask)) {
			config_report(0, 0, "set->adminmask: not configured; using 'server-administrator'");
			DupString(ConfigSet.adminmask, "server-administrator");
		}
		if (BadPtr(ConfigSet.coadminmask)) {
			config_report(0, 0, "set->coadminmask: not configured; using 'server-coadministrator'");
			DupString(ConfigSet.coadminmask, "server-coadministrator");
		}
		if (BadPtr(ConfigSet.opermask)) {
			config_report(0, 0, "set->opermask: not configured; using 'irc-operator'");
			DupString(ConfigSet.opermask, "irc-operator");
		}
		if (BadPtr(ConfigSet.usermask)) {
			config_report(0, 0, "set->usermask: not configured; using 'masked'");
			DupString(ConfigSet.usermask, "masked");
		}
		if (BadPtr(ConfigSet.localklineaddress)) {
			config_report(0, 0, "set->localklineaddress: not configured; using 'admin@poorly-configured-server.com'");
			DupString(ConfigSet.localklineaddress, "admin@poorly-configured-server.com");
		}
		if (BadPtr(ConfigSet.dns_nameserver)) {
			config_report(0, 0, "set->dns: name server not configured; using '127.0.0.1'");
			DupString(ConfigSet.dns_nameserver, "127.0.0.1");
		}
		if (ConfigSet.dns_timeout < 1) {
			config_report(0, 0, "set->dns: timeout invalid (< 1); using 2s");
			ConfigSet.dns_timeout = 2;
		}
		if (ConfigSet.dns_retries < 1) {
			config_report(0, 0, "set->dns: retries invalid (< 1); using 2");
			ConfigSet.dns_retries = 2;
		}
		if (ConfigSet.linkdepth < 2) {
			config_report(0, 0, "set->linkdepth: invalid value (< 2); using 5");
			ConfigSet.linkdepth = 5;
		}
		if (ConfigSet.maxchansperuser < 1) {
			config_report(0, 0, "set->maxchansperuser: invalid value (< 1); using 10");
			ConfigSet.maxchansperuser = 10;
		}
		if (ConfigSet.maxclients < 32 || ConfigSet.maxclients > MASTER_MAX) {
			config_report(0, 0, "set->maxclients: invalid value (< 32 | > %d); using 100", MASTER_MAX);
			ConfigSet.maxclients = 100;
		}
		if (ConfigSet.mask_key1 < 10000) {
			config_report(1, 0, "set->maskingkeys: key 1 invalid (< 10000)");
			ConfigSet.mask_crc = -1;
		}
		if (ConfigSet.mask_key2 < ConfigSet.mask_key1) {
			config_report(1, 0, "set->maskingkeys: key 2 invalid (< first key)");
			ConfigSet.mask_crc = -1;
		}
		if (ConfigSet.mask_key3 < ConfigSet.mask_key2) {
			config_report(1, 0, "set->maskingkeys: key 3 invalid (< second key)");
			ConfigSet.mask_crc = -1;
		}
	}
	return;
}

static void parse_file(char *filename)
{
	FILE *fd;
	char ptr[1024];
	int i;

	config_report(0, 0, "Opening config file %s...", filename);

	fd = fopen(filename, "r");
	if (!fd) {
		config_report(1, 0, "Failed to open config file %s: %s.", filename, strerror(errno));
		return;
	}
	config_report(0, 0, "Successfully opened config file %s.", filename);

	for (i = 0; i < INCLUDE_DEPTH; i++) {
		if (!ConfigFileTable[i]) {
			ConfigFileTable[i] = (ConfigFile *)MyMalloc(sizeof(ConfigFile));
			DupString(ConfigFileTable[i]->filename, filename);
			ConfigFileTable[i]->linenum = 0;
			current_file++;
			break;
		}
	}
	if (i == INCLUDE_DEPTH) {
		config_report(1, 0, "fatal error (config file table limit reached)");
		fclose(fd);
		return;
	}

	while (fgets(ptr, 1024, fd)) {
		ConfigFileTable[i]->linenum++;
		parse_line(ptr);
	}

	fclose(fd);

	if (opencomment > 0) {
		config_report(1, 0, "Unexpected EOF in %s: unterminated comment; possible comment start at line %d.",
			filename, commentstart);
		config_report(1, 0, "Syntax error: unexpected end of file.");
	}

	/*MyFree(ConfigFileTable[i]->filename);
	MyFree(ConfigFileTable[i]);*/
	current_file--;

	return;
}

void init_config2()
{
	if (!rehashing) {
		fprintf(stderr, "-----------------------------------------------------------------------------\n");
		memset(&ConfigFileTable, '\0', sizeof(ConfigFileTable));
	}

	parse_file(CONFIG_FILE);
	check_config();

	if (ConfigSet.mask_crc == -1) {
		char *maskkeyerrtext[] = {
			"***** WARNING *****",
			"You have not configured your servers masking keys correctly. These keys",
			"are crutial for keeping your normal-user maskedhost hashsums secure. The",
			"keys consist of three large integers, the first one above 10000, the",
			"second above the first key, and the third above the second key. Each key",
			"should be chosen completely at random, and should be the same on all the",
			"servers you plan to link to. We strongly recommend you also keep these",
			"keys a secret in order to keep the masked hostnames secure.",
			"***** WARNING *****",
			NULL
		};
		char **text = maskkeyerrtext;
		while (*text) {
			config_report(0, 0, "%s", *text++);
		}
	}

#ifdef HAVE_OPENSSL_ON
	init_openssl();
#endif

	if (!config_errors && !rehashing) {
		config_report(0, 0, "Server configuration successfully initialized.");
		fprintf(stderr, "-----------------------------------------------------------------------------\n");
	}
	else if (config_errors) {
		config_report(0, 0, "Received %d error%s during initialization, terminating program.",
			config_errors, config_errors == 1 ? "" : "s");
		if (!rehashing) {
			fprintf(stderr, "-----------------------------------------------------------------------------\n");
		}
		exit (-1);
	}
	return;
}

void parse_servinfo(char *paras[], int paracount)
{
	int paranum;

	if (!config_servinfo) {
		config_servinfo = (ConfigItem_servinfo *)MyMallocX(sizeof(ConfigItem_servinfo));
	}
	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				if (BadPtr(config_servinfo->name)) {
					if (!strchr(paras[paranum], '.')) {
						config_report(0, 1, "servinfo: invalid server name (contains no '.')");
					}
					DupString(config_servinfo->name, paras[paranum]);
				}
				break;
			case 1:
				MyFree(config_servinfo->info);
				DupString(config_servinfo->info, paras[paranum]);
				break;
			default:
				config_report(0, 1, "servinfo: ignoring extra parameter '%s'",
					paras[paranum]);
		}
	}
	return;
}

void parse_admin(char *paras[], int paracount)
{
	int paranum;
	ConfigItem_admin *admin = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			default:
				admin = (ConfigItem_admin *)MyMallocX(sizeof(ConfigItem_admin));
				DupString(admin->line, paras[paranum]);
				if (config_admin_tail == NULL) {
					config_admin_tail = admin;
				}
				add_ConfigItem((ConfigItem *)admin, (ConfigItem **)&config_admin);
				break;
		}
	}
	return;
}

void parse_class(char *paras[], int paracount)
{
	int paranum, new = 1;
	ConfigItem_class *class = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				class = find_class(paras[paranum]);
				if (!class) {
					class = (ConfigItem_class *)MyMallocX(sizeof(ConfigItem_class));
					DupString(class->name, paras[paranum]);
				}
				else {
					new = 0;
				}
			case 1:
				class->pingfreq = atime(paras[paranum]);
				if (class->pingfreq < 0) {
					config_report(0, 1, "class->%s: invalid ping frequency; using %d",
						class->name, PINGFREQUENCY);
					class->pingfreq = PINGFREQUENCY;
				}
				break;
			case 2:
				class->connfreq = atime(paras[paranum]);
				if (class->connfreq < 0) {
					config_report(0, 1, "class->%s: invalid connect frequeny; using %d",
						class->name, CONNECTFREQUENCY);
					class->connfreq = CONNECTFREQUENCY;
				}
				break;
			case 3:
				class->maxclients = atoi(paras[paranum]);
				if (class->maxclients < 0) {
					config_report(0, 1, "class->%s: invalid maximum clients; using 100",
						class->name);
					class->maxclients = 100;
				}
				break;
			case 4:
				class->sendq = atoi(paras[paranum]);
				if (class->sendq < 0) {
					config_report(0, 1, "class->%s: invalid sendq; using 500000",
						class->name);
					class->sendq = 500000;
				}
				break;
			default:
				config_report(0, 1, "class: ignoring extra parameter '%s'", paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)class, (ConfigItem **)&config_class);
	}
	class->flag.temporary = 0;
	return;
}

void parse_allow(char *paras[], int paracount)
{
	int paranum, new = 1;
	ConfigItem_allow *allow = NULL;
	char *hostname = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				hostname = paras[paranum];
				break;
			case 1:
				allow = find_allow(hostname, paras[paranum]);
				if (!allow) {
					if (!check_ipaddr(paras[paranum])) {
						config_report(0, 1, "allow: invalid ip address; ignoring directive");
						return;
					}
					allow = (ConfigItem_allow *)MyMallocX(sizeof(ConfigItem_allow));
					DupString(allow->hostname, hostname);
					DupString(allow->ip, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			case 2:
				allow->maxperip = atoi(paras[paranum]);
				if (allow->maxperip < 0) {
					config_report(0, 1, "allow: invalid maximum per ip; using 3");
					allow->maxperip = 3;
				}
				break;
			case 3:
				allow->class = find_class(paras[paranum]);
				if (!allow->class) {
					config_report(0, 1, "allow: unknown class entry '%s'; ignoring directive",
						paras[paranum]);
					MyFree(allow->hostname);
					MyFree(allow->ip);
					MyFree(allow);
					return;
				}
				break;
			case 4:
				MyFree(allow->password);
				DupString(allow->password, paras[paranum]);
				break;
			default:
				config_report(0, 1, "allow: ignoring extra parameter '%s'", paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)allow, (ConfigItem **)&config_allow);
	}
	return;
}

void parse_password(char *paras[], int paracount)
{
	int paranum;

	if (!config_password) {
		config_password = (ConfigItem_password *)MyMallocX(sizeof(ConfigItem_password));
	}
	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				if (*paras[paranum] != '$') {
					MyFree(config_password->die);
					DupString(config_password->die, paras[paranum]);
				}
				break;
			case 1:
				if (*paras[paranum] != '$') {
					MyFree(config_password->restart);
					DupString(config_password->restart, paras[paranum]);
				}
				break;
			default:
				config_report(0, 1, "password: ignoring extra parameter '%s'", paras[paranum]);
		}
	}
	return;
}

void parse_oper(char *paras[], int paracount)
{
	int paranum, new = 1, *i, flag;
	ConfigItem_oper *oper = NULL;
	char *m;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				oper = find_oper(paras[paranum]);
				if (!oper) {
					oper = (ConfigItem_oper *)MyMallocX(sizeof(ConfigItem_oper));
					DupString(oper->name, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			case 1:
				MyFree(oper->from);
				if (!strchr(paras[paranum], '@')) {
					char tmp[HOSTLEN + 3];
					ircsprintf(tmp, "*@%s", paras[paranum]);
					DupString(oper->from, tmp);
					config_report(0, 1, "oper->%s: invalid user@host; using *@%s",
						oper->name, tmp);
				}
				else {
					DupString(oper->from, paras[paranum]);
				}
				break;
			case 2:
				MyFree(oper->password);
				DupString(oper->password, paras[paranum]);
				break;
			case 3:
				oper->class = find_class(paras[paranum]);
				if (!oper->class) {
					config_report(0, 1, "oper->%s: unknown class entry '%s'; ignoring "
						"directive", oper->name, paras[paranum]);
					MyFree(oper->name);
					MyFree(oper->from);
					MyFree(oper->password);
					MyFree(oper);
					return;
				}
				break;
			case 4:
				for (m = paras[paranum]; *m; m++) {
					for (i = OperFlags; (flag = *i); i += 2) {
						if (*m == (char)(*(i + 1))) {
							oper->flags |= flag;
							break;
						}
					}
				}
				break;
			default:
				config_report(0, 1, "oper->%s: ignoring extra parameter '%s'",
					oper->name, paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)oper, (ConfigItem **)&config_oper);
	}
	return;
}

void parse_uline(char *paras[], int paracount)
{
	int paranum, new = 1;
	ConfigItem_uline *uline = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				uline = find_uline(paras[paranum]);
				if (!uline) {
					uline = (ConfigItem_uline *)MyMallocX(sizeof(ConfigItem_uline));
					DupString(uline->servername, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			default:
				config_report(0, 1, "uline->%s: ignoring extra parameter '%s'",
					uline->servername, paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)uline, (ConfigItem **)&config_uline);
	}
	return;
}

void parse_link(char *paras[], int paracount)
{
	int paranum, new = 1, *i, flag;
	ConfigItem_link *link = NULL;
	char *m, *servername = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				servername = paras[paranum];
				break;
			case 1:
				link = find_link(servername, paras[paranum]);
				if (!link) {
					if (!check_ipaddr(paras[paranum])) {
						config_report(0, 1, "link->%s: invalid from ip address; ignoring "
							"directive", servername);
						return;
					}
					link = (ConfigItem_link *)MyMallocX(sizeof(ConfigItem_link));
					DupString(link->servername, servername);
					DupString(link->fromip, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			case 2:
				MyFree(link->bindip);
				if (!check_ipaddr(paras[paranum])) {
					config_report(0, 1, "link->%s: invalid bind ip address; using *",
						servername);
					DupString(link->bindip, "*");
				}
				else {
					DupString(link->bindip, paras[paranum]);
				}
				break;
			case 3:
				MyFree(link->password);
				DupString(link->password, paras[paranum]);
				break;
			case 4:
				link->port = atoi(paras[paranum]);
				if (link->port < 1 || link->port > 65535) {
					config_report(0, 1, "link->%s: invalid port; using 7000",
						link->servername);
					link->port = 7000;
				}
				break;
			case 5:
				link->class = find_class(paras[paranum]);
				if (!link->class) {
					config_report(0, 1, "link->%s: unknown class entry '%s'; ignoring "
						"directive", link->servername, paras[paranum]);
					MyFree(link->servername);
					MyFree(link->fromip);
					MyFree(link->bindip);
					MyFree(link->password);
					MyFree(link);
					return;
				}
				break;
			case 6:
				for (m = paras[paranum]; *m; m++) {
					for (i = LinkFlags; (flag = *i); i += 2) {
						if (*m == (char)(*(i + 1))) {
							link->options |= flag;
							break;
						}
					}
				}
				break;
			default:
				config_report(0, 1, "link->%s: ignoring extra parameter '%s'",
					link->servername, paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)link, (ConfigItem **)&config_link);
	}
	link->flag.temporary = 0;
	return;
}

void parse_ban(char *paras[], int paracount)
{
	int paranum, new = 1;
	ConfigItem_ban *ban = NULL;
	ConfigBanType *bt = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				for (bt = ConfigBanTypes; bt->name; bt++) {
					if (!mycmp(paras[paranum], bt->name)) {
						break;
					}
				}
				if (!bt) {
					config_report(0, 1, "ban: unknown type '%s'; ignoring directive",
						paras[paranum]);
					return;
				}
				break;
			case 1:
				ban = find_ban(bt->type, paras[paranum]);
				if (!ban) {
					ban = (ConfigItem_ban *)MyMallocX(sizeof(ConfigItem_ban));
					ban->type = bt->type;
					DupString(ban->mask, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			case 2:
				MyFree(ban->reason);
				DupString(ban->reason, paras[paranum]);
				break;
			default:
				config_report(0, 1, "ban->%s->%s: ignoring extra parameter '%s'",
					bt->name, ban->mask, paras[paranum]);
		}
	}
	if (BadPtr(ban->reason)) {
		DupString(ban->reason, "No reason specified");
	}
	if (new) {
		add_ConfigItem((ConfigItem *)ban, (ConfigItem **)&config_ban);
	}
	ban->flag.temporary = 0;
	return;
}

void parse_listen(char *paras[], int paracount)
{
	int paranum, new = 1, port, *i, flag;
	ConfigItem_listen *listen = NULL;
	char *ip = NULL, *m;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				if (!check_ipaddr(paras[paranum])) {
					config_report(0, 1, "listen: invalid ip address; ignoring directive");
					return;
				}
				ip = paras[paranum];
				break;
			case 1:
				port = atoi(paras[paranum]);
				if (port < 1 || port > 65535) {
					config_report(0, 1, "listen->%s: invalid port; ignoring directive", ip);
					return;
				}

				listen = find_listen(ip, port);
				if (!listen) {
					listen = (ConfigItem_listen *)MyMallocX(sizeof(ConfigItem_listen));
					DupString(listen->ip, ip);
					listen->port = port;
				}
				else {
					new = 0;
				}
				break;
			case 2:
				for (m = paras[paranum]; *m; m++) {
					for (i = ListenFlags; (flag = *i); i += 2) {
						if (*m == (char)(*(i + 1))) {
							listen->options |= flag;
							break;
						}
					}
				}
				break;
			default:
				config_report(0, 1, "listen->%s: ignoring extra parameter '%s'", listen->ip,
					paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)listen, (ConfigItem **)&config_listen);
	}
	listen->flag.temporary = 0;
	return;
}

void parse_vhost(char *paras[], int paracount)
{
	int paranum, new = 1;
	ConfigItem_vhost *vhost = NULL;
	char *host = NULL;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				host = paras[paranum];
				break;
			case 1:
				vhost = find_vhost(paras[paranum]);
				if (!vhost) {
					vhost = (ConfigItem_vhost *)MyMallocX(sizeof(ConfigItem_vhost));
					DupString(vhost->vhost, host);
					DupString(vhost->login, paras[paranum]);
				}
				else {
					new = 0;
				}
				break;
			case 2:
				MyFree(vhost->password);
				DupString(vhost->password, paras[paranum]);
				break;
			case 3:
				MyFree(vhost->from);
				if (!strchr(paras[paranum], '@')) {
					char tmp[HOSTLEN + 3];
					ircsprintf(tmp, "*@%s", paras[paranum]);
					DupString(vhost->from, tmp);
					config_report(0, 1, "vhost->%s: invalid user@host; using %s", vhost->vhost,
						tmp);
				}
				else {
					DupString(vhost->from, paras[paranum]);
				}
				break;
			default:
				config_report(0, 1, "vhost->%s: ignoring extra parameter '%s'", vhost->vhost,
					paras[paranum]);
		}
	}
	if (new) {
		add_ConfigItem((ConfigItem *)vhost, (ConfigItem **)&config_vhost);
	}
	return;
}

void parse_set(char *paras[], int paracount)
{
	int paranum, *i, flag;
	char *m;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				if (!mycmp(paras[paranum], "shortmotd")) {
					ConfigSet.shortmotd = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "showopermotd")) {
					ConfigSet.showopermotd = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "hideulines")) {
					ConfigSet.hideulines = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "encryptedpasswords")) {
					ConfigSet.encryptedpasswords = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "doidentd")) {
					ConfigSet.doidentd = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "showheaders")) {
					ConfigSet.showheaders = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "nomixedcase")) {
					ConfigSet.nomixedcase = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "ignoreident")) {
					ConfigSet.ignoreident = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "failopernotice")) {
					ConfigSet.failopernotice = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "nooperflood")) {
					ConfigSet.nooperflood = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "ignorefirstchar")) {
					ConfigSet.ignorefirstchar = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "dynchans")) {
					ConfigSet.dynchans = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "automaskopers")) {
					ConfigSet.automaskopers = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "hub")) {
					ConfigSet.hub = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "logging")) {
					ConfigSet.logging = config_set(paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "maxchansperuser")) {
					ConfigSet.maxchansperuser = atoi(paras[1]);
					if (ConfigSet.maxchansperuser < 0) {
						config_report(0, 1, "set->maxchansperuser: parameter invalid; using 10");
						ConfigSet.maxchansperuser = 10;
					}
					return;
				}
				if (!mycmp(paras[paranum], "floodcount")) {
					ConfigSet.floodcount = atoi(paras[1]);
					if (ConfigSet.floodcount < 0) {
						config_report(0, 1, "set->floodcount: parameter invalid; using 8");
						ConfigSet.floodcount = 8;
					}
					return;
				}
				if (!mycmp(paras[paranum], "linkdepth")) {
					ConfigSet.linkdepth = atoi(paras[1]);
					if (ConfigSet.linkdepth < 0) {
						config_report(0, 1, "set->linkdepth: parameter invalid; using 5");
						ConfigSet.linkdepth = 5;
					}
					return;
				}
				if (!mycmp(paras[paranum], "maxclients")) {
					ConfigSet.maxclients = atoi(paras[1]);
					if (ConfigSet.maxclients < 0) {
						config_report(0, 1, "set->maxclients: parameter invalid; using 256");
						ConfigSet.maxclients = 256;
					}
					return;
				}
				if (!mycmp(paras[paranum], "pacetime")) {
					ConfigSet.pacetime = atime(paras[1]);
					if (ConfigSet.pacetime < 0) {
						config_report(0, 1, "set->pacetime: parameter invalid; using 10s");
						ConfigSet.pacetime = 10;
					}
					return;
				}
				if (!mycmp(paras[paranum], "knockdelay")) {
					ConfigSet.knockdelay = atime(paras[1]);
					if (ConfigSet.knockdelay < 0) {
						config_report(0, 1, "set->knockdelay: parameter invalid; using 5m");
						ConfigSet.knockdelay = 300;
					}
					return;
				}
				if (!mycmp(paras[paranum], "acceptnoticetime")) {
					ConfigSet.acceptnoticetime = atime(paras[1]);
					if (ConfigSet.acceptnoticetime < 0) {
						config_report(0, 1, "set->acceptnoticetime: parameter invalid; using 3m");
						ConfigSet.acceptnoticetime = 180;
					}
					return;
				}
				if (!mycmp(paras[paranum], "localklineaddress")) {
					MyFree(ConfigSet.localklineaddress);
					if (!strchr(paras[1], '@')) {
						config_report(0, 1, "set->localklineaddress: parameter invalid; using admin@poorly-configured-server.com");
						DupString(ConfigSet.localklineaddress, "admin@poorly-configured-server.com");
					}
					else {
						DupString(ConfigSet.localklineaddress, paras[1]);
					}
					return;
				}
				if (!mycmp(paras[paranum], "klineaddress")) {
					MyFree(ConfigSet.klineaddress);
					if (!strchr(paras[1], '@')) {
						config_report(0, 1, "set->klineaddress: parameter invalid; using admin@poorly-configured-network.com");
						DupString(ConfigSet.klineaddress, "admin@poorly-configured-network.com");
					}
					else {
						DupString(ConfigSet.klineaddress, paras[1]);
					}
					return;
				}
				if (!mycmp(paras[paranum], "networkname")) {
					MyFree(ConfigSet.ircnetwork);
					DupString(ConfigSet.ircnetwork, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "helpchannel")) {
					MyFree(ConfigSet.helpchannel);
					DupString(ConfigSet.helpchannel, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "defaultserver")) {
					MyFree(ConfigSet.defaultserver);
					DupString(ConfigSet.defaultserver, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "services")) {
					MyFree(ConfigSet.services);
					DupString(ConfigSet.services, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "stats")) {
					MyFree(ConfigSet.stats);
					DupString(ConfigSet.stats, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "chanserv")) {
					MyFree(ConfigSet.chanserv);
					DupString(ConfigSet.chanserv, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "nickserv")) {
					MyFree(ConfigSet.nickserv);
					DupString(ConfigSet.nickserv, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "memoserv")) {
					MyFree(ConfigSet.memoserv);
					DupString(ConfigSet.memoserv, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "operserv")) {
					MyFree(ConfigSet.operserv);
					DupString(ConfigSet.operserv, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "statserv")) {
					MyFree(ConfigSet.statserv);
					DupString(ConfigSet.statserv, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "netadminmask")) {
					MyFree(ConfigSet.netadminmask);
					DupString(ConfigSet.netadminmask, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "adminmask")) {
					MyFree(ConfigSet.adminmask);
					DupString(ConfigSet.adminmask, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "coadminmask")) {
					MyFree(ConfigSet.coadminmask);
					DupString(ConfigSet.coadminmask, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "opermask")) {
					MyFree(ConfigSet.opermask);
					DupString(ConfigSet.opermask, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "usermask")) {
					MyFree(ConfigSet.usermask);
					DupString(ConfigSet.usermask, paras[1]);
					return;
				}
				if (!mycmp(paras[paranum], "maskingkeys")) {
					char tmpbuf[256];
					ConfigSet.mask_key1 = atol(paras[1]);
					if (ConfigSet.mask_key1 < 0 || ConfigSet.mask_key1 < 10000) {
						config_report(1, 1, "set->maskingkeys: parameter invalid");
						return;
					}
					ConfigSet.mask_key2 = atol(paras[2]);
					if (ConfigSet.mask_key2 < 0 || ConfigSet.mask_key2 < ConfigSet.mask_key1) {
						config_report(1, 1, "set->maskingkeys: parameter invalid");
						return;
					}
					ConfigSet.mask_key3 = atol(paras[3]);
					if (ConfigSet.mask_key3 < 0 || ConfigSet.mask_key3 < ConfigSet.mask_key2) {
						config_report(1, 1, "set->maskingkeys: parameter invalid");
						return;
					}
					ircsprintf(tmpbuf, "%ld.%ld.%ld", ConfigSet.mask_key1, ConfigSet.mask_key2,
						ConfigSet.mask_key3);
					ConfigSet.mask_crc = crc32(tmpbuf);
					return;
				}
				if (!mycmp(paras[paranum], "dns")) {
					MyFree(ConfigSet.dns_nameserver);
					if (!check_ipaddr(paras[1])) {
						config_report(0, 1, "set->dns: invalid parameter; using 127.0.0.1");
						DupString(ConfigSet.dns_nameserver, "127.0.0.1");
					}
					else {
						DupString(ConfigSet.dns_nameserver, paras[1]);
					}
					ConfigSet.dns_timeout = atime(paras[2]);
					if (ConfigSet.dns_timeout < 0) {
						config_report(0, 1, "set->dns: invalid parameter; using 2s");
						ConfigSet.dns_timeout = 2;
					}
					ConfigSet.dns_retries = atoi(paras[3]);
					if (ConfigSet.dns_timeout < 0) {
						config_report(0, 1, "set->dns: invalid parameter; using 2");
						ConfigSet.dns_timeout = 2;
					}
					return;
				}
				if (!mycmp(paras[paranum], "connectmodes")) {
					ConfigSet.connectmodes = 0;
					for (m = paras[1]; *m; m++) {
						for (i = user_modes; (flag = *i); i += 2) {
							if (*m == (char)(*(i + 1))) {
								ConfigSet.connectmodes |= flag;
								break;
							}
						}
					}
					return;
				}
#ifdef HAVE_THROTTLE_ON
				if (!mycmp(paras[paranum], "throttle")) {
					ConfigSet.throttle_trigcount = atoi(paras[1]);
					if (ConfigSet.throttle_trigcount < 0) {
						config_report(0, 1, "set->throttle: invalid parameter; using 3");
						ConfigSet.throttle_trigcount = 3;
					}
					ConfigSet.throttle_trigtime = atime(paras[2]);
					if (ConfigSet.throttle_trigtime < 0) {
						config_report(0, 1, "set->throttle: invalid parameter; using 15s");
						ConfigSet.throttle_trigtime = 15;
					}
					return;
				}
#endif
				config_report(0, 1, "set: ignoring unknown directive '%s'", paras[paranum]);
				return;
		}
	}
	return;
}

void parse_include(char *paras[], int paracount)
{
	int paranum;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				if (current_file >= INCLUDE_DEPTH) {
					config_report(0, 1, "include->%s: reached maximum amount of config files "
						"open; ignoring directive", paras[paranum]);
					return;
				}
				parse_file(paras[paranum]);
				break;
			default:
				config_report(0, 1, "include->%s: ignoring extra parameter '%s'",
					paras[0], paras[paranum]);
		}
	}
	return;
}

void parse_loadmodule(char *paras[], int paracount)
{
#ifndef STATIC_MODULES
	int paranum;

	for (paranum = 0; paranum < paracount; paranum++) {
		switch (paranum) {
			case 0:
				load_one_module(1, paras[paranum]);
				break;
			default:
				config_report(0, 1, "loadmodule->%s: ignoring extra parameter '%s'",
					paras[0], paras[paranum]);
		}
	}
#endif
	return;
}

ConfigItem_class *find_class(char *name)
{
	ConfigItem_class *class;
	for (class = config_class; class; class = (ConfigItem_class *)class->next) {
		if (!mycmp(class->name, name)) {
			return class;
		}
	}
	return NULL;
}

ConfigItem_allow *find_allow(char *hostname, char *ip)
{
	ConfigItem_allow *allow;
	for (allow = config_allow; allow; allow = (ConfigItem_allow *)allow->next) {
		if (!match(allow->hostname, hostname) || !match(hostname, allow->hostname)) {
			if (!match(allow->ip, ip) || !match(ip, allow->ip)) {
				return allow;
			}
		}
	}
	return NULL;
}

ConfigItem_oper *find_oper(char *name)
{
	ConfigItem_oper *oper;
	for (oper = config_oper; oper; oper = (ConfigItem_oper *)oper->next) {
		if (!mycmp(oper->name, name)) {
			return oper;
		}
	}
	return NULL;
}

ConfigItem_uline *find_uline(char *servername)
{
	ConfigItem_uline *uline;
	for (uline = config_uline; uline; uline = (ConfigItem_uline *)uline->next) {
		if (!mycmp(uline->servername, servername)) {
			return uline;
		}
	}
	return NULL;
}

ConfigItem_link *find_link(char *servername, char *fromip)
{
	ConfigItem_link *link;
	for (link = config_link; link; link = (ConfigItem_link *)link->next) {
		if (!match(link->servername, servername) || !match(servername, link->servername)) {
			if (!match(link->fromip, fromip) || !match(fromip, link->fromip)) {
				return link;
			}
		}
	}
	return NULL;
}

ConfigItem_ban *find_ban(int type, char *mask)
{
	ConfigItem_ban *ban;
	for (ban = config_ban; ban; ban = (ConfigItem_ban *)ban->next) {
		if (ban->type & type) {
			if (!match(ban->mask, mask)) {
				return ban;
			}
		}
	}
	return NULL;
}

ConfigItem_listen *find_listen(char *ip, int port)
{
	ConfigItem_listen *listen;
	for (listen = config_listen; listen; listen = (ConfigItem_listen *)listen->next) {
		if (listen->port == port) {
			if (!match(listen->ip, ip) || !match(ip, listen->ip)) {
				return listen;
			}
		}
	}
	return NULL;
}

ConfigItem_vhost *find_vhost(char *login)
{
	ConfigItem_vhost *vhost;
	for (vhost = config_vhost; vhost; vhost = (ConfigItem_vhost *)vhost->next) {
		if (!mycmp(vhost->login, login)) {
			return vhost;
		}
	}
	return NULL;
}

void run_listeners(void)
{
	ConfigItem_listen *listen;
	for (listen = config_listen; listen; listen = (ConfigItem_listen *)listen->next) {
		if (!(listen->options & LISTEN_BOUND)) {
			if (add_listener(listen) == -1) {
				Debug((DEBUG_DEBUG, "Failed to bind to %s:%d", listen->ip, listen->port));
			}
		}
	}
	return;
}

void config_cleanup(void)
{
	ConfigItem c;
	ConfigItem_class *class;
	ConfigItem_link *link;
	ConfigItem_listen *listen;
	int i = 0;

	for (class = config_class; class; class = (ConfigItem_class *)class->next) {
		if (class->flag.temporary && class->clients <= 0) {
			c.next = del_ConfigItem((ConfigItem *)class, (ConfigItem **)&config_class);
			MyFree(class->name);
			MyFree(class);
			class = (ConfigItem_class *)&c;
		}
	}

	for (link = config_link; link; link = (ConfigItem_link *)link->next) {
		if (link->flag.temporary && link->refcount <= 0) {
			c.next = del_ConfigItem((ConfigItem *)link, (ConfigItem **)&config_link);
			MyFree(link->servername);
			MyFree(link->fromip);
			MyFree(link->bindip);
			MyFree(link->password);
			MyFree(link);
			link = (ConfigItem_link *)&c;
		}
	}

	for (listen = config_listen; listen; listen = (ConfigItem_listen *)listen->next) {
		if (!(listen->options & LISTEN_MAINPORT)) {
			if (listen->flag.temporary && listen->clients <= 0) {
				c.next = del_ConfigItem((ConfigItem *)listen, (ConfigItem **)&config_listen);
				MyFree(listen->ip);
				MyFree(listen);
				listen = (ConfigItem_listen *)&c;
				i++;
			}
		}
	}
	if (i > 0) {
		close_listeners();
	}
	return;
}

int AllowClient(aClient *cptr, struct hostent *hp, char *sockhost)
{
	ConfigItem_allow *allow;
	IP_ENTRY *ip_found;
	char *hname;
	int i, ulen = 0;
	static char uhost[USERLEN + HOSTLEN + 2], uhost2[USERLEN + HOSTLEN + 2];
	static char fullname[HOSTLEN + 1];

	for (allow = config_allow; allow; allow = (ConfigItem_allow *)allow->next) {
		if (allow->class->flag.temporary) {
			continue;
		}
		if (BadPtr(allow->hostname) || BadPtr(allow->ip)) {
			goto attach;
		}

		if (hp) {
			for (i = 0, hname = hp->h_name; hname; hname = hp->h_aliases[i++]) {
				strncpy(fullname, hname, sizeof(fullname) - 1);
				add_local_domain(fullname, HOSTLEN - strlen(fullname));
				Debug((DEBUG_DNS, "ac: %s->%s", sockhost, fullname));

				if (index(allow->hostname, '@')) {
					ulen = ircsprintf(uhost, "%s@", cptr->username);
					strcpy(uhost2, uhost);
				}
				else {
					ulen = 0;
					*uhost = *uhost2 = '\0';
				}
				strncat(uhost, fullname, sizeof(uhost) - ulen);
				strncat(uhost2, sockhost, sizeof(uhost2) - ulen);

				if (!match(allow->hostname, uhost) || !match(allow->hostname, uhost2)) {
					goto attach;
				}
			}
		}

		if (index(allow->ip, '@')) {
			ulen = ircsprintf(uhost, "%s@", cptr->username);
		}
		else {
			ulen = 0;
			*uhost = '\0';
		}
		strncat(uhost, sockhost, sizeof(uhost) - ulen);

		if (!match(allow->ip, uhost)) {
			goto attach;
		}
		continue;

	attach:
		if (index(uhost, '@')) {
			cptr->flags |= FLAGS_DOID;
		}
		get_sockhost(cptr, uhost);

		ip_found = find_or_add_ip(cptr->ip.s_addr);
		cptr->flags |= FLAGS_IPHASH;
		ip_found->count++;

		if (allow->maxperip) {
			if (ip_found->count > allow->maxperip) {
				ircstp->is_ref++;
				exit_client(cptr, cptr, &me, "Too many connections from your IP");
				return -5;
			}
		}
		if (allow->password) {
			if (BadPtr(cptr->passwd)) {
				ircstp->is_ref++;
				exit_client(cptr, cptr, &me, "No password");
				return -5;
			}
			if (strcmp(allow->password, cptr->passwd)) {
				ircstp->is_ref++;
				exit_client(cptr, cptr, &me, "Bad password");
				return -5;
			}
		}
		else if (BadPtr(allow->password) && !BadPtr(cptr->passwd)) {
			memset(cptr->passwd, '\0', sizeof(cptr->passwd));
		}
		if (allow->class->clients + 1 > allow->class->maxclients) {
			ircstp->is_ref++;
			exit_client(cptr, cptr, &me, "Your connection class is full");
			return -3;
		}

		cptr->class = allow->class;
		cptr->class->clients++;
		return 0;
	}
	return -1;
}

int rehash(aClient *cptr, aClient *sptr, int signal)
{
	ConfigItem c;
	ConfigItem_admin *admin;
	ConfigItem_class *class;
	ConfigItem_allow *allow;
	ConfigItem_oper *oper;
	ConfigItem_uline *uline;
	ConfigItem_link *link;
	ConfigItem_ban *ban;
	ConfigItem_listen *listen;
	ConfigItem_vhost *vhost;

	rehashing = 1;

	if (signal == SIGHUP) {
		sendto_ops("Got signal SIGHUP, reloading server configuration files.");
#ifdef ULTRIX
		if (fork() > 0) {
			exit (0);
		}
#endif
		write_pidfile();
	}

	if (config_servinfo) {
		MyFree(config_servinfo->info);
	}

	for (admin = config_admin; admin; admin = (ConfigItem_admin *)admin->next) {
		c.next = del_ConfigItem((ConfigItem *)admin, (ConfigItem **)&config_admin);
		MyFree(admin->line);
		MyFree(admin);
		admin = (ConfigItem_admin *)&c;
	}

	for (class = config_class; class; class = (ConfigItem_class *)class->next) {
		class->flag.temporary = 1;
	}

	for (allow = config_allow; allow; allow = (ConfigItem_allow *)allow->next) {
		c.next = del_ConfigItem((ConfigItem *)allow, (ConfigItem **)&config_allow);
		MyFree(allow->hostname);
		MyFree(allow->ip);
		MyFree(allow->password);
		MyFree(allow);
		allow = (ConfigItem_allow *)&c;
	}

	if (config_password) {
		MyFree(config_password->die);
		MyFree(config_password->restart);
		MyFree(config_password);
	}

	for (oper = config_oper; oper; oper = (ConfigItem_oper *)oper->next) {
		c.next = del_ConfigItem((ConfigItem *)oper, (ConfigItem **)&config_oper);
		MyFree(oper->name);
		MyFree(oper->from);
		MyFree(oper->password);
		MyFree(oper);
		oper = (ConfigItem_oper *)&c;
	}

	for (uline = config_uline; uline; uline = (ConfigItem_uline *)uline->next) {
		c.next = del_ConfigItem((ConfigItem *)uline, (ConfigItem **)&config_uline);
		MyFree(uline->servername);
		MyFree(uline);
		uline = (ConfigItem_uline *)&c;
	}

	for (link = config_link; link; link = (ConfigItem_link *)link->next) {
		link->flag.temporary = 1;
	}

	for (ban = config_ban; ban; ban = (ConfigItem_ban *)ban->next) {
		c.next = del_ConfigItem((ConfigItem *)ban, (ConfigItem **)&config_ban);
		MyFree(ban->mask);
		MyFree(ban->reason);
		MyFree(ban);
		ban = (ConfigItem_ban *)&c;
	}

	for (listen = config_listen; listen; listen = (ConfigItem_listen *)listen->next) {
		if (!(listen->options & LISTEN_MAINPORT)) {
			listen->flag.temporary = 1;
		}
	}

	for (vhost = config_vhost; vhost; vhost = (ConfigItem_vhost *)vhost->next) {
		c.next = del_ConfigItem((ConfigItem *)vhost, (ConfigItem **)&config_vhost);
		MyFree(vhost->vhost);
		MyFree(vhost->login);
		MyFree(vhost->password);
		MyFree(vhost->from);
		MyFree(vhost);
		vhost = (ConfigItem_vhost *)&c;
	}

	if (signal != SIGINT) {
		flush_cache();
	}

	init_config2();
	close_listeners();
	config_cleanup();
	close_listeners();
	run_listeners();
	runHook0(HOOKTYPE_REHASH);
	check_pings(timeofday);

	rehashing = 0;
	sendto_ops("Rehash completed.");

	return 1;
}
