/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_serv.c: server functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_serv.c,v 1.67 2002/04/08 16:00:51 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "patchlevel.h"
#include "zlink.h"
#include "memory.h"
#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined(HAVE_STRING_H)
# include <string.h>
#else
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "fdlist.h"

#ifndef NO_PRIORITY
extern void check_fdlists();
#endif
extern ConfigItem *del_ConfigItem(ConfigItem *, ConfigItem **);

extern fdlist serv_fdlist;
int max_connection_count = 1;
int max_client_count = 1;

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

/* Local function prototypes */
static void sendnick_TS(aClient *, aClient *);

int do_server_estab(aClient *cptr)
{
	char *inpath = get_client_name(cptr, HIDEME), capablist[BUFSIZE];
	aClient *acptr;
	aChannel *chptr;
	int i;
	ConfigItem_ban *ban;

	SetServer(cptr);
	Count.server++;
	Count.myserver++;

	if (IsZipCapable(cptr) && DoZipThis(cptr)) {
		sendto_one(cptr, "SVINFO ZIP");
		SetZipOut(cptr);
		cptr->serv->zip_out = zip_create_output_session();
	}

#ifdef MAXBUFFERS
	reset_sock_opts(cptr->fd, 1);
#endif

	addto_fdlist(cptr->fd, &serv_fdlist);
#ifndef NO_PRIORITY
	check_fdlists();
#endif

	if (find_uline(cptr->name)) {
		Count.myulined++;
		cptr->flags |= FLAGS_ULINE;
	}

	ircsprintf(capablist, "%s%s%s%s",
		DoesTS(cptr) ? " TS" : " Non-TS",
		ZipOut(cptr) ? " OutputZipped" : "",
		RC4EncLink(cptr) ? " Encrypted" : "",
		IsULine(cptr) ? " ULined" : "");

	sendto_gnotice("Link with %s established (capabilities:%s)", inpath, capablist);
	sendto_serv_butone(NULL, ":%s GNOTICE :Link with %s established (capabilities:%s)",
		me.name, inpath, capablist);
	ircdlog(LOG_SERVER, "link with %s established (%s)", get_client_name(cptr, FALSE), capablist);

	add_to_client_hash_table(cptr->name, cptr);
	find_or_add(cptr->name);

	for (i = 0; i <= highest_fd; i++) {
		if (!(acptr = local[i]) || !IsServer(acptr) || acptr == cptr || IsMe(acptr)) {
			continue;
		}
		sendto_one(acptr, ":%s SERVER %s 2 :%s", me.name, cptr->name, cptr->info);
	}

	for (acptr = &me; acptr; acptr = acptr->prev) {
		if (acptr->from == cptr) {
			continue;
		}
		if (IsServer(acptr)) {
			sendto_one(cptr, ":%s SERVER %s %d :%s", acptr->serv->up, acptr->name,
				acptr->hopcount + 1, acptr->info);
		}
	}

	if (ConfigSet.hub) {
		for (ban = config_ban; ban; ban = (ConfigItem_ban *)ban->next) {
			if (ban->type & BAN_NICKNAME && ban->type & BAN_NICKNAME_S) {
				sendto_one(cptr, ":%s SQLINE %s :%s", me.name, ban->mask, ban->reason);
			}
			if (ban->type & BAN_GCOS_S && ban->type & BAN_GCOS_S) {
				sendto_one(cptr, ":%s SGLINE %d :%s:%s", me.name, strlen(ban->mask),
					ban->mask, ban->reason);
			}
		}
	}
	else {
		ConfigItem b;
		for (ban = config_ban; ban; ban = (ConfigItem_ban *)ban->next) {
			b.next = del_ConfigItem((ConfigItem *)ban, (ConfigItem **)&config_ban);
			MyFree(ban->mask);
			MyFree(ban->reason);
			MyFree(ban);
			ban = (ConfigItem_ban *)&b;
		}
	}

	if (IsBurst(cptr)) {
		sendto_one(cptr, "BURST");
	}

	{
		chanMember *cm;
		static char nickissent = 1;

		nickissent = 3 - nickissent;
		for (chptr = channel; chptr; chptr = chptr->nextch) {
			for (cm = chptr->members; cm; cm = cm->next) {
				acptr = cm->cptr;
				if (acptr->nicksent != nickissent) {
					acptr->nicksent = nickissent;
					if (acptr->from != cptr) {
						sendnick_TS(cptr, acptr);
					}
				}
			}
			send_channel_modes(cptr, chptr);
		}

		for (acptr = &me; acptr; acptr = acptr->prev) {
			if (acptr->nicksent != nickissent) {
				acptr->nicksent = nickissent;
				if (acptr->from != cptr) {
					sendnick_TS(cptr, acptr);
				}
			}
		}
	}

	if (ZipOut(cptr)) {
		unsigned long inb, outb;
		double rat;

		zip_out_get_stats(cptr->serv->zip_out, &inb, &outb, &rat);
		if (inb) {
			sendto_gnotice("Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
				get_client_name(cptr, HIDEME), inb, outb, rat);
			sendto_serv_butone(cptr, ":%s GNOTICE :Connect burst to %s: %lu bytes normal, "
				"%lu compressed (%3.2f%%)", me.name, get_client_name(cptr, HIDEME),
				inb, outb, rat);
		}
	}

	cptr->flags |= FLAGS_BURST|FLAGS_PINGSENT;
	if (IsBurst(cptr)) {
		cptr->flags |= FLAGS_SOBSENT;
	}
	sendto_one(cptr, "PING :%s", me.name);

	return 0;
}

static void sendnick_TS(aClient *cptr, aClient *acptr)
{
	static char ubuf[12];

	if (IsPerson(acptr)) {
		send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);
		if (!*ubuf) {
			ubuf[0] = '+';
			ubuf[1] = '\0';
		}
		if (IsNICKIP(cptr)) {
			sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu %lu :%s", acptr->name,
				acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->user->username,
				acptr->user->host, acptr->user->server, acptr->user->servicestamp,
				htonl(acptr->ip.s_addr), acptr->info);
		}
		else {
			sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu :%s", acptr->name,
				acptr->hopcount + 1, acptr->tsinfo, ubuf, acptr->user->username,
				acptr->user->host, acptr->user->server, acptr->user->servicestamp,
				acptr->info);
		}
		if (IsSENDVHOST(cptr) && IsMasked(acptr)) {
			sendto_one(cptr, ":%s SENDVHOST :%s", acptr->name, acptr->user->maskedhost);
		}
	}
}

/*
 * send_lusers
 *	parv[0] = sender prefix
 *	parv[1] = servername
 *
 * Revamped to include cached channel count too.
 *				- TD (October 2001)
 */
int send_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	static long last_time = 0;
	static int s_count = 0, c_count = 0, u_count = 0, i_count = 0;
	static int o_count = 0, m_client = 0, m_server = 0, m_ulined = 0;
	static int ch_count = 0;
	int forced = (IsOper(sptr) && parc > 3);
	aClient *acptr;
	aChannel *chptr;

	Count.unknown = 0;
	m_server = Count.myserver;
	m_ulined = Count.myulined;
	m_client = Count.local;
	i_count = Count.invisi;
	u_count = Count.unknown;
	c_count = Count.total - Count.invisi;
	s_count = Count.server;
	o_count = Count.oper;
	ch_count = Count.chan;

	if (forced || last_time + LUSERS_CACHE_TIME > timeofday) {
		last_time = timeofday;
		s_count = c_count = u_count = i_count = o_count = m_client = m_server = m_ulined = 0;

		for (acptr = client; acptr; acptr = acptr->next) {
			switch (acptr->status) {
				case STAT_SERVER:
					if (MyConnect(acptr)) {
						m_server++;
						if (IsULine(acptr)) {
							m_ulined++;
						}
					}
				case STAT_ME:
					s_count++;
					break;
				case STAT_CLIENT:
					if (IsOper(acptr)) {
						o_count++;
					}
#ifdef SHOW_INVISIBLE_LUSERS
					if (MyConnect(acptr)) {
						m_client++;
					}
					if (!IsInvisible(acptr)) {
						c_count++;
					}
					else {
						i_count++;
					}
#else
					if (MyConnect(acptr)) {
						if (IsInvisible(acptr)) {
							if (IsOper(sptr)) {
								m_client++;
							}
						}
						else {
							m_client++;
						}
					}
					if (!IsInsible(acptr)) {
						c_count++;
					}
					else {
						i_count++;
					}
#endif /* SHOW_INVISIBLE_LUSERS */
					break;
				default:
					u_count++;
					break;
			}
		}

		for (chptr = channel; chptr; chptr = chptr->nextch) {
			ch_count++;
		}

		if (!forced) {
			if (m_server != Count.myserver) {
				sendto_realops_lev(DEBUG_LEV, "Local server count off by %d",
					Count.myserver - m_server);
				Count.myserver = m_server;
			}
			if (m_ulined != Count.myulined) {
				sendto_realops_lev(DEBUG_LEV, "Local ulined server count off by %d",
					Count.myulined - m_ulined);
				Count.myulined = m_ulined;
			}
			if (m_client != Count.local) {
				sendto_realops_lev(DEBUG_LEV, "Local client count off by %d",
					Count.local - m_client);
				Count.local = m_client;
			}
			if (s_count != Count.server) {
				sendto_realops_lev(DEBUG_LEV, "Server count off by %d",
					Count.server - s_count);
				Count.server = s_count;
			}
			if (i_count != Count.invisi) {
				sendto_realops_lev(DEBUG_LEV, "Invisible client count off by %d",
					Count.invisi - i_count);
				Count.invisi = i_count;
			}
			if (c_count + i_count != Count.total) {
				sendto_realops_lev(DEBUG_LEV, "Total client count off by %d",
					Count.total - (c_count + i_count));
				Count.total = c_count + i_count;
			}
			if (o_count != Count.oper) {
				sendto_realops_lev(DEBUG_LEV, "Oper count off by %d",
					Count.oper - o_count);
				Count.oper = o_count;
			}
			if (ch_count != Count.chan) {
				sendto_realops_lev(DEBUG_LEV, "Chan count off by %d",
					Count.chan - ch_count);
				Count.chan = ch_count;
			}
			Count.unknown = u_count;
		}
	}

#ifndef SHOW_INVISIBLE_LUSERS
	if (IsOper(sptr) && i_count) {
#endif
		sendto_one(sptr, rpl_str(RPL_LUSERCLIENT), me.name, parv[0], c_count, i_count,
			s_count);
#ifndef SHOW_INVISIBLE_LUSERS
	}
	else {
		sendto_one(sptr, ":%s %d %s :There are %d users on %d server", me.name,
			RPL_LUSERCLIENT, parv[0], c_count, s_count);
	}
#endif

	if (o_count) {
		sendto_one(sptr, rpl_str(RPL_LUSEROP), me.name, parv[0], o_count);
	}
	if (u_count > 0) {
		sendto_one(sptr, rpl_str(RPL_LUSERUNKNOWN), me.name, parv[0], u_count);
	}
	if (ch_count > 0) {
		sendto_one(sptr, rpl_str(RPL_LUSERCHANNELS), me.name, parv[0], ch_count);
	}

	if (ConfigSet.hideulines) {
		sendto_one(sptr, rpl_str(RPL_LUSERME), me.name, parv[0], m_client,
			IsOper(sptr) ? m_server : m_server - m_ulined);
	}
	else {
		sendto_one(sptr, rpl_str(RPL_LUSERME), me.name, parv[0], m_client, m_server);
	}

	sendto_one(sptr, rpl_str(RPL_LOCALUSERS), me.name, parv[0], Count.local, Count.max_loc);
	sendto_one(sptr, rpl_str(RPL_GLOBALUSERS), me.name, parv[0], Count.total, Count.max_tot);
	sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0], max_connection_count,
		max_client_count);

	if (m_client > max_client_count) {
		max_client_count = m_client;
	}
	if (m_client + m_server > max_connection_count) {
		max_connection_count = m_client + m_server;
		if (max_connection_count % 10 == 0) {
			sendto_ops("New highest connections: %d (%d clients)", max_connection_count,
				max_client_count);
		}
	}
	return 0;
}
