/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/s_user.c: user functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: s_user.c,v 1.81 2002/04/08 16:00:52 rageircd Exp $
 */

#include "struct.h"
#include "patchlevel.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"
#include "memory.h"
#include "openssl.h"
#if defined(HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif

BlockHeap *local_client_heap;
BlockHeap *remote_client_heap;
BlockHeap *local_user_heap;
BlockHeap *remote_user_heap;

static char buf[BUFSIZE];

char usermode_string[30];
int is_silenced(aClient *, aClient *);
int add_silence(aClient *, char *);
int del_silence(aClient *, char *);

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif
extern char chmode_string[512];
extern int add_ConfigItem(ConfigItem *, ConfigItem **);
extern char *find_client_ban(aClient *);

int user_modes[] = {
	UMODE_o,	'o',
	UMODE_i,	'i',
	UMODE_w,	'w',
	UMODE_s,	's',
	UMODE_c,	'c',
	UMODE_r,	'r',
	UMODE_R,	'R',
	UMODE_k,	'k',
	UMODE_n,	'n',
	UMODE_f,	'f',
	UMODE_y,	'y',
	UMODE_d,	'd',
	UMODE_a,	'a',
	UMODE_A,	'A',
	UMODE_b,	'b',
	UMODE_g,	'g',
	UMODE_h,	'h',
	UMODE_m,	'm',
	UMODE_e,	'e',
	UMODE_z,	'z',
	UMODE_N,	'N',
	UMODE_C,	'C',
	UMODE_p,	'p',
	UMODE_S,	'S',
	0,		0
};

#ifdef ANTI_SPAMBOT
int spam_time = MIN_JOIN_LEAVE_TIME;
int spam_num = MAX_JOIN_LEAVE_COUNT;
#endif

void client_heap_garbage_collect()
{
	BlockHeapGarbageCollect(local_client_heap);
	BlockHeapGarbageCollect(remote_client_heap);
	BlockHeapGarbageCollect(local_client_heap);
	BlockHeapGarbageCollect(remote_client_heap);
	return;
}

void init_clients()
{
	local_client_heap = BlockHeapCreate((size_t)CLIENT_LOCAL_SIZE, MAXCONNECTIONS);
	remote_client_heap = BlockHeapCreate((size_t)CLIENT_REMOTE_SIZE, CLIENTS_PREALLOCATE);
	local_user_heap = BlockHeapCreate((size_t)USER_LOCAL_SIZE, MAXCONNECTIONS);
	remote_user_heap = BlockHeapCreate((size_t)USER_REMOTE_SIZE, CLIENTS_PREALLOCATE);
	addEvent("client_heap_garbage_collect", client_heap_garbage_collect, NULL, CLIENT_HEAP_GARBAGE_COLLECT_FREQ, 1);
	return;
}

void build_usermode_string(void)
{
	int *s;
	char *m = usermode_string;

	for (s = user_modes; *s; s += 2) {
		*m++ = (char)*(s + 1);
	}
	*m = '\0';
	return;
}

aClient *next_client(aClient *next, char *ch)
{
	aClient *tmp = next;

	next = find_client(ch, tmp);
	if (tmp && tmp->prev == next) {
		return ((aClient *)NULL);
	}
	if (next != tmp) {
		return (next);
	}
	for (; next; next = next->next) {
		if (!match(ch, next->name)) {
			break;
		}
	}
	return (next);
}

aClient *next_client_double(aClient *next, char *ch)
{
	aClient *tmp = next;

	next = find_client(ch, tmp);
	if (tmp && tmp->prev ==next) {
		return ((aClient *)NULL);
	}
	if (next != tmp) {
		return (next);
	}
	for (; next; next = next->next) {
		if (!match(ch, next->name) || !match(next->name, ch)) {
			break;
		}
	}
	return next;
}

int hunt_server(aClient *cptr, aClient *sptr, char *command, int server, int parc, char *parv[])
{
	aClient *acptr;
	int wilds;

	if (parc <= server || BadPtr(parv[server]) || !match(me.name, parv[server])
	  || !match(parv[server], me.name)) {
		return (HUNTED_ISME);
	}

	acptr = find_client(parv[server], NULL);
	if (acptr) {
		if (acptr->from == sptr->from && !MyConnect(acptr)) {
			acptr = NULL;
		}
	}
	if (!acptr && (acptr = find_server(parv[server], NULL))) {
		if (acptr->from == sptr->from && !MyConnect(acptr)) {
			acptr = NULL;
		}
	}

	collapse(parv[server]);
	wilds = (strchr(parv[server], '?') || strchr(parv[server], '*'));

	if (!acptr) {
		if (!wilds) {
			acptr = find_name(parv[server], (aClient *)NULL);
			if (!acptr || !IsRegistered(acptr) || !IsServer(acptr)) {
				sendto_one(sptr, err_str(ERR_NOSUCHSERVER),me.name, parv[0], parv[server]);
				return (HUNTED_NOSUCH);
			}
		}
		else {
			for (acptr = client; (acptr = next_client(acptr, parv[server])); acptr = acptr->next) {
				if (acptr->from == sptr->from && !MyConnect(acptr)) {
					continue;
				}
				if (IsRegistered(acptr) && acptr != cptr) {
					break;
				}
			}
		}
	}

	if (acptr) {
		if (IsMe(acptr) || MyClient(acptr)) {
			return (HUNTED_ISME);
		}
		if (match(acptr->name, parv[server])) {
			parv[server] = acptr->name;
		}
		sendto_one(acptr, command, parv[0], parv[1], parv[2], parv[3], parv[4], parv[5], parv[6],
			parv[7], parv[8]);
		return (HUNTED_PASS);
	}
	sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[server]);
	return (HUNTED_NOSUCH);
}

char *canonize(char *buffer)
{
	static char cbuf[BUFSIZE];
	char *s, *t, *cp = cbuf, *p = NULL, *p2;
	int l = 0;

	*cp = '\0';
	for (s = strtoken(&p, buffer, ","); s; s = strtoken(&p, NULL, ",")) {
		if (l) {
			for (p2 = NULL, t = strtoken(&p2, cbuf, ","); t; t = strtoken(&p2, NULL, ",")) {
				if (!mycmp(s, t)) {
					break;
				}
				else if (p2) {
					p2[-1] = ',';
				}
			}
		}
		else {
			t = NULL;
		}
		if (!t) {
			if (l) {
				*(cp - 1) = ',';
			}
			else {
				l = 1;
			}
			strcpy(cp, s);
			if (p) {
				cp += (p - s);
			}
		}
		else if (p2) {
			p2[-1] = ',';
		}
	}
	return (cbuf);
}

int register_user(aClient *cptr, aClient *sptr, char *nick, char *username)
{
	static char ubuf[12];
	char *parv[3], *p, tmpstr2[512];
	aClient *nsptr;
	anUser *user = sptr->user;
	int i, dots = 0, bad_dns = 0;
	ConfigItem_ban *ban;
	char mask[USERLEN + HOSTLEN + 2];
#ifdef ANTI_SPAMBOT
	char spamchar = '\0';
#endif
#ifdef HAVE_OPENSSL_ON
	char *opensslinfo = NULL;
#endif

	user->last = timeofday;
	parv[0] = sptr->name;
	parv[1] = parv[2] = NULL;

	if (MyConnect(sptr)) {
		p = inetntoa((char *)&sptr->ip);
		strncpyzt(sptr->hostip, p, HOSTIPLEN + 1);

		i = check_client(sptr);
		if (i) {
			char *hostptr = get_client_host(sptr);
			if (i != -2) {
				if (i == -5) {
					sendto_realops_lev(CCONN_LEV, "Unauthorized client connection "
						"from %s [%s]", hostptr, p);
					ircdlog(LOG_CLIENT, "Unauthorized client connection from %s[%s]: -5",
						hostptr, p);
					return -5;
				}
				else if (i == -3) {
					sendto_realops_lev(SPY_LEV, "Connection class full for %s [%s]",
						hostptr, p);
					ircdlog(LOG_CLIENT, "Unauthorized client connection from %s[%s]: class full {%s}",
						hostptr, p, sptr->class ? sptr->class->name : "default");
					return -3;
				}
			}
			else {
				ircdlog(LOG_ERROR, "Socket error for connecting client %s[%s]",
					hostptr, p);
				return exit_client(cptr, sptr, &me, "Socket Error");
			}
		}

#ifdef ANTI_SPAMBOT
		if (*(user->host + 1) == '\0') {
			spamchar = *user->host;
		}
#endif

		strncpyzt(user->host, sptr->sockhost, HOSTLEN);
		p = user->host;

		while (*p) {
			if (!isalnum(*p)) {
				if (*p != '-' && *p != '.') {
					bad_dns = 1;
				}
			}
			if (*p == '.') {
				dots++;
			}
			p++;
		}

		if (!dots) {
			sendto_realops("Invalid hostname for %s, dumping user %s", sptr->hostip, sptr->name);
			return exit_client(cptr, sptr, &me, "Invalid hostname");
		}
		if (bad_dns) {
			sendto_one(sptr, ":%s NOTICE %s :*** Notice -- You have a bad character in your hostname",
				me.name, sptr->name);
			strcpy(user->host, sptr->hostip);
			strcpy(sptr->sockhost, sptr->hostip);
		}

		if (sptr->flags & FLAGS_DOID && !(sptr->flags & FLAGS_GOTID)) {
			char temp[USERLEN + 1];
			strncpyzt(temp, username, USERLEN + 1);
			*user->username = '~';
			strncpy(&user->username[1], temp, USERLEN);
			user->username[USERLEN] = '\0';
		}
		else if (ConfigSet.ignoreident && (sptr->flags & FLAGS_GOTID) && *sptr->username != '-') {
			strncpyzt(user->username, sptr->username, USERLEN + 1);
		}
		else {
			strncpyzt(user->username, username, USERLEN + 1);
		}

		if (sptr->fd >= ConfigSet.maxclients + MAX_BUFFER || sptr->fd >= ConfigSet.maxclients - 5) {
			sendto_realops_lev(SPY_LEV, "Too many clients, rejecting %s[%s]", nick,
				sptr->sockhost);
			ircstp->is_ref++;
			return exit_client(cptr, sptr, &me, "Sorry, server is full - try later");
		}

		ircsprintf(mask, "%s@%s", user->username, user->host);
		ban = find_ban(BAN_USERHOST, mask);
		if (!ban) {
			ban = find_ban(BAN_AUTOKILL, mask);
			if (!ban) {
				ban = find_ban(BAN_IP, sptr->hostip);
			}
		}
		if (ban) {
			int kflag = ban->type == BAN_IP ? 2 : ban->type == BAN_USERHOST ? 1 : 0;
			sendto_one(sptr, err_str(ERR_YOUREBANNEDCREEP), me.name, sptr->name,
				kflag > 1 ? "Z:Lined" : kflag ? "K:Lined" : "AutoKilled");
			sendto_one(sptr, ":%s NOTICE %s :*** You are not welcome on this %s: %s",
				me.name, sptr->name, kflag ? "server" : "network", ban->reason);
			sendto_one(sptr, ":%s NOTICE %s :*** For more information, mail %s and include "
				"everything shown here.", me.name, sptr->name,
				kflag ? ConfigSet.localklineaddress : ConfigSet.klineaddress);
			ircstp->is_ref++;
			ircdlog(LOG_BAN, "blocking %s client %s (%s)",
				kflag > 1 ? "Z:Lined" : kflag ? "K:Lined" : "AutoKilled",
				get_client_name(cptr, FALSE), ban->reason);
			return exit_client(cptr, sptr, &me, kflag > 1 ? "Z:Lined" : kflag ? "K:Lined" : "AutoKilled");
		}

#ifdef ANTI_SPAMBOT
		if (spamchar == 'x') {
			sendto_realops_lev(REJ_LEV, "Rejecting possible spambot: %s (single char user-given "
				"userhost: %c", get_client_name(sptr, FALSE), spamchar);
			ircstp->is_ref++;
			return exit_client(cptr, sptr, sptr, "Spambot detected and rejected");
		}
#endif

		{
			char *tmpstr;
			u_char c, cc = 0;
			int lower = 0, upper = 0, special = 0;

			if (strchr(user->username, '@') || strchr(username, '@')) {
				sendto_realops_lev(REJ_LEV, "Illegal \"@\" in username: %s (%s)",
					get_client_name(sptr, FALSE), username);
				ircstp->is_ref++;
				ircsprintf(tmpstr2, "Invalid username [%s] - '@' is not allowed!", username);
				return exit_client(cptr, sptr, sptr, tmpstr2);
			}

			if (ConfigSet.ignorefirstchar) {
				tmpstr = (user->username[0] == '~') ? &user->username[2] : &user->username[1];
				cc = (user->username[0] == '~') ? user->username[1] : user->username[0];

				if ((!isalnum(cc) && !strchr(" -_.", cc)) || cc > 127) {
					special++;
				}
			}
			else {
				tmpstr = (user->username[0] == '~') ? &user->username[1] : &user->username[0];
			}
			while (*tmpstr) {
				c = *(tmpstr++);

				if (islower(c)) {
					lower++;
					continue;
				}
				if (isupper(c)) {
					upper++;
					continue;
				}
				if ((!isalnum(c) && !strchr(" -_.", c)) || c > 127 || c < 32) {
					special++;
				}
			}
			if ((ConfigSet.nomixedcase && lower && upper) || special) {
				sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)", nick,
					user->username, user->host);
				ircstp->is_ref++;
				ircsprintf(tmpstr2, "Invalid username [%s]", user->username);
				return exit_client(cptr, sptr, &me, tmpstr2);
			}

			lower = upper = special = cc = 0;

			if (strcmp(user->username, username)) {
				if (ConfigSet.ignorefirstchar) {
					tmpstr = (username[0] == '~') ? &username[2] : &username[1];
					cc = (username[0] == '~') ? username[1] : username[0];

					if ((!isalnum(cc) && !strchr(" -_.", cc)) || cc > 127) {
						special++;
					}
				}
				else {
					tmpstr = (username[0] == '~') ? &username[1] : username;
				}
				while (*tmpstr) {
					c = *(tmpstr++);

					if (islower(c)) {
						lower++;
						continue;
					}
					if (isupper(c)) {
						upper++;
						continue;
					}
					if ((!isalnum(c) && !strchr(" -_.", c)) || c > 127) {
						special++;
					}
				}
				if ((ConfigSet.nomixedcase && lower && upper) || special) {
					sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)", nick,
						username, user->host);
					ircstp->is_ref++;
					ircsprintf(tmpstr2, "Invalid username [%s]", username);
					return exit_client(cptr, sptr, &me, tmpstr2);
				}
			}
		}

		if (user->username[1] == '\0' && !isalpha(user->username[0])) {
			sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)", nick,
				user->username, user->host);
			ircstp->is_ref++;
			ircsprintf(tmpstr2, "Invalid username [%s]", user->username);
			return exit_client(cptr, sptr, &me, tmpstr2);
		}

#ifdef HAVE_OPENSSL_ON
		if (cptr->flags & FLAGS_OPENSSL) {
			if (cptr->openssl) {
				opensslinfo = openssl_get_cipher((SSL *)cptr->openssl);
			}
			ircdlog(LOG_CLIENT, "secure client connecting on port %d: %s [%s]",
				sptr->listener->port, get_client_name(sptr, FALSE), opensslinfo);
			sendto_realops_lev(CCONN_LEV, "Client connecting: %s (%s@%s) [%s/%s] {%s}", nick,
				user->username, user->host, sptr->hostip, opensslinfo,
				sptr->class ? sptr->class->name : "no class");
		}
		else
#endif
		{
			ircdlog(LOG_CLIENT, "client connecting on port %d: %s",
				sptr->listener->port, get_client_name(sptr, FALSE));
			sendto_realops_lev(CCONN_LEV, "Client connecting: %s (%s@%s) [%s] {%s}", nick,
				user->username, user->host, sptr->hostip,
				sptr->class ? sptr->class->name : "no class");

		}

		if (++Count.local > Count.max_loc) {
			Count.max_loc = Count.local;
			if (!(Count.max_loc % 10)) {
				sendto_ops("New max local clients: %d", Count.max_loc);
			}
		}

		if (NOW - Count.day > 86400) {
			Count.today = 0;
			Count.day = NOW;
		}
		if (NOW - Count.week > 604800) {
			Count.weekly = 0;
			Count.week = NOW;
		}
		if (NOW - Count.month > 2592000) {
			Count.monthly = 0;
			Count.month = NOW;
		}
		if (NOW - Count.year > 31536000) {
			Count.yearly = 0;
			Count.year = NOW;
		}
		Count.today++;
		Count.weekly++;
		Count.monthly++;
		Count.yearly++;

		if (sptr->flags & FLAGS_BAD_DNS) {
			sendto_realops_lev(SPY_LEV, "DNS lookup: %s (%s@%s) is a possible cache polluter",
				sptr->name, sptr->user->username,
				IsMasked(sptr) ? sptr->user->maskedhost : sptr->user->host);
		}
	}
	else {
		strncpyzt(user->username, username, USERLEN + 1);
	}

	SetClient(sptr);
	do_maskedhost(sptr);

	if (++Count.total > Count.max_tot) {
		Count.max_tot = Count.total;
	}
	if (IsInvisible(sptr)) {
		Count.invisi++;
	}

	if (MyConnect(sptr)) {
		sptr->pingval = sptr->class->pingfreq;
		sptr->sendqlen = sptr->class->sendq;
#ifdef MAXBUFFERS
		reset_sock_opts(sptr->fd, 0);
#endif

		sendto_one(sptr, rpl_str(RPL_WELCOME), me.name, nick, ConfigSet.ircnetwork, nick,
			sptr->user->username, sptr->user->host);
		sendto_one(sptr, rpl_str(RPL_YOURHOST), me.name, nick, me.name, sptr->listener->port,
			version);
#ifdef IRCII_KLUDGE
		sendto_one(sptr, "NOTICE %s :*** Your host is %s, running version %s", nick,
			get_client_name(&me, TRUE), version);
#endif
		sendto_one(sptr, rpl_str(RPL_CREATED), me.name, nick, creation);
		sendto_one(sptr, rpl_str(RPL_MYINFO), me.name, parv[0], me.name, version,
			usermode_string, chmode_string);
		sendto_one(sptr, rpl_str(RPL_PROTOCTL), me.name, parv[0], MAXWATCH,
			ConfigSet.ircnetwork, PROTOCTL_CHANTYPES, PROTOCTL_PREFIX);

#ifdef HAVE_OPENSSL_ON
		if (cptr->flags & FLAGS_OPENSSL && cptr->openssl != NULL) {
			sendto_one(sptr, "NOTICE %s :*** You are connected to %s via %s.", nick, me.name, opensslinfo);
		}
#endif

		send_lusers(sptr, sptr, 1, parv);
		runHook1(HOOKTYPE_LOCAL_CONNECT, sptr);
	}
	else if (IsServer(cptr)) {
		aClient *acptr;

		acptr = find_server(user->server, NULL);
		if (!acptr) {
			sendto_one(cptr, ":%s KILL %s :%s GHOST (no server %s on this net)", me.name,
				sptr->name, me.name, user->server);
			sendto_realops_lev(DEBUG_LEV, "No server %s for user %s[%s@%s] from %s",
				user->server, sptr->name, user->username, user->host,
				sptr->from->name);
			sptr->flags |= FLAGS_KILLED;
			return exit_client(sptr, sptr, &me, "Ghosted Client");
		}
		else if (acptr && acptr->from != sptr->from) {
			sendto_realops_lev(DEBUG_LEV, "Bad User [%s] :%s USER %s@%s %s, != %s[%s]",
				cptr->name, nick, user->username, user->host, user->server,
				acptr->name, acptr->from->name);
			sendto_one(sptr, ":%s KILL %s :%s (%s != %s[%s] USER from wrong direction)",
				me.name, sptr->name, me.name, user->server, acptr->from->name,
				acptr->from->sockhost);
			sptr->flags |= FLAGS_KILLED;
			return exit_client(sptr, sptr, &me, "USER server wrong direction");
		}
	}

	send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);
	if (!*ubuf) {
		ubuf[0] = '+';
		ubuf[1] = '\0';
	}

	hash_check_watch(sptr, RPL_LOGON);

	sendto_nickip_servs_butone(1, cptr, "NICK %s %d %ld %s %s %s %s %lu %lu :%s", nick,
		sptr->hopcount + 1, sptr->tsinfo, ubuf, user->username, user->host, user->server,
		sptr->user->servicestamp, htonl(sptr->ip.s_addr), sptr->info);
	sendto_nickip_servs_butone(0, cptr, "NICK %s %d %ld %s %s %s %s %lu :%s", nick, sptr->hopcount + 1,
		sptr->tsinfo, ubuf, user->username, user->host, user->server, sptr->user->servicestamp,
		sptr->info);

	if (MyClient(sptr)) {
		nsptr = find_person(ConfigSet.nickserv, NULL);
		if (sptr->passwd[0] && nsptr) {
			sendto_one(nsptr, ":%s PRIVMSG %s@%s :IDENTIFY %s", sptr->name, ConfigSet.nickserv,
				ConfigSet.services, sptr->passwd);
		}
		memset(sptr->passwd, '\0', PASSWDLEN);

		if (ubuf[1]) {
			send_umode(cptr, sptr, 0, ALL_UMODES, ubuf);
		}
	}
	return 0;
}

int check_dccsend(aClient *from, aClient *to, char *filename)
{
	Link *lp;
	ConfigItem_ban *ban;

	if (from == to || !IsPerson(from) || IsOper(from) || !MyClient(to) || !*filename) {
		return 0;
	}

	ban = find_ban(BAN_FILE, filename);
	if (!ban) {
		return 0;
	}

	for (lp = to->user->dccallow; lp; lp = lp->next) {
		if (lp->flags == DCC_LINK_ME && lp->value.cptr == from) {
			break;
		}
	}
	if (!lp) {
		aChannel *chptr = NULL;
		Link *tlp, *flp;

		sendto_one(from, ":%s NOTICE %s :*** The file %s has been blocked on this server, and user "
			"%s (%s@%s) is not currently accepting DCC sends from you. Therefore, the file was "
			"blocked.", me.name, from->name, filename, to->name, to->user->username,
			IsMasked(to) ? to->user->maskedhost : to->user->host);
		sendto_one(to, ":%s NOTICE %s :*** User %s (%s@%s) attempted to DCC send you file %s, which "
			"is banned (%s).", me.name, to->name, from->name, from->user->username,
			IsMasked(from) ? from->user->maskedhost : from->user->host, filename,
			!BadPtr(ban->reason) ? ban->reason : "no reason specified");

		if (!(to->flags & FLAGS_SEENDCCNOTICE)) {
			to->flags |= FLAGS_SEENDCCNOTICE;
			sendto_one(to, ":%s NOTICE %s :*** The majority of files sent of this type are "
				"malicious virri and trojan horses. In order to protect the clients on %s, "
				"the DCC sends of these files are blocked by default.", me.name, to->name);
			sendto_one(to, ":%s NOTICE %s :*** If you trust %s, and want him/her to send you this"
				"file, you can add them to your DCC allow list. To obtain more information on "
				"the DCC allow system, type /DCCALLOW HELP now.", me.name, to->name, from->name);
		}

		for (tlp = to->user->channel; tlp && !chptr; tlp = tlp->next) {
			for (flp = from->user->channel; flp && !chptr; flp = flp->next) {
				if (tlp->value.chptr == flp->value.chptr) {
					chptr = tlp->value.chptr;
				}
			}
		}
		if (chptr) {
			sendto_realops_lev(DCCSEND_LEV, "%s (%s@%s) attempted to send file %s to %s (%s@%s) "
				"[channel %s]", from->name, from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host,
				filename, to->name, to->user->username,
				IsMasked(to) ? to->user->maskedhost : to->user->host, chptr->chname);
		}
		else {
			sendto_realops_lev(DCCSEND_LEV, "%s (%s@%s) attempted to send file %s to %s (%s@%s)",
				from->name, from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host,
				filename, to->name, to->user->username,
				IsMasked(to) ? to->user->maskedhost : to->user->host);
		}
		ircdlog(LOG_DCCALLOW, "client %s attempted to send blocked file %s to %s",
			get_client_name(from, FALSE), filename, get_client_name(to, FALSE));
		return 1;
	}
	return 0;
}

int check_message_accept(aClient *from, aClient *to)
{
	Link *lp;

	if (from == to || !IsPerson(from) || IsOper(from) || !MyClient(to) || !IsUmodep(to)) {
		return 0;
	}

	for (lp = to->user->accept; lp; lp = lp->next) {
		if (lp->flags == ACCEPT_LINK_ME && lp->value.cptr == from) {
			break;
		}
	}
	if (!lp) {
		sendto_one(from, ":%s NOTICE %s :*** User %s (%s@%s) is in private mode (+p), and is not "
			"currently accepting messages from you.", me.name, from->name, to->name,
			to->user->username, IsMasked(to) ? to->user->maskedhost : to->user->host);
		if (timeofday - to->user->last_accept_notice >= ConfigSet.acceptnoticetime) {
			to->user->last_accept_notice = timeofday;
			sendto_one(from, ":%s NOTICE %s :*** They have been informed of your message attempt.",
				me.name, from->name);
			sendto_one(to, ":%s NOTICE %s :*** User %s (%s@%s) attempted to message you. However, "
				"you are currently in private mode (+p) and therefore the message was blocked.",
				me.name, to->name, from->name, from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host);
			sendto_one(to, ":%s NOTICE %s :*** To accept messages from %s, you can add him/her to "
				"your message accept list. Type /ACCEPT HELP for more information.", me.name,
				to->name, from->name);
		}
		return 1;
	}
	return 0;
}

int do_user(char *nick, aClient *cptr, aClient *sptr, char *username, char *host,
  char *server, unsigned long serviceid, unsigned int nickip, char *realname)
{
	anUser *user = make_user(sptr);

	if (!MyConnect(sptr)) {
		user->server = find_or_add(server);
		strncpyzt(user->host, host, sizeof(user->host));
	}
	else {
		if (!IsUnknown(sptr)) {
			sendto_one(sptr, rpl_str(ERR_ALREADYREGISTRED), me.name, nick);
			return 0;
		}
		sptr->umode |= ConfigSet.connectmodes;
		sptr->umode |= ((UMODE_i|UMODE_w|UMODE_s) & atoi(host));
		strncpyzt(user->host, host, sizeof(user->host));
		user->server = me.name;
	}

	strncpyzt(sptr->info, realname, sizeof(sptr->info));
	sptr->user->servicestamp = serviceid;

	if (!MyConnect(sptr)) {
		sptr->ip.s_addr = ntohl(nickip);
	}
	if (MyConnect(sptr)) {
		sptr->oflag = 0;
	}
	if (sptr->name[0]) {
		return register_user(cptr, sptr, sptr->name, username);
	}
	else {
		strncpyzt(sptr->user->username, username, USERLEN + 1);
	}
	return 0;
}

/*
 * user_mode
 *	parv[0] = sender prefix
 *	parv[1] = client to change mode for
 *	parv[2] = modes to change
 */
int user_mode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	int flag, *s, setflags = 0, what = MODE_ADD;
	char *m, **p;
	aClient *acptr;
	short badflag = 0;

	if (parc < 2 || *parv[1] == '\0') {
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
		return 0;
	}

	acptr = find_person(parv[1], NULL);
	if (!acptr) {
		if (MyConnect(sptr)) {
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
				parv[1]);
		}
		return 0;
	}

	if (IsServer(sptr) || sptr != acptr || acptr->from != sptr->from) {
		if (!IsServer(cptr)) {
			sendto_one(sptr, err_str(ERR_USERSDONTMATCH), me.name, parv[0]);
		}
		return 0;
	}

	if (parc < 3) {
		m = buf;
		*m++ = '+';
		for (s = user_modes; (flag = *s) && (m - buf < BUFSIZE - 4); s += 2) {
			if (sptr->umode & flag) {
				*m++ = (char)(*(s + 1));
			}
		}
		*m = '\0';
		sendto_one(sptr, rpl_str(RPL_UMODEIS), me.name, parv[0], buf);
		return 0;
	}

	for (s = user_modes; (flag = *s); s += 2) {
		if (sptr->umode & flag) {
			setflags |= flag;
		}
	}

	for (p = &parv[2]; p && *p; p++) {
		for (m = *p; *m; m++) {
			switch (*m) {
				case '+':
					what = MODE_ADD;
					break;
				case '-':
					what = MODE_DEL;
					break;
				case ' ':
				case '\n':
				case '\r':
				case '\t':
				case 'r':
					break;
				default:
					for (s = user_modes; (flag = *s); s += 2) {
						if (*m == (char)(*(s + 1))) {
							if (what == MODE_ADD) {
								sptr->umode |= flag;
							}
							else {
								sptr->umode &= ~flag;
							}
							break;
						}
					}
					if (!flag && MyConnect(sptr)) {
						badflag = 1;
					}
					break;
			}
		}
	}

	if (badflag) {
		sendto_one(sptr, err_str(ERR_UMODEUNKNOWNFLAG), me.name, parv[0]);
	}

	if (!(setflags & UMODE_o) && IsOper(sptr) && !IsServer(cptr)) {
		ClearOper(sptr);
	}
	if (!(setflags & UMODE_o) && IsOper(sptr)) {
		Count.oper++;
	}
	if ((setflags & UMODE_o) && !IsOper(sptr)) {
		Count.oper--;
		if (MyConnect(sptr)) {
			delfrom_fdlist(sptr->fd, &oper_fdlist);
		}
	}

	if (!IsOper(sptr) && !IsServer(cptr)) {
		if (!(setflags & UMODE_b)) {
			sptr->umode &= ~UMODE_b;
		}
		if (!(setflags & UMODE_R)) {
			sptr->umode &= ~UMODE_R;
		}
		if (!(setflags & UMODE_e)) {
			sptr->umode &= ~UMODE_e;
		}
		if (!(setflags & UMODE_y)) {
			sptr->umode &= ~UMODE_y;
		}
		if (!(setflags & UMODE_n)) {
			sptr->umode &= ~UMODE_n;
		}
		if (!(setflags & UMODE_d)) {
			sptr->umode &= ~UMODE_d;
		}
		if (!(setflags & UMODE_h)) {
			sptr->umode &= ~UMODE_h;
		}
	}
	if (!IsULine(sptr) && !IsServer(cptr)) {
		if (!(setflags & UMODE_S)) {
			sptr->umode &= ~UMODE_S;
		}
	}

	if (!(setflags & UMODE_i) && IsInvisible(sptr)) {
		Count.invisi++;
	}
	if ((setflags & UMODE_i) && !IsInvisible(sptr)) {
		Count.invisi--;
	}

	if (!IsOper(sptr) && !IsServer(sptr)) {
		if (IsNetAdmin(sptr)) {
			ClearNetAdmin(sptr);
		}
		if (IsAdmin(sptr)) {
			ClearAdmin(sptr);
		}
		if (IsCoAdmin(sptr)) {
			ClearCoAdmin(sptr);
		}
		if (IsSAdmin(sptr)) {
			ClearSAdmin(sptr);
		}
		if (IsUmodef(sptr)) {
			ClearUmodef(sptr);
		}
		if (IsUmodec(sptr)) {
			ClearUmodec(sptr);
		}
		if (IsUmodee(sptr)) {
			ClearUmodee(sptr);
		}
		if (IsUmodey(sptr)) {
			ClearUmodey(sptr);
		}
		if (IsUmoded(sptr)) {
			ClearUmoded(sptr);
		}
		if (IsUmodeb(sptr)) {
			ClearUmodeb(sptr);
		}
		if (IsUmoden(sptr)) {
			ClearUmoden(sptr);
		}
		if (IsUmodem(sptr)) {
			ClearUmodem(sptr);
		}
		if (IsRouting(sptr)) {
			ClearRouting(sptr);
		}
	}
	if (!IsULine(sptr) && !IsServer(cptr)) {
		if (IsServices(sptr)) {
			ClearServices(sptr);
		}
	}

	if (MyClient(sptr) && IsOper(sptr)) {
		if (IsNetAdmin(sptr) && !OPIsNetAdmin(sptr)) {
			ClearNetAdmin(sptr);
		}
		if (IsAdmin(sptr) && !OPIsAdmin(sptr)) {
			ClearAdmin(sptr);
		}
		if (IsCoAdmin(sptr) && !OPIsCoAdmin(sptr)) {
			ClearCoAdmin(sptr);
		}
		if (IsSAdmin(sptr) && !OPIsSAdmin(sptr)) {
			ClearSAdmin(sptr);
		}
		if (IsUmodef(sptr) && !OPSetUModeF(sptr)) {
			ClearUmodef(sptr);
		}
		if (IsUmodec(sptr) && !OPSetUModeC(sptr)) {
			ClearUmodec(sptr);
		}
		if (IsRouting(sptr) && !OPIsRStaff(sptr)) {
			ClearRouting(sptr);
		}
	}
	if (IsMasked(sptr)) {
		do_maskedhost(sptr);
	}
	send_umode_out(cptr, sptr, setflags);
	return 0;
}

void send_umode(aClient *cptr, aClient *sptr, int old, int sendmask, char *umode_buf)
{
	int *s, flag, what = MODE_NULL;
	char *m = umode_buf;

	*m = '\0';

	for (s = user_modes; (flag = *s); s += 2) {
		if (MyClient(sptr) && !(flag & sendmask)) {
			continue;
		}
		if ((flag & old) && !(sptr->umode & flag)) {
			if (what == MODE_DEL) {
				*m++ = *(s + 1);
			}
			else {
				what = MODE_DEL;
				*m++ = '-';
				*m++ = *(s + 1);
			}
		}
		else if (!(flag & old) && (sptr->umode & flag)) {
			if (what == MODE_ADD) {
				*m++ = *(s + 1);
			}
			else {
				what = MODE_ADD;
				*m++ = '+';
				*m++ = *(s + 1);
			}
		}
	}

	*m = '\0';

	if (*umode_buf && cptr) {
		sendto_one(cptr, ":%s MODE %s :%s", sptr->name, sptr->name, umode_buf);
	}
}

void send_umode_out(aClient *cptr, aClient *sptr, int old)
{
	int i, j;
	aClient *acptr;
	fdlist listp = serv_fdlist;

	send_umode(NULL, sptr, old, SEND_UMODES, buf);

	for (i = listp.entry[j = 1]; j <= listp.last_entry; i = listp.entry[++j]) {
		if ((acptr = local[i]) && acptr != cptr && acptr != sptr && *buf) {
			sendto_one(acptr, ":%s MODE %s :%s", sptr->name, sptr->name, buf);
		}
	}

	if (cptr && MyClient(cptr)) {
		send_umode(cptr, sptr, old, ALL_UMODES, buf);
	}
}

int is_silenced(aClient *sptr, aClient *acptr)
{
	Link *lp;
	char sender[NICKLEN + USERLEN + HOSTLEN + 5];

	if (!acptr->user || !acptr->user->silence || !sptr->user) {
		return 0;
	}

	ircsprintf(sender, "%s!%s@%s", sptr->name, sptr->user->username, sptr->user->host);
	for (lp = acptr->user->silence; lp; lp = lp->next) {
		if (!match(lp->value.cp, sender)) {
			if (!MyConnect(sptr)) {
				sendto_one(sptr->from, ":%s SILENCE %s :%s", acptr->name, sptr->name,
					lp->value.cp);
				lp->flags = 1;
			}
			return 1;
		}
	}
	return 0;
}

int del_silence(aClient *sptr, char *mask)
{
	Link **lp, *tmp;

	for (lp = &(sptr->user->silence); *lp; lp = &((*lp)->next)) {
		if (!mycmp(mask, (*lp)->value.cp)) {
			tmp = *lp;
			*lp = tmp->next;
			MyFree(tmp->value.cp);
			free_link(tmp);
			return 0;
		}
	}
	return 1;
}

int add_silence(aClient *sptr, char *mask)
{
	Link *lp;
	int cnt = 0, len = 0;

	for (lp = sptr->user->silence; lp; lp = lp->next) {
		len += strlen(lp->value.cp);
		if (MyClient(sptr)) {
			if (len > MAXSILELENGTH || ++cnt >= MAXSILES) {
				sendto_one(sptr, err_str(ERR_SILELISTFULL), me.name, sptr->name, mask);
				return -1;
			}
			else {
				if (!match(lp->value.cp, mask)) {
					return -1;
				}
			}
		}
		else if (!mycmp(lp->value.cp, mask)) {
			return -1;
		}
	}

	lp = make_link();
	lp->next = sptr->user->silence;
	lp->value.cp = (char *)MyMalloc(strlen(mask) + 1);
	strcpy(lp->value.cp, mask);
	sptr->user->silence = lp;
	return 0;
}
