/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/send.c: sending routenes
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: send.c,v 1.23 2002/04/08 16:00:52 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "numeric.h"
#include "dh.h"
#include "zlink.h"
#include <stdio.h>
#include <time.h>
#include "memory.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendbuf[2048];
static char remotebuf[2048];
static int send_message(aClient *, char *, int);

#ifdef HAVE_OPENSSL_ON
/* WARNING -- rc4buf size must be the same as zipOutBuf in zlink.c */
static char rc4buf[16384];
#endif

static int sentalong[MAXCONNECTIONS];
static int sent_serial;

void init_send()
{
	memset(sentalong, '\0', sizeof(int) * MAXCONNECTIONS);
	sent_serial = 0;
	return;
}
#define HIGHEST_SERIAL INT_MAX
#define INC_SERIAL if (sent_serial == HIGHEST_SERIAL) { memset(sentalong, '\0', sizeof(sentalong)); sent_serial = 0; } sent_serial++;

static int dead_link(aClient *to, char *notice, int sockerr)
{
	int errtmp = errno;

	to->sockerr = sockerr;
	to->flags = FLAGS_DEADSOCKET;

	DBufClear(&to->recvQ);
	DBufClear(&to->sendQ);

	if (IsServer(to) && !(to->flags & FLAGS_CLOSING)) {
		char fbuf[512], *path = get_client_name(to, HIDEME), *errtxt = strerror(errtmp);

		ircsprintf(fbuf, "%s", notice);
		sendto_gnotice(fbuf, path, errtxt);
		ircsprintf(fbuf, ":%s GNOTICE :%s", me.name, notice);
		sendto_serv_butone(to, fbuf, path, errtxt);
	}
	return -1;
}

static int send_message(aClient *to, char *msg, int len)
{
	static int SQinK;

#ifdef DUMP_DEBUG
	fprintf(dumpfd, "-> %s: %s\n", to->name ? to->name, : "*", msg);
#endif

	if (to->from) {
		to = to->from;
	}

	if (IsServer(to) || IsNegoServer(to)) {
		if (len > 510) {
			msg[511] = '\n';
			msg[512] = '\0';
			len = 512;
		}
		else {
			msg[len] = '\n';
			msg[++len] = '\0';
		}
	}
	else {
		if (len > 509) {
			msg[510] = '\r';
			msg[511] = '\n';
			msg[512] = '\0';
			len = 512;
		}
		else {
			msg[len] = '\r';
			msg[++len] = '\n';
			msg[++len] = '\0';
		}
	}

	if (IsMe(to)) {
		sendto_ops("Trying to send to myself! [%s]", msg);
		return 0;
	}
	if (IsDead(to)) {
		return 0;
	}

	if (DBufLength(&to->sendQ) > to->sendqlen) {
		if (IsServer(to)) {
			sendto_ops("Max SendQ limit exceeded for %s: %d > %d", get_client_name(to, HIDEME),
				DBufLength(&to->sendQ), to->class ? to->class->sendq : MAXSENDQLENGTH);
		}
		to->flags |= FLAGS_SENDQEX;
		return dead_link(to, "Max SendQ exceeded for %s, closing link", 0);
	}

	if (ZipOut(to)) {
		int ldata = to->flags & FLAGS_BURST;

		msg = zip_output(to->serv->zip_out, msg, &len, 0, &ldata);
		if (len == -1) {
			sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
			return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
		}
		if (!len) {
			return 0;
		}
	}

#ifdef HAVE_OPENSSL_ON
	if (IsRC4OUT(to)) {
		rc4_process_stream_to_buf(to->serv->rc4_out, msg, rc4buf, len);
		msg = rc4buf;
	}
#endif

	if (dbuf_put(&to->sendQ, msg, len) < 0) {
		return dead_link(to, "Buffer allocation error for %s, closing link", IRCERR_BUFALLOC);
	}

	to->sendM += 1;
	me.sendM += 1;
	if (to->listener != &me) {
		to->listener->sendM += 1;
	}

#ifdef ALWAYS_SEND_DURING_SPLIT
	if (currently_processing_netsplit && !(to->flags & FLAGS_BLOCKED)) {
		send_queued(to);
		return 0;
	}
#endif

	SQinK = (DBufLength(&to->sendQ) >> 10);
	if (IsServer(to)) {
		if (SQinK > to->lastsq) {
			send_queued(to);
		}
	}
	else {
		if (SQinK > to->lastsq + 4) {
			send_queued(to);
		}
	}
	return 0;
}

int send_queued(aClient *to)
{
	char *msg;
	int len, rlen, more_data = 0;

	if (IsDead(to)) {
		return -1;
	}

	if (ZipOut(to) && zip_is_data_out(to->serv->zip_out)) {
		if (DBufLength(&to->sendQ)) {
			more_data = 1;
		}
		else {
			int ldata = to->flags & FLAGS_BURST;

			msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);
			if (len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
				return dead_link(to, "Zip output error for %s", IRCERR_ZIP);
			}

#ifdef HAVE_OPENSSL_ON
			if (IsRC4OUT(to)) {
				rc4_process_stream(to->serv->rc4_out, msg, len);
			}
#endif

			if (!dbuf_put(&to->sendQ, msg, len)) {
				return dead_link(to, "Buffer allocation error for %s", IRCERR_BUFALLOC);
			}
		}
	}

	while (DBufLength(&to->sendQ) > 0) {
		msg = dbuf_map(&to->sendQ, (size_t *)&len);

		rlen = deliver_it(to, msg, len);
		if (rlen < 0) {
			return dead_link(to, "Write error to %s, closing link (%s)", errno);
		}

		dbuf_delete(&to->sendQ, rlen);
		to->lastsq = (DBufLength(&to->sendQ) >> 10);
		if (rlen < len) {
			break;
		}

		if (more_data && !DBufLength(&to->sendQ)) {
			int ldata = to->flags & FLAGS_BURST;

			more_data = 0;
			msg = zip_output(to->serv->zip_out, NULL, &len, 1, &ldata);

			if (len == -1) {
				sendto_realops("Zip output error for %s: (%d) %s", to->name, ldata, msg);
				return dead_link(to, "Zip outpu terror for %s", IRCERR_ZIP);
			}

#ifdef HAVE_OPENSSL_ON
			if (IsRC4OUT(to)) {
				rc4_process_stream(to->serv->rc4_out, msg, len);
			}
#endif

			if (!dbuf_put(&to->sendQ, msg, len)) {
				return dead_link(to, "Buffer allocation error for %s", IRCERR_BUFALLOC);
			}
		}
	}

	if ((to->flags & FLAGS_SOBSENT) && IsBurst(to) && DBufLength(&to->sendQ) < 20480) {
		if (!(to->flags & FLAGS_BURST)) {
			to->flags &= ~FLAGS_SOBSENT;
			sendto_one(to, "BURST %d", DBufLength(&to->sendQ));

#ifdef HTM_LOCK_ON_NETBURST
			if (!(to->flags & FLAGS_EOBRECV)) {
				HTMLOCK = NO;
			}
#endif
		}
	}

	return IsDead(to) ? -1 : 0;
}

void sendto_one(aClient *to, char *pattern, ...)
{
	va_list vl;
	int len;

	va_start(vl, pattern);
	len = ircvsprintf(sendbuf, pattern, vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to)) {
		sendto_ops("Trying to send to myself! [%s]", sendbuf);
		return;
	}
	send_message(to, sendbuf, len);
	va_end(vl);
	return;
}

void vsendto_one(aClient *to, char *pattern, va_list vl)
{
	int len = ircvsprintf(sendbuf, pattern, vl);

	if (to->from) {
		to = to->from;
	}
	if (IsMe(to) && to->fd >= 0) {
		sendto_ops("Trying to send to myself! [%s]", sendbuf);
		return;
	}
	send_message(to, sendbuf, len);
	return;
}

static inline int prefix_buffer(int remote, aClient *from, char *prefix, char *buffer, char *pattern, va_list vl)
{
	char *p;
	int msglen, sidx = 1;

	*buffer = ':';

	if (!remote && IsPerson(from)) {
		int flag = 0;
		anUser *user = from->user;

		for (p = from->name; *p; p++) {
			buffer[sidx++] = *p;
		}
		if (user) {
			if (*user->username) {
				buffer[sidx++] = '!';
				for (p = user->username; *p; p++) {
					buffer[sidx++] = *p;
				}
			}
			if (*user->host && !MyConnect(from)) {
				buffer[sidx++] = '@';
				for (p = IsMasked(from) ? user->maskedhost : user->host; *p; p++) {
					buffer[sidx++] = *p;
				}
				flag = 1;
			}
		}

		if (!flag && MyConnect(from) && *user->host) {
			buffer[sidx++] = '@';
			for (p = IsMasked(from) ? from->user->maskedhost : from->sockhost; *p; p++) {
				buffer[sidx++] = *p;
			}
		}
	}
	else {
		for (p = prefix; *p; p++) {
			buffer[sidx++] = *p;
		}
	}

	msglen = ircvsprintf(&buffer[sidx], pattern + 3, vl);
	msglen += sidx;
	return msglen;
}

static inline int check_fake_direction(aClient *from, aClient *to)
{
	if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
		if (IsServer(from)) {
			sendto_ops("Message to %s[%s] dropped from %s (Fake direction)",
				to->name, to->from->name, from->name);
			return -1;
		}

		sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username,
			IsMasked(to) ? to->user->maskedhost : to->user->host, from->name,
			from->user->username, IsMasked(from) ? from->user->maskedhost : from->user->host,
			to->from->name);
		sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
			me.name, to->name, to->user->username,
			IsMasked(to) ? to->user->maskedhost : to->user->host, to->from->name);
		to->flags |= FLAGS_KILLED;
		exit_client(NULL, to, &me, "Ghosted client");

		if (IsPerson(from)) {
			sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
				to->user->username, IsMasked(to) ? to->user->maskedhost : to->user->host,
				to->from);
		}
		return -1;
	}
	return 0;
}

void sendto_channel_butone(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	int i, didlocal = 0, didremote = 0;
	va_list vl;
	char *pfix;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	INC_SERIAL

	for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;
		if (acptr->from == one) {
			continue;
		}
		i = acptr->from->fd;
		if (MyClient(acptr)) {
			if (!didlocal) {
				didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			send_message(acptr, sendbuf, didlocal);
			sentalong[i] = sent_serial;
		}
		else {
			if (!didremote) {
				didremote = prefix_buffer(1, from, pfix, remotebuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			if (sentalong[i] != sent_serial) {
				send_message(acptr, remotebuf, didremote);
				sentalong[i] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}

void sendto_serv_butone(aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);

	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from)) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_nickip_servs_butone(int nickip, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from) || (nickip && !IsNICKIP(cptr)) ||
		  (!nickip && IsNICKIP(cptr))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_ssjoin_servs_butone(aChannel *chptr, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	if (chptr && *chptr->chname != '#') {
		return;
	}

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || (one && cptr == one->from) || (!IsSSJOIN(cptr) && !IsSSJv2(cptr))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_noquit_servs_butone(int noquit, aClient *one, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || one == cptr || (noquit && !IsNoQuit(cptr)) ||
		  (!noquit && IsNoQuit(cptr))) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_common_channels(aClient *from, char *pattern, ...)
{
	Link *channels;
	chanMember *users;
	aClient *cptr;
	va_list vl;
	char *pfix;
	int msglen = 0;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	INC_SERIAL

	if (from->fd >= 0) {
		sentalong[from->fd] = sent_serial;
	}
	if (from->user) {
		for (channels = from->user->channel; channels; channels = channels->next) {
			for (users = channels->value.chptr->members; users; users = users->next) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}

				sentalong[cptr->fd] = sent_serial;
				if (!msglen) {
					msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
				}
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	if (MyConnect(from)) {
		if (!msglen) {
			msglen = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
		}
		send_message(from, sendbuf, msglen);
	}
	va_end(vl);
	return;
}

void send_quit_to_common_channels(aClient *from, char *reason)
{
	Link *channels;
	chanMember *users;
	aClient *cptr;
	int msglen;

	INC_SERIAL

	msglen = sprintf(sendbuf, ":%s!%s@%s QUIT :%s", from->name, from->user->username,
		IsMasked(from) ? from->user->maskedhost : from->user->host, reason);

	if (from->fd >= 0) {
		sentalong[from->fd] = sent_serial;
	}
	for (channels = from->user->channel; channels; channels = channels->next) {
		if (!can_send(from, channels->value.chptr)) {
			for (users = channels->value.chptr->members; users; users = users->next) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}

				sentalong[cptr->fd] = sent_serial;
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	return;
}

void send_part_to_common_channels(aClient *from)
{
	Link *channels;
	chanMember *users;
	aClient *cptr;
	int msglen = 0;

	INC_SERIAL

	if (from->fd >= 0) {
		sentalong[from->fd] = sent_serial;
	}
	for (channels = from->user->channel; channels; channels = channels->next) {
		if (can_send(from, channels->value.chptr)) {
			msglen = sprintf(sendbuf, ":%s!%s@%s PART %s", from->name, from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host,
				channels->value.chptr->chname);
			for (users = channels->value.chptr->members; users; users = users->next) {
				cptr = users->cptr;
				if (!MyConnect(cptr) || sentalong[cptr->fd] == sent_serial) {
					continue;
				}

				sentalong[cptr->fd] = sent_serial;
				if (check_fake_direction(from, cptr)) {
					continue;
				}
				send_message(cptr, sendbuf, msglen);
			}
		}
	}
	return;
}

#ifdef FLUD
void sendto_channel_butlocal(aClient *one, aClient *from, aChannel *chptr, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	int i;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;
		if (acptr->from == one) {
			continue;
		}

		i = acptr->from->fd;
		if (!MyFludConnect(acptr)) {
			if (sentalong[i] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}
#endif

void sendto_channel_butserv(aChannel *chptr, aClient *from, char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	va_list vl;
	int didlocal = 0;
	char *pfix;

	va_start(vl, pattern);
	pfix = va_arg(vl, char *);

	for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;
		if (MyConnect(acptr)) {
			if (!didlocal) {
				didlocal = prefix_buffer(0, from, pfix, sendbuf, pattern, vl);
			}
			if (check_fake_direction(from, acptr)) {
				continue;
			}
			send_message(acptr, sendbuf, didlocal);
		}
	}
	va_end(vl);
	return;
}

static int match_it(aClient *one, char *mask, int what)
{
	if (what == MATCH_HOST) {
		return (match(mask, one->user->host) == 0);
	}
	else {
		return (match(mask, one->user->server) == 0);
	}
}

void sendto_match_servs(aChannel *chptr, aClient *from, char *pattern, ...)
{
	int i, j, k = 0;
	aClient *cptr;
	fdlist send_fdlist;
	va_list vl;

	if (chptr && *chptr->chname != '#') {
		return;
	}

	va_start(vl, pattern);
	for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry; i = serv_fdlist.entry[++j]) {
		if (!(cptr = local[i]) || cptr == from) {
			continue;
		}
		send_fdlist.entry[++k] = i;
	}
	send_fdlist.last_entry = k;
	if (k) {
		vsendto_fdlist(&send_fdlist, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_match_butone(aClient *one, aClient *from, char *mask, int what, char *pattern, ...)
{
	int i;
	aClient *cptr, *acptr;
	int cansendlocal, cansendglobal;
	va_list vl;

	va_start(vl, pattern);
	if (MyConnect(from)) {
		cansendlocal = OPCanLNotice(from) ? 1 : 0;
		cansendglobal = OPCanGNotice(from) ? 1 : 0;
	}
	else {
		cansendlocal = cansendglobal = 1;
	}

	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || cptr == one) {
			continue;
		}
		if (IsServer(cptr)) {
			if (!cansendglobal) {
				continue;
			}
			for (acptr = client; acptr; acptr = acptr->next) {
				if (IsRegisteredUser(acptr) && match_it(acptr, mask, what) &&
				  acptr->from == cptr) {
					break;
				}
			}
			if (!acptr) {
				continue;
			}
		}
		else if (!cansendlocal || !(IsRegisteredUser(cptr) && match_it(cptr, mask, what))) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_all_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsMe(cptr) || one == cptr) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_ops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || IsMe(cptr)) {
			continue;
		}
		switch (lev) {
			case CCONN_LEV:
				if (!SendCConnNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			case REJ_LEV:
				if (!SendRejNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			case SKILL_LEV:
				if (!SendSkillNotice(cptr)) {
					continue;
				}
				break;
			case SPY_LEV:
				if (!SendSpyNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			case DCCSEND_LEV:
				if (!(cptr->flags & FLAGS_SEENDCCNOTICE) || !IsOper(cptr)) {
					continue;
				}
				break;
			case FLOOD_LEV:
				if (!SendFloodNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			case SPAM_LEV:
				if (!SendSpamNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			case DEBUG_LEV:
				if (!SendDebugNotice(cptr) || !IsOper(cptr)) {
					continue;
				}
				break;
			default:
				if (!SendServNotice(cptr)) {
					continue;
				}
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_ops(char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || IsMe(cptr) || !IsOper(cptr) || !SendServNotice(cptr)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_ops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cptr = client; cptr; cptr = cptr->next) {
		i = cptr->from->fd;
		if (!SendWallops(cptr) || sentalong[i] == sent_serial || (one && cptr->from == one)) {
			continue;
		}

		sentalong[i] = sent_serial;
		vsendto_prefix_one(cptr->from, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void sendto_wallops_butone(aClient *one, aClient *from, char *pattern, ...)
{
	int i;
	aClient *cptr;
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || !(IsRegistered(cptr) && (SendWallops(cptr) || IsServer(cptr))) || cptr == one) {
			continue;
		}
		vsendto_prefix_one(cptr, from, pattern, vl);
	}
	va_end(vl);
	return;
}

void send_globops(char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || !IsOper(cptr) || IsMe(cptr) || !SendGlobops(cptr)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Global -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void send_chatops(char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || !IsOper(cptr) || IsMe(cptr) || !SendChatops(cptr)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** ChatOps --- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	anUser *user;
	char *idx, *par;
	int flag = 0, sidx = 0;
	va_list vl, vl2;

	va_start(vl, pattern);
	vl2 = vl;
	par = va_arg(vl, char *);

	if (from) {
		if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl2);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				va_end(vl);
				return;
			}
			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username,
				IsMasked(to) ? to->user->maskedhost : to->user->host, from->name,
				from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host, to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name,
				to->name, me.name, to->name, to->user->username,
				IsMasked(to) ? to->user->maskedhost : to->user->host, to->from->name);
			to->flags |= FLAGS_KILLED;
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
					to->user->username,
					IsMasked(to) ? to->user->maskedhost : to->user->host, to->from);
			}
			va_end(vl);
			return;
		}

		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			user = from->user;
			for (idx = from->name; *idx; idx++) {
				sender[sidx++] = *idx;
			}
			if (user) {
				if (*user->username) {
					sender[sidx++] = '!';
					for (idx = user->username; *idx; idx++) {
						sender[sidx++] = *idx;
					}
				}
				if (*user->host && !MyConnect(from)) {
					sender[sidx++] = '@';
					for (idx = IsMasked(from) ? user->maskedhost : user->host; *idx; idx++) {
						sender[sidx++] = *idx;
					}
					flag = 1;
				}
			}
			if (!flag && MyConnect(from) && *user->host) {
				sender[sidx++] = '@';
				for (idx = IsMasked(from) ? from->user->maskedhost : from->sockhost; *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}

			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;

	for (idx = par; *idx; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx; idx++) {
		temp[sidx++] = *idx;
	}
	temp[sidx] = '\0';

	vsendto_one(to, temp, vl);
	va_end(vl);
	return;
}

void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
	static char sender[HOSTLEN + NICKLEN + USERLEN + 5], temp[1024];
	anUser *user;
	char *idx, *par;
	int flag = 0, sidx = 0;
	va_list vl2 = vl;

	par = va_arg(vl2, char *);

	if (from) {
		if (!MyClient(from) && IsPerson(to) && to->from == from->from) {
			if (IsServer(from)) {
				ircvsprintf(temp, pattern, vl);
				sendto_ops("Send message (%s) to %s[%s] dropped from %s (Fake direction)",
					temp, to->name, to->from->name, from->name);
				return;
			}
			sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name, to->user->username,
				IsMasked(to) ? to->user->maskedhost : to->user->host, from->name,
				from->user->username,
				IsMasked(from) ? from->user->maskedhost : from->user->host, to->from->name);
			sendto_serv_butone(NULL, ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)", me.name, to->name,
				to->user->username, IsMasked(to) ? to->user->maskedhost : to->user->host,
				to->from->name);
			to->flags |= FLAGS_KILLED;
			exit_client(NULL, to, &me, "Ghosted client");

			if (IsPerson(from)) {
				sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name, from->name, to->name,
					to->user->username, IsMasked(to) ? to->user->maskedhost : to->user->host,
					to->from);
			}
			return;
		}
		if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
			user = from->user;
			for (idx = from->name; *idx; idx++) {
				sender[sidx++] = *idx;
			}
			if (user) {
				if (*user->username) {
					sender[sidx++] = '!';
					for (idx = user->username; *idx; idx++) {
						sender[sidx++] = *idx;
					}
				}
				if (*user->host && !MyConnect(from)) {
					sender[sidx++] = '@';
					for (idx = IsMasked(from) ? user->maskedhost : user->host; *idx; idx++) {
						sender[sidx++] = *idx;
					}
					flag = 1;
				}
			}
			if (!flag && MyConnect(from) && *user->host) {
				sender[sidx++] = '@';
				for (idx = IsMasked(from) ? from->user->maskedhost : from->sockhost; *idx; idx++) {
					sender[sidx++] = *idx;
				}
			}

			sender[sidx] = '\0';
			par = sender;
		}
	}

	temp[0] = ':';
	sidx = 1;

	for (idx = par; *idx; idx++) {
		temp[sidx++] = *idx;
	}
	for (idx = (pattern + 3); *idx; idx++) {
		temp[sidx++] = *idx;
	}

	temp[sidx] = '\0';
	vsendto_one(to, temp, vl2);
	return;
}

void sendto_fdlist(fdlist *listp, char *pattern, ...)
{
	int len, j, fd;
	va_list vl;

	va_start(vl, pattern);
	len = ircvsprintf(sendbuf, pattern, vl);
	for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j]) {
		send_message(local[fd], sendbuf, len);
	}
	va_end(vl);
	return;
}

void vsendto_fdlist(fdlist *listp, char *pattern, va_list vl)
{
	int len, j, fd;
	len = ircvsprintf(sendbuf, pattern, vl);

	for (fd = listp->entry[j = 1]; j <= listp->last_entry; fd = listp->entry[++j]) {
		send_message(local[fd], sendbuf, len);
	}
	return;
}

void sendto_realops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	va_start(vl, pattern);
	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !IsOper(cptr)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void vsendto_realops(char *pattern, va_list vl)
{
	aClient *cptr;
	int i, fd;
	char nbuf[1024];
	fdlist *listp;

	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !IsOper(cptr)) {
			continue;
		}
		ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name, cptr->name, pattern);
		vsendto_one(cptr, nbuf, vl);
	}
	return;
}

void sendto_realops_lev(int lev, char *pattern, ...)
{
	aClient *cptr;
	int i, fd, len = 0;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	listp = &oper_fdlist;
	va_start(vl, pattern);

	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd])) {
			continue;
		}
		switch (lev) {
			case CCONN_LEV:
				if (!SendCConnNotice(cptr)) {
					continue;
				}
				break;
			case REJ_LEV:
				if (!SendRejNotice(cptr)) {
					continue;
				}
				break;
			case SKILL_LEV:
				if (!SendSkillNotice(cptr)) {
					continue;
				}
				break;
			case SPY_LEV:
				if (!SendSpyNotice(cptr)) {
					continue;
				}
				break;
			case DCCSEND_LEV:
				if (!(cptr->flags & FLAGS_SEENDCCNOTICE)) {
					continue;
				}
				break;
			case FLOOD_LEV:
				if (!SendFloodNotice(cptr)) {
					continue;
				}
				break;
			case SPAM_LEV:
				if (!SendSpamNotice(cptr)) {
					continue;
				}
				break;
			case DEBUG_LEV:
				if (!SendDebugNotice(cptr)) {
					continue;
				}
				break;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void ts_warn(char *pattern, ...)
{
	static ts_val last = 0;
	static int warnings = 0;
	ts_val now;
	va_list vl;

	va_start(vl, pattern);

	now = time(NULL);
	if (now - last < 5) {
		if (++warnings > 5) {
			return;
		}
	}
	else {
		last = now;
		warnings = 0;
	}
	vsendto_realops(pattern, vl);
	va_end(vl);
	return;
}

void sendto_locops(char *pattern, ...)
{
	aClient *cptr;
	int i, fd, len = 0;
	char nbuf[1024];
	fdlist *listp;
	va_list vl;

	va_start(vl, pattern);
	listp = &oper_fdlist;
	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || !SendGlobops(cptr)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_gnotice(char *pattern, ...)
{
	aClient *cptr;
	int i, len = 0;
	char nbuf[1024];
	va_list vl;

	va_start(vl, pattern);
	for (i = 0; i <= highest_fd; i++) {
		if (!(cptr = local[i]) || IsServer(cptr) || IsMe(cptr) || !SendRnotice(cptr)) {
			continue;
		}
		len = ircsprintf(nbuf, ":%s NOTICE %s :*** Routing -- ", me.name, cptr->name);
		strncat(nbuf, pattern, sizeof(nbuf) - len);
		vsendto_one(cptr, nbuf, vl);
	}
	va_end(vl);
	return;
}

void sendto_channelflag_butone(int flag, aClient *one, aClient *from, aChannel *chptr,
  char *pattern, ...)
{
	chanMember *cm;
	aClient *acptr;
	int i;
	va_list vl;

	va_start(vl, pattern);

	INC_SERIAL

	for (cm = chptr->members; cm; cm = cm->next) {
		acptr = cm->cptr;
		if (acptr->from == one || !(cm->flags & flag)) {
			continue;
		}

		i = acptr->from->fd;
		if (MyConnect(acptr) && IsRegisteredUser(acptr)) {
			vsendto_prefix_one(acptr, from, pattern, vl);
			sentalong[i] = sent_serial;
		}
		else {
			if (sentalong[i] != sent_serial) {
				vsendto_prefix_one(acptr, from, pattern, vl);
				sentalong[i] = sent_serial;
			}
		}
	}
	va_end(vl);
	return;
}

void flush_connections(int fd)
{
	int i;
	aClient *cptr;

	if (fd == me.fd) {
		for (i = highest_fd; i >= 0; i--) {
			if (!(cptr = local[i]) || cptr->flags & FLAGS_BLOCKED || DBufLength(&cptr->sendQ) <= 0) {
				continue;
			}
			send_queued(cptr);
		}
	}
	else if (fd >= 0 && (cptr = local[fd]) && !(cptr->flags & FLAGS_BLOCKED) && DBufLength(&cptr->sendQ) > 0) {
		send_queued(cptr);
	}
	return;
}

void dump_connections(int fd)
{
	int i;
	aClient *cptr;

	if (fd == me.fd) {
		for (i = highest_fd; i >= 0; i--) {
			if (!(cptr = local[i]) || DBufLength(&cptr->sendQ) <= 0) {
				continue;
			}
			send_queued(cptr);
		}
	}
	else if (fd >= 0 && (cptr = local[fd]) && DBufLength(&cptr->sendQ) > 0) {
		send_queued(cptr);
	}
	return;
}

void flush_fdlist_connections(fdlist *listp)
{
	int i, fd;
	aClient *cptr;

	for (fd = listp->entry[i = 1]; i <= listp->last_entry; fd = listp->entry[++i]) {
		if (!(cptr = local[fd]) || cptr->flags & FLAGS_BLOCKED || DBufLength(&cptr->sendQ) <= 0) {
			continue;
		}
		send_queued(cptr);
	}
	return;
}
