/*
 * RageIRCd, an advanced Internet Relay Chat daemon (IRCd)
 * src/support.c: other misc functions
 *
 * Copyright (C) 2002 the past and present IRCd developers, and others.
 * Refer to doc/authors/ for full credits and copyright details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * $Id: support.c,v 1.16 2002/04/08 16:00:52 rageircd Exp $
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "memory.h"

extern int errno;

void strip(char *string)
{
	char *s;

	if (!string) {
		return;
	}

	if ((s = strchr(string, '\n')) || (s = strchr(string, '\r'))) {
		*s = '\0';
	}
	return;
}

/*
 * atime_exp() and atime() - originally by Potvin
 *
 * converts time format of "dhms" (eg, 1d2h3m4s)
 * to seconds.
 */
static time_t atime_exp(char *base, char *ptr)
{
	time_t tmp;
	char *p, c = *ptr;

	p = ptr;
	*ptr-- = '\0';

	while (ptr-- > base) {
		if (isalpha(*ptr)) {
			break;
		}
	}
	tmp = atoi(ptr + 1);
	*p = c;
	return tmp;
}

time_t atime(char *xtime)
{
	char *d, *h, *m, *s;
	time_t D, H, M, S;
	int i = 0;

	d = h = m = s = NULL;
	D = H = M = S = 0;

	for (d = xtime; *d; d++) {
		if (isalpha(*d) && i != 1) {
			i = 1;
		}
	}
	if (!i) {
		return (atol(xtime));
	}

	d = strchr(xtime, 'd');
	h = strchr(xtime, 'h');
	m = strchr(xtime, 'm');
	s = strchr(xtime, 's');

	if (d) {
		D = atime_exp(xtime, d);
	}
	if (h) {
		H = atime_exp(xtime, h);
	}
	if (m) {
		M = atime_exp(xtime, m);
	}
	if (s) {
		S = atime_exp(xtime, s);
	}
	return ((D * 86400) + (H * 3600) + (M * 60) + S);
}

#ifndef HAVE_STRTOKEN
char *strtoken(char **save, char *str, char *fs)
{
	char *pos = *save;
	char *tmp;

	if (str) {
		pos = str;
	}
	while (pos && *pos && strchr(fs, *pos)) {
		pos++;
	}
	if (!pos || !*pos) {
		return (pos = *save = NULL);
	}

	tmp = pos;
	while (*pos && !strchr(fs, *pos)) {
		pos++;
	}
	if (*pos) {
		*pos++ = '\0';
	}
	else {
		pos = NULL;
	}
	*save = pos;
	return (tmp);
}
#endif

#ifndef HAVE_STRTOK
char *strtok(char *str, char *fs)
{
	static char *pos;
	return strtoken(&pos, str, fs);
}
#endif

#ifndef HAVE_STRERROR
char *strerror(int err_no)
{
#if !defined(__FreeBSD__) && !defined(__NetBSD__)
	extern char *str_errlist[];
	extern int sys_nerr;
#endif
	static char buff[40];
	char *errp = (err_no > sys_nerr) ? NULL : sys_errorlist[err_no]);
	if (!errp) {
		ircsprintf(buff, "Unknown error %d", err_no);
		errp = buff;
	}
	return errp;
}
#endif

char *inetntoa(char *in)
{
	static char buf[16];
	u_char *s = (u_char *)in;
	int a, b, c, d;

	a = (int)*s++;
	b = (int)*s++;
	c = (int)*s++;
	d = (int)*s++;
	ircsprintf(buf, "%d.%d.%d.%d", a, b, c, d);
	return buf;
}

#ifndef HAVE_INET_NETOF
int inet_netof(struct in_addr in)
{
	int addr = in.s_net;

	if (!(addr & 0x80)) {
		return (int)in.s_net;
	}
	if (!(addr & 0x40)) {
		return ((int)in.s_net * 256 + in.s_host);
	}
	return ((int)in.s_net * 256 + in.s_host * 256 + in.s_lh);
}
#endif

char *my_itoa(int d)
{
	static char buf[128];
	ircsprintf(buf, "%d", d);
	return buf;
}

int dgets(int fd, char *buf, int num)
{
	static char dgbuf[8192], *head = dgbuf, *tail = dgbuf;
	char *s, *t;
	int n, nr;

	if (head == tail) {
		*head = '\0';
	}
	if (!num) {
		head = tail = dgbuf;
		*head = '\0';
		return 0;
	}
	if (num > sizeof(dgbuf) - 1) {
		num = sizeof(dgbuf) - 1;
	}

	for (;;)
	{
		if (head > dgbuf) {
			for (nr = tail - head, s = head, t = dgbuf; nr > 0; nr--) {
				*t++ = *s++;
			}
			tail = t;
			head = dgbuf;
		}

		if (head < tail && ((s = strchr(head, '\n')) || (s = strchr(head, '\r'))) && s < tail) {
			n = MIN(s - head + 1, num);
			memcpy(buf, head, n);
			head += n;
			if (head == tail) {
				head = tail = dgbuf;
			}
			return n;
		}

		if (tail - head >= num) {
			n = num;
			memcpy(buf, head, n);
			head += n;
			if (head == tail) {
				head = tail = dgbuf;
			}
			return n;
		}

		n = sizeof(dgbuf) - (tail - dgbuf) - 1;
		nr = read(fd, tail, n);
		if (nr == -1) {
			head = tail = dgbuf;
			return -1;
		}

		if (!nr) {
			if (tail > head) {
				n = MIN(tail - head, num);
				memcpy(buf, head, n);
				head += n;
				if (head == tail) {
					head = tail = dgbuf;
				}
				return n;
			}
			head = tail = dgbuf;
			return 0;
		}

		tail += nr;
		*tail = '\0';

		for (t = head; (s = strchr(t, '\n'));) {
			if ((s > head) && (s > dgbuf)) {
				t = s - 1;
				for (nr = 0; *t == '\\'; nr++) {
					t--;
				}
				if (nr & 1) {
					t = s + 1;
					s--;
					nr = tail - t;
					while (nr--) {
						*s++ = *t++;
					}
					tail -= 2;
					*tail = '\0';
				}
				else {
					s++;
				}
			}
			else {
				s++;
			}
			t = s;
		}
		*tail = '\0';
	}
}
